/*	
**	
**	Copyright (C) 1999 Geert Bevin
**	gbevin@thunderstorms.org
**	http://thunderstorms.org/fm
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation; either version 2 of the License, or
**	(at your option) any later version.
**	
**	This program is distributed in the hope that it will be useful,
**	but WITHOUT ANY WARRANTY; without even the implied warranty of
**	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**	GNU General Public License for more details.
**	
**	You should have received a copy of the GNU General Public License
**	along with this program; if not, write to the Free Software
**	
**	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**		
*/

#include <stdio.h>
#include <string.h>

#include <glib.h>
#include <gdk/gdkkeysyms.h>
#include <gdk/gdktypes.h>
#include <gtk/gtk.h>

#include "accelkey.h"
#include "fmwindow.h"
#include "preferences.h"

extern struct Preferences	*pgPreferences;

struct AccelKey* accelKeyFind(GHashTable *accelKeys, gchar *name)
{
	struct AccelKey	*accel_key = NULL;
	
	accel_key = g_hash_table_lookup(accelKeys, name);
	
	return accel_key;
}

struct AccelKey* accelKeyNew(GHashTable *accelKeys, gchar *id, guint key, guint mods)
{
	struct AccelKey	*accel_key = NULL;
	gchar			*real_id = NULL;
	
//	printf("accelKeyNew : id='%s', key=%d, mods=%d.\n", id, key, mods);
	accel_key = g_new(struct AccelKey, 1);
	if(accel_key)
	{
		real_id = g_strdup(id);
		if(real_id)
		{
			accelKeyRemove(accelKeys, id);
			g_hash_table_insert(accelKeys, real_id, accel_key);
			
			accel_key->key = key;
			accel_key->mods = mods;
			
			return accel_key;
		}
		g_free(accel_key);
	}
	
	return NULL;
}

void accelKeyRemove(GHashTable *accelKeys, gchar *name)
{
	gpointer	orig_key;
	gpointer	orig_value;
	
//	printf("Remove accelkey '%s'\n", name);
	if(g_hash_table_lookup_extended(accelKeys, name, &orig_key, &orig_value))
	{
//		printf("It exists, thus removing it\n");
		if(orig_key)
		{
			g_hash_table_remove(accelKeys, orig_key);
			g_free(orig_key);
		}
		if(orig_value)
		{
			g_free(orig_value);
		}
	}
}

void accelKeyUpdateInAllFmWindows(GHashTable *accelKeys, GList *fmWindows, gchar *id)
{
	guint				number_of_fm_windows = g_list_length(fmWindows);
	guint				counter = 0;
	struct	FmWindow	*fm_window = NULL;
	GtkWidget			*registered_widget = NULL;
    struct AccelKey		*global_accel_key = NULL;
	guint				accel_key_key = 0;
	GdkModifierType		accel_key_mods = 0;

//	printf("accelKeyUpdateInAllFmWindows : id='%s'\n", id);
	global_accel_key = accelKeyFind(accelKeys, id);
	for(counter = 0; counter < number_of_fm_windows; counter++)
	{
		fm_window = g_list_nth_data(fmWindows, counter);
		registered_widget = fmWindow_acceleratedWidgetFind(fm_window, id);
		accel_key_key = (guint)gtk_object_get_data(GTK_OBJECT(registered_widget), "accelKeyKey");
		accel_key_mods = (GdkModifierType)gtk_object_get_data(GTK_OBJECT(registered_widget), "accelKeyMods");
//		printf("=> counter=%d, fmWindow=0x%x, registered_widget=0x%x\n", counter, (guint)fm_window, (guint)registered_widget);
		if(!global_accel_key && (accel_key_key || accel_key_mods))
		{
//			printf("==> let's remove it\n");
			fmWindow_accelKeyRemoveFromGtkWidget(fm_window, registered_widget);

			continue;
		}
		else if(global_accel_key && !accel_key_key && !accel_key_mods)
		{
//			printf("==> let's add it\n");
			fmWindow_accelKeyAddToGtkWidget(fm_window, registered_widget);

			continue;
		}
//		printf("==> do nothing\n");
	}
}

void accelKeysRemoveAll(GHashTable *accelKeys)
{
//	printf("pAccelKeys size = %d\n", g_hash_table_size(accelKeys));
	g_hash_table_foreach_remove(accelKeys, accelKeysForEachRemove, NULL);
}

gboolean accelKeysForEachRemove(gpointer key, gpointer value, gpointer user_data)
{
	if(key)
	{
		g_free(key);
	}
	if(value)
	{
		g_free(value);
	}

	return TRUE;
}

