/*	
**	
**	Copyright (C) 1999 Geert Bevin
**	gbevin@thunderstorms.org
**	http://thunderstorms.org/fm
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation; either version 2 of the License, or
**	(at your option) any later version.
**	
**	This program is distributed in the hope that it will be useful,
**	but WITHOUT ANY WARRANTY; without even the implied warranty of
**	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**	GNU General Public License for more details.
**	
**	You should have received a copy of the GNU General Public License
**	along with this program; if not, write to the Free Software
**	
**	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**		
*/

#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include <glib.h>
#include <gdk/gdkkeysyms.h>
#include <gdk/gdktypes.h>
#include <gtk/gtk.h>

#include "filetype.h"
#include "fileutils.h"
#include "fmwindow.h"
#include "rowstyle.h"

gint fileTypeCompare(gconstpointer fileTypeNode, gconstpointer fileTypeToCompare)
{
	const struct FileType	*file_type = fileTypeNode;
	const struct FileType	*file_type_to_compare = fileTypeToCompare;
	guint					return_value = 0;
	
	if(file_type_to_compare->basicType != -1 && file_type->basicType != file_type_to_compare->basicType)
	{
		return_value = -1;
	}
	
	if(file_type_to_compare->pRowStyle != NULL && file_type->pRowStyle != file_type_to_compare->pRowStyle)
	{
		return_value = -1;
	}

	return return_value;
}

struct FileType* fileTypeGet(GList *fileTypes, GNode *currentNode)
{
	struct FileType	file_type_to_compare;
	GList			*file_type_node = NULL;
	struct FileType	*file_type_result = NULL;

//	printf("fileTypeGet '%s'\n", ((struct NodeInfo*)currentNode->data)->pFullName);

    file_type_to_compare.basicType = -1;
    file_type_to_compare.pRowStyle = NULL;

	if(isRegularFile(currentNode))
	{
//		printf("isRegularFile\n");
        file_type_to_compare.basicType = REGULAR_FILE;
		file_type_node = g_list_find_custom(fileTypes, &file_type_to_compare, fileTypeCompare);
	}
	else if(isDirectoryFilled(currentNode))
	{
//		printf("isDirectoryFilled\n");
        file_type_to_compare.basicType = DIRECTORY_FILLED;
		file_type_node = g_list_find_custom(fileTypes, &file_type_to_compare, fileTypeCompare);
	}
	else if(isCharacterDevice(currentNode))
	{
//		printf("isCharacterDevice\n");
        file_type_to_compare.basicType = CHARACTER_DEVICE;
		file_type_node = g_list_find_custom(fileTypes, &file_type_to_compare, fileTypeCompare);
	}
	else if(isBlockDevice(currentNode))
	{
//		printf("isBlockDevice\n");
        file_type_to_compare.basicType = BLOCK_DEVICE;
		file_type_node = g_list_find_custom(fileTypes, &file_type_to_compare, fileTypeCompare);
	}
	else if(isDirectoryEmpty(currentNode))
	{
//		printf("isDirectoryEmpty\n");
        file_type_to_compare.basicType = DIRECTORY_EMPTY;
		file_type_node = g_list_find_custom(fileTypes, &file_type_to_compare, fileTypeCompare);
	}
	else if(isDirectoryInaccessible(currentNode))
	{
//		printf("isDirectoryInaccessible\n");
        file_type_to_compare.basicType = DIRECTORY_INACCESSIBLE;
		file_type_node = g_list_find_custom(fileTypes, &file_type_to_compare, fileTypeCompare);
	}
	else if(isLink(currentNode))
	{
//		printf("isLink\n");
        file_type_to_compare.basicType = BROKEN_LINK;
		file_type_node = g_list_find_custom(fileTypes, &file_type_to_compare, fileTypeCompare);
	}
	else if(isSocket(currentNode))
	{
//		printf("isSocket\n");
        file_type_to_compare.basicType = SOCKET;
		file_type_node = g_list_find_custom(fileTypes, &file_type_to_compare, fileTypeCompare);
	}
	else if(isFifo(currentNode))
	{
//		printf("isFifo\n");
        file_type_to_compare.basicType = FIFO;
		file_type_node = g_list_find_custom(fileTypes, &file_type_to_compare, fileTypeCompare);
	}
	else
	{
//		printf("isUNKNOWN\n");
        file_type_to_compare.basicType = UNKNOWN;
		file_type_node = g_list_find_custom(fileTypes, &file_type_to_compare, fileTypeCompare);
	}
	
//	printf("file_type_node = 0x%x\n", (guint)file_type_node);
	if(file_type_node)
	{
		file_type_result = file_type_node->data;
	}
	
	return file_type_result;
}

struct FileType* fileTypeNew(GList **fileTypes, GHashTable *rowStyles, gboolean deletable, BasicFileType basicType, gchar *rowStyle)
{
	struct RowStyle	*row_style = NULL;
	struct FileType	*new_file_type = NULL;
	GList			*existing_file_type_node = NULL;
	
	row_style = rowStyleFind(rowStyles, rowStyle);
	
//	printf("basicType = %d deletable = %d rowStyle = %s row_style = 0x%x\n", basicType, deletable, rowStyle, (guint)row_style);

	if(row_style)
	{
		new_file_type = g_new(struct FileType, 1);
		if(new_file_type)
		{
			new_file_type->deletable = deletable,
			new_file_type->basicType = basicType;
			new_file_type->pRowStyle = row_style;

			existing_file_type_node = g_list_find_custom(*fileTypes, new_file_type, fileTypeCompare);
			if(existing_file_type_node)
			{
				fileTypeRemove(fileTypes, (struct FileType*)existing_file_type_node->data);
			}
		
			*fileTypes = g_list_prepend(*fileTypes, new_file_type);
			row_style->pReferencingFileTypes = g_list_prepend(row_style->pReferencingFileTypes, new_file_type);
		}
	}

	return new_file_type;
}

void fileTypeRemove(GList **fileTypes, struct FileType *fileType)
{
	GList	*glist_node = NULL;

	glist_node = g_list_find(*fileTypes, fileType);
	
	if(glist_node)
	{
		fileType->pRowStyle->pReferencingFileTypes = g_list_remove(fileType->pRowStyle->pReferencingFileTypes, fileType);
		*fileTypes = g_list_remove(*fileTypes, fileType);
		g_free(fileType);
	}
}

void fileTypesRemoveAll(GList **fileTypes)
{
	GList			*glist_node = NULL;
	struct FileType	*file_type = NULL;
	
	while((glist_node = (GList*)g_list_nth(*fileTypes, 0)))
	{
		file_type = g_list_nth_data(*fileTypes, 0);
		file_type->pRowStyle->pReferencingFileTypes = g_list_remove(file_type->pRowStyle->pReferencingFileTypes, file_type);
		*fileTypes = g_list_remove(*fileTypes, file_type);
		g_free(file_type);
	}
}

