/*	
**	
**	Copyright (C) 1999 Geert Bevin
**	gbevin@thunderstorms.org
**	http://thunderstorms.org/fm
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation; either version 2 of the License, or
**	(at your option) any later version.
**	
**	This program is distributed in the hope that it will be useful,
**	but WITHOUT ANY WARRANTY; without even the implied warranty of
**	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**	GNU General Public License for more details.
**	
**	You should have received a copy of the GNU General Public License
**	along with this program; if not, write to the Free Software
**	
**	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**		
*/

#include <dirent.h>
#include <errno.h>
#include <math.h>
#include <pwd.h>
#include <grp.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include <sys/stat.h>
#include <sys/types.h>

#include <glib.h>
#include <gtk/gtk.h>

#include "fileutils.h"
#include "fmwindow.h"


gchar* compressPath(gchar *path)
{
	gchar	**split_path_elements = NULL;
	gchar	**split_path_element = NULL;
	GList	*list_path_elements = NULL;
	GList	*list_path_element = NULL;
	guint	path_element_id = 0;
	gchar	*path_element = NULL;
	GString	*compressed_path = g_string_new(NULL);
	gchar	*compressed_path_string = NULL;

//	printf("path = %s\n", path);

	// split the path elements and put them in the list
	split_path_elements = g_strsplit(path, "/", 0);
	split_path_element = split_path_elements;
	while(*split_path_element)
	{
		list_path_elements = g_list_prepend(list_path_elements, g_strdup(*split_path_element));
		split_path_element++;
	}
	g_strfreev(split_path_elements);
	list_path_elements = g_list_reverse(list_path_elements);
	
	// remove the unwanted nodes from the list to have only the relevant path parts
	
	path_element_id = 1;
	while(path_element_id < g_list_length(list_path_elements))
	{
		list_path_element = g_list_nth(list_path_elements, path_element_id);
		path_element = (gchar *)list_path_element->data;
//		printf("%s\n", path_element);
		if(strcmp(path_element, "") == 0 || strcmp(path_element, ".") == 0 || strcmp(path_element, "..") == 0)
		{
			if(strcmp(path_element, "..") == 0)
			{
				if(path_element_id > 1)
				{
					path_element_id--;
					list_path_element = g_list_nth(list_path_elements, path_element_id);
//					printf("removing previous %s\n", (gchar*)list_path_element->data);
					list_path_elements = g_list_remove_link(list_path_elements, list_path_element);
					g_free(list_path_element->data);
					g_list_free_1(list_path_element);

					list_path_element = g_list_nth(list_path_elements, path_element_id);
				}
			}
//			printf("removing %s\n", (gchar*)list_path_element->data);
			list_path_elements = g_list_remove_link(list_path_elements, list_path_element);
			g_free(list_path_element->data);
			g_list_free_1(list_path_element);
		}
		else
		{
			path_element_id++;
		}
	}
	
	// create the compressed path
	list_path_element = g_list_first(list_path_elements);
	while(list_path_element)
	{
		if(list_path_element->data && strcmp((gchar*)list_path_element->data, "") != 0)
		{
			compressed_path = g_string_append(compressed_path, list_path_element->data);
		}
		if(list_path_element != g_list_last(list_path_elements))
		{
			compressed_path = g_string_append(compressed_path, "/");
		}
		list_path_element = g_list_next(list_path_element);
	}

	compressed_path_string = compressed_path->str;
	g_string_free(compressed_path, FALSE);
//	printf("compressed_path_string = %s\n", compressed_path_string);

	return compressed_path_string;
}

gchar* getAccessTime(GNode *currentNode)
{
	struct NodeInfo *node_info = NULL;
	time_t	atime = 0;
	gchar	*atime_string = NULL;

	if(currentNode)
	{
		node_info = (struct NodeInfo *)currentNode->data;
		if(node_info->pStat)
		{
			atime = node_info->pStat->st_atime;
    	    atime_string = getFormattedTime(atime);

		    return atime_string;
		}
	}

	atime_string = g_strdup("Unknown");

    return atime_string;
}

gchar* getChangeTime(GNode *currentNode)
{
	struct NodeInfo *node_info = NULL;
	time_t	ctime = 0;
	gchar	*ctime_string = NULL;

	if(currentNode)
	{
		node_info = (struct NodeInfo *)currentNode->data;
	    if(node_info->pStat)
		{
			ctime = node_info->pStat->st_ctime;
	        ctime_string = getFormattedTime(ctime);

		    return ctime_string;
		}
	}

	ctime_string = g_strdup("Unknown");

    return ctime_string;
}

gchar* getGroupName(GNode *currentNode)
{
	struct NodeInfo *node_info = NULL;
	gid_t			gid = 0;
    struct group    *temp_group = NULL;
	guint			gid_length = 0;
	gchar			*gname_string = NULL;

	if(currentNode)
	{
		node_info = (struct NodeInfo *)currentNode->data;
	    if(node_info->pStat)
		{
			gid = node_info->pStat->st_gid;
			temp_group = (struct group *)getgrgid(gid);
			if(temp_group)
			{
				gname_string = g_strdup(temp_group->gr_name);
			}
			else
			{
				if((gid_t)0 == gid)
				{
					return g_strdup("root");
				}
				gid_length = log10(gid)+1;
				gname_string = g_new(gchar, gid_length+1);
				sprintf(gname_string, "%d\n", (gint)gid);
			}

		    return gname_string;
		}
	}
	
	gname_string = g_strdup("Unknown");

    return gname_string;
}

GNode* getLinkSource(GNode* currentNode)
{
	GNode	*previous_gnode = NULL;

	if(currentNode && currentNode->data)
	{
		previous_gnode = currentNode;
//		printf("%s\n", ((struct NodeInfo*)previous_gnode->data)->pFullName);
		while(((struct NodeInfo*)previous_gnode->data)->pLinkFromNode)
		{
			previous_gnode = ((struct NodeInfo*)previous_gnode->data)->pLinkFromNode;
//			printf("%s\n", ((struct NodeInfo*)previous_gnode->data)->pFullName);
		}
	}

	return previous_gnode;
}

GNode* getLinkTarget(GNode* currentNode)
{
	GNode	*next_gnode = NULL;

	if(currentNode && currentNode->data)
	{
		next_gnode = currentNode;
//		printf("%s\n", ((struct NodeInfo*)next_gnode->data)->pFullName);
		while(((struct NodeInfo*)next_gnode->data)->pLinkToNode)
		{
			next_gnode = ((struct NodeInfo*)next_gnode->data)->pLinkToNode;
//			printf("%s\n", ((struct NodeInfo*)next_gnode->data)->pFullName);
		}
	}

	return next_gnode;
}

gchar* getFormattedTime(time_t time)
{
	struct tm	*local = NULL;
	gchar		formatted_time[50];

	local = localtime(&time);
	strftime(formatted_time, 50, "%Y-%m-%d %T", local);

	return g_strdup(formatted_time);
}

gchar* getModificationTime(GNode *currentNode)
{
	struct NodeInfo *node_info = NULL;
	time_t			mtime = 0;
	gchar			*mtime_string = NULL;

	if(currentNode)
	{
		node_info = (struct NodeInfo *)currentNode->data;
	    if(node_info->pStat)
		{
			mtime = node_info->pStat->st_mtime;
	        mtime_string = getFormattedTime(mtime);

		    return mtime_string;
		}
	}

	mtime_string = g_strdup("Unknown");

    return mtime_string;
}

gchar* getPermissions(GNode *currentNode)
{
	struct NodeInfo *node_info = NULL;
	mode_t			mode = 0;
	gchar			mode_string[10];

	if(currentNode)
	{
		node_info = (struct NodeInfo *)currentNode->data;
	    if(node_info->pStat)
		{
            mode_string[9] = 0;
			mode = node_info->pStat->st_mode;
			if(mode & S_IRUSR)
			{
				mode_string[0] = 'r';
			}
			else
			{
				mode_string[0] = '-';
			}
			if(mode & S_IWUSR)
			{
				mode_string[1] = 'w';
			}
			else
			{
				mode_string[1] = '-';
			}
			if(mode & S_IXUSR)
			{
				mode_string[2] = 'x';
			}
			else
			{
				mode_string[2] = '-';
			}
			if(mode & S_IRGRP)
			{
				mode_string[3] = 'r';
			}
			else
			{
				mode_string[3] = '-';
			}
			if(mode & S_IWGRP)
			{
				mode_string[4] = 'w';
			}
			else
			{
				mode_string[4] = '-';
			}
			if(mode & S_IXGRP)
			{
				mode_string[5] = 'x';
			}
			else
			{
				mode_string[5] = '-';
			}
			if(mode & S_IROTH)
			{
				mode_string[6] = 'r';
			}
			else
			{
				mode_string[6] = '-';
			}
			if(mode & S_IWOTH)
			{
				mode_string[7] = 'w';
			}
			else
			{
				mode_string[7] = '-';
			}
			if(mode & S_IXOTH)
			{
				mode_string[8] = 'x';
			}
			else
			{
				mode_string[8] = '-';
			}

		    return g_strdup(mode_string);
		}
	}

    return g_strdup("Unknown");
}

gchar* getSize(GNode *currentNode)
{
	struct NodeInfo *node_info = NULL;
	off_t	size = 0;
	guint	size_length = 0;
	gchar	*size_string = NULL;

	if(currentNode)
	{
		node_info = (struct NodeInfo *)currentNode->data;
	    if(node_info->pStat)
		{
			size = node_info->pStat->st_size;
			// check for bounds due to the size of files on a Solaris 2.5 /tmp directory
			if(size <= G_MAXINT)
			{
				size_length = log10(size)+1;
				if(isDirectory(currentNode))
				{
					size_length = size_length+2;
					size_string = g_new(gchar, size_length+1);
					sprintf(size_string, "(%d)", (int)size);
				}
				else
				{
					size_string = g_new(gchar, size_length+1);
					sprintf(size_string, "%d", (int)size);
				}
			}
			else
			{
                size = G_MAXINT;

				if(isDirectory(currentNode))
				{
					size_length = size_length+2;
					size_string = g_new(gchar, size_length+1);
					sprintf(size_string, "(> %d)", (int)size);
				}
				else
				{
					size_string = g_new(gchar, size_length+1);
					sprintf(size_string, "> %d", (int)size);
				}
			}

		    return size_string;
		}
	}

	size_string = g_strdup("Unknown");

    return size_string;
}

gchar* getUserName(GNode *currentNode)
{
	struct NodeInfo *node_info = NULL;
	uid_t			uid = 0;
	struct passwd   *temp_passwd = NULL;
	guint			uid_length = 0;
	gchar			*uname_string = NULL;

	if(currentNode)
	{
		node_info = (struct NodeInfo *)currentNode->data;
	    if(node_info->pStat)
		{
			uid = node_info->pStat->st_uid;
			temp_passwd = (struct passwd *)getpwuid(uid);
			if(temp_passwd)
			{
				uname_string = g_strdup(temp_passwd->pw_name);
			}
			else
			{
				if((uid_t)0 == uid)
				{
					return g_strdup("root");
				}
				uid_length = log10(uid)+1;
				uname_string = g_new(gchar, uid_length+1);
				sprintf(uname_string, "%d\n", uid);
			}

		    return uname_string;
		}
	}

	uname_string = g_strdup("Unknown");

    return uname_string;
}

gboolean isBlockDevice(GNode *currentNode)
{
    struct NodeInfo	*node_info = NULL;

	if(currentNode)
	{
		node_info = (struct NodeInfo*)currentNode->data;

		if(node_info && node_info->pStat && S_ISBLK(node_info->pStat->st_mode))
		{
			return TRUE;
		}
	}

	return FALSE;
}

gboolean isCharacterDevice(GNode *currentNode)
{
    struct NodeInfo	*node_info = NULL;

	if(currentNode)
	{
		node_info = (struct NodeInfo*)currentNode->data;

		if(node_info && node_info->pStat && S_ISCHR(node_info->pStat->st_mode))
		{
			return TRUE;
		}
	}

	return FALSE;
}

gboolean isDirectory(GNode *currentNode)
{
    struct NodeInfo	*node_info = NULL;

	if(currentNode)
	{
		node_info = (struct NodeInfo*)currentNode->data;

		if(node_info && node_info->pStat && S_ISDIR(node_info->pStat->st_mode))
		{
			return TRUE;
		}
	}

	return FALSE;
}

gboolean isDirectoryFilled(GNode *currentNode)
{
    struct NodeInfo	*node_info = NULL;

	if(currentNode)
	{
		node_info = (struct NodeInfo*)currentNode->data;

		if(node_info && node_info->pStat && S_ISDIR(node_info->pStat->st_mode))
		{
			if(node_info->dirHasChildren)
			{
				return TRUE;
			}
		}
	}

	return FALSE;
}

gboolean isDirectoryEmpty(GNode *currentNode)
{
    struct NodeInfo	*node_info = NULL;

	if(currentNode)
	{
		node_info = (struct NodeInfo*)currentNode->data;

		if(node_info && node_info->pStat && S_ISDIR(node_info->pStat->st_mode))
		{
			if(!node_info->dirHasChildren && node_info->dirIsAccessible)
			{
				return TRUE;
			}
		}
	}

	return FALSE;
}

gboolean isDirectoryInaccessible(GNode *currentNode)
{
    struct NodeInfo	*node_info = NULL;

	if(currentNode)
	{
		node_info = (struct NodeInfo*)currentNode->data;

		if(node_info && node_info->pStat && S_ISDIR(node_info->pStat->st_mode))
		{
			if(!node_info->dirIsAccessible)
			{
				return TRUE;
			}
		}
	}

	return FALSE;
}

gboolean isFifo(GNode *currentNode)
{
    struct NodeInfo	*node_info = NULL;

	if(currentNode)
	{
		node_info = (struct NodeInfo*)currentNode->data;

		if(node_info && node_info->pStat && S_ISFIFO(node_info->pStat->st_mode))
		{
			return TRUE;
		}
	}
	
	return FALSE;
}

gboolean isLink(GNode *currentNode)
{
    struct NodeInfo	*node_info = NULL;

	if(currentNode)
	{
		node_info = (struct NodeInfo*)currentNode->data;

		if(node_info && node_info->pStat && S_ISLNK(node_info->pStat->st_mode))
		{
			return TRUE;
		}
	}
	
	return FALSE;
}

gboolean isRegularFile(GNode *currentNode)
{
    struct NodeInfo	*node_info = NULL;

	if(currentNode)
	{
		node_info = (struct NodeInfo*)currentNode->data;

		if(node_info && node_info->pStat && S_ISREG(node_info->pStat->st_mode))
		{
			return TRUE;
		}
	}
	
	return FALSE;
}

gboolean isSocket(GNode *currentNode)
{
    struct NodeInfo	*node_info = NULL;

	if(currentNode)
	{
		node_info = (struct NodeInfo*)currentNode->data;

		if(node_info && node_info->pStat && S_ISSOCK(node_info->pStat->st_mode))
		{
			return TRUE;
		}
	}
	
	return FALSE;
}

