/*	
**	
**	Copyright (C) 1999 Geert Bevin
**	gbevin@thunderstorms.org
**	http://thunderstorms.org/fm
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation; either version 2 of the License, or
**	(at your option) any later version.
**	
**	This program is distributed in the hope that it will be useful,
**	but WITHOUT ANY WARRANTY; without even the implied warranty of
**	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**	GNU General Public License for more details.
**	
**	You should have received a copy of the GNU General Public License
**	along with this program; if not, write to the Free Software
**	
**	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**		
*/

#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/stat.h>

#include <gtk/gtk.h>

#include "fmwindow.h"
#include "fmwindow_signals.h"
#include "layout.h"
#include "preferences.h"

static char *pgsPanWindowCenterXpm[] =
{
	"18 18 2 1",
	" 	c None",
	".	c #000000",
	"                  ",
	"                  ",
	"                  ",
	"        ..        ",
	"       ....       ",
	"      ......      ",
	"     .  ..  .     ",
	"    ..  ..  ..    ",
	"   ............   ",
	"   ............   ",
	"    ..  ..  ..    ",
	"     .  ..  .     ",
	"      ......      ",
	"       ....       ",
	"        ..        ",
	"                  ",
	"                  ",
	"                  "
};

static char *pgsPanWindowNXpm[] =
{
	"18 18 2 1",
	" 	c None",
	".	c #000000",
	"                  ",
	"        ..        ",
	"       ....       ",
	"      ......      ",
	"        ..        ",
	"        ..        ",
	"     .  ..  .     ",
	"    ..  ..  ..    ",
	"   ............   ",
	"   ............   ",
	"    ..  ..  ..    ",
	"     .  ..  .     ",
	"      ......      ",
	"       ....       ",
	"        ..        ",
	"                  ",
	"                  ",
	"                  "
};

static char *pgsPanWindowNEXpm[] =
{
	"18 18 2 1",
	" 	c None",
	".	c #000000",
	"                  ",
	"        ..        ",
	"       ....       ",
	"      ......      ",
	"        ..        ",
	"        ..        ",
	"     .  ..    .   ",
	"    ..  ..    ..  ",
	"   .............. ",
	"   .............. ",
	"    ..  ..    ..  ",
	"     .  ..    .   ",
	"      ......      ",
	"       ....       ",
	"        ..        ",
	"                  ",
	"                  ",
	"                  "
};

static char *pgsPanWindowEXpm[] =
{
	"18 18 2 1",
	" 	c None",
	".	c #000000",
	"                  ",
	"                  ",
	"                  ",
	"        ..        ",
	"       ....       ",
	"      ......      ",
	"     .  ..    .   ",
	"    ..  ..    ..  ",
	"   .............. ",
	"   .............. ",
	"    ..  ..    ..  ",
	"     .  ..    .   ",
	"      ......      ",
	"       ....       ",
	"        ..        ",
	"                  ",
	"                  ",
	"                  "
};

static char *pgsPanWindowSEXpm[] =
{
	"18 18 2 1",
	" 	c None",
	".	c #000000",
	"                  ",
	"                  ",
	"                  ",
	"        ..        ",
	"       ....       ",
	"      ......      ",
	"     .  ..    .   ",
	"    ..  ..    ..  ",
	"   .............. ",
	"   .............. ",
	"    ..  ..    ..  ",
	"     .  ..    .   ",
	"        ..        ",
	"        ..        ",
	"      ......      ",
	"       ....       ",
	"        ..        ",
	"                  "
};

static char *pgsPanWindowSXpm[] =
{
	"18 18 2 1",
	" 	c None",
	".	c #000000",
	"                  ",
	"                  ",
	"                  ",
	"        ..        ",
	"       ....       ",
	"      ......      ",
	"     .  ..  .     ",
	"    ..  ..  ..    ",
	"   ............   ",
	"   ............   ",
	"    ..  ..  ..    ",
	"     .  ..  .     ",
	"        ..        ",
	"        ..        ",
	"      ......      ",
	"       ....       ",
	"        ..        ",
	"                  "
};

static char *pgsPanWindowSWXpm[] =
{
	"18 18 2 1",
	" 	c None",
	".	c #000000",
	"                  ",
	"                  ",
	"                  ",
	"        ..        ",
	"       ....       ",
	"      ......      ",
	"   .    ..  .     ",
	"  ..    ..  ..    ",
	" ..............   ",
	" ..............   ",
	"  ..    ..  ..    ",
	"   .    ..  .     ",
	"        ..        ",
	"        ..        ",
	"      ......      ",
	"       ....       ",
	"        ..        ",
	"                  "
};

static char *pgsPanWindowWXpm[] =
{
	"18 18 2 1",
	" 	c None",
	".	c #000000",
	"                  ",
	"                  ",
	"                  ",
	"        ..        ",
	"       ....       ",
	"      ......      ",
	"   .    ..  .     ",
	"  ..    ..  ..    ",
	" ..............   ",
	" ..............   ",
	"  ..    ..  ..    ",
	"   .    ..  .     ",
	"      ......      ",
	"       ....       ",
	"        ..        ",
	"                  ",
	"                  ",
	"                  "
};

static char *pgsPanWindowNWXpm[] =
{
	"18 18 2 1",
	" 	c None",
	".	c #000000",
	"                  ",
	"        ..        ",
	"       ....       ",
	"      ......      ",
	"        ..        ",
	"        ..        ",
	"   .    ..  .     ",
	"  ..    ..  ..    ",
	" ..............   ",
	" ..............   ",
	"  ..    ..  ..    ",
	"   .    ..  .     ",
	"      ......      ",
	"       ....       ",
	"        ..        ",
	"                  ",
	"                  ",
	"                  "
};

gchar				*pgConfigPath = NULL;
GList				*pgFmWindows = NULL;
struct Preferences	*pgPreferences = NULL;
gchar				*pgFullProcessPath = NULL;
GdkPixmap 			*pgPixmapPanWindowCenter = NULL;
GdkBitmap 			*pgMaskPanWindowCenter = NULL;
GdkPixmap 			*pgPixmapPanWindowN = NULL;
GdkBitmap 			*pgMaskPanWindowN = NULL;
GdkPixmap 			*pgPixmapPanWindowNE = NULL;
GdkBitmap 			*pgMaskPanWindowNE = NULL;
GdkPixmap 			*pgPixmapPanWindowE = NULL;
GdkBitmap 			*pgMaskPanWindowE = NULL;
GdkPixmap 			*pgPixmapPanWindowSE = NULL;
GdkBitmap 			*pgMaskPanWindowSE = NULL;
GdkPixmap 			*pgPixmapPanWindowS = NULL;
GdkBitmap 			*pgMaskPanWindowS = NULL;
GdkPixmap 			*pgPixmapPanWindowSW = NULL;
GdkBitmap 			*pgMaskPanWindowSW = NULL;
GdkPixmap 			*pgPixmapPanWindowW = NULL;
GdkBitmap 			*pgMaskPanWindowW = NULL;
GdkPixmap 			*pgPixmapPanWindowNW = NULL;
GdkBitmap 			*pgMaskPanWindowNW = NULL;
struct BrowserState	gBrowserState;
int					gFifoFileDescriptor = 0;

typedef enum
{
	IPC_OPENNEWWINDOW
} IPC_ACTIONS;

void processIPCActions()
{
	int 			ipc_action = -1;
	gchar			ipc_argument_char = ' ';
	GString			*ipc_argument = g_string_new("");
	struct FmWindow	*new_fmwindow = NULL;

	while(read(gFifoFileDescriptor, &ipc_action, sizeof(ipc_action)))
	{
//		printf("processIPCActions %d\n", ipc_action);
		switch(ipc_action)
		{
			case IPC_OPENNEWWINDOW:
				new_fmwindow = fmWindow_new(NULL);

				if(ipc_argument)
				{
                    ipc_argument = g_string_truncate(ipc_argument, 0);
					while(read(gFifoFileDescriptor, &ipc_argument_char, sizeof(ipc_argument_char)) && ipc_argument_char != '\n')
					{
						ipc_argument = g_string_append_c(ipc_argument, ipc_argument_char);
					}
					if(ipc_argument->len)
					{
						new_fmwindow->pInitialFocus = g_strdup(ipc_argument->str);
					}
                    buildNewFmWindow(new_fmwindow);
				}
				break;
		}
		ipc_action = -1;
	}
	
	g_string_free(ipc_argument, 1);
}

void signalHandler(int signal)
{
	switch(signal)
	{
		case SIGUSR1:
			processIPCActions();
			break;
	}
}

int main(int argc, char *argv[])
{
	gchar			*home_dir = NULL;
	gchar			*fifo_name = NULL;
	gchar			*pid_name = NULL;
	int				fifo_file_descriptor = -1;
	FILE			*pid_file_stream = NULL;
	int				pid = 0;
	guint			counter = 0;
	struct FmWindow	*new_fmwindow = NULL;
	gchar			linefeed_char = '\n';

	if(!g_path_is_absolute(argv[0]))
	{
		pgFullProcessPath = g_strjoin("/", g_get_current_dir(), g_dirname(argv[0]), NULL);
	}
	else
	{
		pgFullProcessPath = g_strdup(g_dirname(argv[0]));
	}

	gtk_init(&argc, &argv);

	home_dir = g_get_home_dir();
	pgConfigPath = g_strjoin("/", home_dir,".fm", NULL);
    fifo_name = g_strjoin("/", pgConfigPath, "fm_fifo", NULL);
    pid_name = g_strjoin("/", pgConfigPath, "fm_pid", NULL);

	if(pgConfigPath)
	{
//		printf("Config path : %s\n", pgConfigPath);
		mkdir(pgConfigPath, 0755);
		if(fifo_name && pid_name)
		{
			// check if there's not an instance of fm running already
//			printf("Fifo filename : %s\n", fifo_name);
			fifo_file_descriptor = open(fifo_name, O_WRONLY | O_NONBLOCK);
			if(fifo_file_descriptor == - 1)
			{
//				printf("Fm is not yet running\n");

				// check if the fifo is accessible
				if(access(fifo_name, F_OK) != -1)
				{
					// delete it if it exists
					unlink(fifo_name);
				}

				// make a new one if it doesn't exist
				mkfifo(fifo_name, 0755);

				// open the fifo file descriptor for reading
				fifo_file_descriptor = open(fifo_name, O_RDONLY | O_NONBLOCK);
				if(fifo_file_descriptor != -1)
				{
                    gFifoFileDescriptor = fifo_file_descriptor;

					// write the pid to the pid file
//					printf("Pid %d filename : %s\n", getpid(), pid_name);
					unlink(pid_name);
					
					pid_file_stream = fopen(pid_name, "w");
					if(pid_file_stream)
					{
						// writing the pid to the pid_file_stream
						fprintf(pid_file_stream, "%d", getpid());
						fclose(pid_file_stream);

						// connect the signal handler
						signal(SIGUSR1, signalHandler);

						// get the preferences and the browser layout
						pgPreferences = preferences_new();
						if(pgPreferences)
						{
							// initialize the default preferences values and try to load the user defined values that will eventually override the default values
							preferences_initialize(pgPreferences);
                            preferences_load(pgPreferences);

							// initialize the pixmaps
							pgPixmapPanWindowCenter = gdk_pixmap_colormap_create_from_xpm_d(NULL, gdk_colormap_get_system(), &(pgMaskPanWindowCenter), NULL, pgsPanWindowCenterXpm);
							pgPixmapPanWindowN = gdk_pixmap_colormap_create_from_xpm_d(NULL, gdk_colormap_get_system(), &(pgMaskPanWindowN), NULL, pgsPanWindowNXpm);
							pgPixmapPanWindowNE = gdk_pixmap_colormap_create_from_xpm_d(NULL, gdk_colormap_get_system(), &(pgMaskPanWindowNE), NULL, pgsPanWindowNEXpm);
							pgPixmapPanWindowE = gdk_pixmap_colormap_create_from_xpm_d(NULL, gdk_colormap_get_system(), &(pgMaskPanWindowE), NULL, pgsPanWindowEXpm);
							pgPixmapPanWindowSE = gdk_pixmap_colormap_create_from_xpm_d(NULL, gdk_colormap_get_system(), &(pgMaskPanWindowSE), NULL, pgsPanWindowSEXpm);
							pgPixmapPanWindowS = gdk_pixmap_colormap_create_from_xpm_d(NULL, gdk_colormap_get_system(), &(pgMaskPanWindowS), NULL, pgsPanWindowSXpm);
							pgPixmapPanWindowSW = gdk_pixmap_colormap_create_from_xpm_d(NULL, gdk_colormap_get_system(), &(pgMaskPanWindowSW), NULL, pgsPanWindowSWXpm);
							pgPixmapPanWindowW = gdk_pixmap_colormap_create_from_xpm_d(NULL, gdk_colormap_get_system(), &(pgMaskPanWindowW), NULL, pgsPanWindowWXpm);
							pgPixmapPanWindowNW = gdk_pixmap_colormap_create_from_xpm_d(NULL, gdk_colormap_get_system(), &(pgMaskPanWindowNW), NULL, pgsPanWindowNWXpm);

							// initialize the layout
							layout_load();

							// initialize the common browser state structure
							gBrowserState.pExpandedDirs = NULL;
							gBrowserState.pFocusedFile = NULL;
							gBrowserState.pZoomDirectory = NULL;

							// create the first non layouted fm window
							if(argc > 1)
							{
								for(counter = 1; counter < argc; counter++)
								{
                                    new_fmwindow = fmWindow_new(NULL);
									new_fmwindow->pInitialFocus = g_strdup(argv[counter]);
									buildNewFmWindow(new_fmwindow);
								}
							}
							else if(!g_list_length(pgFmWindows))
							{
								new_fmwindow = buildNewFmWindow(fmWindow_new(NULL));
							}

							gtk_main();

							// save, free and clean the window layout and the preferences
							if(g_list_length(pgFmWindows) && pgPreferences->sessionSaveLayoutOnExit)
							{
								layout_save(pgPreferences);
							}
							
							// close all the fm windows
							closeAllFmWindows();
						
							if(pgPreferences->sessionSavePreferencesOnExit)
							{
								preferences_save(pgPreferences);
							}
							preferences_destroy(pgPreferences);
						}
						else
						{
							printf("Can't create the preferences structure !\n");
						}

						unlink(pid_name);
					}
					else
					{
						printf("Can't open the pid file '%s' for writing.\n", pid_name);
					}

					close(fifo_file_descriptor);
					unlink(fifo_name);
				}
				else
				{
					printf("Can't open the fifo filedescriptor '%s' for reading\n", fifo_name);
				}
			}
			else
			{
//				printf("Fm is already running, requesting to open new window\n");
				
				// putting open window message in the fifo
				if(argc > 1)
				{
					for(counter = 1; counter < argc; counter++)
					{
						write(fifo_file_descriptor, IPC_OPENNEWWINDOW, sizeof(IPC_OPENNEWWINDOW));
						printf("writing '%s'\n", argv[counter]);
						write(fifo_file_descriptor, argv[counter], sizeof(gchar)*strlen(argv[counter]));
						write(fifo_file_descriptor, &linefeed_char, sizeof(gchar));
					}
				}
				else
				{
					write(fifo_file_descriptor, IPC_OPENNEWWINDOW, sizeof(IPC_OPENNEWWINDOW));
					write(fifo_file_descriptor, &linefeed_char, sizeof(gchar));
				}
				close(fifo_file_descriptor);

				// getting the pid to send a signal
				pid_file_stream = fopen(pid_name, "r");
				if(pid_file_stream)
				{
					fscanf(pid_file_stream, "%d", &pid);
					fclose(pid_file_stream);

					if(pid)
					{
//						printf("Sending signal to %d\n", pid);
						kill(pid, SIGUSR1);
					}
					else
					{
						printf("Can't get the pid from the original fm process.\n");
					}
				}
				else
				{
					printf("Can't open the pid file '%s' for reading.\n", pid_name);
				}
			}
		}
		if(pid_name)
		{
			g_free(pid_name);
		}
		if(fifo_name)
		{
			g_free(fifo_name);
		}
		g_free(pgConfigPath);
		g_free(home_dir);
	}

	if(pgFullProcessPath)
	{
		g_free(pgFullProcessPath);
	}
	
	return 0;
}

