/*	
**	
**	Copyright (C) 1999 Geert Bevin
**	gbevin@thunderstorms.org
**	http://thunderstorms.org/fm
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation; either version 2 of the License, or
**	(at your option) any later version.
**	
**	This program is distributed in the hope that it will be useful,
**	but WITHOUT ANY WARRANTY; without even the implied warranty of
**	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**	GNU General Public License for more details.
**	
**	You should have received a copy of the GNU General Public License
**	along with this program; if not, write to the Free Software
**	
**	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**		
*/

#include <stdarg.h>
#include <gtk/gtk.h>

int fmDialog_delete(GtkWidget *widget, GdkEvent *event, gpointer data) 
{
  GtkSignalFunc  cancel_callback;
  GtkWidget     *cancel_widget;

  cancel_callback = (GtkSignalFunc)gtk_object_get_data(GTK_OBJECT(widget), "fmDialog_cancelCallback");
  cancel_widget = (GtkWidget*)gtk_object_get_data(GTK_OBJECT(widget), "fmDialog_cancelWidget");

  /* the cancel callback has to destroy the dialog */
  if(cancel_callback)
  {
	  (* cancel_callback)(cancel_widget, data);
  }

  return TRUE;
}


GtkWidget* fmDialog_new(const gchar *title, const gchar *wmclassName, GtkWindowPosition position, gint allowShrink, gint allowGrow, gint autoShrink, ...)
{
	GtkWidget     *dialog = NULL;
	GtkWidget     *button_box = NULL;
	GtkWidget     *button = NULL;
	
	/*  action area variables  */
	gchar         *label = NULL;
	GtkSignalFunc  callback;
	gpointer       data = NULL;
	gboolean       default_action = FALSE;
	gboolean       connect_delete = FALSE;
	
	va_list        args;
	gboolean       delete_connected = FALSE;
	
	if(title == NULL || wmclassName == NULL)
	{
		return NULL;
	}
	else
	{
		dialog = gtk_dialog_new();
		gtk_window_set_wmclass(GTK_WINDOW(dialog), wmclassName, "Fm");
		gtk_window_set_title(GTK_WINDOW(dialog), title);
		gtk_window_set_position(GTK_WINDOW (dialog), position);
		gtk_window_set_policy(GTK_WINDOW(dialog), allowShrink, allowGrow, autoShrink);
		
		/*  prepare the action_area  */
		gtk_container_set_border_width(GTK_CONTAINER(GTK_DIALOG(dialog)->action_area), 2);
		gtk_box_set_homogeneous(GTK_BOX (GTK_DIALOG (dialog)->action_area), FALSE);
		
		button_box = gtk_hbutton_box_new();
		gtk_button_box_set_spacing(GTK_BUTTON_BOX (button_box), 4);
		gtk_box_pack_end(GTK_BOX(GTK_DIALOG (dialog)->action_area), button_box, FALSE, FALSE, 0);
		gtk_widget_show(button_box);
		
		/*  the action_area buttons  */
		va_start(args, autoShrink);
		label = va_arg(args, gchar*);
		while(label)
		{
			callback = va_arg(args, GtkSignalFunc);
			data = va_arg(args, gpointer);
			default_action = va_arg(args, gboolean);
			connect_delete = va_arg(args, gboolean);
			
			button = gtk_button_new_with_label (label);
			GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
			gtk_box_pack_start(GTK_BOX (button_box), button, FALSE, FALSE, 0);
			
			/*  pass data as user_data if data != NULL, or the dialog otherwise  */
			if(callback)
			{
				gtk_signal_connect (GTK_OBJECT (button), "clicked", GTK_SIGNAL_FUNC(callback), data ? data : dialog);
			}
			
			if(connect_delete && callback && !delete_connected)
			{
				gtk_object_set_data(GTK_OBJECT(dialog), "fmDialog_cancelCallback", callback);
				gtk_object_set_data(GTK_OBJECT(dialog), "fmDialog_cancelWidget", button);
			
				/*  catch the WM delete event  */
				gtk_signal_connect (GTK_OBJECT (dialog), "delete_event", (GdkEventFunc)fmDialog_delete, data ? data : dialog);
				
				delete_connected = TRUE;
			}
		
			if(default_action)
			{
				gtk_widget_grab_default(button);
			}
			gtk_widget_show(button);
			
			label = va_arg(args, gchar*);
		}
		va_end (args);
		
		/*  catch the WM delete event if not already done  */
		if(!delete_connected)
		{
			gtk_signal_connect(GTK_OBJECT(dialog), "delete_event", (GdkEventFunc)fmDialog_delete, NULL);
		}
		
		return dialog;
	}
}


