/*	
**	
**	Copyright (C) 1999 Geert Bevin
**	gbevin@thunderstorms.org
**	http://thunderstorms.org/fm
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation; either version 2 of the License, or
**	(at your option) any later version.
**	
**	This program is distributed in the hope that it will be useful,
**	but WITHOUT ANY WARRANTY; without even the implied warranty of
**	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**	GNU General Public License for more details.
**	
**	You should have received a copy of the GNU General Public License
**	along with this program; if not, write to the Free Software
**	
**	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**		
*/

#ifndef __FM_WINDOW_H__
#define __FM_WINDOW_H__

#include <dirent.h>
#include <glib.h>
#include <gtk/gtk.h>
#include <stdio.h>

#define	CELL_HORIZ_PADDING	5
#define	CTREE_INDENT		16
#define	CTREE_SPACING		16

#define PAN_LEFT_EDGE		4
#define PAN_RIGHT_EDGE		14
#define PAN_TOP_EDGE		4
#define PAN_BOTTOM_EDGE		14

#define AUTOSCROLL_BORDER		10
#define AUTOSCROLL_TIME			100
#define AUTOSCROLL_START_TIME	1000

// defined in gtkclist.c
#define	CELL_SPACING		1
#define COLUMN_INSET		3

struct NodeInfo
{
	gchar 			*pChangeTimeString;
	GtkCTreeNode	*pCtreeNode;
	GtkCTreeNode	*pDirEmptyCtreeNode;
	GtkCTreeNode	*pDirFirstCtreeNode;
	gchar 			*pFullName;
	gchar 			*pGroupString;
	gchar			*pLinkRepresentation;
	gchar			*pLinkTarget;
	GNode			*pLinkToNode;
	GNode			*pLinkFromNode;
	gchar 			*pModificationTimeString;
	gchar 			*pPermissionsString;
	gchar 			*pShortName;
	gchar 			*pSizeString;
	struct stat		*pStat;
	gchar 			*pUserString;
	gboolean		dirHasChildren;
	gboolean		dirIsAccessible;
};

struct ButtonBookmark
{
	GtkWidget	*pButtonBookmark;
	GtkWidget	*pPopupMenu;
	GtkWidget	*pPopupMenuItemSet;
	GtkWidget	*pPopupMenuItemGoTo;
	GtkWidget	*pPopupMenuItemClear;
};

typedef enum
{
	SELECTION_TARGET_STRING,
	SELECTION_TARGET_TEXT,
	SELECTION_TARGET_COMPOUND_TEXT
} SelectionTargetTypes;

typedef enum
{
	DRAGDROP_TARGET_FM_SELECTION,
	DRAGDROP_TARGET_STRING,
	DRAGDROP_TARGET_TEXT,
	DRAGDROP_TARGET_URL,
    DRAGDROP_TARGET_NETSCAPE_URL
} DragDropTargetTypes;

typedef enum
{
	SORT_BOOKMARK,
	SORT_FILENAME,
	SORT_SIZE,
	SORT_USER,
	SORT_GROUP,
	SORT_PERMISSIONS,
	SORT_MODIFICATION_TIME,
	SORT_CHANGE_TIME
} SortType;

typedef enum
{
	SORT_ASCENDING,
	SORT_DESCENDING
} SortDirection;

struct FmWindow
{
	gint					build;
	gint					focusedColumn;
	gint					freed;
	gboolean				hideDotFiles;
	gint					initialX;
	gint					initialY;
	gint					initialWidth;
	gint					initialHeight;
	gchar					*pInitialFocus;
	SortDirection			sortDirection;
	SortType				sortType;

	GList		*pComboBoxStrings;
	GList		*pCTreeColumns;
	GList		*pExpandedDirs;
	GNode		*pFocusedNode;
	GNode 		*pRootNode;
	GList		*pSelection;
	gchar		*pSelectionTextSnapshot;
	GString		*pShortcut;
	GArray		*pVisitedLinks;
	gchar		*pZoomDirectory;

	GtkAccelGroup	*pAccelGroup;
	GtkObject		*pAdjustmentHScrollbar;
	GtkObject		*pAdjustmentVScrollbar;
	GtkWidget		*pAlignmentSelectionSize;
	GtkWidget		*pButtonHideDotfiles;
	GtkWidget		*pComboBox;
	GtkWidget		*pComboBoxPopupMenu;
	GtkWidget		*pCTree;
	GtkWidget		*pEventBoxPanWindow;
	GtkWidget		*pFrameGeneralStatus;
	GtkWidget		*pFrameSelectionSize;
	GtkWidget		*pHBoxBottomButtons;
	GtkWidget		*pHBoxBottomStatus;
	GtkWidget		*pHBoxScrolledWindow;
	GtkWidget		*pHBoxTop;
	GtkWidget		*pHScrollbarScrolledWindow;
	GtkWidget		*pLabelSelectionSize;
	GtkWidget		*pPixmapPanWindow;
	GtkWidget		*pPopupMenu;
	GtkWidget		*pPopupMenuItemSelection;
	GtkWidget		*pPopupMenuSelection;
	GtkWidget		*pPopupMenuSelectionMenuItemClear;
	GtkWidget		*pPopupMenuSelectionMenuItemSelectDirectory;
	GtkWidget		*pPopupMenuSelectionMenuItemDeselectDirectory;
	GtkWidget		*pPopupMenuSelectionMenuItemCopyToXClipboard;
	GtkWidget		*pPopupMenuSelectionMenuItemCopyFocusToComboEntry;
	GtkWidget		*pPopupMenuItemDirectory;
	GtkWidget		*pPopupMenuDirectory;
	GtkWidget		*pPopupMenuDirectoryMenuItemRefresh;
	GtkWidget		*pPopupMenuDirectoryMenuItemZoom;
	GtkWidget		*pPopupMenuDirectoryMenuItemOpenInNewWindow;
	GtkWidget		*pPopupMenuItemBrowser;
	GtkWidget		*pPopupMenuBrowser;
	GtkWidget		*pPopupMenuBrowserMenuItemRefresh;
	GtkWidget		*pPopupMenuBrowserMenuItemNew;
	GtkWidget		*pPopupMenuBrowserMenuItemClose;
	GtkWidget		*pPopupMenuBrowserMenuItemCollapseEverything;
	GtkWidget		*pPopupMenuBrowserMenuItemUnZoom;
	GtkWidget		*pPopupMenuBrowserMenuItemCopyState;
	GtkWidget		*pPopupMenuBrowserMenuItemPasteState;
	GtkWidget		*pPopupMenuItemColumns;
	GtkWidget		*pPopupMenuColumns;
	GtkWidget		*pPopupMenuColumnsMenuItemSize;
	GtkWidget		*pPopupMenuColumnsMenuItemUser;
	GtkWidget		*pPopupMenuColumnsMenuItemGroup;
	GtkWidget		*pPopupMenuColumnsMenuItemPermissions;
	GtkWidget		*pPopupMenuColumnsMenuItemModificationTime;
	GtkWidget		*pPopupMenuColumnsMenuItemChangeTime;
	GtkWidget		*pPopupMenuItemSeperator1;
	GtkWidget		*pPopupMenuItemAbout;
	GtkWidget		*pPopupMenuItemPreferences;
	GtkWidget		*pPopupMenuItemSeperator2;
	GtkWidget		*pPopupMenuItemQuit;
	
	GtkWidget		*pLabelGeneralStatus;
	GtkTooltips		*pToolTips;
	GtkWidget		*pVBox;
	GtkWidget		*pVBoxScrolledWindowLeft;
	GtkWidget		*pVBoxScrolledWindowRight;
	GtkWidget		*pVScrollbarScrolledWindow;
	GtkWidget		*pWindow;

	GHashTable				*pAcceleratedWidgets;
	struct ButtonBookmark	pBookmarkButtons[10];
	gchar					**pBookmarks;
};

GtkWidget*				fmWindow_acceleratedWidgetFind(struct FmWindow*, gchar*);
void					fmWindow_acceleratedWidgetForget(struct FmWindow*, GtkWidget*);
void					fmWindow_acceleratedWidgetRemember(struct FmWindow*, GtkWidget*);
void					fmWindow_accelKeyAddToGtkWidget(struct FmWindow*, GtkWidget*);
void					fmWindow_accelKeyRemoveFromGtkWidget(struct FmWindow*, GtkWidget*);
void					fmWindow_bookmarkClear(struct FmWindow*, guint);
void					fmWindow_bookmarkClearIndication(struct FmWindow*, GNode*);
void					fmWindow_bookmarkDrawIndication(struct FmWindow*, GNode*, guint);
gint					fmWindow_bookmarkGetIndex(struct FmWindow*, GNode*);
void					fmWindow_bookmarkGoTo(struct FmWindow*, guint);
void					fmWindow_bookmarkSet(struct FmWindow*, gchar*, guint);
void					fmWindow_bookmarkUpdatePopupMenu(struct FmWindow*, guint);
GtkWidget*				fmWindow_buildCtree(struct FmWindow*);
void					fmWindow_buildDir(struct FmWindow*, GNode*, const char*);
void					fmWindow_buildFile(struct FmWindow*, GNode*, const char*);
void					fmWindow_buildLink(struct FmWindow*, GNode*, const char*);
GNode*					fmWindow_buildNode(struct FmWindow*, const char*);
gboolean				fmWindow_collapseNode(struct FmWindow*, GNode*);
void					fmWindow_createBookmarkButton(struct FmWindow*, guint);
void					fmWindow_ctreeColumnsApplyAlignment(struct FmWindow*);
gint					fmWindow_ctreeHeight(struct FmWindow*);
void					fmWindow_ctreeMoveHorizontal(struct FmWindow*, gint);
void					fmWindow_ctreeMoveVertical(struct FmWindow*, gint);
void					fmWindow_ctreeScrollHorizontal(struct FmWindow*, gint);
gint					fmWindow_ctreeScrollHorizontalTimeout(struct FmWindow*);
gboolean				fmWindow_ctreeScrollHorizontalValid(struct FmWindow*, gint);
void					fmWindow_ctreeScrollVertical(struct FmWindow*, gint);
gint					fmWindow_ctreeScrollVerticalTimeout(struct FmWindow*);
gboolean				fmWindow_ctreeScrollVerticalValid(struct FmWindow*, gint);
gint					fmWindow_ctreeWidth(struct FmWindow*);
void					fmWindow_destroy(struct FmWindow*);
gboolean				fmWindow_expandNode(struct FmWindow*, GNode*);
gboolean				fmWindow_fillDir(struct FmWindow*, GNode*);
void					fmWindow_focusNode(struct FmWindow*, GNode*);
void					fmWindow_focusCTreeRow(struct FmWindow*, guint);
gboolean				fmWindow_focusDeepestExpandedNode(struct FmWindow*, GNode*);
void					fmWindow_focusDrawInfo(struct FmWindow*, guint);
gboolean				fmWindow_focusFileName(struct FmWindow*, gchar*);
gboolean				fmWindow_focusFirstChildNode(struct FmWindow*, GNode*);
gboolean				fmWindow_focusFirstExpandedNode(struct FmWindow*, GNode*);
void					fmWindow_focusFirstColumn(struct FmWindow *fmWindow);
void					fmWindow_focusFirstRow(struct FmWindow *fmWindow);
void					fmWindow_focusLastColumn(struct FmWindow *fmWindow);
void					fmWindow_focusLastRow(struct FmWindow *fmWindow);
void					fmWindow_focusNextColumn(struct FmWindow*);
gboolean				fmWindow_focusNextNode(struct FmWindow*, GNode*);
gboolean				fmWindow_focusNextDirectoryNode(struct FmWindow*, GNode*);
gboolean				fmWindow_focusPageDownNode(struct FmWindow*, GNode*);
gboolean				fmWindow_focusPageUpNode(struct FmWindow*, GNode*);
gboolean				fmWindow_focusParentNode(struct FmWindow*, GNode*);
void					fmWindow_focusPreviousColumn(struct FmWindow*);
gboolean				fmWindow_focusPreviousNode(struct FmWindow*, GNode*);
gboolean				fmWindow_focusPreviousDirectoryNode(struct FmWindow*, GNode*);
gboolean				fmWindow_focusRootNode(struct FmWindow*, GNode*);
void					fmWindow_freeComboBoxPopupMenu(struct FmWindow*);
void					fmWindow_freeDir(struct FmWindow*, GNode*);
void					fmWindow_freeDirEntries(struct FmWindow*, GNode*);
void					fmWindow_freeNode(struct FmWindow*, GNode*);
void					fmWindow_freeFile(struct FmWindow*, GNode*);
void					fmWindow_freeLink(struct FmWindow*, GNode*);
gchar*					fmWindow_getFilenameToRemember(struct FmWindow*, GNode*);
GNode*					fmWindow_getNodeFromFilename(struct FmWindow*, gchar*, gboolean);
gchar*					fmWindow_getSelectionSize(struct FmWindow*);
void					fmWindow_initialize(struct FmWindow*);
void					fmWindow_insertCTreeNode(struct FmWindow*, GNode*, GtkCTreeNode*, GtkCTreeNode*, gchar*);
gboolean				fmWindow_insertCTreeDir(struct FmWindow*, GNode*, GtkCTreeNode*, GtkCTreeNode*, gchar*);
gboolean				fmWindow_insertCTreeFile(struct FmWindow*, GNode*, GtkCTreeNode*, GtkCTreeNode*, gchar*);
gboolean				fmWindow_insertCTreeLink(struct FmWindow*, GNode*, GtkCTreeNode*, GtkCTreeNode*, gchar*);
gboolean				fmWindow_isNodeParentOfNode(struct FmWindow*, GNode*, GNode*);
void					fmWindow_listCTreeDirChildren(struct FmWindow*, GtkCTreeNode*);
struct FmWindow*		fmWindow_new(struct FmWindow*);
void					fmWindow_moveToCTreeRow(struct FmWindow*, guint);
void					fmWindow_moveToNode(struct FmWindow*, GNode*);
void					fmWindow_openDirectoryInNewWindow(struct FmWindow*, GNode*);
gint 					fmWindow_panTimeout(struct FmWindow*);
void					fmWindow_popupMenu(struct FmWindow*, GdkEventButton*);
void					fmWindow_rebuildCTree(struct FmWindow*);
void					fmWindow_refreshCTree(struct FmWindow*);
void					fmWindow_refreshDirectory(struct FmWindow*, GNode*, gchar*);
void					fmWindow_refreshParentDirectory(struct FmWindow*, GNode*);
void					fmWindow_removeCTreeDirChildren(struct FmWindow*, GtkCTreeNode*);
void					fmWindow_resizeColumns(struct FmWindow*);
void					fmWindow_selectionAddNode(struct FmWindow*, GNode*);
void					fmWindow_selectionAddAllDirEntries(struct FmWindow*, GNode*);
void					fmWindow_selectionDestroy(struct FmWindow*);
GList*					fmWindow_selectionGetListNode(struct FmWindow*, GNode*);
void					fmWindow_selectionRemoveAllDirEntries(struct FmWindow*, GNode*);
void					fmWindow_selectionRemoveNode(struct FmWindow*, GNode*);
gchar*					fmWindow_selectionTextSnapshotString(struct FmWindow*, GList*);
gchar*					fmWindow_selectionTextSnapshotURL(struct FmWindow*, GList*);
void					fmWindow_selectionToggleNode(struct FmWindow*, GNode*);
void					fmWindow_selectionUnselectAll(struct FmWindow*);
void					fmWindow_selectionUpdateSizeLabel(struct FmWindow*);
void					fmWindow_selectNode(struct FmWindow*, GNode*);
void					fmWindow_selectCTreeNode(struct FmWindow*, GtkCTreeNode*);
void					fmWindow_selectCTreeRow(struct FmWindow*, guint);
void					fmWindow_selectMoveToCTreeFocus(struct FmWindow*);
void					fmWindow_shortcutClear(struct FmWindow*);
gboolean				fmWindow_shortcutCompare(struct FmWindow*, GNode*);
GNode*					fmWindow_shortcutFindFirstNode(struct FmWindow*, GNode*);
GNode*					fmWindow_shortcutFindFollowingNode(struct FmWindow*, GNode*);
GNode*					fmWindow_shortcutFindPreviousNode(struct FmWindow*, GNode*);
gboolean				fmWindow_toggleNode(struct FmWindow*, GNode*);
void					fmWindow_unselectNode(struct FmWindow*, GNode*);
void					fmWindow_unselectCTreeRow(struct FmWindow*, guint);
void					fmWindow_updateCTreeTitles(struct FmWindow*);
gboolean				fmWindow_useDirEntry(struct FmWindow*, struct dirent*);
GNode*					fmWindow_validateActionOnNode(struct FmWindow*, GNode*);
	
struct FmWindow*		buildNewFmWindow(struct FmWindow*);
void					closeFmWindow(struct FmWindow*);
void					closeAllFmWindows();
GNode*					findGNodeShortName(GNode*, gchar*);
void					rebuildAllFmWindowsCTrees();
gint					sortNodesByChangeTime(GNode*, GNode*);
gint					sortNodesByChangeTimeAscending(gconstpointer, gconstpointer);
gint					sortNodesByChangeTimeDescending(gconstpointer, gconstpointer);
gint					sortNodesByFilename(GNode*, GNode*);
gint					sortNodesByFilenameAscending(gconstpointer, gconstpointer);
gint					sortNodesByFilenameDescending(gconstpointer, gconstpointer);
gint					sortNodesByGroup(GNode*, GNode*);
gint					sortNodesByGroupAscending(gconstpointer, gconstpointer);
gint					sortNodesByGroupDescending(gconstpointer, gconstpointer);
gint					sortNodesByModificationTime(GNode*, GNode*);
gint					sortNodesByModificationTimeAscending(gconstpointer, gconstpointer);
gint					sortNodesByModificationTimeDescending(gconstpointer, gconstpointer);
gint					sortNodesByPermissions(GNode*, GNode*);
gint					sortNodesByPermissionsAscending(gconstpointer, gconstpointer);
gint					sortNodesByPermissionsDescending(gconstpointer, gconstpointer);
gint					sortNodesBySize(GNode*, GNode*);
gint					sortNodesBySizeAscending(gconstpointer, gconstpointer);
gint					sortNodesBySizeDescending(gconstpointer, gconstpointer);
gint					sortNodesByUser(GNode*, GNode*);
gint					sortNodesByUserAscending(gconstpointer, gconstpointer);
gint					sortNodesByUserDescending(gconstpointer, gconstpointer);

#endif /* __FM_WINDOW_H__ */

