/*	
**	
**	Copyright (C) 1999 Geert Bevin
**	gbevin@thunderstorms.org
**	http://thunderstorms.org/fm
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation; either version 2 of the License, or
**	(at your option) any later version.
**	
**	This program is distributed in the hope that it will be useful,
**	but WITHOUT ANY WARRANTY; without even the implied warranty of
**	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**	GNU General Public License for more details.
**	
**	You should have received a copy of the GNU General Public License
**	along with this program; if not, write to the Free Software
**	
**	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**		
*/

#include <stdio.h>
#include <string.h>
#include <sys/stat.h>

#include <glib.h>
#include <gdk/gdkkeysyms.h>
#include <gdk/gdktypes.h>
#include <gtk/gtk.h>

#include "aboutwindow.h"
#include "accelkey.h"
#include "ctreecolumn.h"
#include "filetype.h"
#include "fileutils.h"
#include "fmwindow.h"
#include "layout.h"
#include "preferences.h"
#include "preferenceswindow.h"
#include "rowstyle.h"
#include "fmwindow_signals.h"

extern GList				*pgFmWindows;
extern struct Preferences	*pgPreferences;
extern struct BrowserState	gBrowserState;
extern GtkTargetEntry		gComboEntryDragDropTargets[];
extern guint				gComboEntryNumberOfDragDropTargets;
extern GtkTargetEntry		gCTreeDragDropTargets[];
extern guint				gCTreeNumberOfDragDropTargets;
extern GdkPixmap 			*pgPixmapPanWindowCenter;
extern GdkBitmap 			*pgMaskPanWindowCenter;

static struct FmWindow	*pgsDragFocusedFmWindow = NULL;

void fmWindow_acceleratorAdd(GtkWidget *widget, guint accel_signal_id, GtkAccelGroup *accel_group, guint accel_key, GdkModifierType accel_mods, GtkAccelFlags accel_flags, gpointer userData)
{
	gchar			*accel_key_id = (gchar*)gtk_object_get_data(GTK_OBJECT(widget), "accelKeyId");

//	printf("fmWindow_acceleratorAdd()\n");
	accelKeyNew(pgPreferences->pAccelKeys, accel_key_id, accel_key, accel_mods);
	accelKeyUpdateInAllFmWindows(pgPreferences->pAccelKeys, pgFmWindows, accel_key_id);
}

void fmWindow_acceleratorRemove(GtkWidget *widget, guint accel_signal_id, GtkAccelGroup *accel_group, guint accel_key, GdkModifierType accel_mods, gpointer userData)
{
	gchar			*accel_key_id = (gchar*)gtk_object_get_data(GTK_OBJECT(widget), "accelKeyId");

//	printf("fmWindow_acceleratorRemove()\n");
	accelKeyRemove(pgPreferences->pAccelKeys, accel_key_id);
	accelKeyUpdateInAllFmWindows(pgPreferences->pAccelKeys, pgFmWindows, accel_key_id);
}

gboolean fmWindow_buttonBookmarkButtonPress(GtkWidget *bookmarkButton, GdkEventButton *event, gpointer userData)
{
	struct FmWindow	*fm_window = userData;
	guint 			index = (guint)gtk_object_get_data(GTK_OBJECT(bookmarkButton), "index");
	
	if(event->button == 1)
	{
		if(GTK_BIN(bookmarkButton)->child->state == GTK_STATE_INSENSITIVE)
		{
			// stop the left mouseclick if the label is disactivated
			gtk_signal_emit_stop_by_name(GTK_OBJECT(bookmarkButton), "button_press_event");
			return TRUE;
		}
	}
	else
	{
		// popup the menu
		gtk_menu_popup(GTK_MENU(fm_window->pBookmarkButtons[index].pPopupMenu), NULL, NULL, NULL, NULL, event->button, event->time);

		// stop the right mousebutton click signal
		gtk_signal_emit_stop_by_name(GTK_OBJECT(bookmarkButton), "button_press_event");
		return TRUE;
	}

	return FALSE;
}

void fmWindow_buttonBookmarkClicked(GtkButton *bookmarkButton, gpointer userData)
{
    struct FmWindow	*fm_window = userData;
	guint 			index = (guint)gtk_object_get_data(GTK_OBJECT(bookmarkButton), "index");
	
	fmWindow_bookmarkGoTo(fm_window, index);
}

gboolean fmWindow_buttonBookmarkEnterNotify(GtkWidget *bookmarkButton, GdkEventCrossing *event, gpointer userData)
{
	if(GTK_BIN(bookmarkButton)->child->state == GTK_STATE_INSENSITIVE)
	{
		gtk_signal_emit_stop_by_name(GTK_OBJECT(bookmarkButton), "enter_notify_event");

		return TRUE;		
	}
	
	return FALSE;
}

gboolean fmWindow_buttonBookmarkLeaveNotify(GtkWidget *bookmarkButton, GdkEventCrossing *event, gpointer userData)
{
	if(GTK_BIN(bookmarkButton)->child->state == GTK_STATE_INSENSITIVE)
	{
		gtk_signal_emit_stop_by_name(GTK_OBJECT(bookmarkButton), "leave_notify_event");

		return TRUE;		
	}
	
	return FALSE;
}

void fmWindow_buttonBookmarkClear(GtkMenuItem *menuItem, gpointer userData)
{
	gint			number_of_fmwindows = g_list_length(pgFmWindows);
	gint			counter = 0;
	struct FmWindow	*temp_fm_window = NULL;
	guint 			index = (guint)gtk_object_get_data(GTK_OBJECT(gtk_object_get_data(GTK_OBJECT(menuItem), "bookmarkButton")), "index");

	for(counter = 0; counter < number_of_fmwindows; counter++)
	{
		temp_fm_window = g_list_nth_data(pgFmWindows, counter);
		fmWindow_bookmarkClear(temp_fm_window, index);
	}
}

void fmWindow_buttonBookmarkGoTo(GtkMenuItem *menuItem, gpointer userData)
{
    struct FmWindow	*fm_window = userData;
	guint 			index = (guint)gtk_object_get_data(GTK_OBJECT(gtk_object_get_data(GTK_OBJECT(menuItem), "bookmarkButton")), "index");

	fmWindow_bookmarkGoTo(fm_window, index);
}

void fmWindow_buttonBookmarkSet(GtkMenuItem *menuItem, gpointer userData)
{
	struct FmWindow	*fm_window = (struct FmWindow *)userData;
	GNode			*source_gnode = fm_window->pFocusedNode;
	gint			number_of_fmwindows = g_list_length(pgFmWindows);
	gint			counter = 0;
	struct FmWindow	*temp_fm_window = NULL;
	guint 			index = (guint)gtk_object_get_data(GTK_OBJECT(gtk_object_get_data(GTK_OBJECT(menuItem), "bookmarkButton")), "index");
	gchar			*bookmark_name = NULL;

	if((source_gnode = fmWindow_validateActionOnNode(fm_window, source_gnode)))
	{
		bookmark_name = fmWindow_getFilenameToRemember(fm_window, fm_window->pFocusedNode);
		if(bookmark_name)
		{
			for(counter = 0; counter < number_of_fmwindows; counter++)
			{
				temp_fm_window = g_list_nth_data(pgFmWindows, counter);
				fmWindow_bookmarkSet(temp_fm_window, bookmark_name, index);
			}
			g_free(bookmark_name);
		}
	}
}

void fmWindow_buttonHideDotFilesToggled(GtkToggleButton *togglebutton, gpointer userData)
{
    struct FmWindow	*fm_window = (struct FmWindow *)userData;

	if(fm_window->hideDotFiles)
	{
        fm_window->hideDotFiles = FALSE;
	}
	else
	{
        fm_window->hideDotFiles = TRUE;
	}
	
	if(fm_window->pInitialFocus)
	{
		g_free(fm_window->pInitialFocus);
	}
	fm_window->pInitialFocus = fmWindow_getFilenameToRemember(fm_window, fm_window->pFocusedNode);
	
	if(fm_window->pFocusedNode)
	{
		fm_window->pFocusedNode = NULL;
	}
	fmWindow_rebuildCTree(fm_window);
}

gboolean fmWindow_comboKeyPress(GtkWidget *widget, GdkEventKey *event, gpointer userData)
{
    struct FmWindow	*fm_window = userData;
	gchar			*combo_entry_text = NULL;
	gchar			*combo_entry_text_basename = NULL;
	gchar			*combo_entry_text_dirname = NULL;
	DIR 			*dir = NULL;
	struct dirent	*dir_entry = NULL;
	gchar			*dir_entry_name = NULL;
	GList  			*directory_names = NULL;
	GCompletion		*base_completion = NULL;
	gchar			*temp_filename = NULL;
	GNode			*temp_source_gnode = NULL;
	GList			*matching_items = NULL;
	GList			*matching_items_sorted = NULL;
	gchar			*matching_prefix = NULL;
	GtkWidget		*menu_item = NULL;
	GtkWidget		*menu_item_label = NULL;
	gchar			*new_combo_entry_text = NULL;
	GList			*temp_glist = NULL;

	switch(event->keyval)
	{
		case GDK_Tab:
			// make the general combo box tab completion work
			if(!(event->state & GDK_MOD1_MASK))
			{
				combo_entry_text = gtk_entry_get_text(GTK_ENTRY(GTK_COMBO(fm_window->pComboBox)->entry));
				// only do string completion when the control key is not pressed and when there is text in the combo box and when the directory is an absolute pathname and doesn't end with a slash
				if(!(event->state & GDK_CONTROL_MASK) && strcmp(combo_entry_text, "") != 0 && g_path_is_absolute(combo_entry_text) && combo_entry_text[strlen(combo_entry_text)-1] != '/')
				{
					base_completion = g_completion_new(NULL);
					if(base_completion)
					{
						// get the directory and the basename parts
						combo_entry_text_dirname = g_dirname(combo_entry_text);
						combo_entry_text_basename = g_basename(combo_entry_text);

						// read the directory contents and put them in a GList
						dir = opendir(combo_entry_text_dirname);
						if(dir)
						{
							while((dir_entry = readdir(dir)))
							{
								temp_filename = g_strconcat(combo_entry_text_dirname, "/", dir_entry->d_name, NULL);
								if(temp_filename)
								{
									temp_source_gnode = fmWindow_buildNode(fm_window, temp_filename);
									if(isDirectory(temp_source_gnode))
									{
										dir_entry_name = g_strconcat(dir_entry->d_name, "/", NULL);
									}
									else
									{
										dir_entry_name = g_strdup(dir_entry->d_name);
									}
									directory_names = g_list_prepend(directory_names, dir_entry_name);
									fmWindow_freeNode(fm_window, temp_source_gnode);
									g_free(temp_filename);
								}
							}

							closedir(dir);
						}

						// perform the string completion
						if(g_list_length(directory_names))
						{
							g_completion_add_items(base_completion, directory_names);
							matching_items = g_completion_complete(base_completion, combo_entry_text_basename, &matching_prefix);

							// check if items were matching
							if(matching_items)
							{
								// check if there is a common prefix between the matching items
								if(matching_prefix)
								{
									// check if the match is exactly the same as the final part of the initial path
									if(g_list_length(matching_items) == 1 && strcmp(combo_entry_text_basename, matching_prefix) == 0)
									{
										// if there was only one matching item, let the TAB cycle to the ctree
										gtk_widget_grab_focus(GTK_WIDGET(fm_window->pCTree));
									}
									else
									{
		//								printf("matching_prefix = %s\n", matching_prefix);
										if(combo_entry_text_dirname[strlen(combo_entry_text_dirname)-1] == '/')
										{
											combo_entry_text_dirname[strlen(combo_entry_text_dirname)-1] = 0;
										}
										new_combo_entry_text = g_strconcat(combo_entry_text_dirname, "/", matching_prefix, NULL);
										gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(fm_window->pComboBox)->entry), new_combo_entry_text);
										g_free(new_combo_entry_text);

										// if there was more than one match, display the results in a popup menu
										if(g_list_length(matching_items) > 1)
										{
											// copy the matching filenames list and sort it
											matching_items_sorted = g_list_copy(matching_items);
											matching_items_sorted = g_list_sort(matching_items_sorted, (GCompareFunc)strcmp);

											// create the popup menu of all the matching filenames
											fmWindow_freeComboBoxPopupMenu(fm_window);
											gtk_object_set_data(GTK_OBJECT(fm_window->pComboBoxPopupMenu), "path", (gpointer)g_strdup(combo_entry_text_dirname));

											// free the list of sorted filenames and attach the popup menu items
											while(matching_items_sorted && g_list_length(matching_items_sorted))
											{
												temp_glist = g_list_first(matching_items_sorted);
												matching_items_sorted = g_list_remove_link(matching_items_sorted, temp_glist);

												menu_item = gtk_menu_item_new();
												menu_item_label = gtk_label_new((gchar*)temp_glist->data);
												gtk_misc_set_alignment(GTK_MISC(menu_item_label), 0.0, 0.5);
												gtk_container_add(GTK_CONTAINER(menu_item), GTK_WIDGET(menu_item_label));
												gtk_object_set_data(GTK_OBJECT(menu_item), "path", (gpointer)g_strconcat(combo_entry_text_dirname, "/", (gchar*)temp_glist->data, NULL));
												gtk_signal_connect(GTK_OBJECT(menu_item), "activate", GTK_SIGNAL_FUNC(fmWindow_comboPopupMenuItemSelect), fm_window);
												gtk_widget_show(GTK_WIDGET(menu_item_label));
												gtk_widget_show(GTK_WIDGET(menu_item));
												gtk_menu_append(GTK_MENU(fm_window->pComboBoxPopupMenu), GTK_WIDGET(menu_item));

												g_list_free_1(temp_glist);
											}
											gtk_menu_shell_select_item(GTK_MENU_SHELL(fm_window->pComboBoxPopupMenu), GTK_WIDGET(g_list_first(GTK_MENU_SHELL(fm_window->pComboBoxPopupMenu)->children)->data));
											gtk_menu_popup(GTK_MENU(fm_window->pComboBoxPopupMenu), NULL, NULL, (GtkMenuPositionFunc)fmWindow_comboPositionPopupMenu, fm_window, 0, event->time);
										}
									}
									g_free(matching_prefix);
								}
							}
							else
							{
								// no, emit a beep
								gdk_beep();
							}
						}
						else
						{
							// no, emit a beep
							gdk_beep();
						}

						// free the completion data
						g_completion_free(base_completion);

						// free the list of filenames
						while(directory_names && g_list_length(directory_names))
						{
							temp_glist = g_list_first(directory_names);
							directory_names = g_list_remove_link(directory_names, temp_glist);
							g_free(temp_glist->data);
							g_list_free_1(temp_glist);
						}

						g_free(combo_entry_text_dirname);
					}
				}
				else
				{
					// let the TAB cycle to the ctree
					gtk_widget_grab_focus(GTK_WIDGET(fm_window->pCTree));
				}
				break;
			}
		default:
			return FALSE;
			break;
	}
	gtk_signal_emit_stop_by_name(GTK_OBJECT(widget), "key_press_event");
	return TRUE;
}

void fmWindow_comboPopupMenuItemSelect(GtkMenuItem *menuItem, gpointer userData)
{
    struct FmWindow	*fm_window = userData;
	gchar			*dir_name = (gchar*)gtk_object_get_data(GTK_OBJECT(menuItem), "path");

	gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(fm_window->pComboBox)->entry), dir_name);
}

void fmWindow_comboPositionPopupMenu(GtkMenu *menu, gint *x, gint *y, gpointer userData)
{
    struct FmWindow	*fm_window = userData;
	gchar			*dir_name = (gchar*)gtk_object_get_data(GTK_OBJECT(menu), "path");
	GtkStyle		*style = gtk_widget_get_style(GTK_WIDGET(GTK_COMBO(fm_window->pComboBox)->entry));
	guint			text_area_x = 0;
	guint			text_area_y = 0;

	gdk_window_get_position(fm_window->pWindow->window, x, y);
	*x += GTK_COMBO(fm_window->pComboBox)->entry->allocation.x;
	*y += GTK_COMBO(fm_window->pComboBox)->entry->allocation.y;
	gdk_window_get_position(GTK_ENTRY(GTK_COMBO(fm_window->pComboBox)->entry)->text_area, &text_area_x, &text_area_y);
	*x += text_area_x;
	*y += text_area_y;
	if(style && style->font && dir_name)
	{
		*x += gdk_string_width(style->font, dir_name);
	}
}

void fmWindow_comboTextEntryActivate(GtkEditable *editable, gpointer userData)
{
    struct FmWindow	*fm_window = userData;
	gchar			*string = NULL;
	GList			*last_entry = NULL;
	
	string = g_strdup(gtk_entry_get_text(GTK_ENTRY(editable)));
	fm_window->pComboBoxStrings = g_list_prepend(fm_window->pComboBoxStrings, string);
	if(g_list_length(fm_window->pComboBoxStrings) > pgPreferences->comboPopdownLength)
	{
        last_entry = g_list_last(fm_window->pComboBoxStrings);
		fm_window->pComboBoxStrings = g_list_remove_link(fm_window->pComboBoxStrings, last_entry);
		g_free(last_entry->data);
		g_list_free_1(last_entry);
	}
	gtk_combo_set_popdown_strings(GTK_COMBO(fm_window->pComboBox), fm_window->pComboBoxStrings);
    gtk_entry_set_text(GTK_ENTRY(editable), string);
    fmWindow_focusFileName(fm_window, string);
	fmWindow_selectMoveToCTreeFocus(fm_window);
}

void fmWindow_comboTextEntryDragDataReceived(GtkWidget *widget, GdkDragContext *dragContext, gint x, gint y, GtkSelectionData *selectionData, guint info, guint time, gpointer userData)
{
    struct FmWindow	*fm_window = userData;
	gchar			*drop_data = NULL;
	gchar			*selection_target = gdk_atom_name(selectionData->target);
	guint			counter = 0;

//	printf("data=0x%x length=%d selection=%s target=%s type=%s format=%d\n", (guint)selectionData->data, selectionData->length, gdk_atom_name(selectionData->selection), gdk_atom_name(selectionData->target), gdk_atom_name(selectionData->type), selectionData->format);
	if((selectionData->length >= 0) && (selectionData->format == 8))
	{
		if(strcmp(selection_target, "_NETSCAPE_URL") == 0 || strcmp(selection_target, "text/uri-list") == 0)
		{
			// remove the eventual leading 'file:'
			if(strncmp(selectionData->data, "file:", 5) == 0)
			{
				drop_data = g_strdup(selectionData->data+5);

				// remove the eventual URL anchor
				for(counter = strlen(drop_data)-1; counter >= 0; counter--)
				{
					if(drop_data[counter] == '#')
					{
						drop_data[counter] = 0;
						break;
					}
				}
			}
			else
			{
				drop_data = g_strdup(selectionData->data);
			}
		}
		else
		{
			drop_data = g_strdup(selectionData->data);
		}
		gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(fm_window->pComboBox)->entry), drop_data);
		gtk_widget_grab_focus(GTK_WIDGET(GTK_COMBO(fm_window->pComboBox)->entry));
		gtk_entry_set_position(GTK_ENTRY(GTK_COMBO(fm_window->pComboBox)->entry), strlen(drop_data));
		
		g_free(drop_data);
	}
}

void fmWindow_comboTextEntryDragDrop(GtkWidget *widget, GdkDragContext *dragContext, gint x, gint y, guint time, gpointer userData)
{
	GList	*temp_glist = NULL;
	gint	counter = 0;
	gchar	*orig_target = NULL;
	gchar	*context_target = NULL;

	for(counter = 0; counter < gComboEntryNumberOfDragDropTargets; counter++)
	{
        orig_target = gComboEntryDragDropTargets[counter].target;
		temp_glist = dragContext->targets;
		while(temp_glist)
		{
            context_target = gdk_atom_name((GdkAtom)temp_glist->data);
//			printf("%d %s %s\n", counter, orig_target, context_target);
			if(strcmp(orig_target, context_target) == 0)
			{
				gtk_drag_get_data(widget, dragContext, (GdkAtom)temp_glist->data, time);
				gtk_drag_finish(dragContext, TRUE, FALSE, time);
				return;
			}
			temp_glist = temp_glist->next;
		}
	}
	gtk_drag_finish(dragContext, FALSE, FALSE, time);
}

void fmWindow_comboTextEntryDragMotion(GtkWidget *widget, GdkDragContext *dragContext, gint x, gint y, guint time, gpointer userData)
{
	GList	*temp_glist = NULL;

	printf("DRAG MOTION, TARGETS = { ");
	temp_glist = dragContext->targets;
	while(temp_glist)
	{
		printf("%s ", gdk_atom_name((GdkAtom)temp_glist->data));
		temp_glist = temp_glist->next;
	}
	printf(" }\n");
}

gboolean fmWindow_ctreeButtonPress(GtkWidget *widget, GdkEventButton *event, gpointer userData)
{
    struct FmWindow	*fm_window = userData;
	guint 			row = 0;
	guint 			column = 0;
	gint			click_in_clist = 0;
	GtkCTreeNode	*ctree_node = NULL;
	GtkCTreeRow		*ctree_row = NULL;
	gint			row_depth = 0;
	gint			row_pixmap_width = 0;
	gint			row_pixmap_height = 0;
	gint			hotspot_left = 0;
	gint			hotspot_right = 0;
	gint			focus_row = 0;
	GNode			*source_gnode = NULL;
	GNode			*target_gnode = NULL;
	GtkCTreeNode	*previous_ctree_node = NULL;

	gtk_signal_emit_stop_by_name(GTK_OBJECT(widget), "button_press_event");
	
	click_in_clist = gtk_clist_get_selection_info(GTK_CLIST(fm_window->pCTree), event->x, event->y, &row, &column);
//	printf("selection x = %f y = %f row = %d column = %d\n", event->x, event->y, row, column);

	if(click_in_clist)
	{
		ctree_node = gtk_ctree_node_nth(GTK_CTREE(fm_window->pCTree), row);
		if(ctree_node)
		{
			source_gnode = (GNode*)gtk_ctree_node_get_row_data(GTK_CTREE(fm_window->pCTree), GTK_CTREE_NODE(ctree_node));
			if(source_gnode && (source_gnode = fmWindow_validateActionOnNode(fm_window, source_gnode)))
			{
				// redefine them since the validation function might have changed the tree structure
				row = gtk_clist_find_row_from_data(GTK_CLIST(fm_window->pCTree), source_gnode);
				ctree_node = gtk_ctree_node_nth(GTK_CTREE(fm_window->pCTree), row);
				
				target_gnode = getLinkTarget(source_gnode);
				if(event->button == 1)
				{
					if(event->type == GDK_BUTTON_PRESS || event->type == GDK_2BUTTON_PRESS)
					{
						// needed to be able to generate the motion_notify_event
						gtk_grab_add(widget);
						gdk_pointer_grab(GTK_CTREE(fm_window->pCTree)->clist.clist_window, FALSE, GDK_POINTER_MOTION_MASK | GDK_POINTER_MOTION_HINT_MASK | GDK_BUTTON_RELEASE_MASK, NULL, NULL, event->time);
						gtk_object_set_data(GTK_OBJECT(widget), "drag_is_possible", (gpointer)1);

						if(isDirectoryFilled(target_gnode))
						{
							ctree_row = GTK_CTREE_ROW(((struct NodeInfo*)source_gnode->data)->pCtreeNode);
							if(ctree_row)
							{
								// calculate to place of the expander pixmap pixmap
								gdk_window_get_size((GdkWindow*)ctree_row->pixmap_closed, &row_pixmap_width, &row_pixmap_height);
								row_depth = 0;
								while(ctree_row->parent)
								{
									ctree_row = GTK_CTREE_ROW(ctree_row->parent);
									row_depth++;
								}
								hotspot_left = GTK_CLIST(fm_window->pCTree)->column[1].area.x + GTK_CTREE(fm_window->pCTree)->tree_spacing+(GTK_CTREE(fm_window->pCTree)->tree_indent*row_depth) - GTK_CTREE(fm_window->pCTree)->clist.hadjustment->value;
								hotspot_right = hotspot_left+row_pixmap_width;
//								printf("button pressed x=%f y=%f hotspot_left=%d hotspot_right=%d depth=%d row=%d column=%d\n", event->x, event->y, hotspot_left, hotspot_right,row_depth, row, column);
								// check if the mouseclick was in the hotspot
								if(event->x >= hotspot_left && event->x <= hotspot_right)
								{
									// yes is was, toggle the expansion of the node
									gtk_object_remove_data(GTK_OBJECT(widget), "drag_is_possible");
									fmWindow_toggleNode(fm_window, target_gnode);
									return TRUE;
								}
								else if(event->type == GDK_2BUTTON_PRESS)
								{
									// no, but was if it was a doubleclick, expand / collapse node
									gtk_object_remove_data(GTK_OBJECT(widget), "drag_is_possible");
									fmWindow_toggleNode(fm_window, target_gnode);
									return TRUE;
								}
							}
						}
						if(event->state & GDK_CONTROL_MASK)
						{
							// control key was pressed in the meantime, toggle node in selection
							fmWindow_selectionToggleNode(fm_window, source_gnode);
						}
						else if(event->state & GDK_SHIFT_MASK)
						{
							// shift key was pressed in the meantime, toggle the range of nodes in the selection
                            focus_row = gtk_clist_find_row_from_data(GTK_CLIST(fm_window->pCTree), fm_window->pFocusedNode);
//							printf("focus_row = %d\n", focus_row);
                            previous_ctree_node = gtk_ctree_node_nth(GTK_CTREE(fm_window->pCTree), focus_row);
//							printf("previous_ctree_node = 0x%x\n", (guint)previous_ctree_node);
							if(row < focus_row)
							{
								// shift click above the focus
								while(previous_ctree_node != ctree_node)
								{
									fmWindow_selectionToggleNode(fm_window, (GNode*)gtk_ctree_node_get_row_data(GTK_CTREE(fm_window->pCTree), previous_ctree_node));
									previous_ctree_node = GTK_CTREE_NODE_PREV(previous_ctree_node);
								}
							}
							else
							{
								// shift click under the focus
								while(previous_ctree_node != ctree_node)
								{
									fmWindow_selectionToggleNode(fm_window, (GNode*)gtk_ctree_node_get_row_data(GTK_CTREE(fm_window->pCTree), previous_ctree_node));
									previous_ctree_node = GTK_CTREE_NODE_NEXT(previous_ctree_node);
								}
							}
						}
						fmWindow_focusCTreeRow(fm_window, row);
						fmWindow_selectCTreeRow(fm_window, row);
						fmWindow_moveToCTreeRow(fm_window, row);
					}
				}
				else if(event->button == 2)
				{
					// open new window
					fmWindow_focusCTreeRow(fm_window, row);
					fmWindow_selectCTreeRow(fm_window, row);
					fmWindow_moveToCTreeRow(fm_window, row);
					fmWindow_openDirectoryInNewWindow(fm_window, source_gnode);
				}
				else if(event->button == 3)
				{
					// popup the menu
					fmWindow_focusCTreeRow(fm_window, row);
					fmWindow_selectCTreeRow(fm_window, row);
					fmWindow_moveToCTreeRow(fm_window, row);
					fmWindow_popupMenu(fm_window, event);
				}
			}
		}
	}
	else if(event->button != 1)
	{
		fmWindow_popupMenu(fm_window, event);
	}
	
	return TRUE;
}

gboolean fmWindow_ctreeButtonRelease(GtkWidget *widget, GdkEventButton *event, gpointer userData)
{
//	printf("button released\n");
	gtk_signal_emit_stop_by_name(GTK_OBJECT(widget), "button_release_event");

	// clear the data that indicates a started drag
	if(gtk_object_get_data(GTK_OBJECT(widget), "drag_has_started"))
	{
		gtk_object_remove_data(GTK_OBJECT(widget), "drag_has_started");
		gtk_object_remove_data(GTK_OBJECT(widget), "drag_is_possible");
	}

	// needed to be able to generate the motion_notify_event
	gtk_grab_remove(widget);
	gdk_pointer_ungrab(event->time);
	
	return TRUE;
}

void fmWindow_ctreeClickColumn(GtkCList *clist, gint column, gpointer userData)
{
    struct FmWindow	*fm_window = userData;

	if(column == COLUMN_BOOKMARK)
	{
		fmWindow_refreshCTree(fm_window);
	}
	else
	{
		if(((struct CtreeColumn*)g_list_nth_data(pgPreferences->pCtreeColumns, column))->type == COLUMN_FILENAME)
		{
			if(fm_window->sortType != SORT_FILENAME)
			{
				fm_window->sortType = SORT_FILENAME;
				fm_window->sortDirection = SORT_ASCENDING;
			}
			else if(fm_window->sortDirection == SORT_ASCENDING)
			{
				fm_window->sortDirection = SORT_DESCENDING;
			}
			else
			{
				fm_window->sortDirection = SORT_ASCENDING;
			}
		}
		else if(((struct CtreeColumn*)g_list_nth_data(pgPreferences->pCtreeColumns, column))->type == COLUMN_SIZE)
		{
			if(fm_window->sortType != SORT_SIZE)
			{
				fm_window->sortType = SORT_SIZE;
				fm_window->sortDirection = SORT_ASCENDING;
			}
			else if(fm_window->sortDirection == SORT_ASCENDING)
			{
				fm_window->sortDirection = SORT_DESCENDING;
			}
			else
			{
				fm_window->sortDirection = SORT_ASCENDING;
			}
		}
		else if(((struct CtreeColumn*)g_list_nth_data(pgPreferences->pCtreeColumns, column))->type == COLUMN_USER)
		{
			if(fm_window->sortType != SORT_USER)
			{
				fm_window->sortType = SORT_USER;
				fm_window->sortDirection = SORT_ASCENDING;
			}
			else if(fm_window->sortDirection == SORT_ASCENDING)
			{
				fm_window->sortDirection = SORT_DESCENDING;
			}
			else
			{
				fm_window->sortDirection = SORT_ASCENDING;
			}
		}
		else if(((struct CtreeColumn*)g_list_nth_data(pgPreferences->pCtreeColumns, column))->type == COLUMN_GROUP)
		{
			if(fm_window->sortType != SORT_GROUP)
			{
				fm_window->sortType = SORT_GROUP;
				fm_window->sortDirection = SORT_ASCENDING;
			}
			else if(fm_window->sortDirection == SORT_ASCENDING)
			{
				fm_window->sortDirection = SORT_DESCENDING;
			}
			else
			{
				fm_window->sortDirection = SORT_ASCENDING;
			}
		}
		else if(((struct CtreeColumn*)g_list_nth_data(pgPreferences->pCtreeColumns, column))->type == COLUMN_PERMISSIONS)
		{
			if(fm_window->sortType != SORT_PERMISSIONS)
			{
				fm_window->sortType = SORT_PERMISSIONS;
				fm_window->sortDirection = SORT_ASCENDING;
			}
			else if(fm_window->sortDirection == SORT_ASCENDING)
			{
				fm_window->sortDirection = SORT_DESCENDING;
			}
			else
			{
				fm_window->sortDirection = SORT_ASCENDING;
			}
		}
		else if(((struct CtreeColumn*)g_list_nth_data(pgPreferences->pCtreeColumns, column))->type == COLUMN_MODIFICATION_TIME)
		{
			if(fm_window->sortType != SORT_MODIFICATION_TIME)
			{
				fm_window->sortType = SORT_MODIFICATION_TIME;
				fm_window->sortDirection = SORT_ASCENDING;
			}
			else if(fm_window->sortDirection == SORT_ASCENDING)
			{
				fm_window->sortDirection = SORT_DESCENDING;
			}
			else
			{
				fm_window->sortDirection = SORT_ASCENDING;
			}
		}
		else if(((struct CtreeColumn*)g_list_nth_data(pgPreferences->pCtreeColumns, column))->type == COLUMN_CHANGE_TIME)
		{
			if(fm_window->sortType != SORT_CHANGE_TIME)
			{
				fm_window->sortType = SORT_CHANGE_TIME;
				fm_window->sortDirection = SORT_ASCENDING;
			}
			else if(fm_window->sortDirection == SORT_ASCENDING)
			{
				fm_window->sortDirection = SORT_DESCENDING;
			}
			else
			{
				fm_window->sortDirection = SORT_ASCENDING;
			}
		}
		fmWindow_updateCTreeTitles(fm_window);
		fmWindow_refreshCTree(fm_window);
	}
}

void fmWindow_ctreeCollapse(GtkCTree *ctree, GtkCTreeNode *ctreeNode, gpointer data)
{
    struct FmWindow	*fm_window = data;
	GNode			*source_gnode = NULL;
	GNode			*target_gnode = NULL;
	gboolean		is_node_parent_of_focused_node = FALSE;
	GList			*found_glist = NULL;
	gchar			*name_to_remember = NULL;
	
	source_gnode = (GNode*)gtk_ctree_node_get_row_data(GTK_CTREE(ctree), GTK_CTREE_NODE(ctreeNode));
	target_gnode = getLinkTarget(source_gnode);
	is_node_parent_of_focused_node = fmWindow_isNodeParentOfNode(fm_window, source_gnode, fm_window->pFocusedNode);
	
	if(isDirectory(target_gnode))
	{
		gtk_clist_freeze(GTK_CLIST(fm_window->pCTree));
		// empty the directory node
        fmWindow_removeCTreeDirChildren(fm_window, ctreeNode);
		fmWindow_freeDirEntries(fm_window, target_gnode);
		// remove the directory from the list of expanded ones
		name_to_remember = fmWindow_getFilenameToRemember(fm_window, source_gnode);
		if(name_to_remember)
		{
			do
			{
				found_glist = g_list_find_custom(fm_window->pExpandedDirs, name_to_remember, (GCompareFunc)strcmp);
				if(found_glist)
				{
//					printf("forget %s\n", name_to_remember);
					g_free(found_glist->data);
        	        found_glist->data = NULL;
					fm_window->pExpandedDirs = g_list_remove_link(fm_window->pExpandedDirs, found_glist);
					g_list_free(found_glist);
				}
			}
			while(found_glist);
			g_free(name_to_remember);
		}
		gtk_clist_thaw(GTK_CLIST(fm_window->pCTree));
		// check if the gnode is the parent of the previously focused row, and it thus has disappeared
		if(is_node_parent_of_focused_node)
		{
			// yes, focus collapsed directory entry
			fmWindow_focusNode(fm_window, source_gnode);
			fmWindow_selectNode(fm_window, source_gnode);
		}
	}
}

void fmWindow_ctreeDragBegin(GtkWidget *widget, GdkDragContext *drag_context, gpointer userData)
{
    struct FmWindow	*fm_window = userData;
	GList			*selected_files = NULL;
	GList			*file_to_draw = NULL;
	gchar			*focused_filename = NULL;
	guint			string_width = 0;
	struct RowStyle	*row_style = NULL;
	GdkGC			*pixmap_gc = NULL;
	GdkGC			*mask_gc = NULL;
	GtkStyle		*ctree_style = NULL;
	GdkPixmap		*drag_pixmap = NULL;
	GdkPixmap		*drag_mask = NULL;
	GdkColor		mask_color;
	guint			width = 0;
	guint			height = 0;
	gchar			*too_many_filenames = NULL;
	guint 			counter = 0;
	guint			seperator_line_y = 0;

//	printf("fmWindow_ctreeDragBegin\n");

	gtk_signal_emit_stop_by_name(GTK_OBJECT(widget), "drag_begin");
    
	// get the gtk style of the ctree
	ctree_style = gtk_widget_get_style(GTK_WIDGET(fm_window->pCTree));
    
	// make a sorted GList of the filenames that should be used to construct the drag icon
	if(!g_list_length(fm_window->pSelection))
	{
        focused_filename = g_strdup(fmWindow_getFilenameToRemember(fm_window, fm_window->pFocusedNode));
		selected_files = g_list_append(selected_files, focused_filename);
		row_style = rowStyleFind(pgPreferences->pRowStyles, "Focus");
        gtk_object_set_data(GTK_OBJECT(widget), "free_selection_data", (gpointer)1);
	}
	else
	{
        selected_files = g_list_copy(fm_window->pSelection);
		selected_files = g_list_sort(selected_files, (GCompareFunc)strcmp);
		row_style = rowStyleFind(pgPreferences->pRowStyles, "Selected");
	}

	if(selected_files)
	{
		// store the drag selection data
        gtk_object_set_data(GTK_OBJECT(widget), "drag_selection_data", selected_files);
		
		// calculate the drag icon dimensions
		file_to_draw = selected_files;
		while(file_to_draw)
		{
			string_width = gdk_string_width(ctree_style->font, (gchar*)file_to_draw->data);
			if(string_width > width)
			{
				width = string_width;
			}
			file_to_draw = file_to_draw->next;
		}
		width = width + 4;
		height = g_list_length(selected_files) * (ctree_style->font->ascent + ctree_style->font->descent + 2);

		if(height > (gdk_screen_height() / 3) * 2)
		{
			height = ctree_style->font->ascent + ctree_style->font->descent + 2;
            too_many_filenames = g_new(gchar, 100);
            sprintf(too_many_filenames, "Big selection, dragging %d files", g_list_length(selected_files));
			width = gdk_string_width(ctree_style->font, too_many_filenames) + 4;
		}

		// allocate the pixmaps
		drag_pixmap = gdk_pixmap_new(fm_window->pCTree->window, width, height, -1);
        drag_mask = gdk_pixmap_new(NULL, width, height, 1);
		
		// initialize the graphical contexts
		pixmap_gc = gdk_gc_new(drag_pixmap);
		gdk_gc_set_background(pixmap_gc, &(row_style->background));
		gdk_gc_set_foreground(pixmap_gc, &(row_style->foreground));
		gdk_gc_set_function(pixmap_gc, GDK_COPY);
		gdk_gc_set_font(pixmap_gc, ctree_style->font);
		
		mask_gc = gdk_gc_new(drag_mask);
        mask_color.pixel = 1;
		gdk_gc_set_foreground(mask_gc, &mask_color);
		gdk_gc_set_function(mask_gc, GDK_COPY);

		// set the drag icon backgrounds
		gdk_gc_set_foreground(pixmap_gc, &(row_style->background));
		gdk_draw_rectangle(drag_pixmap, pixmap_gc, TRUE, 0, 0, width, height);
		gdk_gc_set_foreground(pixmap_gc, &(row_style->foreground));

		gdk_draw_rectangle(drag_mask, mask_gc, TRUE, 0, 0, width, height);
        mask_color.pixel = 0;
		gdk_gc_set_foreground(mask_gc, &mask_color);

		if(too_many_filenames)
		{
			gdk_draw_text(drag_pixmap, ctree_style->font, pixmap_gc, 2, ctree_style->font->ascent, too_many_filenames, strlen(too_many_filenames));
			g_free(too_many_filenames);
		}
		else
		{
			// draw the filenames
			file_to_draw = selected_files;
			counter = 0;
			while(file_to_draw)
			{
				gdk_draw_text(drag_pixmap, ctree_style->font, pixmap_gc, 2, ctree_style->font->ascent + counter * (ctree_style->font->ascent + ctree_style->font->descent + 2), (gchar*)file_to_draw->data, strlen((gchar*)file_to_draw->data));
				counter++;
				if(file_to_draw->next)
				{
					seperator_line_y = counter * (ctree_style->font->ascent + ctree_style->font->descent + 2);
					gdk_draw_line(drag_mask, mask_gc, 0, seperator_line_y, width, seperator_line_y);
				}
				file_to_draw = file_to_draw->next;
			}
		}
		gtk_drag_set_icon_pixmap(drag_context, gtk_widget_get_colormap(GTK_WIDGET(fm_window->pWindow)), drag_pixmap, drag_mask, -10, -10);
	}
}

void fmWindow_ctreeDragDataGet(GtkWidget *widget, GdkDragContext *dragContext, GtkSelectionData *selectionData, guint info, guint time, gpointer userData)
{
    struct FmWindow	*fm_window = userData;
    gchar			*text_snapshot = NULL;
	GList			*selection_data = (GList*)gtk_object_get_data(GTK_OBJECT(widget), "drag_selection_data");
	
//	printf("fmWindow_ctreeDragDataGet\n");

	gtk_object_remove_data(GTK_OBJECT(widget), "drag_selection_data");
	if(selection_data)
	{
		if(info == DRAGDROP_TARGET_FM_SELECTION)
		{
//			gtk_selection_data_set(selectionData, selectionData->target, 8, text_snapshot, strlen(text_snapshot));
		}
		else if(info == DRAGDROP_TARGET_STRING || info == DRAGDROP_TARGET_TEXT)
		{
			text_snapshot = fmWindow_selectionTextSnapshotString(fm_window, selection_data);

			if(text_snapshot)
			{
				gtk_selection_data_set(selectionData, selectionData->target, 8, text_snapshot, strlen(text_snapshot));
				g_free(text_snapshot);
			}
		}
		else if(info == DRAGDROP_TARGET_NETSCAPE_URL || info == DRAGDROP_TARGET_URL)
		{
			text_snapshot = fmWindow_selectionTextSnapshotURL(fm_window, selection_data);

			if(text_snapshot)
			{
				gtk_selection_data_set(selectionData, selectionData->target, 8, text_snapshot, strlen(text_snapshot));
				g_free(text_snapshot);
			}
		}
	}
}

void fmWindow_ctreeDragDataReceived(GtkWidget *widget, GdkDragContext *dragContext, gint x, gint y, GtkSelectionData *selectionData, guint info, guint time, gpointer userData)
{
//	printf("fmWindow_ctreeDragDataReceived\n");
	
	gtk_signal_emit_stop_by_name(GTK_OBJECT(widget), "drag_data_received");
//	printf("data=0x%x length=%d selection=%s target=%s type=%s format=%d\n", (guint)selectionData->data, selectionData->length, gdk_atom_name(selectionData->selection), gdk_atom_name(selectionData->target), gdk_atom_name(selectionData->type), selectionData->format);
}

void fmWindow_ctreeDragDrop(GtkWidget *widget, GdkDragContext *dragContext, gint x, gint y, guint time, gpointer userData)
{
//	printf("fmWindow_ctreeDragDrop\n");
	
	gtk_signal_emit_stop_by_name(GTK_OBJECT(widget), "drag_drop");
	gtk_drag_finish(dragContext, FALSE, FALSE, time);
}

void fmWindow_ctreeDragEnd(GtkWidget *widget, GdkDragContext *dragContext, gpointer userData)
{
	GList	*selection_data = (GList*)gtk_object_get_data(GTK_OBJECT(widget), "drag_selection_data");
	GList	*temp_glist = NULL;
	guint	timeout = 0;
	
//	printf("fmWindow_ctreeDragEnd\n");
	
	gtk_signal_emit_stop_by_name(GTK_OBJECT(widget), "drag_end");
	
	if(selection_data)
	{
		if(gtk_object_get_data(GTK_OBJECT(widget), "free_selection_data"))
		{
			while(g_list_length(selection_data))
			{
				temp_glist = g_list_first(selection_data);
				if(temp_glist)
				{
					selection_data = g_list_remove_link(selection_data, temp_glist);
					g_free(temp_glist->data);
					g_list_free_1(temp_glist);
				}
			}
		}
		else
		{
			g_list_free(selection_data);
		}
	}
	timeout = (guint)gtk_object_get_data(GTK_OBJECT(widget), "ctree_timeout");
	if(timeout)
	{
		gtk_timeout_remove(timeout);
		timeout = 0;
	}
	gtk_object_remove_data(GTK_OBJECT(widget), "ctree_timeout");
	gtk_object_remove_data(GTK_OBJECT(widget), "free_selection_data");
	gtk_object_remove_data(GTK_OBJECT(widget), "ctree_drag_focused_row");
}

void fmWindow_ctreeDragLeave(GtkWidget *widget, GdkDragContext *dragContext, guint time, gpointer userData)
{
	guint timeout = 0;

	gtk_signal_emit_stop_by_name(GTK_OBJECT(widget), "drag_leave");
	
	timeout = (guint)gtk_object_get_data(GTK_OBJECT(widget), "ctree_timeout");
	if(timeout)
	{
		gtk_timeout_remove(timeout);
		gtk_object_remove_data(GTK_OBJECT(widget), "ctree_timeout");
	}
}

void fmWindow_ctreeDragMotion(GtkWidget *widget, GdkDragContext *dragContext, gint x, gint y, guint time, gpointer userData)
{
    struct FmWindow	*fm_window = userData;
	//GList			*temp_glist = NULL;
	gint			adapted_y = y - GTK_CTREE(fm_window->pCTree)->clist.column_title_area.height;
	gint 			previous_row = 0;
	gint 			row = 0;
	gint 			column = 0;
	gint			motion_in_clist = 0;
	guint			timeout = 0;

	gtk_signal_emit_stop_by_name(GTK_OBJECT(widget), "drag_motion");
    pgsDragFocusedFmWindow = fm_window;

/*	printf("x=%d y=%d DRAG MOTION, TARGETS = { ", x, y);
	temp_glist = dragContext->targets;
	while(temp_glist)
	{
		printf("%s ", gdk_atom_name((GdkAtom)temp_glist->data));
		temp_glist = temp_glist->next;
	}
	printf(" }\n");*/
	
	if(adapted_y >= 0)
	{
		motion_in_clist = gtk_clist_get_selection_info(GTK_CLIST(fm_window->pCTree), x, adapted_y, &row, &column);
		if(motion_in_clist)
		{
			previous_row = (gint)gtk_object_get_data(GTK_OBJECT(widget), "ctree_drag_focused_row");
//			printf("previous_row=%d row=%d column=%d\n", previous_row, row, column);
			if(previous_row != row)
			{
				gtk_object_set_data(GTK_OBJECT(widget), "ctree_drag_focused_row", (gpointer)row);
				timeout = (guint)gtk_object_get_data(GTK_OBJECT(widget), "ctree_timeout");
				if(timeout)
				{
					gtk_timeout_remove(timeout);
					gtk_object_remove_data(GTK_OBJECT(widget), "ctree_timeout");
				}
				timeout = gtk_timeout_add(COLLAPSE_EXPAND_TIME, (GtkFunction)fmWindow_ctreeDragMotionTimeout, fm_window);
				gtk_object_set_data(GTK_OBJECT(widget), "ctree_timeout", (gpointer)timeout);
				fmWindow_focusCTreeRow(fm_window, row);
                fmWindow_selectCTreeRow(fm_window, row);
			}
		}
	}

	if(fmWindow_ctreeScrollHorizontalValid(fm_window, x))
	{
		gtk_timeout_add(AUTOSCROLL_START_TIME, (GtkFunction)fmWindow_ctreeScrollHorizontalTimeout, fm_window);
	}
	if(fmWindow_ctreeScrollVerticalValid(fm_window, y))
	{
		gtk_timeout_add(AUTOSCROLL_START_TIME, (GtkFunction)fmWindow_ctreeScrollVerticalTimeout, fm_window);
	}
}

gint fmWindow_ctreeDragMotionTimeout(gpointer userData)
{
    struct FmWindow	*fm_window = userData;
	gint x = 0;
	gint y = 0;

    gtk_widget_get_pointer(GTK_WIDGET(fm_window->pCTree), &x, &y);
	if(!fmWindow_ctreeScrollHorizontalValid(fm_window, x) && !fmWindow_ctreeScrollVerticalValid(fm_window, y))
	{
		fmWindow_toggleNode(fm_window, getLinkTarget(fm_window->pFocusedNode));
	}
	
	return FALSE;
}

void fmWindow_ctreeExpand(GtkCTree *ctree, GtkCTreeNode *ctreeNode, gpointer data)
{
    struct FmWindow	*fm_window = data;
	GNode			*source_gnode = NULL;
	GNode			*target_gnode = NULL;
	gchar			*name_to_remember = NULL;

	source_gnode = (GNode*)gtk_ctree_node_get_row_data(GTK_CTREE(ctree), GTK_CTREE_NODE(ctreeNode));
    target_gnode = getLinkTarget(source_gnode);
	
	if(isDirectory(target_gnode))
	{
		gtk_clist_freeze(GTK_CLIST(fm_window->pCTree));
		// add the directory to those that are expanded if it's not already in it
		name_to_remember = fmWindow_getFilenameToRemember(fm_window, source_gnode);
		if(!g_list_find_custom(fm_window->pExpandedDirs, name_to_remember, (GCompareFunc)strcmp))
		{
//			printf("remember %s\n", name_to_remember);
			fm_window->pExpandedDirs = g_list_prepend(fm_window->pExpandedDirs, name_to_remember);
		}
		// fill up the directory node
		fmWindow_fillDir(fm_window, target_gnode);
		gtk_clist_thaw(GTK_CLIST(fm_window->pCTree));
	}
}

void fmWindow_ctreeHAdjustmentChanged(GtkAdjustment *adjustment, gpointer userData)
{
    struct FmWindow	*fm_window = userData;
	
//	printf("pAdjustmentHScrollbar->page_size=%f pAdjustmentHScrollbar->upper=%f\n", GTK_ADJUSTMENT(fm_window->pAdjustmentHScrollbar)->page_size, GTK_ADJUSTMENT(fm_window->pAdjustmentHScrollbar)->upper);
	if(GTK_ADJUSTMENT(fm_window->pAdjustmentHScrollbar)->page_size == GTK_ADJUSTMENT(fm_window->pAdjustmentHScrollbar)->upper)
	{
//		printf("hiding hscrollbar\n");
		gtk_widget_hide(GTK_WIDGET(fm_window->pHScrollbarScrolledWindow));
	}
	else
	{
//		printf("showing hscrollbar\n");
		gtk_widget_show(GTK_WIDGET(fm_window->pHScrollbarScrolledWindow));
	}
}

gboolean fmWindow_ctreeKeyPress(GtkWidget *widget, GdkEventKey *event, gpointer userData)
{
    struct FmWindow	*fm_window = userData;
	GNode			*source_gnode = NULL;
	GNode			*temp_source_gnode = NULL;

	source_gnode = gtk_clist_get_row_data(GTK_CLIST(fm_window->pCTree), GTK_CLIST(fm_window->pCTree)->focus_row);

//	printf("key pressed : 0x%x\n", event->keyval);
	if(source_gnode)
	{
		switch(event->keyval)
		{
			case GDK_Left:
				if( (source_gnode = fmWindow_validateActionOnNode(fm_window, source_gnode)) )
				{
					if(event->state & GDK_MOD1_MASK && event->state & GDK_SHIFT_MASK)
					{
						fmWindow_focusFirstColumn(fm_window);
					}
					else if(event->state & GDK_MOD1_MASK)
					{
						fmWindow_focusPreviousColumn(fm_window);
					}
					else if(event->state & GDK_CONTROL_MASK && event->state & GDK_SHIFT_MASK)
					{
						fmWindow_focusRootNode(fm_window, source_gnode);
					}
					else if(event->state & GDK_CONTROL_MASK)
					{
						fmWindow_focusParentNode(fm_window, source_gnode);
					}
					else
					{
						if(!fmWindow_collapseNode(fm_window, getLinkTarget(source_gnode)))
						{
							fmWindow_focusParentNode(fm_window, source_gnode);
						}
					}
				}
				break;
			case GDK_Right:
				if( (source_gnode = fmWindow_validateActionOnNode(fm_window, source_gnode)) )
				{
					if(event->state & GDK_MOD1_MASK && event->state & GDK_SHIFT_MASK)
					{
						fmWindow_focusLastColumn(fm_window);
					}
					else if(event->state & GDK_MOD1_MASK)
					{
						fmWindow_focusNextColumn(fm_window);
					}
					else if(event->state & GDK_CONTROL_MASK && event->state & GDK_SHIFT_MASK)
					{
						fmWindow_focusDeepestExpandedNode(fm_window, getLinkTarget(source_gnode));
					}
					else if(event->state & GDK_CONTROL_MASK)
					{
						fmWindow_focusFirstExpandedNode(fm_window, getLinkTarget(source_gnode));
					}
					else
					{
						if(!fmWindow_expandNode(fm_window, getLinkTarget(source_gnode)))
						{
							fmWindow_focusFirstChildNode(fm_window, getLinkTarget(source_gnode));
						}
					}
				}
				break;
			case GDK_Up:
				if(event->state & GDK_SHIFT_MASK)
				{
					fmWindow_selectionToggleNode(fm_window, source_gnode);
				}
				if(event->state & GDK_CONTROL_MASK)
				{
					fmWindow_focusPreviousDirectoryNode(fm_window, source_gnode);
				}
				else
				{
					fmWindow_focusPreviousNode(fm_window, source_gnode);
				}
				break;
			case GDK_Down:
				if(event->state & GDK_SHIFT_MASK)
				{
					fmWindow_selectionToggleNode(fm_window, source_gnode);
				}
				if(event->state & GDK_CONTROL_MASK)
				{
					fmWindow_focusNextDirectoryNode(fm_window, source_gnode);
				}
				else
				{
					fmWindow_focusNextNode(fm_window, source_gnode);
				}
				break;
			case GDK_Home:
				if(event->state & GDK_CONTROL_MASK)
				{
					fmWindow_focusFirstRow(fm_window);
				}
				break;
			case GDK_End:
				if(event->state & GDK_CONTROL_MASK)
				{
					fmWindow_focusLastRow(fm_window);
				}
				break;
			case GDK_Page_Down:
				fmWindow_focusPageDownNode(fm_window, source_gnode);
				break;
			case GDK_Page_Up:
				fmWindow_focusPageUpNode(fm_window, source_gnode);
				break;
			case GDK_Escape:
				fmWindow_selectionUnselectAll(fm_window);
				break;
			case GDK_Return:
				if( (source_gnode = fmWindow_validateActionOnNode(fm_window, source_gnode)) )
				{
					if(fmWindow_expandNode(fm_window, getLinkTarget(source_gnode)))
					{
						fmWindow_focusFirstChildNode(fm_window, getLinkTarget(source_gnode));
					}
					else
					{
						fmWindow_collapseNode(fm_window, getLinkTarget(source_gnode));
					}
				}
				break;
			case GDK_Tab:
				gtk_widget_grab_focus(GTK_WIDGET(GTK_COMBO(fm_window->pComboBox)->entry));
				break;
			case GDK_BackSpace:
				if(fm_window->pShortcut->len)
				{
                    fm_window->pShortcut = g_string_truncate(fm_window->pShortcut, fm_window->pShortcut->len-1);
					temp_source_gnode = fmWindow_shortcutFindPreviousNode(fm_window, source_gnode);
					if(temp_source_gnode)
					{
						fmWindow_focusNode(fm_window, temp_source_gnode);
						fmWindow_selectMoveToCTreeFocus(fm_window);
					}
//					printf("fm_window->pShortcut->str = %s\n", fm_window->pShortcut->str);
				}
//				gtk_signal_emit_stop_by_name(GTK_OBJECT(widget), "key_press_event");
				// continue event propogation
				return FALSE;
				break;
			case GDK_space:
				// when the spacebar char browsing is enabled, only catch the spacebar presses with at least one modifier to select/deselect an entry
				// otherwise, plain spacebar works
				if(!pgPreferences->characterBrowsingEnableSpacebar || event->state != 0)
				{
					if( (source_gnode = fmWindow_validateActionOnNode(fm_window, source_gnode)) )
					{
						fmWindow_selectionToggleNode(fm_window, source_gnode);
					}
					break;
				}
			default:
				// when the character browsing is enabled, only use ascii keyvalues without the slash and without the keys that are pressed together with the control or alt key
				if(pgPreferences->characterBrowsingEnabled && (event->keyval <= 255 && event->keyval != '/' && !(event->state & GDK_CONTROL_MASK) && !(event->state & GDK_MOD1_MASK)))
				{
					fm_window->pShortcut = g_string_append_c(fm_window->pShortcut, event->keyval);
					temp_source_gnode = fmWindow_shortcutFindFollowingNode(fm_window, source_gnode);
					
					// focus the first node with the shortcut combination that exists
					if(temp_source_gnode)
					{
						fmWindow_focusNode(fm_window, temp_source_gnode);
						fmWindow_selectMoveToCTreeFocus(fm_window);
					}
					else
					{
						// shortcut combination doesn't exist
						fm_window->pShortcut = g_string_assign(fm_window->pShortcut, "");
						// clean the shortcut and the new char as new start
						fm_window->pShortcut = g_string_append_c(fm_window->pShortcut, event->keyval);
						// find the first following node with the new shortcut or the very first one if it can't find a further one
						temp_source_gnode = fmWindow_shortcutFindFollowingNode(fm_window, source_gnode);
						if(temp_source_gnode)
						{
							fmWindow_focusNode(fm_window, temp_source_gnode);
							fmWindow_selectMoveToCTreeFocus(fm_window);
						}
						else
						{
							// it doesn't exists, clean the shortcut and don't focus anything
							fm_window->pShortcut = g_string_assign(fm_window->pShortcut, "");
						}
					}
//					printf("fm_window->pShortcut->str = %s\n", fm_window->pShortcut->str);
//					gtk_signal_emit_stop_by_name(GTK_OBJECT(widget), "key_press_event");
//					return TRUE;
				}
				return FALSE;
				break;
		}
	}
    fmWindow_shortcutClear(fm_window);
	gtk_signal_emit_stop_by_name(GTK_OBJECT(widget), "key_press_event");
	return TRUE;
}

void fmWindow_ctreeMenuItemAbout(GtkMenuItem *menuItem, gpointer userData)
{
	aboutWindow_new();
}

void fmWindow_ctreeMenuItemBrowserClose(GtkMenuItem *menuItem, gpointer userData)
{
    struct FmWindow	*fm_window = userData;

	gtk_widget_destroy(fm_window->pWindow);
}

void fmWindow_ctreeMenuItemBrowserCollapseEverything(GtkMenuItem *menuItem, gpointer userData)
{
    struct FmWindow	*fm_window = userData;
	
    fmWindow_focusRootNode(fm_window, fm_window->pFocusedNode);
	
	g_list_free(fm_window->pExpandedDirs);
    fm_window->pExpandedDirs = NULL;

	fmWindow_refreshCTree(fm_window);
}

void fmWindow_ctreeMenuItemBrowserCopyState(GtkMenuItem *menuItem, gpointer userData)
{
    struct FmWindow	*fm_window = userData;
	GList			*expanded_dir = NULL;
	gint 			number_of_expanded_dirs = 0;
	gint 			counter = 0;

	// clear the eventual previous focused filename
	if(gBrowserState.pFocusedFile)
	{
		g_free(gBrowserState.pFocusedFile);
	}

	// copy the focused filename
	gBrowserState.pFocusedFile = fmWindow_getFilenameToRemember(fm_window, fm_window->pFocusedNode);

	// clear the eventual previous zoomed directory
	if(gBrowserState.pZoomDirectory)
	{
		g_free(gBrowserState.pZoomDirectory);
	}

	// copy the zoomed directory
    gBrowserState.pZoomDirectory = g_strdup(fm_window->pZoomDirectory);

	// clear the eventual previous expanded dirs
	while(g_list_length(gBrowserState.pExpandedDirs))
	{
		expanded_dir = g_list_first(gBrowserState.pExpandedDirs);
		gBrowserState.pExpandedDirs = g_list_remove_link(gBrowserState.pExpandedDirs, expanded_dir);
		g_free(expanded_dir->data);
		g_list_free_1(expanded_dir);
	}

	// copy the expanded dirs
	number_of_expanded_dirs = g_list_length(fm_window->pExpandedDirs);
	counter = 0;
	while(counter < number_of_expanded_dirs)
	{
		gBrowserState.pExpandedDirs = g_list_append(gBrowserState.pExpandedDirs, g_strdup(g_list_nth_data(fm_window->pExpandedDirs, counter)));
		counter++;
	}
}

void fmWindow_ctreeMenuItemBrowserNew(GtkMenuItem *menuItem, gpointer userData)
{
    struct FmWindow	*fm_window = userData;
	
	buildNewFmWindow(fmWindow_new(fm_window));
}

void fmWindow_ctreeMenuItemBrowserPasteState(GtkMenuItem *menuItem, gpointer userData)
{
    struct FmWindow	*fm_window = userData;
	GList			*expanded_dir = NULL;
	gint 			number_of_expanded_dirs = 0;
	gint 			counter = 0;

	if(gBrowserState.pExpandedDirs != NULL)
	{
		// clear the eventual previous expanded dirs
		while(g_list_length(fm_window->pExpandedDirs))
		{
			expanded_dir = g_list_first(fm_window->pExpandedDirs);
			fm_window->pExpandedDirs = g_list_remove_link(fm_window->pExpandedDirs, expanded_dir);
			g_free(expanded_dir->data);
			g_list_free_1(expanded_dir);
		}
		
		// copy the expanded dirs
		number_of_expanded_dirs = g_list_length(gBrowserState.pExpandedDirs);
		counter = 0;
		while(counter < number_of_expanded_dirs)
		{
			fm_window->pExpandedDirs = g_list_append(fm_window->pExpandedDirs, g_strdup(g_list_nth_data(gBrowserState.pExpandedDirs, counter)));
			counter++;
		}

	}

	if(gBrowserState.pZoomDirectory != NULL)
	{
		// copy the zoomed directory
		if(fm_window->pZoomDirectory)
		{
			g_free(fm_window->pZoomDirectory);
		}
		fm_window->pZoomDirectory = g_strdup(gBrowserState.pZoomDirectory);
	}

	if(gBrowserState.pFocusedFile != NULL)
	{
		// copy the focused file
		if(fm_window->pInitialFocus)
		{
			g_free(fm_window->pInitialFocus);
		}
		fm_window->pInitialFocus = g_strdup(gBrowserState.pFocusedFile);
		if(fm_window->pFocusedNode)
		{
			fm_window->pFocusedNode = NULL;
		}
	}
	
	fmWindow_rebuildCTree(fm_window);
}

void fmWindow_ctreeMenuItemBrowserRefresh(GtkMenuItem *menuItem, gpointer userData)
{
    struct FmWindow	*fm_window = userData;

	fmWindow_refreshCTree(fm_window);
}

void fmWindow_ctreeMenuItemBrowserUnZoom(GtkMenuItem *menuItem, gpointer userData)
{
    struct FmWindow	*fm_window = userData;

	if(fm_window->pInitialFocus)
	{
		g_free(fm_window->pInitialFocus);
	}
	fm_window->pInitialFocus = fmWindow_getFilenameToRemember(fm_window, fm_window->pFocusedNode);
	
	if(fm_window->pZoomDirectory)
	{
		g_free(fm_window->pZoomDirectory);
	}
	fm_window->pZoomDirectory = g_strdup("/");
	if(fm_window->pFocusedNode)
	{
		fm_window->pFocusedNode = NULL;
	}
	fmWindow_rebuildCTree(fm_window);
}

void fmWindow_ctreeMenuItemColumnsToggleColumn(GtkMenuItem *menuItem, gpointer userData)
{
    struct FmWindow	*fm_window = userData;
	guint			column_id = (guint)gtk_object_get_data(GTK_OBJECT(menuItem), "columnId");

	if(g_list_find(fm_window->pCTreeColumns, (gpointer)column_id))
	{
		fm_window->pCTreeColumns = g_list_remove(fm_window->pCTreeColumns, (gpointer)column_id);
	}
	else
	{
		fm_window->pCTreeColumns = g_list_prepend(fm_window->pCTreeColumns, (gpointer)column_id);
	}

    fmWindow_rebuildCTree(fm_window);
}

void fmWindow_ctreeMenuItemDirectoryOpenInNewWindow(GtkMenuItem *menuItem, gpointer userData)
{
    struct FmWindow	*fm_window = userData;
	GNode			*source_gnode = getLinkSource(fm_window->pFocusedNode);

	if((source_gnode = fmWindow_validateActionOnNode(fm_window, source_gnode)))
	{
        fmWindow_openDirectoryInNewWindow(fm_window, source_gnode);
	}
}

void fmWindow_ctreeMenuItemDirectoryRefresh(GtkMenuItem *menuItem, gpointer userData)
{
    struct FmWindow	*fm_window = userData;
	GNode			*source_gnode = NULL;
	
	source_gnode = gtk_clist_get_row_data(GTK_CLIST(fm_window->pCTree), GTK_CLIST(fm_window->pCTree)->focus_row);
	if((source_gnode = fmWindow_validateActionOnNode(fm_window, source_gnode)))
	{
		fmWindow_refreshParentDirectory(fm_window, source_gnode);
	}
}

void fmWindow_ctreeMenuItemDirectoryZoom(GtkMenuItem *menuItem, gpointer userData)
{
    struct FmWindow	*fm_window = userData;
	GNode			*source_gnode = getLinkSource(fm_window->pFocusedNode);
	gchar			*temp_zoom_dir = NULL;
	gchar			*temp_string1 = NULL;
	gchar			*temp_string2 = NULL;

	if((source_gnode = fmWindow_validateActionOnNode(fm_window, source_gnode)))
	{
		if(isDirectoryFilled(getLinkTarget(source_gnode)))
		{
			if(fm_window->pInitialFocus)
			{
				g_free(fm_window->pInitialFocus);
			}
			fm_window->pInitialFocus = NULL;
			if(isLink(source_gnode))
			{
				if(g_path_is_absolute(((struct NodeInfo*)source_gnode->data)->pLinkTarget))
				{
					temp_zoom_dir = compressPath(((struct NodeInfo*)source_gnode->data)->pLinkTarget);
				}
				else
				{
					temp_string1 = fmWindow_getFilenameToRemember(fm_window, source_gnode);
					temp_string2 = g_dirname(temp_string1);
					g_free(temp_string1);
					temp_string1 = g_strjoin("/", temp_string2, ((struct NodeInfo*)source_gnode->data)->pLinkTarget, NULL);
					g_free(temp_string2);
					temp_zoom_dir = compressPath(temp_string1);
					g_free(temp_string1);
				}
			}
			else
			{
				temp_zoom_dir = fmWindow_getFilenameToRemember(fm_window, source_gnode);
			}
			if(fm_window->pZoomDirectory)
			{
				g_free(fm_window->pZoomDirectory);
			}
            fm_window->pZoomDirectory = temp_zoom_dir;
			if(fm_window->pFocusedNode)
			{
				fm_window->pFocusedNode = NULL;
			}
			fmWindow_rebuildCTree(fm_window);
		}
	}
}

void fmWindow_ctreeMenuItemSelectionClear(GtkMenuItem *menuItem, gpointer userData)
{
    struct FmWindow	*fm_window = userData;

    fmWindow_selectionUnselectAll(fm_window);
}

void fmWindow_ctreeMenuItemSelectionCopyFocusToComboEntry(GtkMenuItem *menuItem, gpointer userData)
{
    struct FmWindow	*fm_window = userData;
	GNode			*source_gnode = NULL;
	
	source_gnode = gtk_clist_get_row_data(GTK_CLIST(fm_window->pCTree), GTK_CLIST(fm_window->pCTree)->focus_row);
	gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(fm_window->pComboBox)->entry), fmWindow_getFilenameToRemember(fm_window, source_gnode));
}

void fmWindow_ctreeMenuItemSelectionCopyToXClipboard(GtkMenuItem *menuItem, gpointer userData)
{
    struct FmWindow	*fm_window = userData;

	if(fm_window->pSelectionTextSnapshot)
	{
		g_free(fm_window->pSelectionTextSnapshot);
		fm_window->pSelectionTextSnapshot = NULL;
	}
	fm_window->pSelectionTextSnapshot = fmWindow_selectionTextSnapshotString(fm_window, fm_window->pSelection);

	gtk_selection_owner_set(GTK_WIDGET(fm_window->pCTree), GDK_SELECTION_PRIMARY, GDK_CURRENT_TIME);
}

void fmWindow_ctreeMenuItemSelectionDeselectDirectory(GtkMenuItem *menuItem, gpointer userData)
{
    struct FmWindow	*fm_window = userData;
	GNode			*source_gnode = NULL;
	
	source_gnode = gtk_clist_get_row_data(GTK_CLIST(fm_window->pCTree), GTK_CLIST(fm_window->pCTree)->focus_row);
	fmWindow_selectionRemoveAllDirEntries(fm_window, source_gnode);
}

void fmWindow_ctreeMenuItemSelectionSelectDirectory(GtkMenuItem *menuItem, gpointer userData)
{
    struct FmWindow	*fm_window = userData;
	GNode			*source_gnode = NULL;
	
	source_gnode = gtk_clist_get_row_data(GTK_CLIST(fm_window->pCTree), GTK_CLIST(fm_window->pCTree)->focus_row);
	fmWindow_selectionAddAllDirEntries(fm_window, source_gnode);
}

void fmWindow_ctreeMenuItemPreferences(GtkMenuItem *menuItem, gpointer userData)
{
    preferencesWindow_new();
}

void fmWindow_ctreeMenuItemQuit(GtkMenuItem *menuItem, gpointer userData)
{
	gtk_main_quit();
}

gboolean fmWindow_ctreeMotionNotify(GtkWidget *widget, GdkEventMotion *event, gpointer userData)
{
//	GtkWidget		*grabbed_widget = NULL;
	
	if(!gtk_object_get_data(GTK_OBJECT(widget), "drag_has_started") && gtk_object_get_data(GTK_OBJECT(widget), "drag_is_possible"))
	{
		gtk_object_set_data(GTK_OBJECT(widget), "drag_has_started", (gpointer)1);
		gtk_object_set_data(GTK_OBJECT(widget), "ctree_drag_focused_row", (gpointer)-1);
		gtk_drag_begin(widget, gtk_target_list_new(gCTreeDragDropTargets, gCTreeNumberOfDragDropTargets), GDK_ACTION_COPY | GDK_ACTION_MOVE, 1, (GdkEvent*)event);
	}
	
// discouraged by owen
/*	grabbed_widget = gtk_grab_get_current();
	if(grabbed_widget)
	{
//		gtk_signal_connect(GTK_OBJECT(grabbed_widget), "key_press_event", GTK_SIGNAL_FUNC(fmWindow_grabbedWidgetKeyPress), NULL);
		gtk_signal_connect(GTK_OBJECT(grabbed_widget), "button_release_event", GTK_SIGNAL_FUNC(fmWindow_grabbedWidgetButtonRelease), NULL);
	}*/
	
	return FALSE;
}

void fmWindow_ctreeSelectionClear(GtkWidget *widget, GdkEventSelection *event, gpointer userData)
{
    struct FmWindow	*fm_window = userData;
	
	if(fm_window->pSelectionTextSnapshot)
	{
		g_free(fm_window->pSelectionTextSnapshot);
	}
}

void fmWindow_ctreeSelectionGet(GtkWidget *widget, GtkSelectionData *selection_data, guint info, guint time, gpointer userData)
{
    struct FmWindow	*fm_window = userData;
	guchar			*compound_text = NULL;
	GdkAtom 		compound_text_encoding = GDK_NONE;
	gint 			compound_text_format = 0;
	guint 			compound_text_length = 0;
	
	if(fm_window->pSelectionTextSnapshot)
	{
		switch(info)
		{
			case SELECTION_TARGET_COMPOUND_TEXT:
			case SELECTION_TARGET_TEXT:
				gdk_string_to_compound_text(fm_window->pSelectionTextSnapshot, &compound_text_encoding, &compound_text_format, &compound_text, &compound_text_length);
				gtk_selection_data_set(selection_data, compound_text_encoding, compound_text_format, compound_text, compound_text_length);
				gdk_free_compound_text(compound_text);
				break;
			case SELECTION_TARGET_STRING:
				gtk_selection_data_set(selection_data, GDK_SELECTION_TYPE_STRING, 8, fm_window->pSelectionTextSnapshot, strlen(fm_window->pSelectionTextSnapshot));
				break;
		}
	}
}

void fmWindow_ctreeSizeAllocate(GtkWidget *widget, GtkAllocation *allocation, gpointer userData)
{
	fmWindow_resizeColumns((struct FmWindow*)userData);
}

gboolean fmWindow_destroyButtonBookmark(GtkWidget *widget, gpointer userData)
{
    struct FmWindow	*fm_window = userData;
	guint 			index = (guint)gtk_object_get_data(GTK_OBJECT(widget), "index");

	gtk_widget_destroy(fm_window->pBookmarkButtons[index].pPopupMenuItemSet);
	gtk_widget_destroy(fm_window->pBookmarkButtons[index].pPopupMenuItemGoTo);
	gtk_widget_destroy(fm_window->pBookmarkButtons[index].pPopupMenuItemClear);
	gtk_widget_destroy(fm_window->pBookmarkButtons[index].pPopupMenu);

//	printf("button destroyed\n");
	return FALSE;
}

gboolean fmWindow_destroyMenuItem(GtkWidget *widget, gpointer userData)
{
	gchar *accel_key_id = (gchar*)gtk_object_get_data(GTK_OBJECT(widget), "accelKeyId");
	
//	printf("destroy menuitem\n");
	if(accel_key_id)
	{
//		printf("menuitem accelerator '%s'\n", accel_key_id);
		g_free(accel_key_id);
	}
	
	return FALSE;
}

gboolean fmWindow_deleteWindow(GtkWidget *widget, GdkEvent *event, gpointer userData)
{
	// save the layout of the last window if layout saving is enable
	if(g_list_length(pgFmWindows) == 1 && pgPreferences->sessionSaveLayoutOnExit)
	{
		layout_save(pgPreferences);
	}
	
	return FALSE;
}

gboolean fmWindow_destroyWindow(GtkWidget *widget, gpointer userData)
{
    struct FmWindow	*fm_window = userData;

//	printf("destroy window\n");
	closeFmWindow(fm_window);

	if(g_list_length(pgFmWindows) == 0)
	{
		gtk_main_quit();
	}
	
	return FALSE;
}

gboolean fmWindow_grabbedWidgetButtonRelease(GtkWidget *widget, GdkEventButton *event, gpointer userData)
{
    struct FmWindow	*fm_window = pgsDragFocusedFmWindow;
	GNode			*temp_source_gnode = NULL;
	
	if(event->button != 1)
	{
		if((temp_source_gnode = fmWindow_validateActionOnNode(fm_window, fm_window->pFocusedNode)))
		{
			if(temp_source_gnode)
			{
				fmWindow_toggleNode(fm_window, getLinkTarget(temp_source_gnode));
				gtk_signal_emit_stop_by_name(GTK_OBJECT(widget), "button_release_event");
				return TRUE;
			}
		}
	}

	return FALSE;
}

gboolean fmWindow_grabbedWidgetKeyPress(GtkWidget *widget, GdkEventKey *event, gpointer userData)
{
    struct FmWindow	*fm_window = pgsDragFocusedFmWindow;
	GNode			*source_gnode = NULL;

	source_gnode = gtk_clist_get_row_data(GTK_CLIST(fm_window->pCTree), GTK_CLIST(fm_window->pCTree)->focus_row);

//	printf("key pressed : 0x%x\n", event->keyval);
	if(source_gnode)
	{
		switch(event->keyval)
		{
			case GDK_Left:
				if( (source_gnode = fmWindow_validateActionOnNode(fm_window, source_gnode)) )
				{
					fmWindow_collapseNode(fm_window, getLinkTarget(source_gnode));
				}
				break;
			case GDK_Right:
				if( (source_gnode = fmWindow_validateActionOnNode(fm_window, source_gnode)) )
				{
					fmWindow_expandNode(fm_window, getLinkTarget(source_gnode));
				}
				break;
			case GDK_Return:
				if( (source_gnode = fmWindow_validateActionOnNode(fm_window, source_gnode)) )
				{
					fmWindow_toggleNode(fm_window, getLinkTarget(source_gnode));
				}
				break;
		}
	}
	gtk_signal_emit_stop_by_name(GTK_OBJECT(widget), "key_press_event");
	return TRUE;
}

gboolean fmWindow_panButtonPress(GtkWidget *widget, GdkEventButton *event, gpointer userData)
{
    struct FmWindow	*fm_window = userData;
	gtk_grab_add(widget);
	
	gtk_pixmap_set(GTK_PIXMAP(fm_window->pPixmapPanWindow), pgPixmapPanWindowCenter, pgMaskPanWindowCenter);

	gtk_object_set_data(GTK_OBJECT(widget), "timeout", (gpointer)gtk_timeout_add(PAN_TIME, (GtkFunction)fmWindow_panTimeout, fm_window));
	
	return FALSE;
}

gboolean fmWindow_panButtonRelease(GtkWidget *widget, GdkEventButton *event, gpointer userData)
{
    struct FmWindow	*fm_window = userData;
	gtk_pixmap_set(GTK_PIXMAP(fm_window->pPixmapPanWindow), pgPixmapPanWindowCenter, pgMaskPanWindowCenter);

	gtk_timeout_remove((guint)gtk_object_get_data(GTK_OBJECT(widget), "timeout"));
    gtk_object_remove_data(GTK_OBJECT(widget), "timeout");
	gtk_grab_remove(widget);
	
	return FALSE;
}

gboolean fmWindow_panMotionNotify(GtkWidget *widget, GdkEventMotion *event, gpointer userData)
{
//    struct FmWindow	*fm_window = userData;

	return TRUE;
}


