/*	
**	
**	Copyright (C) 1999 Geert Bevin
**	gbevin@thunderstorms.org
**	http://thunderstorms.org/fm
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation; either version 2 of the License, or
**	(at your option) any later version.
**	
**	This program is distributed in the hope that it will be useful,
**	but WITHOUT ANY WARRANTY; without even the implied warranty of
**	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**	GNU General Public License for more details.
**	
**	You should have received a copy of the GNU General Public License
**	along with this program; if not, write to the Free Software
**	
**	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**		
*/

#include <unistd.h>
#include <stdio.h>
#include <gdk/gdkkeysyms.h>
#include <gdk/gdktypes.h>
#include <glib.h>
#include <math.h>
#include <fcntl.h>
#include <sys/stat.h>

#include "accelkey.h"
#include "ctreecolumn.h"
#include "filetype.h"
#include "preferences.h"
#include "rowstyle.h"

extern gchar			*pgConfigPath;
extern gchar			*pgFullProcessPath;
FILE					*pgConfigFileStream = NULL;

static void 					preferences_duplicatePixmapPaths(gpointer, gpointer);
static void 					preferences_duplicateCtreeColumns(gpointer, gpointer);
static void 					preferences_duplicateAccelKey(gpointer, gpointer, gpointer);
static void 					preferences_duplicateRowStyle(gpointer, gpointer, gpointer);
static void 					preferences_duplicateFileTypes(struct Preferences*, struct Preferences*);
static void 					preferences_entryClear(struct PreferencesEntry*);
static void 					preferences_entryDestroy(struct PreferencesEntry*);
static struct PreferencesEntry*	preferences_entryNew();
static void						preferences_entryProcess(struct Preferences*, struct PreferencesEntry*);
static gchar*					preferences_getFileName();
static guint					preferences_parseSymbol(struct Preferences*, GScanner*);

enum
{
	SYMBOL_CTREE_COLUMNS					= G_TOKEN_LAST + 1,
	SYMBOL_ACCEL_KEYS						= G_TOKEN_LAST + 2,
	SYMBOL_ROW_STYLES						= G_TOKEN_LAST + 3,
	SYMBOL_FILE_TYPES						= G_TOKEN_LAST + 4,
	SYMBOL_GENERAL							= G_TOKEN_LAST + 5,
				
	SYMBOL_CTREE_COLUMN						= G_TOKEN_LAST + 6,
			
	SYMBOL_FILE_TYPE						= G_TOKEN_LAST + 7,
			
	SYMBOL_CHARACTERBROWSING				= G_TOKEN_LAST + 8,
	SYMBOL_COMBOBOX							= G_TOKEN_LAST + 9,
	SYMBOL_SELECTION						= G_TOKEN_LAST + 10,
	SYMBOL_SESSION							= G_TOKEN_LAST + 11,
	SYMBOL_SORT								= G_TOKEN_LAST + 12,
	SYMBOL_FMWINDOW							= G_TOKEN_LAST + 13,
				
	SYMBOL_TYPE								= G_TOKEN_LAST + 14,
	SYMBOL_ALIGNMENT						= G_TOKEN_LAST + 15,
	
	SYMBOL_KEY								= G_TOKEN_LAST + 16,
	SYMBOL_MODS								= G_TOKEN_LAST + 17,
		
	SYMBOL_FGRED							= G_TOKEN_LAST + 18,
	SYMBOL_FGGREEN							= G_TOKEN_LAST + 19,
	SYMBOL_FGBLUE							= G_TOKEN_LAST + 20,
	SYMBOL_BGRED							= G_TOKEN_LAST + 21,
	SYMBOL_BGGREEN							= G_TOKEN_LAST + 22,
	SYMBOL_BGBLUE							= G_TOKEN_LAST + 23,
	SYMBOL_FILENAMECLOSED					= G_TOKEN_LAST + 24,
	SYMBOL_FILENAMEOPENED					= G_TOKEN_LAST + 25,
	SYMBOL_DELETABLE						= G_TOKEN_LAST + 26,
		
	SYMBOL_BASICFILETYPE					= G_TOKEN_LAST + 27,
	SYMBOL_ROWSTYLE							= G_TOKEN_LAST + 28,
		
	SYMBOL_ENABLED							= G_TOKEN_LAST + 29,
	SYMBOL_ENABLESPACEBAR					= G_TOKEN_LAST + 30,
	SYMBOL_IGNORECASE						= G_TOKEN_LAST + 31,
			
	SYMBOL_POPDOWNLENGTH					= G_TOKEN_LAST + 32,
			
	SYMBOL_DIRECTORIESFIRST					= G_TOKEN_LAST + 33,
			
	SYMBOL_SAVEPREFERENCESONEXIT			= G_TOKEN_LAST + 34,
	SYMBOL_SAVELAYOUTONEXIT					= G_TOKEN_LAST + 35,
	SYMBOL_SAVELAYOUTBOOKMARKS				= G_TOKEN_LAST + 36,
	SYMBOL_SAVELAYOUTCOMBOHISTORY			= G_TOKEN_LAST + 37,
	SYMBOL_SAVELAYOUTEXPANDEDDIRS			= G_TOKEN_LAST + 38,
	SYMBOL_SAVELAYOUTONGEOMETRY				= G_TOKEN_LAST + 39,
	SYMBOL_SAVELAYOUTONBROWSERAPPEARANCE	= G_TOKEN_LAST + 40,
	
	SYMBOL_TEXT_ELEMENT_SEPERATOR			= G_TOKEN_LAST + 41,
		
	SYMBOL_DEFAULT_POSITION					= G_TOKEN_LAST + 42,
	SYMBOL_X								= G_TOKEN_LAST + 43,
	SYMBOL_Y								= G_TOKEN_LAST + 44,
	SYMBOL_WIDTH							= G_TOKEN_LAST + 45,
	SYMBOL_HEIGHT							= G_TOKEN_LAST + 46
};

static const struct SymbolDefinition	psSymbolTablePreferences[] =
{
	{ "CTREE_COLUMNS",	 				SYMBOL_CTREE_COLUMNS },
	{ "ACCEL_KEYS",			  			SYMBOL_ACCEL_KEYS },
	{ "ROW_STYLES", 		  			SYMBOL_ROW_STYLES },
	{ "FILE_TYPES",			  			SYMBOL_FILE_TYPES },
	{ "GENERAL", 			  			SYMBOL_GENERAL },
		
	{ "ctree_column", 			  		SYMBOL_CTREE_COLUMN },
		
	{ "file_type", 			  			SYMBOL_FILE_TYPE },
		
	{ "characterbrowsing", 	  			SYMBOL_CHARACTERBROWSING },
	{ "combobox", 	  					SYMBOL_COMBOBOX },
	{ "selection", 			  			SYMBOL_SELECTION },
	{ "session", 			  			SYMBOL_SESSION },
	{ "sort", 				  			SYMBOL_SORT },
	{ "fmwindow", 			  			SYMBOL_FMWINDOW },
		
	{ "Type", 				  			SYMBOL_TYPE },
	{ "Alignment", 				  		SYMBOL_ALIGNMENT },
		
	{ "Key", 				  			SYMBOL_KEY },
	{ "Mods", 				  			SYMBOL_MODS },
	
	{ "FgRed", 				  			SYMBOL_FGRED },
	{ "FgGreen", 			  			SYMBOL_FGGREEN },
	{ "FgBlue", 			  			SYMBOL_FGBLUE },
	{ "BgRed", 				  			SYMBOL_BGRED },
	{ "BgGreen", 			  			SYMBOL_BGGREEN },
	{ "BgBlue", 			  			SYMBOL_BGBLUE },
	{ "FilenameClosed", 	  			SYMBOL_FILENAMECLOSED },
	{ "FilenameOpened", 	  			SYMBOL_FILENAMEOPENED },
	{ "Deletable", 			  			SYMBOL_DELETABLE },
		
	{ "BasicFileType", 		  			SYMBOL_BASICFILETYPE },
	{ "RowStyle", 			  			SYMBOL_ROWSTYLE },
		
	{ "Enabled", 			  			SYMBOL_ENABLED },
	{ "EnableSpacebar", 	  			SYMBOL_ENABLESPACEBAR },
	{ "IgnoreCase", 		  			SYMBOL_IGNORECASE },
		
	{ "PopdownLength", 		  			SYMBOL_POPDOWNLENGTH },
		
	{ "DirectoriesFirst", 	  			SYMBOL_DIRECTORIESFIRST },
		
	{ "SavePreferencesOnExit",			SYMBOL_SAVEPREFERENCESONEXIT },
	{ "SaveLayoutOnExit", 	  			SYMBOL_SAVELAYOUTONEXIT },
	{ "SaveLayoutBookmarks", 			SYMBOL_SAVELAYOUTBOOKMARKS },
	{ "SaveLayoutComboHistory", 		SYMBOL_SAVELAYOUTCOMBOHISTORY },
	{ "SaveLayoutExpandedDirs", 		SYMBOL_SAVELAYOUTEXPANDEDDIRS },
	{ "SaveLayoutGeometry", 			SYMBOL_SAVELAYOUTONGEOMETRY },
	{ "SaveLayoutBrowserAppearance",	SYMBOL_SAVELAYOUTONBROWSERAPPEARANCE },
	
	{ "TextElementSeperator", 			SYMBOL_TEXT_ELEMENT_SEPERATOR },
			
	{ "DefaultPosition",				SYMBOL_DEFAULT_POSITION },
	{ "Xpos", 							SYMBOL_X },
	{ "Ypos", 							SYMBOL_Y },
	{ "Width", 							SYMBOL_WIDTH },
	{ "Height", 						SYMBOL_HEIGHT },
			
	{ NULL, 							0 } 
};

void preferences_clear()
{
	gchar	*preferences_filename = preferences_getFileName();

	if(preferences_filename)
	{
		unlink(preferences_filename);
		g_free(preferences_filename);
	}
}

void preferences_destroy(struct Preferences *preferences)
{
	if(preferences)
	{
		if(preferences->pSelectionTextElementSeperator)
		{
			g_free(preferences->pSelectionTextElementSeperator);
		}

		ctreeColumnsFree(preferences);
		
		fileTypesRemoveAll(&preferences->pFileTypes);

		accelKeysRemoveAll(preferences->pAccelKeys);
		g_hash_table_destroy(preferences->pAccelKeys);

		rowStylesRemoveAll(preferences->pRowStyles);
		g_hash_table_destroy(preferences->pRowStyles);

		preferences_pixmapSearchPathsRemoveAll(preferences);
	}
}

static void preferences_duplicatePixmapPaths(gpointer data, gpointer userData)
{
	struct Preferences	*new_preferences = (struct Preferences *)userData;
	gchar				*pixmap_path = (gchar *)data;

	new_preferences->pPixmapSearchPaths = g_list_prepend(new_preferences->pPixmapSearchPaths, pixmap_path ? g_strdup(pixmap_path) : NULL);
}

static void preferences_duplicateCtreeColumns(gpointer data, gpointer userData)
{
	struct Preferences	*new_preferences = (struct Preferences *)userData;
	struct CtreeColumn	*ctree_column = (struct CtreeColumn *)data;
	struct CtreeColumn	*new_ctree_column = g_new0(struct CtreeColumn, 1);

	if(ctree_column)
	{
		new_ctree_column->alignment = ctree_column->alignment;
		new_ctree_column->type = ctree_column->type;
		new_preferences->pCtreeColumns = g_list_append(new_preferences->pCtreeColumns, new_ctree_column);
	}

}

static void preferences_duplicateAccelKey(gpointer key, gpointer value, gpointer userData)
{
	struct Preferences	*new_preferences = (struct Preferences *)userData;
	gchar				*accel_key_name = key;
	struct AccelKey		*accel_key = (struct AccelKey *)value;

	accelKeyNew(new_preferences->pAccelKeys, accel_key_name, accel_key->key, accel_key->mods);
}

static void preferences_duplicateRowStyle(gpointer key, gpointer value, gpointer userData)
{
	struct Preferences	*new_preferences = (struct Preferences *)userData;
	gchar				*row_style_name = key;
	struct RowStyle		*row_style = (struct RowStyle *)value;

	rowStyleNew(new_preferences->pRowStyles, new_preferences->pPixmapSearchPaths, row_style_name, row_style->foreground.red, row_style->foreground.green, row_style->foreground.blue, row_style->background.red, row_style->background.green, row_style->background.blue, row_style->pFilenameClosed ? g_strdup(row_style->pFilenameClosed) : NULL, row_style->pFilenameOpened ? g_strdup(row_style->pFilenameOpened) : NULL, row_style->deletable);
}

static void preferences_duplicateFileTypes(struct Preferences *preferences, struct Preferences *newPreferences)
{
	guint				counter = 0;
	struct FileType		*file_type = NULL;
	gchar				*row_style_name = NULL;

	for(counter = 0; counter < g_list_length(preferences->pFileTypes); counter++)
	{
		file_type = g_list_nth_data(preferences->pFileTypes, counter);
        row_style_name = rowStyleFindName(preferences->pRowStyles, file_type->pRowStyle);
		if(row_style_name)
		{
			fileTypeNew(&newPreferences->pFileTypes, newPreferences->pRowStyles, file_type->deletable, file_type->basicType, g_strdup(row_style_name));
		}
	}
}

struct Preferences*	preferences_duplicate(struct Preferences *preferences)
{
	struct Preferences	*new_preferences = preferences_new();

	if(new_preferences)
	{
		// duplicate the pixmap paths
		g_list_foreach(preferences->pPixmapSearchPaths, (GFunc)preferences_duplicatePixmapPaths, new_preferences);
	
		// duplicate the ctree columns
		g_list_foreach(preferences->pCtreeColumns, (GFunc)preferences_duplicateCtreeColumns, new_preferences);
	
		// duplicate the accel keys
		g_hash_table_foreach(preferences->pAccelKeys, (GHFunc)preferences_duplicateAccelKey, new_preferences);
		
		// duplicate the row styles
		g_hash_table_foreach(preferences->pRowStyles, (GHFunc)preferences_duplicateRowStyle, new_preferences);
		
		// duplicate the file types
		preferences_duplicateFileTypes(preferences, new_preferences);
		
		// duplicate the general parameters
		new_preferences->characterBrowsingEnabled = preferences->characterBrowsingEnabled;
		new_preferences->characterBrowsingEnableSpacebar = preferences->characterBrowsingEnableSpacebar;
		new_preferences->characterBrowsingIgnoreCase = preferences->characterBrowsingIgnoreCase;
		new_preferences->comboPopdownLength = preferences->comboPopdownLength;
		new_preferences->fmWindowDefaultPosition = preferences->fmWindowDefaultPosition;
		new_preferences->fmWindowX = preferences->fmWindowX;
		new_preferences->fmWindowY = preferences->fmWindowY;
		new_preferences->fmWindowWidth = preferences->fmWindowWidth;
		new_preferences->fmWindowHeight = preferences->fmWindowHeight;
		new_preferences->sessionSavePreferencesOnExit = preferences->sessionSavePreferencesOnExit;
		new_preferences->sessionSaveLayoutOnExit = preferences->sessionSaveLayoutOnExit;
		new_preferences->sessionSaveLayoutBookmarks = preferences->sessionSaveLayoutBookmarks;
		new_preferences->sessionSaveLayoutComboHistory = preferences->sessionSaveLayoutComboHistory;
		new_preferences->sessionSaveLayoutExpandedDirs = preferences->sessionSaveLayoutExpandedDirs;
		new_preferences->sessionSaveLayoutGeometry = preferences->sessionSaveLayoutGeometry;
		new_preferences->sessionSaveLayoutBrowserAppearance = preferences->sessionSaveLayoutBrowserAppearance;
		new_preferences->sortDirectoriesFirst = preferences->sortDirectoriesFirst;
		new_preferences->sortIgnoreCase = preferences->sortIgnoreCase;
		new_preferences->pSelectionTextElementSeperator = g_strdup(preferences->pSelectionTextElementSeperator);
	}

	return new_preferences;

}

static void preferences_entryClear(struct PreferencesEntry *preferencesEntry)
{
	if(preferencesEntry->pNameString)
	{
		g_free(preferencesEntry->pNameString);
        preferencesEntry->pNameString = NULL;
	}
	preferencesEntry->nameToken = G_TOKEN_NONE;
	
	if(preferencesEntry->section == SYMBOL_CTREE_COLUMNS)
	{
        preferencesEntry->values.ctreeColumn.type = 0;
        preferencesEntry->values.ctreeColumn.alignment = 0;
	}
	else if(preferencesEntry->section == SYMBOL_ACCEL_KEYS)
	{
        preferencesEntry->values.accelKey.key = 0;
        preferencesEntry->values.accelKey.mods = 0;
	}
	else if(preferencesEntry->section == SYMBOL_ROW_STYLES)
	{
		if(preferencesEntry->values.rowStyle.pFilenameOpened)
		{
			g_free(preferencesEntry->values.rowStyle.pFilenameOpened);
            preferencesEntry->values.rowStyle.pFilenameOpened = NULL;
		}
		if(preferencesEntry->values.rowStyle.pFilenameClosed)
		{
			g_free(preferencesEntry->values.rowStyle.pFilenameClosed);
            preferencesEntry->values.rowStyle.pFilenameClosed = NULL;
		}
		preferencesEntry->values.rowStyle.fgRed = 0;
		preferencesEntry->values.rowStyle.fgGreen = 0;
		preferencesEntry->values.rowStyle.fgBlue = 0;
		preferencesEntry->values.rowStyle.bgRed = 0;
		preferencesEntry->values.rowStyle.bgGreen = 0;
		preferencesEntry->values.rowStyle.bgBlue = 0;
		preferencesEntry->values.rowStyle.deletable = TRUE;
	}
	else if(preferencesEntry->section == SYMBOL_FILE_TYPES)
	{
        preferencesEntry->values.fileType.deletable = TRUE;
		if(preferencesEntry->values.fileType.pRowStyle)
		{
			g_free(preferencesEntry->values.fileType.pRowStyle);
            preferencesEntry->values.fileType.pRowStyle = NULL;
		}
        preferencesEntry->values.fileType.basicType = UNKNOWN;
	}
	else if(preferencesEntry->section == SYMBOL_GENERAL)
	{
		if(preferencesEntry->nameToken == SYMBOL_CHARACTERBROWSING)
		{
			preferencesEntry->values.general.characterBrowsing.enabled = TRUE;
			preferencesEntry->values.general.characterBrowsing.enableSpacebar = FALSE;
			preferencesEntry->values.general.characterBrowsing.ignoreCase = TRUE;
		}
		else if(preferencesEntry->nameToken == SYMBOL_COMBOBOX)
		{
			preferencesEntry->values.general.comboBox.popdownLength = 0;
		}
		else if(preferencesEntry->nameToken == SYMBOL_SORT)
		{
			preferencesEntry->values.general.sort.directoriesFirst = TRUE;
			preferencesEntry->values.general.sort.ignoreCase = TRUE;
		}
		else if(preferencesEntry->nameToken == SYMBOL_SELECTION)
		{
			if(preferencesEntry->values.general.selection.pTextElementSeperator)
			{
				g_free(preferencesEntry->values.general.selection.pTextElementSeperator);
				preferencesEntry->values.general.selection.pTextElementSeperator = NULL;
			}
		}
		else if(preferencesEntry->nameToken == SYMBOL_SESSION)
		{
            preferencesEntry->values.general.session.savePreferencesOnExit = TRUE;
            preferencesEntry->values.general.session.saveLayoutOnExit = TRUE;
            preferencesEntry->values.general.session.saveLayoutBookmarks = TRUE;
            preferencesEntry->values.general.session.saveLayoutBrowserAppearance = TRUE;
            preferencesEntry->values.general.session.saveLayoutComboHistory = TRUE;
            preferencesEntry->values.general.session.saveLayoutExpandedDirs = TRUE;
            preferencesEntry->values.general.session.saveLayoutGeometry = TRUE;
		}
		else if(preferencesEntry->nameToken == SYMBOL_FMWINDOW)
		{
            preferencesEntry->values.general.fmwindow.defaultPosition = FALSE;
            preferencesEntry->values.general.fmwindow.x = -1;
            preferencesEntry->values.general.fmwindow.y = -1;
            preferencesEntry->values.general.fmwindow.width = 400;
            preferencesEntry->values.general.fmwindow.height = 600;
		}
	}
}

static void preferences_entryDestroy(struct PreferencesEntry *preferencesEntry)
{
    preferences_entryClear(preferencesEntry);
	g_free(preferencesEntry);
}

static struct PreferencesEntry* preferences_entryNew()
{
	struct PreferencesEntry *preferences_entry = NULL;

    preferences_entry = g_new0(struct PreferencesEntry, 1);
    preferences_entryClear(preferences_entry);

	return preferences_entry;
}

static void preferences_entryProcess(struct Preferences *preferences, struct PreferencesEntry *preferencesEntry)
{
	// check if there is a valid previous preference entry
	if(preferencesEntry)
	{
		if(preferencesEntry->section == SYMBOL_CTREE_COLUMNS)
		{
			if(preferencesEntry->nameToken == SYMBOL_CTREE_COLUMN)
			{
                ctreeColumnRemove(preferences, preferencesEntry->values.ctreeColumn.type);
                ctreeColumnAdd(preferences, preferencesEntry->values.ctreeColumn.type, preferencesEntry->values.ctreeColumn.alignment);
//				printf("Defined CTREE_COLUMN : %d, %d\n", preferencesEntry->values.ctreeColumn.type, preferencesEntry->values.ctreeColumn.alignment);
			}
		}
		else if(preferencesEntry->section == SYMBOL_ACCEL_KEYS)
		{
			if(preferencesEntry->pNameString)
			{
				accelKeyNew(preferences->pAccelKeys, preferencesEntry->pNameString, preferencesEntry->values.accelKey.key, preferencesEntry->values.accelKey.mods);
//				printf("Defined ACCEL_KEY : %s, %d, %d\n", preferencesEntry->pNameString, preferencesEntry->values.accelKey.key, preferencesEntry->values.accelKey.mods);
			}
		}
		else if(preferencesEntry->section == SYMBOL_ROW_STYLES)
		{
			if(preferencesEntry->pNameString &&
			   preferencesEntry->values.rowStyle.pFilenameClosed != NULL)
			{
				rowStyleNew(preferences->pRowStyles, preferences->pPixmapSearchPaths, preferencesEntry->pNameString, preferencesEntry->values.rowStyle.fgRed, preferencesEntry->values.rowStyle.fgGreen, preferencesEntry->values.rowStyle.fgBlue, preferencesEntry->values.rowStyle.bgRed, preferencesEntry->values.rowStyle.bgGreen, preferencesEntry->values.rowStyle.bgBlue, preferencesEntry->values.rowStyle.pFilenameClosed, preferencesEntry->values.rowStyle.pFilenameOpened, preferencesEntry->values.rowStyle.deletable);
//				printf("Defined ROW_STYLE : %s, 0x%x, 0x%x, 0x%x, 0x%x, 0x%x, 0x%x, %s, %s, %d\n", preferencesEntry->pNameString, preferencesEntry->values.rowStyle.fgRed, preferencesEntry->values.rowStyle.fgGreen, preferencesEntry->values.rowStyle.fgBlue, preferencesEntry->values.rowStyle.bgRed, preferencesEntry->values.rowStyle.bgGreen, preferencesEntry->values.rowStyle.bgBlue, preferencesEntry->values.rowStyle.pFilenameClosed, preferencesEntry->values.rowStyle.pFilenameOpened, preferencesEntry->values.rowStyle.deletable);
			}
		}
		else if(preferencesEntry->section == SYMBOL_FILE_TYPES)
		{
			if(preferencesEntry->nameToken == SYMBOL_FILE_TYPE &&
			   preferencesEntry->values.fileType.pRowStyle != NULL)
			{
				fileTypeNew(&preferences->pFileTypes, preferences->pRowStyles, preferencesEntry->values.fileType.deletable, preferencesEntry->values.fileType.basicType, preferencesEntry->values.fileType.pRowStyle);
//				printf("Defined FILE_TYPE : %d, %d, %s\n",preferencesEntry->values.fileType.deletable, preferencesEntry->values.fileType.basicType, preferencesEntry->values.fileType.pRowStyle);
			}
		}
		else if(preferencesEntry->section == SYMBOL_GENERAL)
		{
			if(preferencesEntry->nameToken == SYMBOL_CHARACTERBROWSING)
			{
				preferences->characterBrowsingEnabled = preferencesEntry->values.general.characterBrowsing.enabled;
				preferences->characterBrowsingEnableSpacebar = preferencesEntry->values.general.characterBrowsing.enableSpacebar;
				preferences->characterBrowsingIgnoreCase = preferencesEntry->values.general.characterBrowsing.ignoreCase;
//				printf("Defined GENERAL.CHARACTERBROWSING : %d, %d, %d\n", preferencesEntry->values.general.characterBrowsing.enabled, preferencesEntry->values.general.characterBrowsing.enableSpacebar, preferencesEntry->values.general.characterBrowsing.ignoreCase);
			}
			else if(preferencesEntry->nameToken == SYMBOL_COMBOBOX)
			{
				preferences->comboPopdownLength = preferencesEntry->values.general.comboBox.popdownLength;
//				printf("Defined GENERAL.COMBOBOX : %d\n", preferencesEntry->values.general.comboBox.popdownLength);
			}
			else if(preferencesEntry->nameToken == SYMBOL_SORT)
			{
				preferences->sortDirectoriesFirst = preferencesEntry->values.general.sort.directoriesFirst;
				preferences->sortIgnoreCase = preferencesEntry->values.general.sort.ignoreCase;
//				printf("Defined GENERAL.SORT : %d, %d\n", preferencesEntry->values.general.sort.directoriesFirst, preferencesEntry->values.general.sort.ignoreCase);
			}
			else if(preferencesEntry->nameToken == SYMBOL_SESSION)
			{
				preferences->sessionSavePreferencesOnExit = preferencesEntry->values.general.session.savePreferencesOnExit;
				preferences->sessionSaveLayoutOnExit = preferencesEntry->values.general.session.saveLayoutOnExit;
				preferences->sessionSaveLayoutBookmarks = preferencesEntry->values.general.session.saveLayoutBookmarks;
				preferences->sessionSaveLayoutBrowserAppearance = preferencesEntry->values.general.session.saveLayoutBrowserAppearance;
				preferences->sessionSaveLayoutComboHistory = preferencesEntry->values.general.session.saveLayoutComboHistory;
				preferences->sessionSaveLayoutExpandedDirs = preferencesEntry->values.general.session.saveLayoutExpandedDirs;
				preferences->sessionSaveLayoutGeometry = preferencesEntry->values.general.session.saveLayoutGeometry;
//				printf("Defined GENERAL.SESSION : %d, %d, %d, %d, %d, %d, %d\n", preferencesEntry->values.general.session.savePreferencesOnExit, preferencesEntry->values.general.session.saveLayoutOnExit, preferencesEntry->values.general.session.saveLayoutBookmarks, preferencesEntry->values.general.session.saveLayoutBrowserAppearance, preferencesEntry->values.general.session.saveLayoutComboHistory, preferencesEntry->values.general.session.saveLayoutExpandedDirs, preferencesEntry->values.general.session.saveLayoutGeometry);
			}
			else if(preferencesEntry->nameToken == SYMBOL_SELECTION && 
					preferencesEntry->values.general.selection.pTextElementSeperator != NULL)
			{
				if(preferences->pSelectionTextElementSeperator)
				{
					g_free(preferences->pSelectionTextElementSeperator);
				}
				preferences->pSelectionTextElementSeperator = g_strdup(preferencesEntry->values.general.selection.pTextElementSeperator);
//				printf("Defined GENERAL.SELECTION : %s\n", preferencesEntry->values.general.selection.pTextElementSeperator);
			}
			else if(preferencesEntry->nameToken == SYMBOL_FMWINDOW)
			{
				preferences->fmWindowDefaultPosition = preferencesEntry->values.general.fmwindow.defaultPosition;
				preferences->fmWindowX = preferencesEntry->values.general.fmwindow.x;
				preferences->fmWindowY = preferencesEntry->values.general.fmwindow.y;
				preferences->fmWindowWidth = preferencesEntry->values.general.fmwindow.width;
				preferences->fmWindowHeight = preferencesEntry->values.general.fmwindow.height;
//				printf("Defined GENERAL.FMWINDOW : %d %d %d %d %d\n", preferencesEntry->values.general.fmwindow.defaultPosition, preferencesEntry->values.general.fmwindow.x, preferencesEntry->values.general.fmwindow.y, preferencesEntry->values.general.fmwindow.width, preferencesEntry->values.general.fmwindow.height);
			}
		}
	}
}

static gchar* preferences_getFileName()
{
	gchar	*preferences_filename = g_strjoin("/", pgConfigPath, "preferences", NULL);

	return preferences_filename;
}

void preferences_initialize(struct Preferences *preferences)
{
	if(preferences)
	{
		// default pixmap search paths
		if(pgFullProcessPath)
		{
			preferences_pixmapSearchPathsAdd(preferences, pgFullProcessPath);
		}

		// default accel keys
		accelKeyNew(preferences->pAccelKeys, "SelectionClear", GDK_Escape, 0);
		accelKeyNew(preferences->pAccelKeys, "SelectionSelectDirectory", GDK_a, GDK_CONTROL_MASK);
		accelKeyNew(preferences->pAccelKeys, "SelectionDeselectDirectory", GDK_a, GDK_CONTROL_MASK | GDK_SHIFT_MASK);
		accelKeyNew(preferences->pAccelKeys, "SelectionCopyToXClipboard", GDK_x, GDK_CONTROL_MASK | GDK_SHIFT_MASK);
		accelKeyNew(preferences->pAccelKeys, "SelectionCopyFocusToComboEntry", GDK_f, GDK_CONTROL_MASK);
		accelKeyNew(preferences->pAccelKeys, "DirectoryRefresh", GDK_r, GDK_CONTROL_MASK);
		accelKeyNew(preferences->pAccelKeys, "DirectoryZoom", GDK_z, GDK_CONTROL_MASK);
		accelKeyNew(preferences->pAccelKeys, "DirectoryOpenInNewWindow", GDK_o, GDK_CONTROL_MASK);
		accelKeyNew(preferences->pAccelKeys, "BrowserRefresh", GDK_r, GDK_CONTROL_MASK | GDK_SHIFT_MASK);
		accelKeyNew(preferences->pAccelKeys, "BrowserNew", GDK_b, GDK_CONTROL_MASK);
		accelKeyNew(preferences->pAccelKeys, "BrowserClose", GDK_b, GDK_CONTROL_MASK | GDK_SHIFT_MASK);
		accelKeyNew(preferences->pAccelKeys, "BrowserCollapseEverything", GDK_k, GDK_CONTROL_MASK);
		accelKeyNew(preferences->pAccelKeys, "BrowserUnZoom", GDK_z, GDK_CONTROL_MASK | GDK_SHIFT_MASK);
		accelKeyNew(preferences->pAccelKeys, "BrowserCopyState", GDK_c, GDK_CONTROL_MASK | GDK_SHIFT_MASK);
		accelKeyNew(preferences->pAccelKeys, "BrowserPasteState", GDK_v, GDK_CONTROL_MASK | GDK_SHIFT_MASK);
		accelKeyNew(preferences->pAccelKeys, "ColumnSize", GDK_s, GDK_CONTROL_MASK);
		accelKeyNew(preferences->pAccelKeys, "ColumnUser", GDK_u, GDK_CONTROL_MASK);
		accelKeyNew(preferences->pAccelKeys, "ColumnGroup", GDK_g, GDK_CONTROL_MASK);
		accelKeyNew(preferences->pAccelKeys, "ColumnPermissions", GDK_p, GDK_CONTROL_MASK);
		accelKeyNew(preferences->pAccelKeys, "ColumnModificationTime", GDK_m, GDK_CONTROL_MASK);
		accelKeyNew(preferences->pAccelKeys, "ColumnChangeTime", GDK_h, GDK_CONTROL_MASK);
		accelKeyNew(preferences->pAccelKeys, "About", GDK_a, GDK_MOD1_MASK);
		accelKeyNew(preferences->pAccelKeys, "Preferences", GDK_p, GDK_MOD1_MASK);
		accelKeyNew(preferences->pAccelKeys, "Quit", GDK_q, GDK_MOD1_MASK);
		accelKeyNew(preferences->pAccelKeys, "BookmarkGoTo1", GDK_ampersand, 0);
		accelKeyNew(preferences->pAccelKeys, "BookmarkGoTo2", GDK_eacute, 0);
		accelKeyNew(preferences->pAccelKeys, "BookmarkGoTo3", GDK_quotedbl, 0);
		accelKeyNew(preferences->pAccelKeys, "BookmarkGoTo4", GDK_apostrophe, 0);
		accelKeyNew(preferences->pAccelKeys, "BookmarkGoTo5", GDK_parenleft, 0);
		accelKeyNew(preferences->pAccelKeys, "BookmarkGoTo6", GDK_section, 0);
		accelKeyNew(preferences->pAccelKeys, "BookmarkGoTo7", GDK_egrave, 0);
		accelKeyNew(preferences->pAccelKeys, "BookmarkGoTo8", GDK_exclam, 0);
		accelKeyNew(preferences->pAccelKeys, "BookmarkGoTo9", GDK_ccedilla, 0);
		accelKeyNew(preferences->pAccelKeys, "BookmarkGoTo0", GDK_agrave, 0);
		accelKeyNew(preferences->pAccelKeys, "BookmarkSet1", GDK_ampersand, GDK_CONTROL_MASK);
		accelKeyNew(preferences->pAccelKeys, "BookmarkSet2", GDK_eacute, GDK_CONTROL_MASK);
		accelKeyNew(preferences->pAccelKeys, "BookmarkSet3", GDK_quotedbl, GDK_CONTROL_MASK);
		accelKeyNew(preferences->pAccelKeys, "BookmarkSet4", GDK_apostrophe, GDK_CONTROL_MASK);
		accelKeyNew(preferences->pAccelKeys, "BookmarkSet5", GDK_parenleft, GDK_CONTROL_MASK);
		accelKeyNew(preferences->pAccelKeys, "BookmarkSet6", GDK_section, GDK_CONTROL_MASK);
		accelKeyNew(preferences->pAccelKeys, "BookmarkSet7", GDK_egrave, GDK_CONTROL_MASK);
		accelKeyNew(preferences->pAccelKeys, "BookmarkSet8", GDK_exclam, GDK_CONTROL_MASK);
		accelKeyNew(preferences->pAccelKeys, "BookmarkSet9", GDK_ccedilla, GDK_CONTROL_MASK);
		accelKeyNew(preferences->pAccelKeys, "BookmarkSet0", GDK_agrave, GDK_CONTROL_MASK);
		accelKeyNew(preferences->pAccelKeys, "BookmarkClear1", GDK_1, GDK_CONTROL_MASK | GDK_SHIFT_MASK);
		accelKeyNew(preferences->pAccelKeys, "BookmarkClear2", GDK_2, GDK_CONTROL_MASK | GDK_SHIFT_MASK);
		accelKeyNew(preferences->pAccelKeys, "BookmarkClear3", GDK_3, GDK_CONTROL_MASK | GDK_SHIFT_MASK);
		accelKeyNew(preferences->pAccelKeys, "BookmarkClear4", GDK_4, GDK_CONTROL_MASK | GDK_SHIFT_MASK);
		accelKeyNew(preferences->pAccelKeys, "BookmarkClear5", GDK_5, GDK_CONTROL_MASK | GDK_SHIFT_MASK);
		accelKeyNew(preferences->pAccelKeys, "BookmarkClear6", GDK_6, GDK_CONTROL_MASK | GDK_SHIFT_MASK);
		accelKeyNew(preferences->pAccelKeys, "BookmarkClear7", GDK_7, GDK_CONTROL_MASK | GDK_SHIFT_MASK);
		accelKeyNew(preferences->pAccelKeys, "BookmarkClear8", GDK_8, GDK_CONTROL_MASK | GDK_SHIFT_MASK);
		accelKeyNew(preferences->pAccelKeys, "BookmarkClear9", GDK_9, GDK_CONTROL_MASK | GDK_SHIFT_MASK);
		accelKeyNew(preferences->pAccelKeys, "BookmarkClear0", GDK_0, GDK_CONTROL_MASK | GDK_SHIFT_MASK);

		// default rowstyles
		rowStyleNew(preferences->pRowStyles, preferences->pPixmapSearchPaths, "RegularFile", 0x0000, 0x0000, 0x0000, 0xE800, 0xE800, 0xE800, "regular_file.xpm", NULL, FALSE);
		rowStyleNew(preferences->pRowStyles, preferences->pPixmapSearchPaths, "DirectoryFilled", 0x0000, 0x0000, 0x6666, 0xD800, 0xD800, 0xD800, "directory_filled.xpm", "directory_filled_opened.xpm", FALSE);
		rowStyleNew(preferences->pRowStyles, preferences->pPixmapSearchPaths, "DirectoryEmpty", 0x6666, 0x6666, 0x6666, 0xD800, 0xD800, 0xD800, "directory_empty.xpm", NULL, FALSE);
		rowStyleNew(preferences->pRowStyles, preferences->pPixmapSearchPaths, "DirectoryInaccessible", 0x9999, 0x3333, 0x3333, 0xD800, 0xD800, 0xD800, "directory_inaccessible.xpm", NULL, FALSE);
		rowStyleNew(preferences->pRowStyles, preferences->pPixmapSearchPaths, "CharacterDevice", 0x4444, 0x4444, 0x0000, 0xE800, 0xE800, 0xE800, "character_device.xpm", NULL, FALSE);
		rowStyleNew(preferences->pRowStyles, preferences->pPixmapSearchPaths, "BlockDevice", 0x4444, 0x4444, 0x0000, 0xE800, 0xE800, 0xE800, "block_device.xpm", NULL, FALSE);
		rowStyleNew(preferences->pRowStyles, preferences->pPixmapSearchPaths, "BrokenLink", 0x6666, 0x0000, 0x0000, 0xE800, 0xE800, 0xE800, "broken_link.xpm", NULL, FALSE);
		rowStyleNew(preferences->pRowStyles, preferences->pPixmapSearchPaths, "Socket", 0x0000, 0x4444, 0x0000, 0xE800, 0xE800, 0xE800, "socket.xpm", NULL, FALSE);
		rowStyleNew(preferences->pRowStyles, preferences->pPixmapSearchPaths, "Fifo", 0x4444, 0x1111, 0x4444, 0xE800, 0xE800, 0xE800, "fifo.xpm", NULL, FALSE);
		rowStyleNew(preferences->pRowStyles, preferences->pPixmapSearchPaths, "Unknown", 0x6666, 0x0000, 0x0000, 0xE800, 0xE800, 0xE800, "unknown.xpm", NULL, FALSE);
		rowStyleNew(preferences->pRowStyles, preferences->pPixmapSearchPaths, "Selected", 0xFFFF, 0xFFFF, 0xFFFF, 0x6666, 0x6666, 0x9999, NULL, NULL, FALSE);
		rowStyleNew(preferences->pRowStyles, preferences->pPixmapSearchPaths, "SelectedFocus", 0xFFFF, 0xFFFF, 0xFFFF, 0x3333, 0x3333, 0x6666, NULL, NULL, FALSE);
		rowStyleNew(preferences->pRowStyles, preferences->pPixmapSearchPaths, "Focus", 0xFFFF, 0xFFFF, 0xFFFF, 0x6666, 0x6666, 0x6666, NULL, NULL, FALSE);
		rowStyleNew(preferences->pRowStyles, preferences->pPixmapSearchPaths, "Bookmark", 0xFFFF, 0xFFFF, 0xFFFF, 0x9999, 0x3333, 0x3333, NULL, NULL, FALSE);

		// default filetypes
		fileTypeNew(&preferences->pFileTypes, preferences->pRowStyles, FALSE, REGULAR_FILE, "RegularFile");
		fileTypeNew(&preferences->pFileTypes, preferences->pRowStyles, FALSE, DIRECTORY_FILLED, "DirectoryFilled");
		fileTypeNew(&preferences->pFileTypes, preferences->pRowStyles, FALSE, DIRECTORY_EMPTY, "DirectoryEmpty");
		fileTypeNew(&preferences->pFileTypes, preferences->pRowStyles, FALSE, DIRECTORY_INACCESSIBLE, "DirectoryInaccessible");
		fileTypeNew(&preferences->pFileTypes, preferences->pRowStyles, FALSE, CHARACTER_DEVICE, "CharacterDevice");
		fileTypeNew(&preferences->pFileTypes, preferences->pRowStyles, FALSE, BLOCK_DEVICE, "BlockDevice");
		fileTypeNew(&preferences->pFileTypes, preferences->pRowStyles, FALSE, BROKEN_LINK, "BrokenLink");
		fileTypeNew(&preferences->pFileTypes, preferences->pRowStyles, FALSE, SOCKET, "Socket");
		fileTypeNew(&preferences->pFileTypes, preferences->pRowStyles, FALSE, FIFO, "Fifo");
		fileTypeNew(&preferences->pFileTypes, preferences->pRowStyles, FALSE, UNKNOWN, "Unknown");

		// default ctree columns
        ctreeColumnAdd(preferences, COLUMN_BOOKMARK, GTK_JUSTIFY_CENTER);
        ctreeColumnAdd(preferences, COLUMN_FILENAME, GTK_JUSTIFY_CENTER);
        ctreeColumnAdd(preferences, COLUMN_SIZE, GTK_JUSTIFY_RIGHT);
        ctreeColumnAdd(preferences, COLUMN_USER, GTK_JUSTIFY_CENTER);
        ctreeColumnAdd(preferences, COLUMN_GROUP, GTK_JUSTIFY_CENTER);
        ctreeColumnAdd(preferences, COLUMN_PERMISSIONS, GTK_JUSTIFY_LEFT );
        ctreeColumnAdd(preferences, COLUMN_MODIFICATION_TIME, GTK_JUSTIFY_CENTER);
        ctreeColumnAdd(preferences, COLUMN_CHANGE_TIME, GTK_JUSTIFY_CENTER);
		
		// default character browsing parameters
		preferences->characterBrowsingEnabled = TRUE;
		preferences->characterBrowsingEnableSpacebar = FALSE;
		preferences->characterBrowsingIgnoreCase = TRUE;

		// default session parameters
		preferences->sessionSavePreferencesOnExit = TRUE;
		preferences->sessionSaveLayoutOnExit = TRUE;
		preferences->sessionSaveLayoutBookmarks = TRUE;
		preferences->sessionSaveLayoutComboHistory = TRUE;
		preferences->sessionSaveLayoutExpandedDirs = TRUE;
		preferences->sessionSaveLayoutGeometry = TRUE;
		preferences->sessionSaveLayoutBrowserAppearance = TRUE;
		
		// default sort parameters
		preferences->sortDirectoriesFirst = TRUE;
		preferences->sortIgnoreCase = TRUE;

		// default selection text parameters
		preferences->pSelectionTextElementSeperator = g_strdup(" ");

		// default fmwindow size
		preferences->fmWindowDefaultPosition = FALSE;
		preferences->fmWindowX = 0;
		preferences->fmWindowY = 0;
		preferences->fmWindowWidth = 400;
		preferences->fmWindowHeight = 600;

		// default maximum combo box entries
        preferences->comboPopdownLength = 20;
	}
}

void preferences_load(struct Preferences *preferences)
{
	gchar							*preferences_filename = NULL;
	int								preferences_file_descriptor = -1;
	GScanner						*scanner = NULL;
	gchar							*cset_identifier_nth = NULL;
	const struct SymbolDefinition	*symbol_table_pointer = psSymbolTablePreferences;
	guint							expected_token = G_TOKEN_NONE;


	preferences_filename = preferences_getFileName();
	if(preferences_filename)
	{
		preferences_file_descriptor = open(preferences_filename, O_RDONLY);
		if(preferences_file_descriptor != -1)
		{
			scanner = g_scanner_new(NULL);
			if(scanner)
			{
                scanner->config->case_sensitive = FALSE;
                scanner->config->numbers_2_int = TRUE;
                scanner->config->symbol_2_token = TRUE;
				if(scanner->config->cset_identifier_nth)
				{
                    cset_identifier_nth = g_strjoin("", scanner->config->cset_identifier_nth, ".", NULL);
                    scanner->config->cset_identifier_nth = cset_identifier_nth;
				}


				while(symbol_table_pointer->name)
				{
//					printf("SYMBOL %s=%d\n", symbol_table_pointer->name, symbol_table_pointer->token);
					g_scanner_add_symbol(scanner, symbol_table_pointer->name, GINT_TO_POINTER(symbol_table_pointer->token));
                    symbol_table_pointer++;
				}

				g_scanner_input_file(scanner, preferences_file_descriptor);

				// parse the file
				do
				{
					expected_token = preferences_parseSymbol(preferences, scanner);
					g_scanner_peek_next_token(scanner);
				}
				while(expected_token != G_TOKEN_NONE && expected_token != G_TOKEN_EOF && expected_token != G_TOKEN_ERROR);

				g_scanner_destroy(scanner);
			}
			else
			{
				printf("preferences_load : Can't create a new GScanner.\n");
			}
			close(preferences_file_descriptor);
			g_free(preferences_filename);
		}
 		else
		{
//			printf("preferences_load : Can't open the preferences file '%s' for reading.\n", preferences_filename);
			g_free(preferences_filename);
		}
	}
	else
	{
		printf("preferences_load : Can't create the preferences filename\n");
	}
}

struct Preferences* preferences_new()
{
	struct Preferences	*preferences = NULL;

	preferences = g_new(struct Preferences, 1);
	if(preferences)
	{
		// initialize the preference values
		preferences->pAccelKeys = NULL;
		preferences->pCtreeColumns = NULL;
		preferences->pFileTypes = NULL;
		preferences->pRowStyles = NULL;
		preferences->pPixmapSearchPaths = NULL;
		preferences->pSelectionTextElementSeperator = NULL;
		preferences->characterBrowsingEnabled = FALSE;
		preferences->characterBrowsingEnableSpacebar = FALSE;
		preferences->characterBrowsingIgnoreCase = FALSE;
        preferences->comboPopdownLength = 0;
		preferences->fmWindowDefaultPosition = FALSE;
		preferences->fmWindowX = -1;
		preferences->fmWindowY = -1;
		preferences->fmWindowWidth = 400;
		preferences->fmWindowHeight = 600;
		preferences->sessionSavePreferencesOnExit = FALSE;
		preferences->sessionSaveLayoutOnExit = FALSE;
		preferences->sortDirectoriesFirst = FALSE;
		preferences->sortIgnoreCase = FALSE;

		preferences->pAccelKeys = g_hash_table_new(g_str_hash, g_str_equal);
		preferences->pRowStyles = g_hash_table_new(g_str_hash, g_str_equal);
	}

	return preferences;
}

static guint preferences_parseSymbol(struct Preferences *preferences, GScanner *scanner)
{
	static struct PreferencesEntry	*preferences_entry = NULL;
	guint  			 				previous_token = G_TOKEN_NONE;

	g_scanner_get_next_token(scanner);
//	printf("scanner->token %d\n", scanner->token);
	
	// check for end of file and for error
	if(scanner->token == G_TOKEN_EOF || scanner->token == G_TOKEN_ERROR)
	{
		preferences_entryProcess(preferences, preferences_entry);
		if(preferences_entry)
		{
			preferences_entryDestroy(preferences_entry);
		}
		return scanner->token;
	}
	else if(scanner->token == G_TOKEN_COMMENT_SINGLE || scanner->token == G_TOKEN_COMMENT_MULTI)
	{
		return scanner->token;
	}
	// get section
	else if(scanner->token == G_TOKEN_LEFT_BRACE)
	{
		g_scanner_get_next_token(scanner);
//		printf("scanner->token %d\n", scanner->token);
		if(scanner->token == SYMBOL_CTREE_COLUMNS ||
           scanner->token == SYMBOL_ACCEL_KEYS ||
           scanner->token == SYMBOL_FILE_TYPES ||
           scanner->token == SYMBOL_ROW_STYLES ||
		   scanner->token == SYMBOL_GENERAL)
		{
			g_scanner_peek_next_token(scanner);
			if(scanner->next_token == G_TOKEN_RIGHT_BRACE)
			{
                preferences_entryProcess(preferences, preferences_entry);
				if(preferences_entry)
				{
					preferences_entryDestroy(preferences_entry);
				}
				preferences_entry = preferences_entryNew();
				if(!preferences_entry)
				{
					printf("preferences_parseSymbol : FATAL ERROR, can't allocate a new preferences entry, quitting.\n");
					exit(1);
				}
				preferences_entry->section = scanner->token;

				g_scanner_get_next_token(scanner);
//				printf("scanner->token %d\n", scanner->token);

//				printf("Started new SECTION %d\n", preferences_entry->section);
				
				return preferences_entry->section;
			}
		}
		// get name
		else if(scanner->token == G_TOKEN_IDENTIFIER || 
				scanner->token == SYMBOL_CTREE_COLUMN ||
				scanner->token == SYMBOL_FILE_TYPE ||
				scanner->token == SYMBOL_CHARACTERBROWSING ||
				scanner->token == SYMBOL_COMBOBOX ||
				scanner->token == SYMBOL_SORT ||
				scanner->token == SYMBOL_SESSION ||
				scanner->token == SYMBOL_SELECTION ||
				scanner->token == SYMBOL_FMWINDOW)
		{
			g_scanner_peek_next_token(scanner);
			if(scanner->next_token == G_TOKEN_RIGHT_BRACE)
			{
                preferences_entryProcess(preferences, preferences_entry);
				// initialize a new preference entry
				preferences_entryClear(preferences_entry);
				preferences_entry->nameToken = scanner->token;
				if(preferences_entry->nameToken == G_TOKEN_IDENTIFIER)
				{
					preferences_entry->pNameString = g_strdup(scanner->value.v_string);
				}

				g_scanner_get_next_token(scanner);
//				printf("scanner->token %d\n", scanner->token);

//				printf("Started new NAME %d %s\n", preferences_entry->nameToken, preferences_entry->pNameString);

				return preferences_entry->nameToken;
			}
		}
	}
	// get the name values
	else if(preferences_entry->nameToken != G_TOKEN_NONE)
	{
		if(preferences_entry->section == SYMBOL_CTREE_COLUMNS)
		{
			if(scanner->token == SYMBOL_TYPE || 
			   scanner->token == SYMBOL_ALIGNMENT)
			{
				g_scanner_peek_next_token(scanner);
				if(scanner->next_token == G_TOKEN_EQUAL_SIGN)
				{
					previous_token = scanner->token;
					g_scanner_get_next_token(scanner);
//					printf("scanner->token %d\n", scanner->token);
					g_scanner_get_next_token(scanner);
//					printf("scanner->token %d\n", scanner->token);
					switch(previous_token)
					{
						case SYMBOL_TYPE :
							if(scanner->token == G_TOKEN_INT)
							{
								preferences_entry->values.ctreeColumn.type = scanner->value.v_int;
								return previous_token;
							}
							break;
						case SYMBOL_ALIGNMENT :
							if(scanner->token == G_TOKEN_INT)
							{
								preferences_entry->values.ctreeColumn.alignment = scanner->value.v_int;
								return previous_token;
							}
							break;
					}
				}
			}
		}
		else if(preferences_entry->section == SYMBOL_ACCEL_KEYS)
		{
			if(scanner->token == SYMBOL_KEY || 
			   scanner->token == SYMBOL_MODS)
			{
				g_scanner_peek_next_token(scanner);
				if(scanner->next_token == G_TOKEN_EQUAL_SIGN)
				{
					previous_token = scanner->token;
					g_scanner_get_next_token(scanner);
//					printf("scanner->token %d\n", scanner->token);
					g_scanner_get_next_token(scanner);
//					printf("scanner->token %d\n", scanner->token);
					switch(previous_token)
					{
						case SYMBOL_KEY :
							if(scanner->token == G_TOKEN_INT)
							{
								preferences_entry->values.accelKey.key = scanner->value.v_int;
								return previous_token;
							}
							break;
						case SYMBOL_MODS :
							if(scanner->token == G_TOKEN_INT)
							{
								preferences_entry->values.accelKey.mods = scanner->value.v_int;
								return previous_token;
							}
							break;
					}
				}
			}
		}
		else if(preferences_entry->section == SYMBOL_ROW_STYLES)
		{
			if(scanner->token == SYMBOL_FGRED || 
			   scanner->token == SYMBOL_FGGREEN ||
			   scanner->token == SYMBOL_FGBLUE ||
			   scanner->token == SYMBOL_BGRED ||
			   scanner->token == SYMBOL_BGGREEN ||
			   scanner->token == SYMBOL_BGBLUE ||
			   scanner->token == SYMBOL_FILENAMECLOSED ||
			   scanner->token == SYMBOL_FILENAMEOPENED ||
			   scanner->token == SYMBOL_DELETABLE)
			{
				g_scanner_peek_next_token(scanner);
				if(scanner->next_token == G_TOKEN_EQUAL_SIGN)
				{
					previous_token = scanner->token;
					g_scanner_get_next_token(scanner);
//					printf("scanner->token %d\n", scanner->token);
					g_scanner_get_next_token(scanner);
//					printf("scanner->token %d\n", scanner->token);
					switch(previous_token)
					{
						case SYMBOL_FGRED :
							if(scanner->token == G_TOKEN_INT)
							{
								preferences_entry->values.rowStyle.fgRed = scanner->value.v_int;
								return previous_token;
							}
							break;
						case SYMBOL_FGGREEN :
							if(scanner->token == G_TOKEN_INT)
							{
								preferences_entry->values.rowStyle.fgGreen = scanner->value.v_int;
								return previous_token;
							}
							break;
						case SYMBOL_FGBLUE :
							if(scanner->token == G_TOKEN_INT)
							{
								preferences_entry->values.rowStyle.fgBlue = scanner->value.v_int;
								return previous_token;
							}
							break;
						case SYMBOL_BGRED :
							if(scanner->token == G_TOKEN_INT)
							{
								preferences_entry->values.rowStyle.bgRed = scanner->value.v_int;
								return previous_token;
							}
							break;
						case SYMBOL_BGGREEN :
							if(scanner->token == G_TOKEN_INT)
							{
								preferences_entry->values.rowStyle.bgGreen = scanner->value.v_int;
								return previous_token;
							}
							break;
						case SYMBOL_BGBLUE :
							if(scanner->token == G_TOKEN_INT)
							{
								preferences_entry->values.rowStyle.bgBlue = scanner->value.v_int;
								return previous_token;
							}
							break;
						case SYMBOL_FILENAMECLOSED :
							if(scanner->token == G_TOKEN_STRING)
							{
								preferences_entry->values.rowStyle.pFilenameClosed = g_strdup(scanner->value.v_string);
								return previous_token;
							}
							break;
						case SYMBOL_FILENAMEOPENED :
							if(scanner->token == G_TOKEN_STRING)
							{
								preferences_entry->values.rowStyle.pFilenameOpened = g_strdup(scanner->value.v_string);
								return previous_token;
							}
							break;
						case SYMBOL_DELETABLE :
							if(scanner->token == G_TOKEN_INT)
							{
								preferences_entry->values.rowStyle.deletable = scanner->value.v_int;
								return previous_token;
							}
							break;
					}
				}
			}
		}
		else if(preferences_entry->section == SYMBOL_FILE_TYPES)
		{
			if(scanner->token == SYMBOL_DELETABLE ||
               scanner->token == SYMBOL_BASICFILETYPE ||
			   scanner->token == SYMBOL_ROWSTYLE)
			{
				g_scanner_peek_next_token(scanner);
				if(scanner->next_token == G_TOKEN_EQUAL_SIGN)
				{
					previous_token = scanner->token;
					g_scanner_get_next_token(scanner);
//					printf("scanner->token %d\n", scanner->token);
					g_scanner_get_next_token(scanner);
//					printf("scanner->token %d\n", scanner->token);
					switch(previous_token)
					{
						case SYMBOL_DELETABLE :
							if(scanner->token == G_TOKEN_INT)
							{
								preferences_entry->values.fileType.deletable = scanner->value.v_int;
								return previous_token;
							}
							break;
						case SYMBOL_BASICFILETYPE :
							if(scanner->token == G_TOKEN_INT)
							{
								preferences_entry->values.fileType.basicType = scanner->value.v_int;
								return previous_token;
							}
							break;
						case SYMBOL_ROWSTYLE :
							if(scanner->token == G_TOKEN_STRING)
							{
								preferences_entry->values.fileType.pRowStyle = g_strdup(scanner->value.v_string);
								return previous_token;
							}
							break;
					}
				}
			}
		}
		else if(preferences_entry->section == SYMBOL_GENERAL && preferences_entry->nameToken == SYMBOL_CHARACTERBROWSING)
		{
			if(scanner->token == SYMBOL_ENABLED || scanner->token == SYMBOL_ENABLESPACEBAR || scanner->token == SYMBOL_IGNORECASE)
			{
				g_scanner_peek_next_token(scanner);
				if(scanner->next_token == G_TOKEN_EQUAL_SIGN)
				{
					previous_token = scanner->token;
					g_scanner_get_next_token(scanner);
//					printf("scanner->token %d\n", scanner->token);
					g_scanner_get_next_token(scanner);
//					printf("scanner->token %d\n", scanner->token);
					switch(previous_token)
					{
						case SYMBOL_ENABLED :
							if(scanner->token == G_TOKEN_INT)
							{
								preferences_entry->values.general.characterBrowsing.enabled = scanner->value.v_int;
								return previous_token;
							}
							break;
						case SYMBOL_ENABLESPACEBAR :
							if(scanner->token == G_TOKEN_INT)
							{
								preferences_entry->values.general.characterBrowsing.enableSpacebar = scanner->value.v_int;
								return previous_token;
							}
							break;
						case SYMBOL_IGNORECASE :
							if(scanner->token == G_TOKEN_INT)
							{
								preferences_entry->values.general.characterBrowsing.ignoreCase = scanner->value.v_int;
								return previous_token;
							}
							break;
					}
				}
			}
		}
		else if(preferences_entry->section == SYMBOL_GENERAL && preferences_entry->nameToken == SYMBOL_COMBOBOX)
		{
			if(scanner->token == SYMBOL_POPDOWNLENGTH)
			{
				g_scanner_peek_next_token(scanner);
				if(scanner->next_token == G_TOKEN_EQUAL_SIGN)
				{
					previous_token = scanner->token;
					g_scanner_get_next_token(scanner);
//					printf("scanner->token %d\n", scanner->token);
					g_scanner_get_next_token(scanner);
//					printf("scanner->token %d\n", scanner->token);
					switch(previous_token)
					{
						case SYMBOL_POPDOWNLENGTH :
							if(scanner->token == G_TOKEN_INT)
							{
								preferences_entry->values.general.comboBox.popdownLength = scanner->value.v_int;
								return previous_token;
							}
							break;
					}
				}
			}
		}
		else if(preferences_entry->section == SYMBOL_GENERAL && preferences_entry->nameToken == SYMBOL_SORT)
		{
			if(scanner->token == SYMBOL_DIRECTORIESFIRST || scanner->token == SYMBOL_IGNORECASE)
			{
				g_scanner_peek_next_token(scanner);
				if(scanner->next_token == G_TOKEN_EQUAL_SIGN)
				{
					previous_token = scanner->token;
					g_scanner_get_next_token(scanner);
//					printf("scanner->token %d\n", scanner->token);
					g_scanner_get_next_token(scanner);
//					printf("scanner->token %d\n", scanner->token);
					switch(previous_token)
					{
						case SYMBOL_DIRECTORIESFIRST :
							if(scanner->token == G_TOKEN_INT)
							{
								preferences_entry->values.general.sort.directoriesFirst = scanner->value.v_int;
								return previous_token;
							}
							break;
						case SYMBOL_IGNORECASE :
							if(scanner->token == G_TOKEN_INT)
							{
								preferences_entry->values.general.sort.ignoreCase = scanner->value.v_int;
								return previous_token;
							}
							break;
					}
				}
			}
		}
		else if(preferences_entry->section == SYMBOL_GENERAL && preferences_entry->nameToken == SYMBOL_SESSION)
		{
			if(scanner->token == SYMBOL_SAVEPREFERENCESONEXIT ||
			   scanner->token == SYMBOL_SAVELAYOUTONEXIT ||
               scanner->token == SYMBOL_SAVELAYOUTBOOKMARKS ||
               scanner->token == SYMBOL_SAVELAYOUTCOMBOHISTORY ||
               scanner->token == SYMBOL_SAVELAYOUTONBROWSERAPPEARANCE ||
               scanner->token == SYMBOL_SAVELAYOUTEXPANDEDDIRS ||
               scanner->token == SYMBOL_SAVELAYOUTONGEOMETRY)
			{
				g_scanner_peek_next_token(scanner);
				if(scanner->next_token == G_TOKEN_EQUAL_SIGN)
				{
					previous_token = scanner->token;
					g_scanner_get_next_token(scanner);
//					printf("scanner->token %d\n", scanner->token);
					g_scanner_get_next_token(scanner);
//					printf("scanner->token %d\n", scanner->token);
					switch(previous_token)
					{
						case SYMBOL_SAVEPREFERENCESONEXIT :
							if(scanner->token == G_TOKEN_INT)
							{
								preferences_entry->values.general.session.savePreferencesOnExit = scanner->value.v_int;
								return previous_token;
							}
							break;
						case SYMBOL_SAVELAYOUTONEXIT :
							if(scanner->token == G_TOKEN_INT)
							{
								preferences_entry->values.general.session.saveLayoutOnExit = scanner->value.v_int;
								return previous_token;
							}
							break;
						case SYMBOL_SAVELAYOUTBOOKMARKS :
							if(scanner->token == G_TOKEN_INT)
							{
								preferences_entry->values.general.session.saveLayoutBookmarks = scanner->value.v_int;
								return previous_token;
							}
							break;
						case SYMBOL_SAVELAYOUTONBROWSERAPPEARANCE :
							if(scanner->token == G_TOKEN_INT)
							{
								preferences_entry->values.general.session.saveLayoutBrowserAppearance = scanner->value.v_int;
								return previous_token;
							}
							break;
						case SYMBOL_SAVELAYOUTCOMBOHISTORY :
							if(scanner->token == G_TOKEN_INT)
							{
								preferences_entry->values.general.session.saveLayoutComboHistory = scanner->value.v_int;
								return previous_token;
							}
							break;
						case SYMBOL_SAVELAYOUTEXPANDEDDIRS :
							if(scanner->token == G_TOKEN_INT)
							{
								preferences_entry->values.general.session.saveLayoutExpandedDirs = scanner->value.v_int;
								return previous_token;
							}
							break;
						case SYMBOL_SAVELAYOUTONGEOMETRY :
							if(scanner->token == G_TOKEN_INT)
							{
								preferences_entry->values.general.session.saveLayoutGeometry = scanner->value.v_int;
								return previous_token;
							}
							break;
					}
				}
			}
		}
		else if(preferences_entry->section == SYMBOL_GENERAL && preferences_entry->nameToken == SYMBOL_SELECTION)
		{
			if(scanner->token == SYMBOL_TEXT_ELEMENT_SEPERATOR)
			{
				g_scanner_peek_next_token(scanner);
				if(scanner->next_token == G_TOKEN_EQUAL_SIGN)
				{
					previous_token = scanner->token;
					g_scanner_get_next_token(scanner);
//					printf("scanner->token %d\n", scanner->token);
					g_scanner_get_next_token(scanner);
//					printf("scanner->token %d\n", scanner->token);
					switch(previous_token)
					{
						case SYMBOL_TEXT_ELEMENT_SEPERATOR :
							if(scanner->token == G_TOKEN_STRING)
							{
								preferences_entry->values.general.selection.pTextElementSeperator = g_strdup(scanner->value.v_string);
								return previous_token;
							}
							break;
					}
				}
			}
		}
		else if(preferences_entry->section == SYMBOL_GENERAL && preferences_entry->nameToken == SYMBOL_FMWINDOW)
		{
			if(scanner->token == SYMBOL_DEFAULT_POSITION ||
			   scanner->token == SYMBOL_X ||
               scanner->token == SYMBOL_Y ||
               scanner->token == SYMBOL_WIDTH ||
			   scanner->token == SYMBOL_HEIGHT)
			{
				g_scanner_peek_next_token(scanner);
				if(scanner->next_token == G_TOKEN_EQUAL_SIGN)
				{
					previous_token = scanner->token;
					g_scanner_get_next_token(scanner);
//					printf("scanner->token %d\n", scanner->token);
					g_scanner_get_next_token(scanner);
//					printf("scanner->token %d\n", scanner->token);
					switch(previous_token)
					{
						case SYMBOL_DEFAULT_POSITION :
							if(scanner->token == G_TOKEN_INT)
							{
								preferences_entry->values.general.fmwindow.defaultPosition = scanner->value.v_int;
								return previous_token;
							}
							break;
						case SYMBOL_X :
							if(scanner->token == G_TOKEN_INT)
							{
								preferences_entry->values.general.fmwindow.x = scanner->value.v_int;
								return previous_token;
							}
							break;
						case SYMBOL_Y :
							if(scanner->token == G_TOKEN_INT)
							{
								preferences_entry->values.general.fmwindow.y = scanner->value.v_int;
								return previous_token;
							}
							break;
						case SYMBOL_WIDTH :
							if(scanner->token == G_TOKEN_INT)
							{
								preferences_entry->values.general.fmwindow.width = scanner->value.v_int;
								return previous_token;
							}
							break;
						case SYMBOL_HEIGHT :
							if(scanner->token == G_TOKEN_INT)
							{
								preferences_entry->values.general.fmwindow.height = scanner->value.v_int;
								return previous_token;
							}
							break;
					}
				}
			}
		}
	}

	return G_TOKEN_NONE;
}

void preferences_pixmapSearchPathsAdd(struct Preferences *preferences, gchar *path)
{
	if(preferences)
	{
		preferences->pPixmapSearchPaths = g_list_prepend(preferences->pPixmapSearchPaths, g_strdup(path));
	}
}

void preferences_pixmapSearchPathsRemoveAll(struct Preferences *preferences)
{
	GList	*list_entry = NULL;

	if(preferences)
	{
		while(g_list_length(preferences->pPixmapSearchPaths))
		{
			list_entry = g_list_nth(preferences->pPixmapSearchPaths, 0);
			preferences->pPixmapSearchPaths = g_list_remove_link(preferences->pPixmapSearchPaths, list_entry); 
			g_free(list_entry->data);
			g_list_free_1(list_entry);
		}
	}
}

void preferences_save(struct Preferences *preferences)
{
	gchar	*preferences_filename = NULL;

	guint	number_of_accel_keys = 0;
	gchar	**accel_keys_names = NULL;
	gchar	**accel_keys_names_iterator = NULL;
	
	guint	number_of_row_styles = 0;
	gchar	**row_styles_names = NULL;
	gchar	**row_styles_names_iterator = NULL;
	
	guint			number_of_file_types = 0;
	guint			file_type_id = 0;
	struct FileType	*file_type = NULL;
	
	guint				number_of_ctree_columns = 0;
	guint				ctree_column_id = 0;
	struct CtreeColumn	*ctree_column = NULL;
	
	if(preferences)
	{
		preferences_filename = preferences_getFileName();
		if(preferences_filename)
		{
			unlink(preferences_filename);
			pgConfigFileStream = fopen(preferences_filename, "w");
			if(pgConfigFileStream)
			{
				// save the ctree columns
				fprintf(pgConfigFileStream, "[CTREE_COLUMNS]\n");
				number_of_ctree_columns = g_list_length(preferences->pCtreeColumns);
				for(ctree_column_id = 0; ctree_column_id < number_of_ctree_columns; ctree_column_id++)
				{
					ctree_column = (struct CtreeColumn*)g_list_nth_data(preferences->pCtreeColumns, ctree_column_id);
					
					fprintf(pgConfigFileStream, "\t[ctree_column]\n");
					fprintf(pgConfigFileStream, "\t\tType=%d\n", ctree_column->type);
					fprintf(pgConfigFileStream, "\t\tAlignment=%d\n", ctree_column->alignment);
				}
				
				// save the accelkeys
				fprintf(pgConfigFileStream, "[ACCEL_KEYS]\n");
				number_of_accel_keys = g_hash_table_size(preferences->pAccelKeys);
				accel_keys_names = g_new(gchar*, number_of_accel_keys);
				accel_keys_names_iterator = accel_keys_names;
				g_hash_table_foreach(preferences->pAccelKeys, preferences_storeAccelKey, &accel_keys_names_iterator);
                g_free(accel_keys_names);
	
				// save the row styles
				fprintf(pgConfigFileStream, "[ROW_STYLES]\n");
				number_of_row_styles = g_hash_table_size(preferences->pRowStyles);
				row_styles_names = g_new(gchar*, number_of_row_styles);
				row_styles_names_iterator = row_styles_names;
				g_hash_table_foreach(preferences->pRowStyles, preferences_storeRowStyle, &row_styles_names_iterator);
				g_free(row_styles_names);
	
				// save the file types
				fprintf(pgConfigFileStream, "[FILE_TYPES]\n");
				number_of_file_types = g_list_length(preferences->pFileTypes);
				for(file_type_id = 0; file_type_id < number_of_file_types; file_type_id++)
				{
					file_type = (struct FileType*)g_list_nth_data(preferences->pFileTypes, file_type_id);
					
					fprintf(pgConfigFileStream, "\t[file_type]\n");
					fprintf(pgConfigFileStream, "\t\tDeletable=%d\n", file_type->deletable);
					fprintf(pgConfigFileStream, "\t\tBasicFileType=%d\n", file_type->basicType);
					fprintf(pgConfigFileStream, "\t\tRowStyle='%s'\n", file_type->pRowStyle->pName);
				}
	
				// save the general preferences
				fprintf(pgConfigFileStream, "[GENERAL]\n");
				fprintf(pgConfigFileStream, "\t[characterbrowsing]\n");
				fprintf(pgConfigFileStream, "\t\tEnabled=%d\n", preferences->characterBrowsingEnabled);
				fprintf(pgConfigFileStream, "\t\tEnableSpacebar=%d\n", preferences->characterBrowsingEnableSpacebar);
				fprintf(pgConfigFileStream, "\t\tIgnoreCase=%d\n", preferences->characterBrowsingIgnoreCase);
				fprintf(pgConfigFileStream, "\t[combobox]\n");
				fprintf(pgConfigFileStream, "\t\tPopdownLength=%d\n", preferences->comboPopdownLength);
				fprintf(pgConfigFileStream, "\t[session]\n");
				fprintf(pgConfigFileStream, "\t\tSavePreferencesOnExit=%d\n", preferences->sessionSavePreferencesOnExit);
				fprintf(pgConfigFileStream, "\t\tSaveLayoutOnExit=%d\n", preferences->sessionSaveLayoutOnExit);
				fprintf(pgConfigFileStream, "\t\tSaveLayoutBookmarks=%d\n", preferences->sessionSaveLayoutBookmarks);
				fprintf(pgConfigFileStream, "\t\tSaveLayoutBrowserAppearance=%d\n", preferences->sessionSaveLayoutBrowserAppearance);
				fprintf(pgConfigFileStream, "\t\tSaveLayoutComboHistory=%d\n", preferences->sessionSaveLayoutComboHistory);
				fprintf(pgConfigFileStream, "\t\tSaveLayoutExpandedDirs=%d\n", preferences->sessionSaveLayoutExpandedDirs);
				fprintf(pgConfigFileStream, "\t\tSaveLayoutGeometry=%d\n", preferences->sessionSaveLayoutGeometry);
				fprintf(pgConfigFileStream, "\t[sort]\n");
				fprintf(pgConfigFileStream, "\t\tDirectoriesFirst=%d\n", preferences->sortDirectoriesFirst);
				fprintf(pgConfigFileStream, "\t\tIgnoreCase=%d\n", preferences->sortDirectoriesFirst);
				fprintf(pgConfigFileStream, "\t[selection]\n");
				fprintf(pgConfigFileStream, "\t\tTextElementSeperator='%s'\n", preferences->pSelectionTextElementSeperator);
				fprintf(pgConfigFileStream, "\t[fmwindow]\n");
				fprintf(pgConfigFileStream, "\t\tDefaultPosition=%d\n", preferences->fmWindowDefaultPosition);
				fprintf(pgConfigFileStream, "\t\tXpos=%d\n", preferences->fmWindowX);
				fprintf(pgConfigFileStream, "\t\tYpos=%d\n", preferences->fmWindowY);
				fprintf(pgConfigFileStream, "\t\tWidth=%d\n", preferences->fmWindowWidth);
				fprintf(pgConfigFileStream, "\t\tHeight=%d\n", preferences->fmWindowHeight);
	
                fclose(pgConfigFileStream);
				g_free(preferences_filename);
			}
			else
			{
				printf("preferences_save : Can't open the preferences file '%s' for writing.\n", preferences_filename);
				g_free(preferences_filename);
			}
		}
		else
		{
			printf("preferences_save : Can't create the preferences filename\n");
		}
	}
}

void preferences_storeAccelKey(gpointer key, gpointer value, gpointer user_data)
{
	gchar			***accel_keys_names_iterator = (gchar ***)user_data;
    struct AccelKey	*accel_key = value;

	fprintf(pgConfigFileStream, "\t[%s]\n", (gchar*)key);
	fprintf(pgConfigFileStream, "\t\tKey=%d\n", accel_key->key);
	fprintf(pgConfigFileStream, "\t\tMods=%d\n", accel_key->mods);
	
	**accel_keys_names_iterator = key;
    (*accel_keys_names_iterator)++;
}

void preferences_storeRowStyle(gpointer key, gpointer value, gpointer user_data)
{
	gchar			***row_styles_names_iterator = (gchar ***)user_data;
    struct RowStyle	*row_style = value;

	fprintf(pgConfigFileStream, "\t[%s]\n", (gchar*)key);
	fprintf(pgConfigFileStream, "\t\tFgRed=%d\n", row_style->foreground.red);
	fprintf(pgConfigFileStream, "\t\tFgGreen=%d\n", row_style->foreground.green);
	fprintf(pgConfigFileStream, "\t\tFgBlue=%d\n", row_style->foreground.blue);
	fprintf(pgConfigFileStream, "\t\tBgRed=%d\n", row_style->background.red);
	fprintf(pgConfigFileStream, "\t\tBgGreen=%d\n", row_style->background.green);
	fprintf(pgConfigFileStream, "\t\tBgBlue=%d\n", row_style->background.blue);
	if(row_style->pFilenameClosed)
	{
		fprintf(pgConfigFileStream, "\t\tFilenameClosed='%s'\n", row_style->pFilenameClosed);
	}
    
	if(row_style->pFilenameOpened)
	{
		fprintf(pgConfigFileStream, "\t\tFilenameOpened='%s'\n", row_style->pFilenameOpened);
	}
	fprintf(pgConfigFileStream, "\t\tDeletable=%d\n", row_style->deletable);

	**row_styles_names_iterator = key;
    (*row_styles_names_iterator)++;
}


