/*	
**	
**	Copyright (C) 1999 Geert Bevin
**	gbevin@thunderstorms.org
**	http://thunderstorms.org/fm
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation; either version 2 of the License, or
**	(at your option) any later version.
**	
**	This program is distributed in the hope that it will be useful,
**	but WITHOUT ANY WARRANTY; without even the implied warranty of
**	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**	GNU General Public License for more details.
**	
**	You should have received a copy of the GNU General Public License
**	along with this program; if not, write to the Free Software
**	
**	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**		
*/

#ifndef __PREFERENCES_H__
#define __PREFERENCES_H__

#include "filetype.h"

struct Preferences
{
	GHashTable		*pAccelKeys;
	GList			*pCtreeColumns;
	GList			*pFileTypes;
	GHashTable		*pRowStyles;
	GList			*pPixmapSearchPaths;
	gchar			*pSelectionTextElementSeperator;
	gboolean		characterBrowsingEnabled;
	gboolean		characterBrowsingEnableSpacebar;
	gboolean		characterBrowsingIgnoreCase;
	guint			comboPopdownLength;
	gboolean		fmWindowDefaultPosition;
	guint			fmWindowX;
	guint			fmWindowY;
	guint			fmWindowHeight;
	guint			fmWindowWidth;
	gboolean		sessionSavePreferencesOnExit;
	gboolean		sessionSaveLayoutBookmarks;
	gboolean		sessionSaveLayoutBrowserAppearance;
	gboolean		sessionSaveLayoutComboHistory;
	gboolean		sessionSaveLayoutExpandedDirs;
	gboolean		sessionSaveLayoutGeometry;
	gboolean		sessionSaveLayoutOnExit;
	gboolean		sortDirectoriesFirst;
	gboolean		sortIgnoreCase;
};

struct SymbolDefinition
{
	gchar	*name;
	guint	token;
};

struct PreferencesEntry
{
	guint		section;
	guint		nameToken;
	gchar*		pNameString;
	union
	{
		struct
		{
			guint	type;
			guint	alignment;
		} ctreeColumn;
		
		struct
		{
			guint	key;
			guint	mods;
		} accelKey;
		
		struct
		{
			gushort		fgRed;
			gushort		fgGreen;
			gushort		fgBlue;
			gushort		bgRed;
			gushort		bgGreen;
			gushort		bgBlue;
			gchar 		*pFilenameClosed;
			gchar 		*pFilenameOpened;
			gboolean 	deletable;
		} rowStyle;

		struct
		{
			gboolean		deletable;
			BasicFileType	basicType;
			gchar			*pRowStyle;
		} fileType;

		union
		{
			struct
			{
				gboolean	enabled;
				gboolean	enableSpacebar;
				gboolean	ignoreCase;
			} characterBrowsing;

			struct
			{
				guint	popdownLength;
			} comboBox;

			struct
			{
				gboolean	directoriesFirst;
				gboolean	ignoreCase;
			} sort;

			struct
			{
				gboolean	savePreferencesOnExit;
				gboolean	saveLayoutOnExit;
				gboolean	saveLayoutBookmarks;
				gboolean	saveLayoutBrowserAppearance;
				gboolean	saveLayoutComboHistory;
				gboolean	saveLayoutExpandedDirs;
				gboolean	saveLayoutGeometry;
			} session;

			struct
			{
				gchar	*pTextElementSeperator;
			} selection;
			
			struct
			{
				gboolean	defaultPosition;
				guint		x;
				guint		y;
				guint		width;
				guint		height;
			} fmwindow;
		} general;
	} values;
};

void						preferences_clear();
void						preferences_destroy(struct Preferences*);
struct Preferences*			preferences_duplicate(struct Preferences*);
void						preferences_initialize(struct Preferences*);
void						preferences_load(struct Preferences*);
struct Preferences*			preferences_new();
void						preferences_pixmapSearchPathsAdd(struct Preferences*, gchar*);
void						preferences_pixmapSearchPathsRemoveAll(struct Preferences*);
void						preferences_save(struct Preferences*);
void						preferences_storeAccelKey(gpointer, gpointer, gpointer);
void						preferences_storeRowStyle(gpointer, gpointer, gpointer);

#endif /* __PREFERENCES_H__ */

