/*	
**	
**	Copyright (C) 1999 Geert Bevin
**	gbevin@thunderstorms.org
**	http://thunderstorms.org/fm
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation; either version 2 of the License, or
**	(at your option) any later version.
**	
**	This program is distributed in the hope that it will be useful,
**	but WITHOUT ANY WARRANTY; without even the implied warranty of
**	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**	GNU General Public License for more details.
**	
**	You should have received a copy of the GNU General Public License
**	along with this program; if not, write to the Free Software
**	
**	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**		
*/

#include <stdio.h>
#include <stdlib.h>

#include "ctreecolumn.h"
#include "fmwindow.h"
#include "layout.h"
#include "preferences.h"
#include "preferenceswindow.h"
#include "preferenceswindow_signals.h"

extern struct Preferences	*pgPreferences;

void preferencesWindow_buttonCancel(GtkWidget *widget, gpointer userData)
{
	struct PreferencesWindow	**preferences_window = (struct PreferencesWindow **)userData;

	if(preferences_window)
	{
		preferencesWindow_destroy(preferences_window);
	}
}

void preferencesWindow_buttonClearLayoutNow(GtkWidget *widget, gpointer userData)
{
	layout_clear();
}

void preferencesWindow_buttonClearPreferencesNow(GtkWidget *widget, gpointer userData)
{
	preferences_clear();
}

void preferencesWindow_buttonSave(GtkWidget *widget, gpointer userData)
{
	struct PreferencesWindow	**preferences_window = (struct PreferencesWindow **)userData;

	if(preferences_window)
	{
        preferences_save((*preferences_window)->pPreferences);
        preferencesWindow_buttonOk(widget, userData);
	}
}

void preferencesWindow_buttonSaveLayoutNow(GtkWidget *widget, gpointer userData)
{
	struct Preferences	*preferences = (struct Preferences *)userData;

	if(preferences)
	{
        layout_save(preferences);
	}
}

void preferencesWindow_buttonSavePreferencesNow(GtkWidget *widget, gpointer userData)
{
	struct Preferences	*preferences = (struct Preferences *)userData;

	if(preferences)
	{
        preferences_save(preferences);
	}
}

void preferencesWindow_buttonOk(GtkWidget *widget, gpointer userData)
{
	struct PreferencesWindow	**preferences_window = (struct PreferencesWindow **)userData;
	struct Preferences			*old_preferences = pgPreferences;

	if(preferences_window)
	{
        pgPreferences = (*preferences_window)->pPreferences;
        rebuildAllFmWindowsCTrees();
		
		(*preferences_window)->pPreferences = NULL;
		preferencesWindow_destroy(preferences_window);
		preferences_destroy(old_preferences);
	}
}

void preferencesWindow_buttonToggled(GtkWidget *widget, gpointer userData)
{
	gboolean	*value = userData;

	*value = GTK_TOGGLE_BUTTON(widget)->active;
}

void preferencesWindow_buttonToggledDefaultPosition(GtkWidget *widget, gpointer userData)
{
	struct PreferencesWindow	*preferences_window = (struct PreferencesWindow *)userData;
	
	gtk_widget_set_sensitive(preferences_window->pEntryX, GTK_TOGGLE_BUTTON(widget)->active);
	gtk_widget_set_sensitive(preferences_window->pEntryY, GTK_TOGGLE_BUTTON(widget)->active);
	preferences_window->pPreferences->fmWindowDefaultPosition = GTK_TOGGLE_BUTTON(widget)->active;
}

void preferencesWindow_buttonToggledSaveLayout(GtkWidget *widget, gpointer userData)
{
	struct PreferencesWindow	*preferences_window = (struct PreferencesWindow *)userData;

	gtk_widget_set_sensitive(preferences_window->pCheckButtonLayoutSaveBookmarks, GTK_TOGGLE_BUTTON(widget)->active);
	gtk_widget_set_sensitive(preferences_window->pCheckButtonLayoutSaveExpandedDirs, GTK_TOGGLE_BUTTON(widget)->active);
	gtk_widget_set_sensitive(preferences_window->pCheckButtonLayoutSaveGeometry, GTK_TOGGLE_BUTTON(widget)->active);
	gtk_widget_set_sensitive(preferences_window->pCheckButtonLayoutSaveBrowserAppearance, GTK_TOGGLE_BUTTON(widget)->active);
	gtk_widget_set_sensitive(preferences_window->pCheckButtonLayoutSaveComboHistory, GTK_TOGGLE_BUTTON(widget)->active);
	preferences_window->pPreferences->sessionSaveLayoutOnExit = GTK_TOGGLE_BUTTON(widget)->active;
}

void preferencesWindow_clistColumnsRowMove(GtkCList *clist, gint arg1, gint arg2, gpointer userData)
{
	struct PreferencesWindow	*preferences_window = (struct PreferencesWindow *)userData;
	GList						*temp_glist = NULL;

//    printf("arg1=%d arg2=%d\n", arg1, arg2);
	temp_glist = g_list_nth(preferences_window->pPreferences->pCtreeColumns, arg1+COLUMN_SIZE);
	preferences_window->pPreferences->pCtreeColumns = g_list_remove_link(preferences_window->pPreferences->pCtreeColumns, temp_glist);
	preferences_window->pPreferences->pCtreeColumns = g_list_insert(preferences_window->pPreferences->pCtreeColumns, temp_glist->data, arg2+COLUMN_SIZE);

	g_list_free_1(temp_glist);
}

void preferencesWindow_ctreeSelect(GtkWidget *widget, GtkCTreeNode *node)
{
	GtkNotebook	*notebook;
	gint			page;
	
	if(GTK_CLIST(widget)->selection)
	{
		notebook = (GtkNotebook*)gtk_object_get_user_data(GTK_OBJECT(widget));
		page = (gint)gtk_ctree_node_get_row_data(GTK_CTREE(widget), node);
		
		gtk_notebook_set_page(notebook, page);
	}
}

void preferencesWindow_entryUnsignedIntegerChange(GtkWidget *widget, gpointer userData)
{
	gint	*value = userData;

	*value = atoi(gtk_entry_get_text(GTK_ENTRY(widget)));
	if(*value < 0)
	{
		value = 0;
	}
}

