/*	
**	
**	Copyright (C) 1999 Geert Bevin
**	gbevin@thunderstorms.org
**	http://thunderstorms.org/fm
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation; either version 2 of the License, or
**	(at your option) any later version.
**	
**	This program is distributed in the hope that it will be useful,
**	but WITHOUT ANY WARRANTY; without even the implied warranty of
**	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**	GNU General Public License for more details.
**	
**	You should have received a copy of the GNU General Public License
**	along with this program; if not, write to the Free Software
**	
**	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**		
*/

#include <dirent.h>
#include <errno.h>
#include <math.h>
#include <pwd.h>
#include <grp.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include <sys/stat.h>

#include <glib.h>
#include <gdk/gdkkeysyms.h>
#include <gdk/gdktypes.h>
#include <gtk/gtk.h>

#include "filetype.h"
#include "fmwindow.h"
#include "rowstyle.h"
#include "fmwindow_signals.h"

struct RowStyleAndName
{
	struct RowStyle	*pRowStyle;
	gchar			*pName;
};

struct RowStyle* rowStyleFind(GHashTable *rowStyles, gchar *name)
{
	struct RowStyle	*row_style = NULL;
	
	row_style = g_hash_table_lookup(rowStyles, name);
	
	return row_style;
}

static void rowStyleHastableCompareRowStyle(gpointer key, gpointer value, gpointer userData)
{
	struct RowStyleAndName	*rowstyle_and_name = (struct RowStyleAndName *)userData;
	gchar					*row_style_name = key;
	struct RowStyle			*row_style = (struct RowStyle *)value;

	if(rowstyle_and_name->pName == NULL && rowstyle_and_name->pRowStyle == row_style)
	{
		rowstyle_and_name->pName = row_style_name;
	}
}

gchar* rowStyleFindName(GHashTable *rowStyles, struct RowStyle *row_style)
{
    struct RowStyleAndName	rowstyle_and_name = {row_style, NULL};
	
	g_hash_table_foreach(rowStyles, (GHFunc)rowStyleHastableCompareRowStyle, &rowstyle_and_name);

	return rowstyle_and_name.pName;
}

GdkColor* rowStyleGetBackground(GHashTable *rowStyles, gchar *name)
{
	struct RowStyle	*row_style = NULL;
	
	row_style = g_hash_table_lookup(rowStyles, name);
	if(row_style)
	{
		return &(row_style->background);
	}
	else
	{
		return NULL;
	}
}

GdkColor* rowStyleGetForeground(GHashTable *rowStyles, gchar *name)
{
	struct RowStyle	*row_style = NULL;
	
	row_style = g_hash_table_lookup(rowStyles, name);
	if(row_style)
	{
		return &(row_style->foreground);
	}
	else
	{
		return NULL;
	}
}

struct RowStyle* rowStyleNew(GHashTable *rowStyles, GList *pixmapSearchPaths, gchar *name, gushort fg_red, gushort fg_green, gushort fg_blue, gushort bg_red, gushort bg_green, gushort bg_blue, gchar *filenameClosed, gchar *filenameOpened, gboolean deletable)
{
	struct RowStyle	*row_style = NULL;
	GList			*referencing_filetypes = NULL;
	gchar			*real_name = NULL;
	guint			counter = 0;
	gchar			*temp_string = NULL;
	struct stat		temp_stat;
	
//	printf("New rowstyle %s %d %d %d %d %d %d %s %s\n", name, fg_red, fg_green, fg_blue, bg_red, bg_green, bg_blue, filenameClosed, filenameOpened);
	row_style = rowStyleFind(rowStyles, name);
	if(row_style)
	{
		referencing_filetypes = row_style->pReferencingFileTypes;
	}

	row_style = g_new(struct RowStyle, 1);
	if(row_style)
	{
		real_name = g_strdup(name);
		if(real_name)
		{
			rowStyleRemove(rowStyles, real_name);
			g_hash_table_insert(rowStyles, real_name, row_style);
			
			row_style->pName = real_name;
			row_style->foreground.red = fg_red;
			row_style->foreground.green = fg_green;
			row_style->foreground.blue = fg_blue;
			row_style->background.red = bg_red;
			row_style->background.green = bg_green;
			row_style->background.blue = bg_blue;
			row_style->pFilenameClosed = NULL;
			row_style->pFilenameOpened = NULL;
			row_style->pPixmapClosed = NULL;
			row_style->pPixmapOpened = NULL;
			row_style->deletable = deletable;
			row_style->pReferencingFileTypes = NULL;

			gdk_colormap_alloc_color(gdk_colormap_get_system(), &row_style->foreground, FALSE, TRUE);
			gdk_colormap_alloc_color(gdk_colormap_get_system(), &row_style->background, FALSE, TRUE);
			
			if(filenameClosed)
			{
				row_style->pFilenameClosed = g_strdup(filenameClosed);
				for(counter = 0; counter < g_list_length(pixmapSearchPaths); counter++)
				{
                    temp_string = g_strconcat((gchar*)g_list_nth_data(pixmapSearchPaths, counter), "/", filenameClosed, NULL);
					if(!lstat(temp_string, &temp_stat))
					{
						row_style->pPixmapClosed = gdk_pixmap_colormap_create_from_xpm(NULL, gdk_colormap_get_system(), &(row_style->pMaskClosed), NULL, temp_string);
						g_free(temp_string);
						break;
					}
					g_free(temp_string);
				}
			}
			if(filenameOpened)
			{
				row_style->pFilenameOpened = g_strdup(filenameOpened);
				for(counter = 0; counter < g_list_length(pixmapSearchPaths); counter++)
				{
                    temp_string = g_strconcat((gchar*)g_list_nth_data(pixmapSearchPaths, counter), "/", filenameOpened, NULL);
					if(!lstat(temp_string, &temp_stat))
					{
						row_style->pPixmapOpened = gdk_pixmap_colormap_create_from_xpm(NULL, gdk_colormap_get_system(), &(row_style->pMaskOpened), NULL, temp_string);
						g_free(temp_string);
						break;
					}
					g_free(temp_string);
				}
			}

			if(g_list_length(referencing_filetypes))
			{
				for(counter = 0; counter < g_list_length(referencing_filetypes); counter++)
				{
					((struct FileType*)g_list_nth_data(referencing_filetypes, counter))->pRowStyle = row_style;
				}
			}

			return row_style;
		}
		g_free(row_style);
	}
	
	return NULL;
}

void rowStyleRemove(GHashTable *rowStyles, gchar *name)
{
	gpointer	orig_key;
	gpointer	orig_value;
	
	if(g_hash_table_lookup_extended(rowStyles, name, &orig_key, &orig_value))
	{
		if(orig_key)
		{
			g_hash_table_remove(rowStyles, orig_key);
			g_free(orig_key);
		}
		if(orig_value)
		{
			if(((struct RowStyle*)orig_value)->pFilenameClosed)
			{
				g_free(((struct RowStyle*)orig_value)->pFilenameClosed);
			}
			if(((struct RowStyle*)orig_value)->pFilenameOpened)
			{
				g_free(((struct RowStyle*)orig_value)->pFilenameOpened);
			}
			g_free(orig_value);
		}
	}
}

void rowStylesRemoveAll(GHashTable *rowStyles)
{
	g_hash_table_foreach_remove(rowStyles, rowStylesForEachRemove, NULL);
}

gboolean rowStylesForEachRemove(gpointer key, gpointer value, gpointer user_data)
{
	if(key)
	{
		g_free(key);
	}
	if(value)
	{
		if(((struct RowStyle*)value)->pFilenameClosed)
		{
			g_free(((struct RowStyle*)value)->pFilenameClosed);
		}
		if(((struct RowStyle*)value)->pFilenameOpened)
		{
			g_free(((struct RowStyle*)value)->pFilenameOpened);
		}
		g_free(value);
	}

	return TRUE;
}

