/*	
**	
**	Copyright (C) 1999 Geert Bevin
**	gbevin@thunderstorms.org
**	http://thunderstorms.org/fm
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation; either version 2 of the License, or
**	(at your option) any later version.
**	
**	This program is distributed in the hope that it will be useful,
**	but WITHOUT ANY WARRANTY; without even the implied warranty of
**	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**	GNU General Public License for more details.
**	
**	You should have received a copy of the GNU General Public License
**	along with this program; if not, write to the Free Software
**	
**	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**		
*/

#include <dirent.h>
#include <errno.h>
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include <sys/stat.h>

#include <glib.h>
#include <gdk/gdkkeysyms.h>
#include <gdk/gdktypes.h>
#include <gtk/gtk.h>

#include "accelkey.h"
#include "ctreecolumn.h"
#include "filetype.h"
#include "fileutils.h"
#include "fmwindow.h"
#include "preferences.h"
#include "rowstyle.h"
#include "fmwindow_signals.h"

static char *pgsReloadXpm[] =
{
	"16 13 3 1",
	" 	c None",
	".	c #838183",
	"+	c #000000",
	"     ..+++.     ",
	"    .+++++++    ",
	"   .++.    .+   ",
	"   ++.          ",
	"  +++       +   ",
	"+++++++    +++  ",
	" +++++    +++++ ",
	"  +++    +++++++",
	"   +       +++  ",
	"          .++   ",
	"   +.    .++.   ",
	"    +++++++.    ",
	"     .+++..     "
};

gchar *pgColumnTitles[8] =
{
	"",
	"files",
	"size",
	"user",
	"group",
	"permissions",
	"modification time",
	"change time"
};

GtkTargetEntry 	gSelectionTargetList[] =
{
	{ "STRING",        0, SELECTION_TARGET_STRING },
	{ "TEXT",          0, SELECTION_TARGET_TEXT },
	{ "COMPOUND_TEXT", 0, SELECTION_TARGET_COMPOUND_TEXT }
};

gint			gSelectionNumberOfTargets = sizeof(gSelectionTargetList) / sizeof(GtkTargetEntry);

GtkTargetEntry	gComboEntryDragDropTargets[] =
{
	{ "_NETSCAPE_URL",	0,	DRAGDROP_TARGET_NETSCAPE_URL },
	{ "text/uri-list",	0,	DRAGDROP_TARGET_URL },
	{ "STRING",			0,	DRAGDROP_TARGET_STRING },
	{ "text/plain",		0,	DRAGDROP_TARGET_TEXT },
	{ "FM_SELECTION",	0,	DRAGDROP_TARGET_FM_SELECTION }
};

guint			gComboEntryNumberOfDragDropTargets = sizeof(gComboEntryDragDropTargets) / sizeof(GtkTargetEntry);

GtkTargetEntry	gCTreeDragDropTargets[] =
{
	{ "STRING",			0,	DRAGDROP_TARGET_STRING },
	{ "text/plain",		0,	DRAGDROP_TARGET_TEXT },
	{ "_NETSCAPE_URL",	0,	DRAGDROP_TARGET_NETSCAPE_URL },
	{ "text/uri-list",	0,	DRAGDROP_TARGET_URL },
	{ "FM_SELECTION",	0,	DRAGDROP_TARGET_FM_SELECTION }
};

guint			gCTreeNumberOfDragDropTargets = sizeof(gCTreeDragDropTargets) / sizeof(GtkTargetEntry);

extern GList				*pgFmWindows;
extern struct Preferences	*pgPreferences;
extern struct BrowserState	gBrowserState;
extern GdkPixmap 			*pgPixmapPanWindowCenter;
extern GdkBitmap 			*pgMaskPanWindowCenter;
extern GdkPixmap 			*pgPixmapPanWindowN;
extern GdkBitmap 			*pgMaskPanWindowN;
extern GdkPixmap 			*pgPixmapPanWindowNE;
extern GdkBitmap 			*pgMaskPanWindowNE;
extern GdkPixmap 			*pgPixmapPanWindowE;
extern GdkBitmap 			*pgMaskPanWindowE;
extern GdkPixmap 			*pgPixmapPanWindowSE;
extern GdkBitmap 			*pgMaskPanWindowSE;
extern GdkPixmap 			*pgPixmapPanWindowS;
extern GdkBitmap 			*pgMaskPanWindowS;
extern GdkPixmap 			*pgPixmapPanWindowSW;
extern GdkBitmap 			*pgMaskPanWindowSW;
extern GdkPixmap 			*pgPixmapPanWindowW;
extern GdkBitmap 			*pgMaskPanWindowW;
extern GdkPixmap 			*pgPixmapPanWindowNW;
extern GdkBitmap 			*pgMaskPanWindowNW;

GtkWidget* fmWindow_acceleratedWidgetFind(struct FmWindow *fmWindow, gchar *accelKeyId)
{
	GtkWidget	*widget = NULL;
	
//	printf("find accelerated widget for fmWindow 0x%x with accelKeyId '%s'\n", (guint)fmWindow, accelKeyId);
	widget = g_hash_table_lookup(fmWindow->pAcceleratedWidgets, accelKeyId);

	return widget;
}

void fmWindow_acceleratedWidgetForget(struct FmWindow *fmWindow, GtkWidget *widget)
{
	gchar			*accel_key_id = NULL;
	
//	printf("fmWindow_acceleratedWidgetForget : fmWindow=0x%x, widget=0x%x\n", (guint)fmWindow, (guint)widget);
	accel_key_id = (gchar*)gtk_object_get_data(GTK_OBJECT(widget), "accelKeyId");
	if(accel_key_id)
	{
//		printf("=> unregistering\n");
		gtk_object_remove_data(GTK_OBJECT(widget), "accelKeyKey");
		gtk_object_remove_data(GTK_OBJECT(widget), "accelKeyMods");
	}
}

void fmWindow_acceleratedWidgetRemember(struct FmWindow *fmWindow, GtkWidget *widget)
{
	gchar			*accel_key_id = NULL;
	struct AccelKey	*accel_key = NULL;
	
//	printf("fmWindow_acceleratedWidgetRemember : fmWindow=0x%x, widget=0x%x\n", (guint)fmWindow, (guint)widget);
	accel_key_id = (gchar*)gtk_object_get_data(GTK_OBJECT(widget), "accelKeyId");
	if(accel_key_id)
	{
		accel_key = accelKeyFind(pgPreferences->pAccelKeys, accel_key_id);
		if(accel_key)
		{
//			printf("=> registering\n");
			gtk_object_set_data(GTK_OBJECT(widget), "accelKeyKey", (gpointer)accel_key->key);
			gtk_object_set_data(GTK_OBJECT(widget), "accelKeyMods", (gpointer)accel_key->mods);
		}
	}
}

void fmWindow_accelKeyAddToGtkWidget(struct FmWindow *fmWindow, GtkWidget *widget)
{
	gchar			*accel_key_id = NULL;
	struct AccelKey	*accel_key = NULL;

	accel_key_id = (gchar*)gtk_object_get_data(GTK_OBJECT(widget), "accelKeyId");
	if(accel_key_id)
	{
//		printf("fmWindow_accelKeyAddToGtkWidget : fmWindow=0x%x, widget=0x%x, accelKeyId='%s'\n", (guint)fmWindow, (guint)widget, accel_key_id);
		accel_key = accelKeyFind(pgPreferences->pAccelKeys, accel_key_id);
		if(accel_key)
		{
			g_hash_table_insert(fmWindow->pAcceleratedWidgets, accel_key_id, widget); 
			fmWindow_acceleratedWidgetRemember(fmWindow, widget);
			gtk_widget_add_accelerator(widget, "activate", fmWindow->pAccelGroup, accel_key->key, accel_key->mods, GTK_ACCEL_VISIBLE);
		}
	}
}

void fmWindow_accelKeyRemoveFromGtkWidget(struct FmWindow *fmWindow, GtkWidget *widget)
{
	guint			accel_key_key = (guint)gtk_object_get_data(GTK_OBJECT(widget), "accelKeyKey");
	GdkModifierType	accel_key_mods = (guint)gtk_object_get_data(GTK_OBJECT(widget), "accelKeyMods");
	
//	printf("fmWindow_accelKeyRemoveFromGtkWidget : fmWindow=0x%x, widget=0x%x, accelKeyKey=%d= accelKeyMods=%d\n", (guint)fmWindow, (guint)widget, accel_key_key, accel_key_mods);
	fmWindow_acceleratedWidgetForget(fmWindow, widget);
	gtk_widget_remove_accelerator(widget, fmWindow->pAccelGroup, accel_key_key, accel_key_mods);
}

void fmWindow_bookmarkClear(struct FmWindow *fmWindow, guint index)
{
	gchar			*bookmark_name = NULL;
	GNode			*source_gnode = NULL;

	bookmark_name = fmWindow->pBookmarks[index];
	if(bookmark_name)
	{
//		printf("clear bookmark %s\n", bookmark_name);
		// clear the previous bookmark indication
		source_gnode = fmWindow_getNodeFromFilename(fmWindow, bookmark_name, FALSE);
		if(source_gnode)
		{
			fmWindow_bookmarkClearIndication(fmWindow, source_gnode);
		}
		fmWindow->pBookmarks[index] = NULL;
		gtk_widget_set_sensitive(GTK_WIDGET(GTK_BIN(fmWindow->pBookmarkButtons[index].pButtonBookmark)->child), FALSE);
		gtk_tooltips_set_tip(GTK_TOOLTIPS(fmWindow->pToolTips), GTK_WIDGET(fmWindow->pBookmarkButtons[index].pButtonBookmark), "", "");
		g_free(bookmark_name);
		fmWindow_bookmarkUpdatePopupMenu(fmWindow, index);
	}
}

void fmWindow_bookmarkClearIndication(struct FmWindow *fmWindow, GNode *sourceNode)
{
	if(sourceNode)
	{
		gtk_ctree_node_set_cell_style(GTK_CTREE(fmWindow->pCTree), GTK_CTREE_NODE(((struct NodeInfo*)sourceNode->data)->pCtreeNode), 0, NULL);
		gtk_ctree_node_set_text(GTK_CTREE(fmWindow->pCTree), GTK_CTREE_NODE(((struct NodeInfo*)sourceNode->data)->pCtreeNode), 0, "");
	}
}

void fmWindow_bookmarkDrawIndication(struct FmWindow *fmWindow, GNode *targetNode, guint index)
{
	GtkStyle		*cell_style = NULL;
	GdkColor		*bookmark_background = rowStyleGetBackground(pgPreferences->pRowStyles, "Bookmark");
	GdkColor		*bookmark_foreground = rowStyleGetForeground(pgPreferences->pRowStyles, "Bookmark");
	gchar			index_text[2];
    
	// place the bookmark indication
	if(targetNode)
	{
		cell_style = gtk_style_copy(gtk_widget_get_style(GTK_WIDGET(fmWindow->pCTree)));
		if(cell_style && bookmark_background && bookmark_foreground)
		{
			cell_style->base->red = bookmark_background->red;
			cell_style->base->green = bookmark_background->green;
			cell_style->base->blue = bookmark_background->blue;
			cell_style->fg->red = bookmark_foreground->red;
			cell_style->fg->green = bookmark_foreground->green;
			cell_style->fg->blue = bookmark_foreground->blue;
			gtk_ctree_node_set_cell_style(GTK_CTREE(fmWindow->pCTree), GTK_CTREE_NODE(((struct NodeInfo*)targetNode->data)->pCtreeNode), 0, cell_style);
		}
		else
		{
			printf("cell_syle == NULL\n");
		}
		index_text[0] = index+'0';
		index_text[1] = 0;
		gtk_ctree_node_set_text(GTK_CTREE(fmWindow->pCTree), GTK_CTREE_NODE(((struct NodeInfo*)targetNode->data)->pCtreeNode), 0, index_text);
	}
}

gint fmWindow_bookmarkGetIndex(struct FmWindow *fmWindow, GNode *sourceNode)
{
	int		counter = 0;
	gchar	*temp_filename = NULL;

	for(counter = 0; counter < 10; counter++)
	{
		if(fmWindow->pBookmarks[counter])
		{
            temp_filename = fmWindow_getFilenameToRemember(fmWindow, sourceNode);
			if(strcmp(fmWindow->pBookmarks[counter], temp_filename) == 0)
			{
				g_free(temp_filename);
				return counter;
			}
		}
	}
	g_free(temp_filename);

	return -1;
}


void fmWindow_bookmarkGoTo(struct FmWindow *fmWindow, guint index)
{
	gchar			*bookmark_name = NULL;

	bookmark_name = fmWindow->pBookmarks[index];
	if(bookmark_name)
	{
//		printf("go to bookmark %s\n", bookmark_name);
		fmWindow_focusFileName(fmWindow, bookmark_name);
		fmWindow_selectMoveToCTreeFocus(fmWindow);
	}
}

void fmWindow_bookmarkSet(struct FmWindow *fmWindow, gchar *bookmarkName, guint index)
{
	gint			counter = 0;
	gchar			*previous_bookmark_name = NULL;
	GNode			*previous_gnode = NULL;
	GNode			*source_gnode = NULL;

//	printf("FMWINDOW_BOOKMARKSET : fmWindow=0x%x\n", (guint)fmWindow);

	if(bookmarkName)
	{
		// check if there not already a bookmark on this filename
		for(counter = 0; counter < 10; counter++)
		{
			if(fmWindow->pBookmarks[counter] && strcmp(bookmarkName, fmWindow->pBookmarks[counter]) == 0)
			{
				fmWindow_bookmarkClear(fmWindow, counter);
				break;
			}
		}

		previous_bookmark_name = fmWindow->pBookmarks[index];
		fmWindow->pBookmarks[index] = g_strdup(bookmarkName);
//		printf("previous_bookmark_name=0x%x\n", (guint)previous_bookmark_name);
		if(previous_bookmark_name)
		{
			// clear the previous bookmark indication
			previous_gnode = fmWindow_getNodeFromFilename(fmWindow, previous_bookmark_name, FALSE);
//			printf("previous_gnode=0x%x\n", (guint)previous_gnode);
			if(previous_gnode)
			{
    			fmWindow_bookmarkClearIndication(fmWindow, previous_gnode);
			}
			g_free(previous_bookmark_name);
		}
//		printf("set bookmark %s\n\n", bookmarkName);
		gtk_widget_set_sensitive(GTK_WIDGET(GTK_BIN(fmWindow->pBookmarkButtons[index].pButtonBookmark)->child), TRUE);
		gtk_tooltips_set_tip(GTK_TOOLTIPS(fmWindow->pToolTips), GTK_WIDGET(fmWindow->pBookmarkButtons[index].pButtonBookmark), bookmarkName, "");
		fmWindow_bookmarkUpdatePopupMenu(fmWindow, index);
        source_gnode = fmWindow_getNodeFromFilename(fmWindow, bookmarkName, FALSE);
		if(source_gnode)
		{
			fmWindow_bookmarkDrawIndication(fmWindow, getLinkTarget(source_gnode), index);
		}
	}
}

void fmWindow_bookmarkUpdatePopupMenu(struct FmWindow *fmWindow, guint index)
{
	
	if(GTK_BIN(fmWindow->pBookmarkButtons[index].pButtonBookmark)->child->state == GTK_STATE_INSENSITIVE)
	{
		// when the label is disactivated and thus the bookmark clear
		gtk_widget_set_sensitive(GTK_WIDGET(fmWindow->pBookmarkButtons[index].pPopupMenuItemSet), TRUE);
		gtk_widget_set_sensitive(GTK_WIDGET(fmWindow->pBookmarkButtons[index].pPopupMenuItemGoTo), FALSE);
		gtk_widget_set_sensitive(GTK_WIDGET(fmWindow->pBookmarkButtons[index].pPopupMenuItemClear), FALSE);
	}
	else
	{
		// when the label is activated and thus the bookmark set
		gtk_widget_set_sensitive(GTK_WIDGET(fmWindow->pBookmarkButtons[index].pPopupMenuItemSet), TRUE);
		gtk_widget_set_sensitive(GTK_WIDGET(fmWindow->pBookmarkButtons[index].pPopupMenuItemGoTo), TRUE);
		gtk_widget_set_sensitive(GTK_WIDGET(fmWindow->pBookmarkButtons[index].pPopupMenuItemClear), TRUE);
	}
}

GtkWidget* fmWindow_buildCtree(struct FmWindow *fmWindow)
{
	GtkWidget		*ctree = NULL;
	GtkBindingSet	*clist_binding_set = NULL;
	GtkBindingSet	*ctree_binding_set = NULL;
	
	// create the tree
	ctree = gtk_ctree_new(g_list_length(fmWindow->pCTreeColumns), 1);
	if(ctree)
	{
		gtk_clist_column_titles_show(GTK_CLIST(ctree));
		gtk_clist_column_titles_active(GTK_CLIST(ctree));
		gtk_ctree_set_expander_style(GTK_CTREE(ctree), GTK_CTREE_EXPANDER_NONE);
		gtk_ctree_set_line_style(GTK_CTREE(ctree), GTK_CTREE_LINES_DOTTED);
		gtk_ctree_set_indent(GTK_CTREE(ctree), CTREE_INDENT);
		gtk_ctree_set_spacing(GTK_CTREE(ctree), CTREE_SPACING);
		gtk_clist_set_selection_mode(GTK_CLIST(ctree), GTK_SELECTION_EXTENDED);

		// remove the keybindings that should not be issued anymore
		clist_binding_set = gtk_binding_set_by_class(gtk_type_class (GTK_TYPE_CLIST));
		gtk_binding_entry_remove(clist_binding_set, GDK_Up, 0);
		gtk_binding_entry_remove(clist_binding_set, GDK_Down, 0);
		gtk_binding_entry_remove(clist_binding_set, GDK_Page_Up, 0);
		gtk_binding_entry_remove(clist_binding_set, GDK_Page_Down, 0);
		gtk_binding_entry_remove(clist_binding_set, GDK_Home, GDK_CONTROL_MASK);
		gtk_binding_entry_remove(clist_binding_set, GDK_End, GDK_CONTROL_MASK);
		gtk_binding_entry_remove(clist_binding_set, GDK_Up, GDK_SHIFT_MASK);
		gtk_binding_entry_remove(clist_binding_set, GDK_Down, GDK_SHIFT_MASK);
		gtk_binding_entry_remove(clist_binding_set, GDK_Page_Up, GDK_SHIFT_MASK);
		gtk_binding_entry_remove(clist_binding_set, GDK_Page_Down, GDK_SHIFT_MASK);
		gtk_binding_entry_remove(clist_binding_set, GDK_Home, GDK_SHIFT_MASK | GDK_CONTROL_MASK);
		gtk_binding_entry_remove(clist_binding_set, GDK_End, GDK_SHIFT_MASK | GDK_CONTROL_MASK);
		gtk_binding_entry_remove(clist_binding_set, GDK_Left, 0);
		gtk_binding_entry_remove(clist_binding_set, GDK_Right, 0);
		gtk_binding_entry_remove(clist_binding_set, GDK_Home, 0);
		gtk_binding_entry_remove(clist_binding_set, GDK_End, 0);
		gtk_binding_entry_remove(clist_binding_set, GDK_Escape, 0);
		gtk_binding_entry_remove(clist_binding_set, GDK_space, 0);
		gtk_binding_entry_remove(clist_binding_set, GDK_space, GDK_CONTROL_MASK);
		gtk_binding_entry_remove(clist_binding_set, '/', GDK_CONTROL_MASK);
		gtk_binding_entry_remove(clist_binding_set, '\\', GDK_CONTROL_MASK);
		gtk_binding_entry_remove(clist_binding_set, GDK_Shift_L, GDK_RELEASE_MASK | GDK_SHIFT_MASK);
		gtk_binding_entry_remove(clist_binding_set, GDK_Shift_R, GDK_RELEASE_MASK | GDK_SHIFT_MASK);
		gtk_binding_entry_remove(clist_binding_set, GDK_Shift_L, GDK_RELEASE_MASK | GDK_SHIFT_MASK | GDK_CONTROL_MASK);
		gtk_binding_entry_remove(clist_binding_set, GDK_Shift_R, GDK_RELEASE_MASK | GDK_SHIFT_MASK | GDK_CONTROL_MASK);
		ctree_binding_set = gtk_binding_set_by_class(gtk_type_class (GTK_TYPE_CTREE));
		gtk_binding_entry_remove(ctree_binding_set, '+', GDK_SHIFT_MASK);
		gtk_binding_entry_remove(ctree_binding_set, GDK_KP_Add, 0);
		gtk_binding_entry_remove(ctree_binding_set, GDK_KP_Add, GDK_CONTROL_MASK);
		gtk_binding_entry_remove(ctree_binding_set, '-', 0);
		gtk_binding_entry_remove(ctree_binding_set, GDK_KP_Subtract, 0);
		gtk_binding_entry_remove(ctree_binding_set, GDK_KP_Subtract, GDK_CONTROL_MASK);
		gtk_binding_entry_remove(ctree_binding_set, '=', 0);
		gtk_binding_entry_remove(ctree_binding_set, GDK_KP_Multiply, 0);
		gtk_binding_entry_remove(ctree_binding_set, GDK_KP_Multiply, GDK_CONTROL_MASK);

		gtk_selection_add_targets(GTK_WIDGET(ctree), GDK_SELECTION_PRIMARY, gSelectionTargetList, gSelectionNumberOfTargets);

		gtk_signal_connect_after(GTK_OBJECT(ctree), "size_allocate", GTK_SIGNAL_FUNC(fmWindow_ctreeSizeAllocate), fmWindow);
		gtk_signal_connect(GTK_OBJECT(ctree), "tree_expand", GTK_SIGNAL_FUNC(fmWindow_ctreeExpand), fmWindow);
		gtk_signal_connect_after(GTK_OBJECT(ctree), "tree_collapse", GTK_SIGNAL_FUNC(fmWindow_ctreeCollapse), fmWindow);
		gtk_signal_connect(GTK_OBJECT(ctree), "key_press_event", GTK_SIGNAL_FUNC(fmWindow_ctreeKeyPress), fmWindow);
		gtk_signal_connect(GTK_OBJECT(ctree), "button_press_event", GTK_SIGNAL_FUNC(fmWindow_ctreeButtonPress), fmWindow);
		gtk_signal_connect(GTK_OBJECT(ctree), "button_release_event", GTK_SIGNAL_FUNC(fmWindow_ctreeButtonRelease), fmWindow);
		gtk_signal_connect(GTK_OBJECT(ctree), "motion_notify_event", GTK_SIGNAL_FUNC(fmWindow_ctreeMotionNotify), fmWindow);
		gtk_signal_connect(GTK_OBJECT(ctree), "click_column", GTK_SIGNAL_FUNC(fmWindow_ctreeClickColumn), fmWindow);
		gtk_signal_connect(GTK_OBJECT(ctree), "selection_clear_event", GTK_SIGNAL_FUNC(fmWindow_ctreeSelectionClear), fmWindow);
		gtk_signal_connect(GTK_OBJECT(ctree), "selection_get", GTK_SIGNAL_FUNC(fmWindow_ctreeSelectionGet), fmWindow);

		gtk_drag_dest_set(GTK_WIDGET(ctree), GTK_DEST_DEFAULT_MOTION, gCTreeDragDropTargets, gCTreeNumberOfDragDropTargets, GDK_ACTION_COPY | GDK_ACTION_MOVE);
		gtk_signal_connect(GTK_OBJECT(ctree), "drag_begin", GTK_SIGNAL_FUNC(fmWindow_ctreeDragBegin), fmWindow);
		gtk_signal_connect(GTK_OBJECT(ctree), "drag_motion", GTK_SIGNAL_FUNC(fmWindow_ctreeDragMotion), fmWindow);
		gtk_signal_connect(GTK_OBJECT(ctree), "drag_drop", GTK_SIGNAL_FUNC(fmWindow_ctreeDragDrop), fmWindow);
		gtk_signal_connect(GTK_OBJECT(ctree), "drag_data_received", GTK_SIGNAL_FUNC(fmWindow_ctreeDragDataReceived), fmWindow);
		gtk_signal_connect(GTK_OBJECT(ctree), "drag_data_get", GTK_SIGNAL_FUNC(fmWindow_ctreeDragDataGet), fmWindow);
		gtk_signal_connect(GTK_OBJECT(ctree), "drag_end", GTK_SIGNAL_FUNC(fmWindow_ctreeDragEnd), fmWindow);
		gtk_signal_connect(GTK_OBJECT(ctree), "drag_leave", GTK_SIGNAL_FUNC(fmWindow_ctreeDragLeave), fmWindow);
	}

	return ctree;
}

void fmWindow_buildDir(struct FmWindow *fmWindow, GNode *dirNode, const char *dirName)
{
	struct NodeInfo		*dir_node_info = dirNode->data;
	DIR 				*dir = NULL;
	struct dirent		*dir_entry = NULL;
	gboolean			dir_contents = FALSE;

//	printf("creating dir : %s\n", dirName);

	// try to open the directory with the given name
	dir = opendir(dirName);
	
	if(dir)
	{
		dir_node_info->dirIsAccessible = TRUE;

		// check if there are directory entries (thus more than 2 '.' and '..' are always present
		while((dir_entry = readdir(dir)) && !dir_contents)
		{
			if(fmWindow_useDirEntry(fmWindow, dir_entry))
			{
				dir_contents = TRUE;
			}
		}
		if(dir_contents)
		{
			dir_node_info->dirHasChildren = TRUE;
		}
		closedir(dir);
	}
}

void fmWindow_buildFile(struct FmWindow *fmWindow, GNode *fileNode, const char *fileName)
{
//	printf("creating file : %s\n", fileName);
}

void fmWindow_buildLink(struct FmWindow *fmWindow, GNode *linkNode, const char *linkName)
{
	struct NodeInfo		*link_node_info = linkNode->data;
	gchar				link_representation[PATH_MAX + 1];
	gint				link_representation_bytes = 0;
	gchar				*link_seperator = " -> ";
	gchar				*link_actual_full_name = NULL;
	gchar				*temp_dir = NULL;
	
	glong				link_counter = 0;
	gboolean			link_is_unique_in_chain = TRUE;

//	printf("creating link : %s\n", linkName);

	// try to get the link information
	if( (link_representation_bytes = readlink(link_node_info->pFullName, link_representation, sizeof(link_representation)-1) ) < 0)
	{
		// link is invalid
		link_node_info->pLinkRepresentation = g_strdup(link_node_info->pShortName);
	}
	else
	{
		// create the representation
		link_representation[link_representation_bytes] = '\0';
		if(!g_path_is_absolute(link_representation))
		{
            temp_dir = g_dirname(link_node_info->pFullName);
			link_actual_full_name = g_strconcat(temp_dir, "/", link_representation, NULL);
			g_free(temp_dir);
		}
		else
		{
			link_actual_full_name = g_strdup(link_representation);
		}

		// check if the link isn't already present in the chain to avoid circular references
//		printf("link_representation=%s link_node_info->pFullName=%s link_actual_full_name=%s\n", link_representation, link_node_info->pFullName, link_actual_full_name);
		while(link_counter < fmWindow->pVisitedLinks->len && link_is_unique_in_chain)
		{
//			printf("%ld %ld\n", link_node_info->pStat->st_ino, g_array_index(psVisitedLinks, glong, link_counter));
			if(link_node_info->pStat->st_ino == g_array_index(fmWindow->pVisitedLinks, glong, link_counter))
			{
				link_is_unique_in_chain = FALSE;
			}
            link_counter++;
		}
		
		if(link_is_unique_in_chain)
		{
			// link is valid
			fmWindow->pVisitedLinks = g_array_append_val(fmWindow->pVisitedLinks, link_node_info->pStat->st_ino);
			
			link_node_info->pLinkToNode = fmWindow_buildNode(fmWindow, link_actual_full_name);
			// check if the link isn't broken
			if(link_node_info->pLinkToNode && link_node_info->pLinkToNode->data && ((struct NodeInfo*)link_node_info->pLinkToNode->data)->pStat)
			{
				// link is not broken
//				printf("%s %d 0x%x\n", linkName, isLink(link_node_info->pLinkToNode), ((struct NodeInfo*)link_node_info->pLinkToNode->data)->pLinkToNode);
				// check if the link doesn't link to a link that is the last in the chain
				if(isLink(link_node_info->pLinkToNode) && !((struct NodeInfo*)link_node_info->pLinkToNode->data)->pLinkToNode)
				{
					// free the last constructed gnode and clear the link to it
//					printf("freeing link : %s\n", ((struct NodeInfo*)link_node_info->pLinkToNode->data)->pFullName);

					fmWindow_freeNode(fmWindow, link_node_info->pLinkToNode);
                    link_node_info->pLinkToNode = NULL;
				}
				else
				{
					// connect the new node to the link in the other direction
					((struct NodeInfo*)link_node_info->pLinkToNode->data)->pLinkFromNode = linkNode;
				}
			}
			else
			{
				// link is broken, remove last node
				fmWindow_freeNode(fmWindow, link_node_info->pLinkToNode);
                link_node_info->pLinkToNode = NULL;
			}
			
			fmWindow->pVisitedLinks = g_array_remove_index(fmWindow->pVisitedLinks, fmWindow->pVisitedLinks->len - 1);
		}

		link_node_info->pLinkRepresentation = g_strconcat(link_node_info->pShortName, link_seperator, link_representation, NULL);
        link_node_info->pLinkTarget = g_strdup(link_representation);
		
		g_free(link_actual_full_name);
	}
//	printf("finished building link\n");
}

GNode* fmWindow_buildNode(struct FmWindow *fmWindow, const char *nodeName)
{
	GNode			*source_gnode = NULL;
	struct NodeInfo	*source_node_info = NULL;
	struct stat		*temp_stat = NULL;
	
	// create the entry dirNode
	source_gnode = g_node_new(NULL);

	if(source_gnode)
	{
		source_node_info = g_malloc(sizeof(struct NodeInfo));

		if(source_node_info)
		{
            source_gnode->data = source_node_info;

			source_node_info->pChangeTimeString = NULL;
			source_node_info->pCtreeNode = NULL;
			source_node_info->pDirEmptyCtreeNode = NULL;
			source_node_info->pDirFirstCtreeNode = NULL;
			source_node_info->pFullName = g_strdup(nodeName);
			source_node_info->pGroupString = NULL;
			source_node_info->pLinkRepresentation = NULL;
			source_node_info->pLinkTarget = NULL;
			source_node_info->pLinkToNode = NULL;
			source_node_info->pLinkFromNode = NULL;
			source_node_info->pModificationTimeString = NULL;
			source_node_info->pPermissionsString = NULL;
			if(strcmp(nodeName, fmWindow->pZoomDirectory) == 0)
			{
				source_node_info->pShortName = g_strdup(nodeName+1);
			}
			else
			{
				source_node_info->pShortName = g_strdup(g_basename(nodeName));
			}
			source_node_info->pSizeString = NULL;
			source_node_info->pStat = NULL;
			source_node_info->pUserString = NULL;
			source_node_info->dirHasChildren = FALSE;
			source_node_info->dirIsAccessible = FALSE;
			
			temp_stat = g_malloc(sizeof(struct stat));

			if(temp_stat)
			{
				if(!lstat(nodeName, temp_stat))
				{
                    source_node_info->pStat = temp_stat;

					if(isLink(source_gnode))
					{
						fmWindow_buildLink(fmWindow, source_gnode, nodeName);
					}
					else if(isDirectory(source_gnode))
					{
						fmWindow_buildDir(fmWindow, source_gnode, nodeName);
					}
					else
					{
						fmWindow_buildFile(fmWindow, source_gnode, nodeName);
					}
                    
					if(g_list_find(fmWindow->pCTreeColumns,(gpointer)COLUMN_CHANGE_TIME))
					{
						source_node_info->pChangeTimeString = getChangeTime(source_gnode);
					}
					if(g_list_find(fmWindow->pCTreeColumns,(gpointer)COLUMN_GROUP))
					{
						source_node_info->pGroupString = getGroupName(source_gnode);
					}
					if(g_list_find(fmWindow->pCTreeColumns,(gpointer)COLUMN_MODIFICATION_TIME))
					{
						source_node_info->pModificationTimeString = getModificationTime(source_gnode);
					}
					if(g_list_find(fmWindow->pCTreeColumns,(gpointer)COLUMN_PERMISSIONS))
					{
						source_node_info->pPermissionsString = getPermissions(source_gnode);
					}
					if(g_list_find(fmWindow->pCTreeColumns,(gpointer)COLUMN_SIZE))
					{
						source_node_info->pSizeString = getSize(source_gnode);
					}
					if(g_list_find(fmWindow->pCTreeColumns,(gpointer)COLUMN_USER))
					{
						source_node_info->pUserString = getUserName(source_gnode);
					}

					fmWindow->build++;
					return source_gnode;
				}
				g_free(temp_stat);
				fmWindow->build++;
				return source_gnode;
			}

			printf("ERROR : Can't allocate memory for stat structure.\n");
			g_free(source_node_info->pFullName);
			g_free(source_node_info->pShortName);
			g_free(source_node_info);
		}
		printf("ERROR : Can't allocate memory for NodeInfo structure.\n");
		g_node_destroy(source_gnode);
	}
	printf("ERROR : Can't allocate memory for GNode structure.\n");

	return NULL;
}

gboolean fmWindow_collapseNode(struct FmWindow *fmWindow, GNode *targetNode)
{
	struct NodeInfo *target_node_info = NULL;
	
//	printf("fmWindow_collapseNode()\n");
	if(fmWindow && targetNode)
	{
		target_node_info = (struct NodeInfo*)targetNode->data;
		if(target_node_info->pDirFirstCtreeNode)
		{
			gtk_ctree_collapse(GTK_CTREE(fmWindow->pCTree), GTK_CTREE_NODE(target_node_info->pCtreeNode));
			fmWindow_resizeColumns(fmWindow);
			return TRUE;
		}
	}

	return FALSE;
}

void fmWindow_createBookmarkButton(struct FmWindow *fmWindow, guint index)
{
	gchar			*index_string = NULL;
	
	index_string = g_new0(gchar, 2);
	index_string[0] = '0'+index;
	fmWindow->pBookmarkButtons[index].pButtonBookmark = gtk_button_new_with_label(index_string);
	if(fmWindow->pBookmarks[index])
	{
		gtk_widget_set_sensitive(GTK_WIDGET(GTK_BIN(fmWindow->pBookmarkButtons[index].pButtonBookmark)->child), TRUE);
		gtk_tooltips_set_tip(GTK_TOOLTIPS(fmWindow->pToolTips), GTK_WIDGET(fmWindow->pBookmarkButtons[index].pButtonBookmark), fmWindow->pBookmarks[index], "");
	}
	else
	{
		gtk_widget_set_sensitive(GTK_WIDGET(GTK_BIN(fmWindow->pBookmarkButtons[index].pButtonBookmark)->child), FALSE);
	}
	gtk_box_pack_start(GTK_BOX(fmWindow->pHBoxBottomButtons), GTK_WIDGET(fmWindow->pBookmarkButtons[index].pButtonBookmark), TRUE, TRUE, 0);
	gtk_object_set_data(GTK_OBJECT(fmWindow->pBookmarkButtons[index].pButtonBookmark), "index", (gpointer)index);
	gtk_object_set_data(GTK_OBJECT(fmWindow->pBookmarkButtons[index].pButtonBookmark), "fmWindow", (gpointer)fmWindow);
	gtk_signal_connect(GTK_OBJECT(fmWindow->pBookmarkButtons[index].pButtonBookmark), "destroy", GTK_SIGNAL_FUNC(fmWindow_destroyButtonBookmark), fmWindow);
	gtk_signal_connect(GTK_OBJECT(fmWindow->pBookmarkButtons[index].pButtonBookmark), "button_press_event", GTK_SIGNAL_FUNC(fmWindow_buttonBookmarkButtonPress), fmWindow);
	gtk_signal_connect(GTK_OBJECT(fmWindow->pBookmarkButtons[index].pButtonBookmark), "clicked", GTK_SIGNAL_FUNC(fmWindow_buttonBookmarkClicked), fmWindow);
	gtk_signal_connect(GTK_OBJECT(fmWindow->pBookmarkButtons[index].pButtonBookmark), "enter_notify_event", GTK_SIGNAL_FUNC(fmWindow_buttonBookmarkEnterNotify), fmWindow);
	gtk_signal_connect(GTK_OBJECT(fmWindow->pBookmarkButtons[index].pButtonBookmark), "leave_notify_event", GTK_SIGNAL_FUNC(fmWindow_buttonBookmarkLeaveNotify), fmWindow);

	// create the bookmark button popup menu
	fmWindow->pBookmarkButtons[index].pPopupMenu = gtk_menu_new();
	gtk_menu_set_accel_group(GTK_MENU(fmWindow->pBookmarkButtons[index].pPopupMenu), fmWindow->pAccelGroup);

	// create the popup menu items
	fmWindow->pBookmarkButtons[index].pPopupMenuItemGoTo = gtk_menu_item_new_with_label("Go to");
	gtk_signal_connect(GTK_OBJECT(fmWindow->pBookmarkButtons[index].pPopupMenuItemGoTo), "destroy", GTK_SIGNAL_FUNC(fmWindow_destroyMenuItem), fmWindow);
	gtk_signal_connect(GTK_OBJECT(fmWindow->pBookmarkButtons[index].pPopupMenuItemGoTo), "activate", GTK_SIGNAL_FUNC(fmWindow_buttonBookmarkGoTo), fmWindow);
	gtk_signal_connect(GTK_OBJECT(fmWindow->pBookmarkButtons[index].pPopupMenuItemGoTo), "add_accelerator", GTK_SIGNAL_FUNC(fmWindow_acceleratorAdd), fmWindow);
	gtk_signal_connect_after(GTK_OBJECT(fmWindow->pBookmarkButtons[index].pPopupMenuItemGoTo), "remove_accelerator", GTK_SIGNAL_FUNC(fmWindow_acceleratorRemove), fmWindow);
	gtk_object_set_data(GTK_OBJECT(fmWindow->pBookmarkButtons[index].pPopupMenuItemGoTo), "bookmarkButton", fmWindow->pBookmarkButtons[index].pButtonBookmark);
	gtk_object_set_data(GTK_OBJECT(fmWindow->pBookmarkButtons[index].pPopupMenuItemGoTo), "accelKeyId", (gpointer)g_strconcat("BookmarkGoTo", index_string, NULL));
	fmWindow_accelKeyAddToGtkWidget(fmWindow, GTK_WIDGET(fmWindow->pBookmarkButtons[index].pPopupMenuItemGoTo));
	
	fmWindow->pBookmarkButtons[index].pPopupMenuItemSet = gtk_menu_item_new_with_label("Set");
	gtk_signal_connect(GTK_OBJECT(fmWindow->pBookmarkButtons[index].pPopupMenuItemSet), "destroy", GTK_SIGNAL_FUNC(fmWindow_destroyMenuItem), fmWindow);
	gtk_signal_connect(GTK_OBJECT(fmWindow->pBookmarkButtons[index].pPopupMenuItemSet), "activate", GTK_SIGNAL_FUNC(fmWindow_buttonBookmarkSet), fmWindow);
	gtk_signal_connect(GTK_OBJECT(fmWindow->pBookmarkButtons[index].pPopupMenuItemSet), "add_accelerator", GTK_SIGNAL_FUNC(fmWindow_acceleratorAdd), fmWindow);
	gtk_signal_connect_after(GTK_OBJECT(fmWindow->pBookmarkButtons[index].pPopupMenuItemSet), "remove_accelerator", GTK_SIGNAL_FUNC(fmWindow_acceleratorRemove), fmWindow);
	gtk_object_set_data(GTK_OBJECT(fmWindow->pBookmarkButtons[index].pPopupMenuItemSet), "bookmarkButton", fmWindow->pBookmarkButtons[index].pButtonBookmark);
	gtk_object_set_data(GTK_OBJECT(fmWindow->pBookmarkButtons[index].pPopupMenuItemSet), "accelKeyId", (gpointer)g_strconcat("BookmarkSet", index_string, NULL));
	fmWindow_accelKeyAddToGtkWidget(fmWindow, GTK_WIDGET(fmWindow->pBookmarkButtons[index].pPopupMenuItemSet));

	fmWindow->pBookmarkButtons[index].pPopupMenuItemClear = gtk_menu_item_new_with_label("Clear");
	gtk_signal_connect(GTK_OBJECT(fmWindow->pBookmarkButtons[index].pPopupMenuItemClear), "destroy", GTK_SIGNAL_FUNC(fmWindow_destroyMenuItem), fmWindow);
	gtk_signal_connect(GTK_OBJECT(fmWindow->pBookmarkButtons[index].pPopupMenuItemClear), "activate", GTK_SIGNAL_FUNC(fmWindow_buttonBookmarkClear), fmWindow);
	gtk_signal_connect(GTK_OBJECT(fmWindow->pBookmarkButtons[index].pPopupMenuItemClear), "add_accelerator", GTK_SIGNAL_FUNC(fmWindow_acceleratorAdd), fmWindow);
	gtk_signal_connect_after(GTK_OBJECT(fmWindow->pBookmarkButtons[index].pPopupMenuItemClear), "remove_accelerator", GTK_SIGNAL_FUNC(fmWindow_acceleratorRemove), fmWindow);
	gtk_object_set_data(GTK_OBJECT(fmWindow->pBookmarkButtons[index].pPopupMenuItemClear), "bookmarkButton", fmWindow->pBookmarkButtons[index].pButtonBookmark);
	gtk_object_set_data(GTK_OBJECT(fmWindow->pBookmarkButtons[index].pPopupMenuItemClear), "accelKeyId", (gpointer)g_strconcat("BookmarkClear", index_string, NULL));
	fmWindow_accelKeyAddToGtkWidget(fmWindow, GTK_WIDGET(fmWindow->pBookmarkButtons[index].pPopupMenuItemClear));

	// add the items to the popup menu
	gtk_menu_append(GTK_MENU(fmWindow->pBookmarkButtons[index].pPopupMenu), fmWindow->pBookmarkButtons[index].pPopupMenuItemGoTo);
	gtk_menu_append(GTK_MENU(fmWindow->pBookmarkButtons[index].pPopupMenu), fmWindow->pBookmarkButtons[index].pPopupMenuItemSet);
	gtk_menu_append(GTK_MENU(fmWindow->pBookmarkButtons[index].pPopupMenu), fmWindow->pBookmarkButtons[index].pPopupMenuItemClear);

	// show the popup menu and its items
	gtk_widget_show(GTK_WIDGET(fmWindow->pBookmarkButtons[index].pPopupMenuItemGoTo));
	gtk_widget_show(GTK_WIDGET(fmWindow->pBookmarkButtons[index].pPopupMenuItemSet));
	gtk_widget_show(GTK_WIDGET(fmWindow->pBookmarkButtons[index].pPopupMenuItemClear));
	gtk_widget_show(GTK_WIDGET(fmWindow->pBookmarkButtons[index].pPopupMenu));

    // set the correct popup menu disabled state
	fmWindow_bookmarkUpdatePopupMenu(fmWindow, index);
	
	g_free(index_string);
}

void fmWindow_ctreeColumnsApplyAlignment(struct FmWindow *fmWindow)
{
	GList	*ctree_column = pgPreferences->pCtreeColumns;
	gint	counter = 0;

	while(ctree_column)
	{
		if(g_list_find(fmWindow->pCTreeColumns,(gpointer)((struct CtreeColumn*)ctree_column->data)->type))
		{
			gtk_clist_set_column_justification(GTK_CLIST(fmWindow->pCTree), counter, ((struct CtreeColumn*)ctree_column->data)->alignment);
			counter++;
		}
		ctree_column = ctree_column->next;
	}
}

gint fmWindow_ctreeHeight(struct FmWindow *fmWindow)
{
	return ( (GTK_CTREE(fmWindow->pCTree)->clist.row_height * GTK_CTREE(fmWindow->pCTree)->clist.rows) + (CELL_SPACING * (GTK_CTREE(fmWindow->pCTree)->clist.rows + 1)) );
}

void fmWindow_ctreeMoveHorizontal(struct FmWindow *fmWindow, gint offset)
{
	gfloat value;
	
	if(GTK_CTREE(fmWindow->pCTree)->clist.hadjustment)
	{
		value = CLAMP(GTK_CTREE(fmWindow->pCTree)->clist.hadjustment->value + offset, 0.0, GTK_CTREE(fmWindow->pCTree)->clist.hadjustment->upper - GTK_CTREE(fmWindow->pCTree)->clist.hadjustment->page_size);
		gtk_adjustment_set_value(GTK_CTREE(fmWindow->pCTree)->clist.hadjustment, value);
	}
}

void fmWindow_ctreeMoveVertical(struct FmWindow *fmWindow, gint offset)
{
	gfloat value;
	
	if(GTK_CTREE(fmWindow->pCTree)->clist.hadjustment)
	{
		value = CLAMP(GTK_CTREE(fmWindow->pCTree)->clist.vadjustment->value + offset, 0.0, GTK_CTREE(fmWindow->pCTree)->clist.vadjustment->upper - GTK_CTREE(fmWindow->pCTree)->clist.vadjustment->page_size);
		gtk_adjustment_set_value(GTK_CTREE(fmWindow->pCTree)->clist.vadjustment, value);
	}
}

void fmWindow_ctreeScrollHorizontal(struct FmWindow *fmWindow, gint x)
{
	if(!GTK_CTREE(fmWindow->pCTree)->clist.htimer)
	{
		if(fmWindow_ctreeScrollHorizontalValid(fmWindow, x))
		{
			GTK_CTREE(fmWindow->pCTree)->clist.htimer = gtk_timeout_add(AUTOSCROLL_TIME, (GtkFunction)fmWindow_ctreeScrollHorizontalTimeout, fmWindow);

			// let all the changements execute before continuing
			while(gtk_events_pending())
			{
				gtk_main_iteration();
			}
			
			if (x < AUTOSCROLL_BORDER)
			{
				fmWindow_ctreeMoveHorizontal(fmWindow, (AUTOSCROLL_BORDER - x) * -3);
			}
			else
			{
				fmWindow_ctreeMoveHorizontal(fmWindow, (AUTOSCROLL_BORDER - (GTK_CTREE(fmWindow->pCTree)->clist.clist_window_width - x)) * 3);
			}
		}
	}
}

gint fmWindow_ctreeScrollHorizontalTimeout(struct FmWindow *fmWindow)
{
	gint x = 0;
	gint y = 0;

    GTK_CTREE(fmWindow->pCTree)->clist.htimer = 0;
    gtk_widget_get_pointer(GTK_WIDGET(fmWindow->pCTree), &x, &y);
	fmWindow_ctreeScrollHorizontal(fmWindow, x);
  
	return FALSE;
}

gboolean fmWindow_ctreeScrollHorizontalValid(struct FmWindow *fmWindow, gint x)
{
	if(GTK_CTREE(fmWindow->pCTree)->clist.hadjustment &&
	   (fmWindow_ctreeWidth(fmWindow) > GTK_CTREE(fmWindow->pCTree)->clist.clist_window_width) &&
	   ((x >= 0 && x <= AUTOSCROLL_BORDER) || (x >= GTK_CTREE(fmWindow->pCTree)->clist.clist_window_width - AUTOSCROLL_BORDER && x <= GTK_CTREE(fmWindow->pCTree)->clist.clist_window_width) ))
	{
		return TRUE;
	}
	else
	{
		return FALSE;
	}
}

void fmWindow_ctreeScrollVertical(struct FmWindow *fmWindow, gint y)
{
	gint title_height = GTK_CTREE(fmWindow->pCTree)->clist.column_title_area.height;

	if(!GTK_CTREE(fmWindow->pCTree)->clist.vtimer)
	{
		if(fmWindow_ctreeScrollVerticalValid(fmWindow, y))
		{
			GTK_CTREE(fmWindow->pCTree)->clist.vtimer = gtk_timeout_add(AUTOSCROLL_TIME, (GtkFunction)fmWindow_ctreeScrollVerticalTimeout, fmWindow);

			// let all the changements execute before continuing
			while(gtk_events_pending())
			{
				gtk_main_iteration();
			}
			
			if (y - title_height < AUTOSCROLL_BORDER)
			{
				fmWindow_ctreeMoveVertical(fmWindow, (AUTOSCROLL_BORDER - (y - title_height)) * -3);
			}
			else
			{
				fmWindow_ctreeMoveVertical(fmWindow, (AUTOSCROLL_BORDER - ((title_height + GTK_CTREE(fmWindow->pCTree)->clist.clist_window_height) - y)) * 3);
			}
		}
	}
}

gint fmWindow_ctreeScrollVerticalTimeout(struct FmWindow *fmWindow)
{
	gint x = 0;
	gint y = 0;

    GTK_CTREE(fmWindow->pCTree)->clist.vtimer = 0;
    gtk_widget_get_pointer(GTK_WIDGET(fmWindow->pCTree), &x, &y);
	fmWindow_ctreeScrollVertical(fmWindow, y);
  
	return FALSE;
}

gboolean fmWindow_ctreeScrollVerticalValid(struct FmWindow *fmWindow, gint y)
{
	gint title_height = GTK_CTREE(fmWindow->pCTree)->clist.column_title_area.height;
	
	if(GTK_CTREE(fmWindow->pCTree)->clist.vadjustment &&
	   (fmWindow_ctreeHeight(fmWindow) > GTK_CTREE(fmWindow->pCTree)->clist.clist_window_height) &&
	   ((y >= 0 && y <= AUTOSCROLL_BORDER + title_height) || (y >= (title_height + GTK_CTREE(fmWindow->pCTree)->clist.clist_window_height) - AUTOSCROLL_BORDER && y <= (title_height + GTK_CTREE(fmWindow->pCTree)->clist.clist_window_height))))
	{
		return TRUE;
	}
	else
	{
		return FALSE;
	}
}

gint fmWindow_ctreeWidth(struct FmWindow *fmWindow)
{
	gint last_column;

	for(last_column = GTK_CTREE(fmWindow->pCTree)->clist.columns - 1; last_column >= 0 && !GTK_CTREE(fmWindow->pCTree)->clist.column[last_column].visible; last_column--);

	if (last_column >= 0)
	{
		return (GTK_CTREE(fmWindow->pCTree)->clist.column[last_column].area.x + GTK_CTREE(fmWindow->pCTree)->clist.column[last_column].area.width + COLUMN_INSET + CELL_SPACING);
	}
	
	return 0;
}

void fmWindow_destroy(struct FmWindow *fmWindow)
{
	// destroy the dynamically generated combo box popup menu data
    fmWindow_freeComboBoxPopupMenu(fmWindow);

	// destroy all the gtk widgets
	gtk_widget_destroy(fmWindow->pPopupMenuItemQuit);
	gtk_widget_destroy(fmWindow->pPopupMenuItemSeperator2);
	gtk_widget_destroy(fmWindow->pPopupMenuItemPreferences);
	gtk_widget_destroy(fmWindow->pPopupMenuItemAbout);
	gtk_widget_destroy(fmWindow->pPopupMenuItemSeperator1);
	gtk_widget_destroy(fmWindow->pPopupMenuColumnsMenuItemChangeTime);
	gtk_widget_destroy(fmWindow->pPopupMenuColumnsMenuItemModificationTime);
	gtk_widget_destroy(fmWindow->pPopupMenuColumnsMenuItemPermissions);
	gtk_widget_destroy(fmWindow->pPopupMenuColumnsMenuItemGroup);
	gtk_widget_destroy(fmWindow->pPopupMenuColumnsMenuItemUser);
	gtk_widget_destroy(fmWindow->pPopupMenuColumnsMenuItemSize);
	gtk_widget_destroy(fmWindow->pPopupMenuColumns);
	gtk_widget_destroy(fmWindow->pPopupMenuItemColumns);
	gtk_widget_destroy(fmWindow->pPopupMenuBrowserMenuItemPasteState);
	gtk_widget_destroy(fmWindow->pPopupMenuBrowserMenuItemCopyState);
	gtk_widget_destroy(fmWindow->pPopupMenuBrowserMenuItemUnZoom);
	gtk_widget_destroy(fmWindow->pPopupMenuBrowserMenuItemCollapseEverything);
	gtk_widget_destroy(fmWindow->pPopupMenuBrowserMenuItemClose);
	gtk_widget_destroy(fmWindow->pPopupMenuBrowserMenuItemNew);
	gtk_widget_destroy(fmWindow->pPopupMenuBrowserMenuItemRefresh);
	gtk_widget_destroy(fmWindow->pPopupMenuBrowser);
	gtk_widget_destroy(fmWindow->pPopupMenuItemBrowser);
	gtk_widget_destroy(fmWindow->pPopupMenuDirectoryMenuItemOpenInNewWindow);
	gtk_widget_destroy(fmWindow->pPopupMenuDirectoryMenuItemZoom);
	gtk_widget_destroy(fmWindow->pPopupMenuDirectoryMenuItemRefresh);
	gtk_widget_destroy(fmWindow->pPopupMenuDirectory);
	gtk_widget_destroy(fmWindow->pPopupMenuItemDirectory);
	gtk_widget_destroy(fmWindow->pPopupMenuSelectionMenuItemCopyFocusToComboEntry);
	gtk_widget_destroy(fmWindow->pPopupMenuSelectionMenuItemCopyToXClipboard);
	gtk_widget_destroy(fmWindow->pPopupMenuSelectionMenuItemDeselectDirectory);
	gtk_widget_destroy(fmWindow->pPopupMenuSelectionMenuItemSelectDirectory);
	gtk_widget_destroy(fmWindow->pPopupMenuSelectionMenuItemClear);
	gtk_widget_destroy(fmWindow->pPopupMenuSelection);
	gtk_widget_destroy(fmWindow->pPopupMenuItemSelection);
	gtk_widget_destroy(fmWindow->pPopupMenu);
	gtk_widget_destroy(fmWindow->pLabelSelectionSize);
	gtk_widget_destroy(fmWindow->pAlignmentSelectionSize);
	gtk_widget_destroy(fmWindow->pFrameSelectionSize);
	gtk_widget_destroy(fmWindow->pLabelGeneralStatus);
	gtk_widget_destroy(fmWindow->pFrameGeneralStatus);
	gtk_widget_destroy(fmWindow->pHBoxBottomStatus);
	gtk_widget_destroy(fmWindow->pBookmarkButtons[0].pButtonBookmark);
	gtk_widget_destroy(fmWindow->pBookmarkButtons[9].pButtonBookmark);
	gtk_widget_destroy(fmWindow->pBookmarkButtons[8].pButtonBookmark);
	gtk_widget_destroy(fmWindow->pBookmarkButtons[7].pButtonBookmark);
	gtk_widget_destroy(fmWindow->pBookmarkButtons[6].pButtonBookmark);
	gtk_widget_destroy(fmWindow->pBookmarkButtons[5].pButtonBookmark);
	gtk_widget_destroy(fmWindow->pBookmarkButtons[4].pButtonBookmark);
	gtk_widget_destroy(fmWindow->pBookmarkButtons[3].pButtonBookmark);
	gtk_widget_destroy(fmWindow->pBookmarkButtons[2].pButtonBookmark);
	gtk_widget_destroy(fmWindow->pBookmarkButtons[1].pButtonBookmark);
	gtk_widget_destroy(fmWindow->pHBoxBottomButtons);
	gtk_widget_destroy(fmWindow->pPixmapPanWindow);
	gtk_widget_destroy(fmWindow->pEventBoxPanWindow);
	gtk_widget_destroy(fmWindow->pVScrollbarScrolledWindow);
	gtk_widget_destroy(fmWindow->pVBoxScrolledWindowRight);
	gtk_widget_destroy(fmWindow->pCTree);
	gtk_widget_destroy(fmWindow->pHScrollbarScrolledWindow);
	gtk_widget_destroy(fmWindow->pVBoxScrolledWindowLeft);
	gtk_widget_destroy(fmWindow->pHBoxScrolledWindow);
	gtk_widget_destroy(fmWindow->pComboBoxPopupMenu);
	gtk_widget_destroy(fmWindow->pComboBox);
	gtk_widget_destroy(fmWindow->pVBox);
/*
	gtk_widget_destroy(fmWindow->pButtonHideDotfiles);
	gtk_widget_destroy(fmWindow->pHBoxTop);
	gtk_widget_destroy(fmWindow->pWindow);
*/

	// free the zoom directory
	if(fmWindow->pZoomDirectory)
	{
		g_free(fmWindow->pZoomDirectory);
	}

	// free the shortcut string
    g_string_free(fmWindow->pShortcut, TRUE);
	
	// free the bookmarks array
    g_free(fmWindow->pBookmarks);

	// free the hastable of accelerated widgets
	g_hash_table_destroy(fmWindow->pAcceleratedWidgets);

	// free the directory nodes
	fmWindow_freeNode(fmWindow, fmWindow->pRootNode);
	
	// free the list of selected nodes
	fmWindow_selectionDestroy(fmWindow);

	// free the list of dir entries
	g_list_free(fmWindow->pExpandedDirs);

	// free the visited links array
	g_array_free(fmWindow->pVisitedLinks, TRUE);
	
	// free the eventual selection text snapshot
	if(fmWindow->pSelectionTextSnapshot)
	{
		g_free(fmWindow->pSelectionTextSnapshot);
	}

	// free the eventual initial focus
	if(fmWindow->pInitialFocus)
	{
		g_free(fmWindow->pInitialFocus);
	}

//	printf("Build %i, Freed %i\n", fmWindow->build, fmWindow->freed);
	g_free(fmWindow);
}

gboolean fmWindow_expandNode(struct FmWindow *fmWindow, GNode *targetNode)
{
	struct NodeInfo *target_node_info = NULL;
	
//	printf("fmWindow_expandNode()\n");
	if(fmWindow && targetNode)
	{
		target_node_info = (struct NodeInfo*)targetNode->data;
		if(target_node_info->pDirEmptyCtreeNode)
		{
			gtk_ctree_expand(GTK_CTREE(fmWindow->pCTree), GTK_CTREE_NODE(target_node_info->pCtreeNode));
			fmWindow_resizeColumns(fmWindow);
			return TRUE;
		}
	}

	return FALSE;
}

gboolean fmWindow_fillDir(struct FmWindow *fmWindow, GNode *targetNode)
{
	DIR 				*dir = NULL;
	struct NodeInfo		*target_node_info = NULL;
	gchar				*dir_name = NULL;
	struct dirent		*dir_entry = NULL;
	gchar				*dir_entry_name = NULL;
	GList  				*directory_nodes = NULL;
	
	GList				*dir_glist_entry = NULL;
	GNode				*entry_gnode = NULL;
	GString				*temp_fullname = NULL;
	GNode				*last_inserted_gnode = NULL;

	gchar				*name_to_remember = NULL;
    guint				number_of_children = 0;
	guint				child_id = 0;
	GNode				*child_source_gnode = NULL;
	GNode				*child_target_gnode = NULL;

	// get the directory name
	target_node_info = (struct NodeInfo*)targetNode->data;
	dir_name = ((struct NodeInfo*)targetNode->data)->pFullName;
	
	// try to open the directory with the given name
//	printf("reading %s\n", dir_name);
	dir = opendir(dir_name);
	if(dir)
	{
		while((dir_entry = readdir(dir)))
		{
			if(fmWindow_useDirEntry(fmWindow, dir_entry))
			{
				// create the full directory name to store in the NodeInfo structure
				temp_fullname = g_string_new(dir_name);
				if(temp_fullname)
				{
					if(dir_name[strlen(dir_name)-1] != '/')
					{
						g_string_append(temp_fullname, "/");
					}
					g_string_append(temp_fullname, dir_entry->d_name);
					g_free(dir_entry_name);

					// build the corresponding gnode
//					printf("temp_fullname->str = %s\n", temp_fullname->str);
					entry_gnode = fmWindow_buildNode(fmWindow, temp_fullname->str);
//					printf("entry_gnode=0x%x\n", (guint)entry_gnode);
					if(entry_gnode)
					{
//						printf("((struct NodeInfo*)entry_gnode->data)->pShortName = %s\n", ((struct NodeInfo*)entry_gnode->data)->pShortName);
						directory_nodes = g_list_prepend(directory_nodes, entry_gnode);
					}

					g_string_free(temp_fullname, TRUE);
				}
			}
		}

		closedir(dir);
	}
	else
	{
		return FALSE;
	}
//	printf("finished reading %s\n", dir_name);

//	printf("sorting directory entries\n");
	if(fmWindow->sortDirection == SORT_ASCENDING)
	{
		if(fmWindow->sortType == SORT_FILENAME)
		{
			directory_nodes = g_list_sort(directory_nodes, (GCompareFunc)sortNodesByFilenameAscending);
		}
		else if(fmWindow->sortType == SORT_SIZE)
		{
			directory_nodes = g_list_sort(directory_nodes, (GCompareFunc)sortNodesBySizeAscending);
		}
		else if(fmWindow->sortType == SORT_USER)
		{
			directory_nodes = g_list_sort(directory_nodes, (GCompareFunc)sortNodesByUserAscending);
		}
		else if(fmWindow->sortType == SORT_GROUP)
		{
			directory_nodes = g_list_sort(directory_nodes, (GCompareFunc)sortNodesByGroupAscending);
		}
		else if(fmWindow->sortType == SORT_PERMISSIONS)
		{
			directory_nodes = g_list_sort(directory_nodes, (GCompareFunc)sortNodesByPermissionsAscending);
		}
		else if(fmWindow->sortType == SORT_MODIFICATION_TIME)
		{
			directory_nodes = g_list_sort(directory_nodes, (GCompareFunc)sortNodesByModificationTimeAscending);
		}
		else if(fmWindow->sortType == SORT_CHANGE_TIME)
		{
			directory_nodes = g_list_sort(directory_nodes, (GCompareFunc)sortNodesByChangeTimeAscending);
		}
	}
	else
	{
		if(fmWindow->sortType == SORT_FILENAME)
		{
			directory_nodes = g_list_sort(directory_nodes, (GCompareFunc)sortNodesByFilenameDescending);
		}
		else if(fmWindow->sortType == SORT_SIZE)
		{
			directory_nodes = g_list_sort(directory_nodes, (GCompareFunc)sortNodesBySizeDescending);
		}
		else if(fmWindow->sortType == SORT_USER)
		{
			directory_nodes = g_list_sort(directory_nodes, (GCompareFunc)sortNodesByUserDescending);
		}
		else if(fmWindow->sortType == SORT_GROUP)
		{
			directory_nodes = g_list_sort(directory_nodes, (GCompareFunc)sortNodesByGroupDescending);
		}
		else if(fmWindow->sortType == SORT_PERMISSIONS)
		{
			directory_nodes = g_list_sort(directory_nodes, (GCompareFunc)sortNodesByPermissionsDescending);
		}
		else if(fmWindow->sortType == SORT_MODIFICATION_TIME)
		{
			directory_nodes = g_list_sort(directory_nodes, (GCompareFunc)sortNodesByModificationTimeDescending);
		}
		else if(fmWindow->sortType == SORT_CHANGE_TIME)
		{
			directory_nodes = g_list_sort(directory_nodes, (GCompareFunc)sortNodesByChangeTimeDescending);
		}
	}
//	printf("finished sorting directory entries\n");

//	printf("creating glist branch\n");
	for(dir_glist_entry = g_list_first(directory_nodes); dir_glist_entry; dir_glist_entry = g_list_next(dir_glist_entry))
	{
		entry_gnode = dir_glist_entry->data;
		if(entry_gnode)
		{
			if(pgPreferences->sortDirectoriesFirst)
			{
				if(fmWindow->sortDirection == SORT_ASCENDING)
				{
					if(isDirectory(getLinkTarget(entry_gnode)))
					{
						g_node_prepend(targetNode, entry_gnode);
					}
					else if(!last_inserted_gnode)
					{
						last_inserted_gnode = g_node_append(targetNode, entry_gnode);
					}
					else
					{
						last_inserted_gnode = g_node_insert_before(targetNode, last_inserted_gnode, entry_gnode);
					}
				}
				else
				{
					if(isDirectory(getLinkTarget(entry_gnode)))
					{
						if(!last_inserted_gnode)
						{
							last_inserted_gnode = g_node_append(targetNode, entry_gnode);
						}
						else
						{
							last_inserted_gnode = g_node_insert_before(targetNode, last_inserted_gnode, entry_gnode);
						}
					}
					else
					{
						g_node_prepend(targetNode, entry_gnode);
					}
				}
			}
			else
			{
				g_node_prepend(targetNode, entry_gnode);
			}
		}
	}
	g_list_free(directory_nodes);
//	printf("finished creating glist branch\n");

//	printf("filling in gtk ctree branch\n");
	number_of_children = g_node_n_children(targetNode);
//	printf("number_of_children = %d\n", number_of_children);
	if(number_of_children)
	{
		child_id = 0;
		while(child_id < number_of_children)
		{
			child_source_gnode = g_node_nth_child(targetNode, child_id);

			if(child_source_gnode)
			{
				child_target_gnode = getLinkTarget(child_source_gnode);
				fmWindow_insertCTreeNode(fmWindow, child_source_gnode, target_node_info->pCtreeNode, NULL, NULL);

				// check if it's a directory and if it's already expanded
				name_to_remember = fmWindow_getFilenameToRemember(fmWindow, child_source_gnode);
				
				if(name_to_remember)
				{
//					printf("finding %s\n", name_to_remember);
					if(g_list_find_custom(fmWindow->pExpandedDirs, name_to_remember, (GCompareFunc)strcmp))
					{
//						printf("additionally expand %s\n", name_to_remember);
						// only expand if the node really has children
						if(((struct NodeInfo*)child_target_gnode->data)->dirHasChildren)
						{
							gtk_ctree_expand(GTK_CTREE(fmWindow->pCTree), ((struct NodeInfo*)child_source_gnode->data)->pCtreeNode);
						}
					}
					g_free(name_to_remember);
					name_to_remember = NULL;
				}
				
				if(target_node_info->pDirEmptyCtreeNode)
				{
					gtk_ctree_remove_node(GTK_CTREE(fmWindow->pCTree), GTK_CTREE_NODE(target_node_info->pDirEmptyCtreeNode));
					target_node_info->pDirFirstCtreeNode = ((struct NodeInfo*)child_target_gnode->data)->pCtreeNode;
					target_node_info->pDirEmptyCtreeNode = NULL;
				}
			}

			child_id++;
		}
	}
//	printf("finished filling in gtk ctree branch\n");

	return TRUE;
}

void fmWindow_focusCTreeRow(struct FmWindow *fmWindow, guint child_row)
{
	guint		previously_focused_row = GTK_CLIST(fmWindow->pCTree)->focus_row;
	GNode		*previously_focused_source_gnode = NULL;

	gtk_widget_grab_focus(GTK_WIDGET(fmWindow->pCTree));
	GTK_CLIST(fmWindow->pCTree)->focus_row = child_row;

	// check the previously focused row is part of the selection
	previously_focused_source_gnode = gtk_ctree_node_get_row_data(GTK_CTREE(fmWindow->pCTree), gtk_ctree_node_nth(GTK_CTREE(fmWindow->pCTree), previously_focused_row)); 
	if(fmWindow_selectionGetListNode(fmWindow, previously_focused_source_gnode) == NULL)
	{
		// no, unselect it
		fmWindow_unselectCTreeRow(fmWindow, previously_focused_row);
	}
	else
	{
		// yes, apply the default selection style
		fmWindow_selectCTreeRow(fmWindow, previously_focused_row);
	}
	gtk_widget_draw_focus(GTK_WIDGET(fmWindow->pCTree));
    fmWindow_focusDrawInfo(fmWindow, child_row);
}

gboolean fmWindow_focusDeepestExpandedNode(struct FmWindow *fmWindow, GNode *targetNode)
{
	struct NodeInfo *node_info = NULL;
	GNode			*temp_gnode = NULL;
	
	if(fmWindow && targetNode)
	{
		node_info = (struct NodeInfo*)targetNode->data;
		while(node_info->pDirFirstCtreeNode)
		{
			temp_gnode = gtk_ctree_node_get_row_data(GTK_CTREE(fmWindow->pCTree), node_info->pDirFirstCtreeNode);
			while(temp_gnode)
			{
				if(((struct NodeInfo*)getLinkTarget(temp_gnode)->data)->pDirFirstCtreeNode)
				{
					break;
				}
				temp_gnode = temp_gnode->next;
			}
			if(!temp_gnode)
			{
				temp_gnode = gtk_ctree_node_get_row_data(GTK_CTREE(fmWindow->pCTree), node_info->pDirFirstCtreeNode);
			}
			
			targetNode = getLinkTarget(temp_gnode);
			node_info = (struct NodeInfo*)targetNode->data;
		}
		fmWindow_focusNode(fmWindow, getLinkSource(targetNode));
		fmWindow_selectMoveToCTreeFocus(fmWindow);
		return TRUE;
	}

	return FALSE;
}

void fmWindow_focusDrawInfo(struct FmWindow *fmWindow, guint child_row)
{
	GNode		*source_gnode = NULL;
	gchar		*ellipsis = "...";
	gint		ellipsis_width = 0;
	gchar		*new_filename = NULL;
	gint		new_filename_width = 0;
	gchar		*new_filename_bis = NULL;
	gchar		*new_filename_last_part = NULL;
	GtkStyle	*statusbar_label_style = NULL;
	gint		label_width = 0;
	gchar		*displayed_filename = NULL;

	source_gnode = gtk_ctree_node_get_row_data(GTK_CTREE(fmWindow->pCTree), gtk_ctree_node_nth(GTK_CTREE(fmWindow->pCTree), child_row));
	fmWindow->pFocusedNode = source_gnode;
	if(source_gnode)
	{
		// put the parent directory filename in the window title
        new_filename = fmWindow_getFilenameToRemember(fmWindow, getLinkSource(source_gnode->parent));
        new_filename_bis = g_strconcat("Fm : ", new_filename, NULL);
        gtk_window_set_title(GTK_WINDOW(fmWindow->pWindow), new_filename_bis);
        g_free(new_filename_bis);
		g_free(new_filename);

		// put the focused filename in the statusbar
        new_filename = fmWindow_getFilenameToRemember(fmWindow, source_gnode);
		if(new_filename)
		{
			statusbar_label_style = gtk_widget_get_style(GTK_WIDGET(fmWindow->pLabelGeneralStatus));
            new_filename_width = gdk_string_width(statusbar_label_style->font, new_filename);
            ellipsis_width = gdk_string_width(statusbar_label_style->font, ellipsis);
            label_width = GTK_WIDGET(fmWindow->pLabelGeneralStatus)->allocation.width;
			if(label_width > 1)
			{
//				printf("new_filename_width = %d %d\n", new_filename_width, label_width);
				// go over the path elements to get the maximum visible path elements
				new_filename_bis = new_filename;
				while(new_filename_width+ellipsis_width > label_width)
				{
					new_filename_last_part = new_filename_bis;
					new_filename_bis++;
					while(*new_filename_bis != '/' && *new_filename_bis != 0)
					{
                        new_filename_bis++;
					}
					new_filename_width = gdk_string_width(statusbar_label_style->font, new_filename_bis);
				}
				// if the last path element is still too long, reduce it to get the maximum number of visible chars
				if(new_filename_width == 0)
				{
					new_filename_bis = new_filename_last_part;
					do
					{
						new_filename_bis++;
						new_filename_width = gdk_string_width(statusbar_label_style->font, new_filename_bis);
					}
					while(new_filename_width+ellipsis_width > label_width);
				}
				if(new_filename_bis != new_filename)
				{
					displayed_filename = g_strconcat(ellipsis, new_filename_bis, NULL);
				}
				else
				{
					displayed_filename = new_filename;
				}
				gtk_label_set(GTK_LABEL(fmWindow->pLabelGeneralStatus), displayed_filename);
				if(displayed_filename != new_filename)
				{
					g_free(displayed_filename);
				}
			}
			else
			{
				gtk_label_set(GTK_LABEL(fmWindow->pLabelGeneralStatus), new_filename);
			}
			g_free(new_filename);
		}
	}
}

gboolean fmWindow_focusFileName(struct FmWindow *fmWindow, gchar* filename)
{
	GNode	*source_gnode = fmWindow_getNodeFromFilename(fmWindow, filename, TRUE);
	if(source_gnode)
	{
		fmWindow_focusNode(fmWindow, source_gnode);
		// expand the node if it's a directory and the name ends with the slash
        if(filename[strlen(filename)-1] == '/')
		{
			fmWindow_expandNode(fmWindow, getLinkTarget(source_gnode));
		}
		fmWindow_resizeColumns(fmWindow);
		return TRUE;
	}
	else
	{
		return FALSE;
	}
}

gboolean fmWindow_focusFirstExpandedNode(struct FmWindow *fmWindow, GNode *targetNode)
{
	struct NodeInfo *target_node_info = NULL;
	GNode			*temp_source_gnode = NULL;
	
	if(fmWindow && targetNode)
	{
		target_node_info = (struct NodeInfo*)targetNode->data;
		if(target_node_info->pDirFirstCtreeNode)
		{
			temp_source_gnode = gtk_ctree_node_get_row_data(GTK_CTREE(fmWindow->pCTree), target_node_info->pDirFirstCtreeNode);
			while(temp_source_gnode)
			{
				if(((struct NodeInfo*)getLinkTarget(temp_source_gnode)->data)->pDirFirstCtreeNode)
				{
					break;
				}
				temp_source_gnode = temp_source_gnode->next;
			}
			if(!temp_source_gnode)
			{
				temp_source_gnode = gtk_ctree_node_get_row_data(GTK_CTREE(fmWindow->pCTree), target_node_info->pDirFirstCtreeNode);
			}
			fmWindow_focusNode(fmWindow, temp_source_gnode);
			fmWindow_selectMoveToCTreeFocus(fmWindow);

			return TRUE;
		}
	}

	return FALSE;
}

gboolean fmWindow_focusFirstChildNode(struct FmWindow *fmWindow, GNode *targetNode)
{
	struct NodeInfo *target_node_info = NULL;
	GNode			*temp_source_gnode = NULL;
	
	if(fmWindow && targetNode)
	{
		target_node_info = (struct NodeInfo*)targetNode->data;
		if(target_node_info->pDirFirstCtreeNode)
		{
			temp_source_gnode = gtk_ctree_node_get_row_data(GTK_CTREE(fmWindow->pCTree), target_node_info->pDirFirstCtreeNode);
			fmWindow_focusNode(fmWindow, temp_source_gnode);
			fmWindow_selectMoveToCTreeFocus(fmWindow);

			return TRUE;
		}
	}

	return FALSE;
}

void fmWindow_focusFirstColumn(struct FmWindow *fmWindow)
{
	if(fmWindow)
	{
		fmWindow->focusedColumn = 0;
		fmWindow_moveToNode(fmWindow, fmWindow->pFocusedNode);
	}
}

void fmWindow_focusFirstRow(struct FmWindow *fmWindow)
{
    fmWindow_focusCTreeRow(fmWindow, 0);
	fmWindow_selectMoveToCTreeFocus(fmWindow);
}

void fmWindow_focusLastColumn(struct FmWindow *fmWindow)
{
	if(fmWindow)
	{
		fmWindow->focusedColumn = g_list_length(pgPreferences->pCtreeColumns)-1;
		fmWindow_moveToNode(fmWindow, fmWindow->pFocusedNode);
	}
}

void fmWindow_focusLastRow(struct FmWindow *fmWindow)
{
    fmWindow_focusCTreeRow(fmWindow, GTK_CLIST(fmWindow->pCTree)->rows - 1);
	fmWindow_selectMoveToCTreeFocus(fmWindow);
}

void fmWindow_focusNextColumn(struct FmWindow *fmWindow)
{
	if(fmWindow)
	{
		if(fmWindow->focusedColumn < g_list_length(pgPreferences->pCtreeColumns)-1)
		{
			fmWindow->focusedColumn++;
			fmWindow_moveToNode(fmWindow, fmWindow->pFocusedNode);
		}
	}
}

gboolean fmWindow_focusNextNode(struct FmWindow *fmWindow, GNode *sourceNode)
{
	struct NodeInfo *source_node_info = NULL;
	GtkCTreeNode	*next_ctree_node = NULL;
	
	if(fmWindow && sourceNode)
	{
		source_node_info = (struct NodeInfo*)sourceNode->data;
		if(source_node_info->pCtreeNode)
		{
            next_ctree_node = GTK_CTREE_NODE_NEXT(source_node_info->pCtreeNode);
			if(next_ctree_node)
			{
				fmWindow_focusNode(fmWindow, gtk_ctree_node_get_row_data(GTK_CTREE(fmWindow->pCTree), next_ctree_node));
				fmWindow_selectMoveToCTreeFocus(fmWindow);
	
				return TRUE;
			}
		}
	}
	return FALSE;
}

gboolean fmWindow_focusNextDirectoryNode(struct FmWindow *fmWindow, GNode *sourceNode)
{
	if(fmWindow && sourceNode)
	{
		if(sourceNode->next)
		{
			fmWindow_focusNode(fmWindow, sourceNode->next);
			fmWindow_selectMoveToCTreeFocus(fmWindow);

			return TRUE;
		}
	}
	return FALSE;
}

void fmWindow_focusNode(struct FmWindow *fmWindow, GNode *sourceNode)
{
	guint	child_row = 0;
	
    child_row = gtk_clist_find_row_from_data(GTK_CLIST(fmWindow->pCTree), sourceNode);
	fmWindow_focusCTreeRow(fmWindow, child_row);
}

gboolean fmWindow_focusPageDownNode(struct FmWindow *fmWindow, GNode *sourceNode)
{
	guint	clist_height = GTK_CLIST(fmWindow->pCTree)->clist_window_height;
	guint	row_height = GTK_CLIST(fmWindow->pCTree)->row_height;
	guint	row_per_clist = clist_height / (row_height + CELL_SPACING);
	gint	new_focus = GTK_CLIST(fmWindow->pCTree)->focus_row + row_per_clist/2 + 1;

	if(new_focus > GTK_CLIST(fmWindow->pCTree)->rows - 1)
	{
        new_focus = GTK_CLIST(fmWindow->pCTree)->rows - 1;
	}

	fmWindow_focusCTreeRow(fmWindow, new_focus);
	fmWindow_selectMoveToCTreeFocus(fmWindow);

	return TRUE;
}

gboolean fmWindow_focusPageUpNode(struct FmWindow *fmWindow, GNode *sourceNode)
{
	guint	clist_height = GTK_CLIST(fmWindow->pCTree)->clist_window_height;
	guint	row_height = GTK_CLIST(fmWindow->pCTree)->row_height;
	guint	row_per_clist = clist_height / (row_height + CELL_SPACING);
	gint	new_focus = GTK_CLIST(fmWindow->pCTree)->focus_row - row_per_clist/2 - 1;

	if(new_focus < 0)
	{
        new_focus = 0;
	}

	fmWindow_focusCTreeRow(fmWindow, new_focus);
	fmWindow_selectMoveToCTreeFocus(fmWindow);

	return TRUE;
}

gboolean fmWindow_focusParentNode(struct FmWindow *fmWindow, GNode *sourceNode)
{
	if(fmWindow && sourceNode)
	{
		if(sourceNode->parent && sourceNode->parent != fmWindow->pRootNode)
		{
			fmWindow_focusNode(fmWindow, getLinkSource(sourceNode->parent));
			fmWindow_selectMoveToCTreeFocus(fmWindow);

			return TRUE;
		}
	}
	return FALSE;
}
			
gboolean fmWindow_focusPreviousNode(struct FmWindow *fmWindow, GNode *sourceNode)
{
	struct NodeInfo *source_node_info = NULL;
	GtkCTreeNode	*previous_ctree_node = NULL;
	
	if(fmWindow && sourceNode)
	{
		source_node_info = (struct NodeInfo*)sourceNode->data;
		if(source_node_info->pCtreeNode)
		{
            previous_ctree_node = GTK_CTREE_NODE_PREV(source_node_info->pCtreeNode);
			if(previous_ctree_node)
			{
				fmWindow_focusNode(fmWindow, gtk_ctree_node_get_row_data(GTK_CTREE(fmWindow->pCTree), previous_ctree_node));
				fmWindow_selectMoveToCTreeFocus(fmWindow);

				return TRUE;
			}
		}
	}
	return FALSE;
}

void fmWindow_focusPreviousColumn(struct FmWindow *fmWindow)
{
	if(fmWindow)
	{
		if(fmWindow->focusedColumn > 0)
		{
			fmWindow->focusedColumn--;
			fmWindow_moveToNode(fmWindow, fmWindow->pFocusedNode);
		}
	}
}

gboolean fmWindow_focusPreviousDirectoryNode(struct FmWindow *fmWindow, GNode *sourceNode)
{
	if(fmWindow && sourceNode)
	{
		if(sourceNode->prev)
		{
			fmWindow_focusNode(fmWindow, sourceNode->prev);
			fmWindow_selectMoveToCTreeFocus(fmWindow);

			return TRUE;
		}
	}
	return FALSE;
}

gboolean fmWindow_focusRootNode(struct FmWindow *fmWindow, GNode *sourceNode)
{
	if(fmWindow && sourceNode)
	{
		while(sourceNode->parent && sourceNode->parent != fmWindow->pRootNode)
		{
			sourceNode = getLinkSource(sourceNode->parent);
		}
		
		fmWindow_focusNode(fmWindow, sourceNode);
		fmWindow_selectMoveToCTreeFocus(fmWindow);

		return TRUE;
	}

	return FALSE;
}

void fmWindow_freeComboBoxPopupMenu(struct FmWindow *fmWindow)
{
	gchar		*temp_string = NULL;
	GList		*temp_glist = NULL;

    if(fmWindow->pComboBoxPopupMenu)
	{
		// free the path data
		temp_string = gtk_object_get_data(GTK_OBJECT(fmWindow->pComboBoxPopupMenu), "path");
		gtk_object_remove_data(GTK_OBJECT(fmWindow->pComboBoxPopupMenu), "path");
		if(temp_string)
		{
			g_free(temp_string);
		}

		// remove the menuitems
		while(g_list_length(GTK_MENU_SHELL(fmWindow->pComboBoxPopupMenu)->children))
		{
			temp_glist = g_list_first(GTK_MENU_SHELL(fmWindow->pComboBoxPopupMenu)->children);
			GTK_MENU_SHELL(fmWindow->pComboBoxPopupMenu)->children = g_list_remove_link(GTK_MENU_SHELL(fmWindow->pComboBoxPopupMenu)->children, temp_glist);
			// free the path data
			temp_string = gtk_object_get_data(GTK_OBJECT(temp_glist->data), "path");
			if(temp_string)
			{
				g_free(temp_string);
			}
			gtk_widget_destroy(GTK_WIDGET(GTK_BIN(temp_glist->data)->child));
			gtk_widget_destroy(GTK_WIDGET(temp_glist->data));
			g_list_free_1(temp_glist);
		}
	}
}

void fmWindow_freeDirEntries(struct FmWindow *fmWindow, GNode *targetNode)
{
	GNode	*child_source_gnode = NULL;
	
	while(g_node_n_children(targetNode))
	{
		child_source_gnode = g_node_first_child(targetNode);
		fmWindow_freeNode(fmWindow, child_source_gnode);
	}
}

void fmWindow_freeDir(struct FmWindow *fmWindow, GNode *sourceNode)
{
//	struct NodeInfo	*source_node_info = (struct NodeInfo *)sourceNode->data;
	
//	printf("freeing dir %s\n", source_node_info->pFullName);
	fmWindow_freeDirEntries(fmWindow, getLinkTarget(sourceNode));
}

void fmWindow_freeFile(struct FmWindow *fmWindow, GNode *sourceNode)
{
//	struct NodeInfo	*source_node_info = (struct NodeInfo *)sourceNode->data;
	
//	printf("freeing file %s\n", source_node_info->pFullName);
}

void fmWindow_freeLink(struct FmWindow *fmWindow, GNode *sourceNode)
{
	struct NodeInfo	*source_node_info = (struct NodeInfo *)sourceNode->data;

//	printf("freeing link %s\n", source_node_info->pFullName);
	if(source_node_info->pLinkRepresentation)
	{
		g_free(source_node_info->pLinkRepresentation);
	}
	if(source_node_info->pLinkTarget)
	{
		g_free(source_node_info->pLinkTarget);
	}
}

void fmWindow_freeNode(struct FmWindow *fmWindow, GNode *sourceNode)
{
	struct NodeInfo	*source_node_info = NULL;

	if(sourceNode)
	{
		if(sourceNode->data)
		{
            source_node_info = (struct NodeInfo *)sourceNode->data;

			if(source_node_info->pLinkToNode)
			{
				fmWindow_freeNode(fmWindow, source_node_info->pLinkToNode);
			}

			if(source_node_info->pStat)
			{
				if(isLink(sourceNode))
				{
					fmWindow_freeLink(fmWindow, sourceNode);
				}
				else if(isDirectory(sourceNode))
				{
					fmWindow_freeDir(fmWindow, sourceNode);
				}
				else
				{
					fmWindow_freeFile(fmWindow, sourceNode);
				}
				
				g_free(source_node_info->pStat);
				
				if(source_node_info->pFullName)
				{
					g_free(source_node_info->pFullName);
				}
				if(source_node_info->pShortName)
				{
					g_free(source_node_info->pShortName);
				}
				if(source_node_info->pChangeTimeString)
				{
					g_free(source_node_info->pChangeTimeString);
				}
				if(source_node_info->pGroupString)
				{
					g_free(source_node_info->pGroupString);
				}
				if(source_node_info->pModificationTimeString)
				{
					g_free(source_node_info->pModificationTimeString);
				}
				if(source_node_info->pPermissionsString)
				{
					g_free(source_node_info->pPermissionsString);
				}
				if(source_node_info->pSizeString)
				{
					g_free(source_node_info->pSizeString);
				}
				if(source_node_info->pUserString)
				{
					g_free(source_node_info->pUserString);
				}
			}
			g_free(source_node_info);
		}
		g_node_destroy(sourceNode);
	}
	
	fmWindow->freed++;
}

gchar* fmWindow_getFilenameToRemember(struct FmWindow *fmWindow, GNode *sourceNode)
{
	gchar	*filename_to_remember = NULL;
	gchar	*temp_filename = NULL;
	gchar	*current_filename = NULL;

	if(sourceNode && sourceNode->data)
	{
//		printf("getFilenameToRemember : %s\n", ((struct NodeInfo*)sourceNode->data)->pShortName);
		
		current_filename = ((struct NodeInfo*)sourceNode->data)->pShortName;
		
		if(sourceNode->parent)
		{
			temp_filename = fmWindow_getFilenameToRemember(fmWindow, getLinkSource(sourceNode->parent));
			if(strlen(temp_filename) && temp_filename[strlen(temp_filename)-1] == '/')
			{
				filename_to_remember = g_strconcat(temp_filename, current_filename, NULL);
			}
			else
			{
				filename_to_remember = g_strconcat(temp_filename, "/", current_filename, NULL);
			}
			g_free(temp_filename);
		}
		else
		{
			filename_to_remember = g_strdup(fmWindow->pZoomDirectory);
		}
	}

	return filename_to_remember;
}

GNode* fmWindow_getNodeFromFilename(struct FmWindow *fmWindow, gchar *filename, gboolean expandStructure)
{
	gchar		*compressed_filename = NULL;
	gchar		*aligned_compressed_filename = NULL;
	struct stat	temp_stat;
	gchar		**filename_parts;
	gchar		**filename_part;
	GNode		*parent_target_gnode = NULL;
	GNode		*child_source_gnode = NULL;

	// check if the path is absolute
	if(filename && g_path_is_absolute(filename))
	{
		compressed_filename = compressPath(filename);
//		printf("compressed_filename = '%s'\n", compressed_filename);
		// check if file exists
		if(!lstat(compressed_filename, &temp_stat))
		{
			// file exists, check if it's a falls under the zoomed directory
			if(strncmp(fmWindow->pZoomDirectory, compressed_filename, strlen(fmWindow->pZoomDirectory)) == 0)
			{
				if(strcmp(fmWindow->pZoomDirectory, "/") == 0)
				{
                    aligned_compressed_filename = compressed_filename;
				}
				else
				{
					aligned_compressed_filename = compressed_filename+strlen(fmWindow->pZoomDirectory);
				}
				if(strlen(aligned_compressed_filename))
				{
					filename_parts = g_strsplit(aligned_compressed_filename, "/", 0);
					filename_part = filename_parts;
					parent_target_gnode = fmWindow->pRootNode;

					gtk_clist_freeze(GTK_CLIST(fmWindow->pCTree));
					while(*(++filename_part))
					{
//						printf("filename_part = '%s'\n", *filename_part);
						// check if it's a hidden file and if hidden files are visible
						if((*filename_part[0] == '.' && !fmWindow->hideDotFiles) || (*filename_part[0] != '.'))
						{
//							printf("looking for child '%s'\n", *filename_part);
							child_source_gnode = findGNodeShortName(parent_target_gnode, *filename_part);
							if(!child_source_gnode)
							{
								if(expandStructure)
								{
//									printf("creating child for parent '%s'\n", ((struct NodeInfo*)parent_target_gnode->data)->pShortName);
									gtk_ctree_expand(GTK_CTREE(fmWindow->pCTree), GTK_CTREE_NODE(((struct NodeInfo*)parent_target_gnode->data)->pCtreeNode));
									child_source_gnode = findGNodeShortName(parent_target_gnode, *filename_part);
								}
								else
								{
									child_source_gnode = NULL;
								}
							}
							else
							{
//								printf("found child '%s'\n", ((struct NodeInfo*)child_source_gnode->data)->pShortName);
							}
							parent_target_gnode = getLinkTarget(child_source_gnode);
						}
						else
						{
							break;
						}
					}
					gtk_clist_thaw(GTK_CLIST(fmWindow->pCTree));

					g_strfreev(filename_parts);

					if(child_source_gnode != NULL)
					{
						g_free(compressed_filename);
						return child_source_gnode;
					}
				}
			}
		}
		g_free(compressed_filename);
	}
	
	return NULL;
}

gchar* fmWindow_getSelectionSize(struct FmWindow *fmWindow)
{
	guint	size = 0;
	guint	size_length = 0;
	gchar	*size_string = NULL;

	size = g_list_length(fmWindow->pSelection);
	size_length = log10(size)+1;
	size_string = g_new(gchar, size_length+1);
	sprintf(size_string, "%d", size);
	
	return size_string;
}

void fmWindow_initialize(struct FmWindow *fmWindow)
{
	guint counter = 0;

	if(fmWindow)
	{
        // initializes the tooltips
        fmWindow->pToolTips = gtk_tooltips_new();
		
		// set the window accel group
		fmWindow->pAccelGroup = gtk_accel_group_new();

		// create the hashtable of accellerated widgets
		fmWindow->pAcceleratedWidgets = g_hash_table_new(g_str_hash, g_str_equal);
		
		// create a new window
		fmWindow->pWindow = gtk_window_new(GTK_WINDOW_TOPLEVEL);
		gtk_window_set_wmclass(GTK_WINDOW(fmWindow->pWindow), "Main", "Fm");
		if(fmWindow->initialWidth >= 0 && fmWindow->initialHeight >= 0)
		{
			gtk_window_set_default_size(GTK_WINDOW(fmWindow->pWindow), fmWindow->initialWidth, fmWindow->initialHeight);
			gtk_widget_set_usize(GTK_WIDGET(fmWindow->pWindow), fmWindow->initialWidth, fmWindow->initialHeight);
		}
		gtk_window_set_policy(GTK_WINDOW(fmWindow->pWindow), TRUE, TRUE, FALSE);
		gtk_container_set_border_width(GTK_CONTAINER(fmWindow->pWindow), 5);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pWindow), "delete_event", GTK_SIGNAL_FUNC(fmWindow_deleteWindow), fmWindow);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pWindow), "destroy", GTK_SIGNAL_FUNC(fmWindow_destroyWindow), fmWindow);
		gtk_accel_group_attach(fmWindow->pAccelGroup, GTK_OBJECT(fmWindow->pWindow));

		// create the vbox
		fmWindow->pVBox = gtk_vbox_new(FALSE, 2);
		gtk_container_add(GTK_CONTAINER(fmWindow->pWindow), GTK_WIDGET(fmWindow->pVBox));

		// create the top hbox
		fmWindow->pHBoxTop = gtk_hbox_new(FALSE, 2);
		gtk_box_pack_start(GTK_BOX(fmWindow->pVBox), GTK_WIDGET(fmWindow->pHBoxTop), FALSE, FALSE, 0);

		// create the hide dotfiles button
		fmWindow->pButtonHideDotfiles = gtk_toggle_button_new_with_label("H");
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(fmWindow->pButtonHideDotfiles), TRUE);
		gtk_box_pack_start(GTK_BOX(fmWindow->pHBoxTop), GTK_WIDGET(fmWindow->pButtonHideDotfiles), FALSE, FALSE, 0);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pButtonHideDotfiles), "toggled", GTK_SIGNAL_FUNC(fmWindow_buttonHideDotFilesToggled), fmWindow);
		gtk_tooltips_set_tip(GTK_TOOLTIPS(fmWindow->pToolTips), GTK_WIDGET(fmWindow->pButtonHideDotfiles), "Toggle hidden files", "");

		// create the combo box
		fmWindow->pComboBox = gtk_combo_new();
		gtk_combo_disable_activate(GTK_COMBO(fmWindow->pComboBox));
		gtk_combo_set_use_arrows(GTK_COMBO(fmWindow->pComboBox), TRUE);
		gtk_combo_set_use_arrows_always(GTK_COMBO(fmWindow->pComboBox), TRUE);
		if(fmWindow->pComboBoxStrings != NULL)
		{
			gtk_combo_set_popdown_strings(GTK_COMBO(fmWindow->pComboBox), fmWindow->pComboBoxStrings);
		}
		gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(fmWindow->pComboBox)->entry), "");
		gtk_signal_connect(GTK_OBJECT(GTK_COMBO(fmWindow->pComboBox)->entry), "activate", GTK_SIGNAL_FUNC(fmWindow_comboTextEntryActivate), fmWindow);
		gtk_signal_connect(GTK_OBJECT(GTK_COMBO(fmWindow->pComboBox)->entry), "key_press_event", GTK_SIGNAL_FUNC(fmWindow_comboKeyPress), fmWindow);
		gtk_box_pack_start(GTK_BOX(fmWindow->pHBoxTop), GTK_WIDGET(fmWindow->pComboBox), TRUE, TRUE, 0);

		// connect the drag and drop handlers for the entry widget
		gtk_drag_dest_set(GTK_WIDGET(GTK_COMBO(fmWindow->pComboBox)->entry), GTK_DEST_DEFAULT_MOTION|GTK_DEST_DEFAULT_HIGHLIGHT, gComboEntryDragDropTargets, gComboEntryNumberOfDragDropTargets, GDK_ACTION_COPY | GDK_ACTION_MOVE);
		gtk_signal_connect(GTK_OBJECT(GTK_COMBO(fmWindow->pComboBox)->entry), "drag_drop", GTK_SIGNAL_FUNC(fmWindow_comboTextEntryDragDrop), fmWindow);
		gtk_signal_connect(GTK_OBJECT(GTK_COMBO(fmWindow->pComboBox)->entry), "drag_data_received", GTK_SIGNAL_FUNC(fmWindow_comboTextEntryDragDataReceived), fmWindow);
//		gtk_signal_connect(GTK_OBJECT(GTK_COMBO(fmWindow->pComboBox)->entry), "drag_motion", GTK_SIGNAL_FUNC(fmWindow_comboTextEntryDragMotion), fmWindow);

		// create the combo box popup menu
		fmWindow->pComboBoxPopupMenu = gtk_menu_new();

		// create the ctree hbox
		fmWindow->pHBoxScrolledWindow = gtk_hbox_new(FALSE, 2);
		gtk_box_pack_start(GTK_BOX(fmWindow->pVBox), GTK_WIDGET(fmWindow->pHBoxScrolledWindow), TRUE, TRUE, 0);
		
		// create the ctree left vbox
		fmWindow->pVBoxScrolledWindowLeft = gtk_vbox_new(FALSE, 2);
		gtk_box_pack_start(GTK_BOX(fmWindow->pHBoxScrolledWindow), GTK_WIDGET(fmWindow->pVBoxScrolledWindowLeft), TRUE, TRUE, 0);
		
		// build the ctree adjustments
        fmWindow->pAdjustmentHScrollbar = gtk_adjustment_new(0, 0, 0, 0, 0, 0);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pAdjustmentHScrollbar), "changed", GTK_SIGNAL_FUNC(fmWindow_ctreeHAdjustmentChanged), fmWindow);
        fmWindow->pAdjustmentVScrollbar = gtk_adjustment_new(0, 0, 0, 0, 0, 0);

		// build the ctree
		fmWindow->pCTree = fmWindow_buildCtree(fmWindow);
		gtk_clist_set_hadjustment(GTK_CLIST(fmWindow->pCTree), GTK_ADJUSTMENT(fmWindow->pAdjustmentHScrollbar));
		gtk_clist_set_vadjustment(GTK_CLIST(fmWindow->pCTree), GTK_ADJUSTMENT(fmWindow->pAdjustmentVScrollbar));
		gtk_box_pack_start(GTK_BOX(fmWindow->pVBoxScrolledWindowLeft), GTK_WIDGET(fmWindow->pCTree), TRUE, TRUE, 0);
		fmWindow_updateCTreeTitles(fmWindow);
		fmWindow_ctreeColumnsApplyAlignment(fmWindow);

		// create the ctree horizontal scrollbar
		fmWindow->pHScrollbarScrolledWindow = gtk_hscrollbar_new(GTK_ADJUSTMENT(fmWindow->pAdjustmentHScrollbar));
		gtk_widget_set_usize(GTK_WIDGET(fmWindow->pHScrollbarScrolledWindow), 0, 18);
		gtk_box_pack_end(GTK_BOX(fmWindow->pVBoxScrolledWindowLeft), GTK_WIDGET(fmWindow->pHScrollbarScrolledWindow), FALSE, FALSE, 0);
		
		// create the ctree right vbox
		fmWindow->pVBoxScrolledWindowRight = gtk_vbox_new(FALSE, 2);
		gtk_box_pack_start(GTK_BOX(fmWindow->pHBoxScrolledWindow), GTK_WIDGET(fmWindow->pVBoxScrolledWindowRight), FALSE, TRUE, 0);
		
		// create the ctree vertical scrollbar
		fmWindow->pVScrollbarScrolledWindow = gtk_vscrollbar_new(GTK_ADJUSTMENT(fmWindow->pAdjustmentVScrollbar));
		gtk_widget_set_usize(GTK_WIDGET(fmWindow->pVScrollbarScrolledWindow), 18, 0);
		gtk_box_pack_start(GTK_BOX(fmWindow->pVBoxScrolledWindowRight), GTK_WIDGET(fmWindow->pVScrollbarScrolledWindow), TRUE, TRUE, 0);
		
		// create the pan window widget
		fmWindow->pEventBoxPanWindow = gtk_event_box_new();
		gtk_box_pack_start(GTK_BOX(fmWindow->pVBoxScrolledWindowRight), GTK_WIDGET(fmWindow->pEventBoxPanWindow), FALSE, FALSE, 0);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pEventBoxPanWindow), "button_press_event", GTK_SIGNAL_FUNC(fmWindow_panButtonPress), fmWindow);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pEventBoxPanWindow), "button_release_event", GTK_SIGNAL_FUNC(fmWindow_panButtonRelease), fmWindow);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pEventBoxPanWindow), "motion_notify_event", GTK_SIGNAL_FUNC(fmWindow_panMotionNotify), fmWindow);
		// don't enable this until there will be an option the show/hide the tooltips and when they'll be applied everywhere
//		gtk_tooltips_set_tip(GTK_TOOLTIPS(fmWindow->pToolTips), GTK_WIDGET(fmWindow->pEventBoxPanWindow), "Pan window", "");

		fmWindow->pPixmapPanWindow = gtk_pixmap_new(pgPixmapPanWindowCenter, pgMaskPanWindowCenter);
		gtk_container_add(GTK_CONTAINER(fmWindow->pEventBoxPanWindow), GTK_WIDGET(fmWindow->pPixmapPanWindow));
		
		// create the buttons hbox
		fmWindow->pHBoxBottomButtons = gtk_hbox_new(FALSE, 2);
		gtk_box_pack_start(GTK_BOX(fmWindow->pVBox), GTK_WIDGET(fmWindow->pHBoxBottomButtons), FALSE, FALSE, 0);

		// create the bottom buttons
		fmWindow_createBookmarkButton(fmWindow, 1);
		fmWindow_createBookmarkButton(fmWindow, 2);
		fmWindow_createBookmarkButton(fmWindow, 3);
		fmWindow_createBookmarkButton(fmWindow, 4);
		fmWindow_createBookmarkButton(fmWindow, 5);
		fmWindow_createBookmarkButton(fmWindow, 6);
		fmWindow_createBookmarkButton(fmWindow, 7);
		fmWindow_createBookmarkButton(fmWindow, 8);
		fmWindow_createBookmarkButton(fmWindow, 9);
		fmWindow_createBookmarkButton(fmWindow, 0);

		// create the bottom hbox
		fmWindow->pHBoxBottomStatus = gtk_hbox_new(FALSE, 2);
		gtk_box_pack_start(GTK_BOX(fmWindow->pVBox), GTK_WIDGET(fmWindow->pHBoxBottomStatus), FALSE, FALSE, 0);

		// create the general status label frame
		fmWindow->pFrameGeneralStatus = gtk_frame_new(NULL);
		gtk_frame_set_shadow_type(GTK_FRAME(fmWindow->pFrameGeneralStatus), GTK_SHADOW_ETCHED_IN);
		gtk_box_pack_start(GTK_BOX(fmWindow->pHBoxBottomStatus), GTK_WIDGET(fmWindow->pFrameGeneralStatus), TRUE, TRUE, 0);

		// create the general status label
		fmWindow->pLabelGeneralStatus = gtk_label_new("");
		gtk_container_add(GTK_CONTAINER(fmWindow->pFrameGeneralStatus), GTK_WIDGET(fmWindow->pLabelGeneralStatus));

		// create the selection size frame
		fmWindow->pFrameSelectionSize = gtk_frame_new(NULL);
		gtk_frame_set_shadow_type(GTK_FRAME(fmWindow->pFrameSelectionSize), GTK_SHADOW_ETCHED_IN);
		gtk_box_pack_start(GTK_BOX(fmWindow->pHBoxBottomStatus), GTK_WIDGET(fmWindow->pFrameSelectionSize), FALSE, FALSE, 0);

		// create the selection size alignment
		fmWindow->pAlignmentSelectionSize = gtk_alignment_new(0.5, 0.5, 1, 1);
		// set the size to contain about ten million selected nodes
		gtk_widget_set_usize(GTK_WIDGET(fmWindow->pAlignmentSelectionSize), gdk_string_width(gtk_widget_get_style(GTK_WIDGET(fmWindow->pAlignmentSelectionSize))->font, "0000000"), 0);
		gtk_container_add(GTK_CONTAINER(fmWindow->pFrameSelectionSize), GTK_WIDGET(fmWindow->pAlignmentSelectionSize));

		// create the selection size label
		fmWindow->pLabelSelectionSize = gtk_label_new("0");
		gtk_container_add(GTK_CONTAINER(fmWindow->pAlignmentSelectionSize), GTK_WIDGET(fmWindow->pLabelSelectionSize));

		// create the popup menu
		fmWindow->pPopupMenu = gtk_menu_new();
		gtk_menu_set_accel_group(GTK_MENU(fmWindow->pPopupMenu), fmWindow->pAccelGroup);

		fmWindow->pPopupMenuItemSelection = gtk_menu_item_new_with_label("Selection");
		fmWindow->pPopupMenuSelection = gtk_menu_new();
		gtk_menu_set_accel_group(GTK_MENU(fmWindow->pPopupMenuSelection), fmWindow->pAccelGroup);
		
		fmWindow->pPopupMenuSelectionMenuItemClear = gtk_menu_item_new_with_label("Clear");
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuSelectionMenuItemClear), "activate", GTK_SIGNAL_FUNC(fmWindow_ctreeMenuItemSelectionClear), fmWindow);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuSelectionMenuItemClear), "destroy", GTK_SIGNAL_FUNC(fmWindow_destroyMenuItem), fmWindow);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuSelectionMenuItemClear), "add_accelerator", GTK_SIGNAL_FUNC(fmWindow_acceleratorAdd), fmWindow);
		gtk_signal_connect_after(GTK_OBJECT(fmWindow->pPopupMenuSelectionMenuItemClear), "remove_accelerator", GTK_SIGNAL_FUNC(fmWindow_acceleratorRemove), fmWindow);
		gtk_object_set_data(GTK_OBJECT(fmWindow->pPopupMenuSelectionMenuItemClear), "accelKeyId", (gpointer)g_strdup("SelectionClear"));
		fmWindow_accelKeyAddToGtkWidget(fmWindow, GTK_WIDGET(fmWindow->pPopupMenuSelectionMenuItemClear));

		fmWindow->pPopupMenuSelectionMenuItemSelectDirectory = gtk_menu_item_new_with_label("Select Directory");
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuSelectionMenuItemSelectDirectory), "activate", GTK_SIGNAL_FUNC(fmWindow_ctreeMenuItemSelectionSelectDirectory), fmWindow);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuSelectionMenuItemSelectDirectory), "destroy", GTK_SIGNAL_FUNC(fmWindow_destroyMenuItem), fmWindow);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuSelectionMenuItemSelectDirectory), "add_accelerator", GTK_SIGNAL_FUNC(fmWindow_acceleratorAdd), fmWindow);
		gtk_signal_connect_after(GTK_OBJECT(fmWindow->pPopupMenuSelectionMenuItemSelectDirectory), "remove_accelerator", GTK_SIGNAL_FUNC(fmWindow_acceleratorRemove), fmWindow);
		gtk_object_set_data(GTK_OBJECT(fmWindow->pPopupMenuSelectionMenuItemSelectDirectory), "accelKeyId", (gpointer)g_strdup("SelectionSelectDirectory"));
		fmWindow_accelKeyAddToGtkWidget(fmWindow, GTK_WIDGET(fmWindow->pPopupMenuSelectionMenuItemSelectDirectory));

		fmWindow->pPopupMenuSelectionMenuItemDeselectDirectory = gtk_menu_item_new_with_label("Deselect Directory");
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuSelectionMenuItemDeselectDirectory), "activate", GTK_SIGNAL_FUNC(fmWindow_ctreeMenuItemSelectionDeselectDirectory), fmWindow);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuSelectionMenuItemDeselectDirectory), "destroy", GTK_SIGNAL_FUNC(fmWindow_destroyMenuItem), fmWindow);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuSelectionMenuItemDeselectDirectory), "add_accelerator", GTK_SIGNAL_FUNC(fmWindow_acceleratorAdd), fmWindow);
		gtk_signal_connect_after(GTK_OBJECT(fmWindow->pPopupMenuSelectionMenuItemDeselectDirectory), "remove_accelerator", GTK_SIGNAL_FUNC(fmWindow_acceleratorRemove), fmWindow);
		gtk_object_set_data(GTK_OBJECT(fmWindow->pPopupMenuSelectionMenuItemDeselectDirectory), "accelKeyId", (gpointer)g_strdup("SelectionDeselectDirectory"));
		fmWindow_accelKeyAddToGtkWidget(fmWindow, GTK_WIDGET(fmWindow->pPopupMenuSelectionMenuItemDeselectDirectory));

		fmWindow->pPopupMenuSelectionMenuItemCopyToXClipboard = gtk_menu_item_new_with_label("Copy  To X Clipboard");
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuSelectionMenuItemCopyToXClipboard), "activate", GTK_SIGNAL_FUNC(fmWindow_ctreeMenuItemSelectionCopyToXClipboard), fmWindow);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuSelectionMenuItemCopyToXClipboard), "destroy", GTK_SIGNAL_FUNC(fmWindow_destroyMenuItem), fmWindow);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuSelectionMenuItemCopyToXClipboard), "add_accelerator", GTK_SIGNAL_FUNC(fmWindow_acceleratorAdd), fmWindow);
		gtk_signal_connect_after(GTK_OBJECT(fmWindow->pPopupMenuSelectionMenuItemCopyToXClipboard), "remove_accelerator", GTK_SIGNAL_FUNC(fmWindow_acceleratorRemove), fmWindow);
		gtk_object_set_data(GTK_OBJECT(fmWindow->pPopupMenuSelectionMenuItemCopyToXClipboard), "accelKeyId", (gpointer)g_strdup("SelectionCopyToXClipboard"));
		fmWindow_accelKeyAddToGtkWidget(fmWindow, GTK_WIDGET(fmWindow->pPopupMenuSelectionMenuItemCopyToXClipboard));

		fmWindow->pPopupMenuSelectionMenuItemCopyFocusToComboEntry = gtk_menu_item_new_with_label("Copy Focus To Combo Entry");
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuSelectionMenuItemCopyFocusToComboEntry), "activate", GTK_SIGNAL_FUNC(fmWindow_ctreeMenuItemSelectionCopyFocusToComboEntry), fmWindow);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuSelectionMenuItemCopyFocusToComboEntry), "destroy", GTK_SIGNAL_FUNC(fmWindow_destroyMenuItem), fmWindow);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuSelectionMenuItemCopyFocusToComboEntry), "add_accelerator", GTK_SIGNAL_FUNC(fmWindow_acceleratorAdd), fmWindow);
		gtk_signal_connect_after(GTK_OBJECT(fmWindow->pPopupMenuSelectionMenuItemCopyFocusToComboEntry), "remove_accelerator", GTK_SIGNAL_FUNC(fmWindow_acceleratorRemove), fmWindow);
		gtk_object_set_data(GTK_OBJECT(fmWindow->pPopupMenuSelectionMenuItemCopyFocusToComboEntry), "accelKeyId", (gpointer)g_strdup("SelectionCopyFocusToComboEntry"));
		fmWindow_accelKeyAddToGtkWidget(fmWindow, GTK_WIDGET(fmWindow->pPopupMenuSelectionMenuItemCopyFocusToComboEntry));

		fmWindow->pPopupMenuItemDirectory = gtk_menu_item_new_with_label("Directory");
		fmWindow->pPopupMenuDirectory = gtk_menu_new();
		gtk_menu_set_accel_group(GTK_MENU(fmWindow->pPopupMenuDirectory), fmWindow->pAccelGroup);
		
		fmWindow->pPopupMenuDirectoryMenuItemRefresh = gtk_menu_item_new_with_label("Refresh");
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuDirectoryMenuItemRefresh), "activate", GTK_SIGNAL_FUNC(fmWindow_ctreeMenuItemDirectoryRefresh), fmWindow);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuDirectoryMenuItemRefresh), "destroy", GTK_SIGNAL_FUNC(fmWindow_destroyMenuItem), fmWindow);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuDirectoryMenuItemRefresh), "add_accelerator", GTK_SIGNAL_FUNC(fmWindow_acceleratorAdd), fmWindow);
		gtk_signal_connect_after(GTK_OBJECT(fmWindow->pPopupMenuDirectoryMenuItemRefresh), "remove_accelerator", GTK_SIGNAL_FUNC(fmWindow_acceleratorRemove), fmWindow);
		gtk_object_set_data(GTK_OBJECT(fmWindow->pPopupMenuDirectoryMenuItemRefresh), "accelKeyId", (gpointer)g_strdup("DirectoryRefresh"));
		fmWindow_accelKeyAddToGtkWidget(fmWindow, GTK_WIDGET(fmWindow->pPopupMenuDirectoryMenuItemRefresh));

		fmWindow->pPopupMenuDirectoryMenuItemZoom = gtk_menu_item_new_with_label("Zoom");
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuDirectoryMenuItemZoom), "activate", GTK_SIGNAL_FUNC(fmWindow_ctreeMenuItemDirectoryZoom), fmWindow);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuDirectoryMenuItemZoom), "destroy", GTK_SIGNAL_FUNC(fmWindow_destroyMenuItem), fmWindow);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuDirectoryMenuItemZoom), "add_accelerator", GTK_SIGNAL_FUNC(fmWindow_acceleratorAdd), fmWindow);
		gtk_signal_connect_after(GTK_OBJECT(fmWindow->pPopupMenuDirectoryMenuItemZoom), "remove_accelerator", GTK_SIGNAL_FUNC(fmWindow_acceleratorRemove), fmWindow);
		gtk_object_set_data(GTK_OBJECT(fmWindow->pPopupMenuDirectoryMenuItemZoom), "accelKeyId", (gpointer)g_strdup("DirectoryZoom"));
		fmWindow_accelKeyAddToGtkWidget(fmWindow, GTK_WIDGET(fmWindow->pPopupMenuDirectoryMenuItemZoom));

		fmWindow->pPopupMenuDirectoryMenuItemOpenInNewWindow = gtk_menu_item_new_with_label("Open In New Window");
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuDirectoryMenuItemOpenInNewWindow), "activate", GTK_SIGNAL_FUNC(fmWindow_ctreeMenuItemDirectoryOpenInNewWindow), fmWindow);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuDirectoryMenuItemOpenInNewWindow), "destroy", GTK_SIGNAL_FUNC(fmWindow_destroyMenuItem), fmWindow);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuDirectoryMenuItemOpenInNewWindow), "add_accelerator", GTK_SIGNAL_FUNC(fmWindow_acceleratorAdd), fmWindow);
		gtk_signal_connect_after(GTK_OBJECT(fmWindow->pPopupMenuDirectoryMenuItemOpenInNewWindow), "remove_accelerator", GTK_SIGNAL_FUNC(fmWindow_acceleratorRemove), fmWindow);
		gtk_object_set_data(GTK_OBJECT(fmWindow->pPopupMenuDirectoryMenuItemOpenInNewWindow), "accelKeyId", (gpointer)g_strdup("DirectoryOpenInNewWindow"));
		fmWindow_accelKeyAddToGtkWidget(fmWindow, GTK_WIDGET(fmWindow->pPopupMenuDirectoryMenuItemOpenInNewWindow));

		fmWindow->pPopupMenuItemBrowser = gtk_menu_item_new_with_label("Browser");
		fmWindow->pPopupMenuBrowser = gtk_menu_new();
		gtk_menu_set_accel_group(GTK_MENU(fmWindow->pPopupMenuBrowser), fmWindow->pAccelGroup);
		
		fmWindow->pPopupMenuBrowserMenuItemRefresh = gtk_menu_item_new_with_label("Refresh");
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuBrowserMenuItemRefresh), "activate", GTK_SIGNAL_FUNC(fmWindow_ctreeMenuItemBrowserRefresh), fmWindow);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuBrowserMenuItemRefresh), "destroy", GTK_SIGNAL_FUNC(fmWindow_destroyMenuItem), fmWindow);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuBrowserMenuItemRefresh), "add_accelerator", GTK_SIGNAL_FUNC(fmWindow_acceleratorAdd), fmWindow);
		gtk_signal_connect_after(GTK_OBJECT(fmWindow->pPopupMenuBrowserMenuItemRefresh), "remove_accelerator", GTK_SIGNAL_FUNC(fmWindow_acceleratorRemove), fmWindow);
		gtk_object_set_data(GTK_OBJECT(fmWindow->pPopupMenuBrowserMenuItemRefresh), "accelKeyId", (gpointer)g_strdup("BrowserRefresh"));
		fmWindow_accelKeyAddToGtkWidget(fmWindow, GTK_WIDGET(fmWindow->pPopupMenuBrowserMenuItemRefresh));

		fmWindow->pPopupMenuBrowserMenuItemNew = gtk_menu_item_new_with_label("New");
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuBrowserMenuItemNew), "activate", GTK_SIGNAL_FUNC(fmWindow_ctreeMenuItemBrowserNew), fmWindow);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuBrowserMenuItemNew), "destroy", GTK_SIGNAL_FUNC(fmWindow_destroyMenuItem), fmWindow);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuBrowserMenuItemNew), "add_accelerator", GTK_SIGNAL_FUNC(fmWindow_acceleratorAdd), fmWindow);
		gtk_signal_connect_after(GTK_OBJECT(fmWindow->pPopupMenuBrowserMenuItemNew), "remove_accelerator", GTK_SIGNAL_FUNC(fmWindow_acceleratorRemove), fmWindow);
		gtk_object_set_data(GTK_OBJECT(fmWindow->pPopupMenuBrowserMenuItemNew), "accelKeyId", (gpointer)g_strdup("BrowserNew"));
		fmWindow_accelKeyAddToGtkWidget(fmWindow, GTK_WIDGET(fmWindow->pPopupMenuBrowserMenuItemNew));

		fmWindow->pPopupMenuBrowserMenuItemClose = gtk_menu_item_new_with_label("Close");
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuBrowserMenuItemClose), "activate", GTK_SIGNAL_FUNC(fmWindow_ctreeMenuItemBrowserClose), fmWindow);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuBrowserMenuItemClose), "destroy", GTK_SIGNAL_FUNC(fmWindow_destroyMenuItem), fmWindow);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuBrowserMenuItemClose), "add_accelerator", GTK_SIGNAL_FUNC(fmWindow_acceleratorAdd), fmWindow);
		gtk_signal_connect_after(GTK_OBJECT(fmWindow->pPopupMenuBrowserMenuItemClose), "remove_accelerator", GTK_SIGNAL_FUNC(fmWindow_acceleratorRemove), fmWindow);
		gtk_object_set_data(GTK_OBJECT(fmWindow->pPopupMenuBrowserMenuItemClose), "accelKeyId", (gpointer)g_strdup("BrowserClose"));
		fmWindow_accelKeyAddToGtkWidget(fmWindow, GTK_WIDGET(fmWindow->pPopupMenuBrowserMenuItemClose));

		fmWindow->pPopupMenuBrowserMenuItemCollapseEverything = gtk_menu_item_new_with_label("Collapse Everything");
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuBrowserMenuItemCollapseEverything), "activate", GTK_SIGNAL_FUNC(fmWindow_ctreeMenuItemBrowserCollapseEverything), fmWindow);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuBrowserMenuItemCollapseEverything), "destroy", GTK_SIGNAL_FUNC(fmWindow_destroyMenuItem), fmWindow);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuBrowserMenuItemCollapseEverything), "add_accelerator", GTK_SIGNAL_FUNC(fmWindow_acceleratorAdd), fmWindow);
		gtk_signal_connect_after(GTK_OBJECT(fmWindow->pPopupMenuBrowserMenuItemCollapseEverything), "remove_accelerator", GTK_SIGNAL_FUNC(fmWindow_acceleratorRemove), fmWindow);
		gtk_object_set_data(GTK_OBJECT(fmWindow->pPopupMenuBrowserMenuItemCollapseEverything), "accelKeyId", (gpointer)g_strdup("BrowserCollapseEverything"));
		fmWindow_accelKeyAddToGtkWidget(fmWindow, GTK_WIDGET(fmWindow->pPopupMenuBrowserMenuItemCollapseEverything));

		fmWindow->pPopupMenuBrowserMenuItemUnZoom = gtk_menu_item_new_with_label("Un-Zoom");
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuBrowserMenuItemUnZoom), "activate", GTK_SIGNAL_FUNC(fmWindow_ctreeMenuItemBrowserUnZoom), fmWindow);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuBrowserMenuItemUnZoom), "destroy", GTK_SIGNAL_FUNC(fmWindow_destroyMenuItem), fmWindow);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuBrowserMenuItemUnZoom), "add_accelerator", GTK_SIGNAL_FUNC(fmWindow_acceleratorAdd), fmWindow);
		gtk_signal_connect_after(GTK_OBJECT(fmWindow->pPopupMenuBrowserMenuItemUnZoom), "remove_accelerator", GTK_SIGNAL_FUNC(fmWindow_acceleratorRemove), fmWindow);
		gtk_object_set_data(GTK_OBJECT(fmWindow->pPopupMenuBrowserMenuItemUnZoom), "accelKeyId", (gpointer)g_strdup("BrowserUnZoom"));
		fmWindow_accelKeyAddToGtkWidget(fmWindow, GTK_WIDGET(fmWindow->pPopupMenuBrowserMenuItemUnZoom));

		fmWindow->pPopupMenuBrowserMenuItemCopyState = gtk_menu_item_new_with_label("Copy State");
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuBrowserMenuItemCopyState), "activate", GTK_SIGNAL_FUNC(fmWindow_ctreeMenuItemBrowserCopyState), fmWindow);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuBrowserMenuItemCopyState), "destroy", GTK_SIGNAL_FUNC(fmWindow_destroyMenuItem), fmWindow);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuBrowserMenuItemCopyState), "add_accelerator", GTK_SIGNAL_FUNC(fmWindow_acceleratorAdd), fmWindow);
		gtk_signal_connect_after(GTK_OBJECT(fmWindow->pPopupMenuBrowserMenuItemCopyState), "remove_accelerator", GTK_SIGNAL_FUNC(fmWindow_acceleratorRemove), fmWindow);
		gtk_object_set_data(GTK_OBJECT(fmWindow->pPopupMenuBrowserMenuItemCopyState), "accelKeyId", (gpointer)g_strdup("BrowserCopyState"));
		fmWindow_accelKeyAddToGtkWidget(fmWindow, GTK_WIDGET(fmWindow->pPopupMenuBrowserMenuItemCopyState));

		fmWindow->pPopupMenuBrowserMenuItemPasteState = gtk_menu_item_new_with_label("Paste State");
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuBrowserMenuItemPasteState), "activate", GTK_SIGNAL_FUNC(fmWindow_ctreeMenuItemBrowserPasteState), fmWindow);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuBrowserMenuItemPasteState), "destroy", GTK_SIGNAL_FUNC(fmWindow_destroyMenuItem), fmWindow);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuBrowserMenuItemPasteState), "add_accelerator", GTK_SIGNAL_FUNC(fmWindow_acceleratorAdd), fmWindow);
		gtk_signal_connect_after(GTK_OBJECT(fmWindow->pPopupMenuBrowserMenuItemPasteState), "remove_accelerator", GTK_SIGNAL_FUNC(fmWindow_acceleratorRemove), fmWindow);
		gtk_object_set_data(GTK_OBJECT(fmWindow->pPopupMenuBrowserMenuItemPasteState), "accelKeyId", (gpointer)g_strdup("BrowserPasteState"));
		fmWindow_accelKeyAddToGtkWidget(fmWindow, GTK_WIDGET(fmWindow->pPopupMenuBrowserMenuItemPasteState));

		fmWindow->pPopupMenuItemColumns = gtk_menu_item_new_with_label("Columns");
		fmWindow->pPopupMenuColumns = gtk_menu_new();
		gtk_menu_set_accel_group(GTK_MENU(fmWindow->pPopupMenuColumns), fmWindow->pAccelGroup);

		fmWindow->pPopupMenuColumnsMenuItemSize = gtk_check_menu_item_new_with_label("Size");
		gtk_check_menu_item_set_show_toggle(GTK_CHECK_MENU_ITEM(fmWindow->pPopupMenuColumnsMenuItemSize), TRUE);
		if(g_list_find(fmWindow->pCTreeColumns, (gpointer)COLUMN_SIZE))
		{
			gtk_check_menu_item_set_state(GTK_CHECK_MENU_ITEM(fmWindow->pPopupMenuColumnsMenuItemSize), TRUE);
		}
		gtk_object_set_data(GTK_OBJECT(fmWindow->pPopupMenuColumnsMenuItemSize), "columnId", (gpointer)COLUMN_SIZE);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuColumnsMenuItemSize), "toggled", GTK_SIGNAL_FUNC(fmWindow_ctreeMenuItemColumnsToggleColumn), fmWindow);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuColumnsMenuItemSize), "destroy", GTK_SIGNAL_FUNC(fmWindow_destroyMenuItem), fmWindow);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuColumnsMenuItemSize), "add_accelerator", GTK_SIGNAL_FUNC(fmWindow_acceleratorAdd), fmWindow);
		gtk_signal_connect_after(GTK_OBJECT(fmWindow->pPopupMenuColumnsMenuItemSize), "remove_accelerator", GTK_SIGNAL_FUNC(fmWindow_acceleratorRemove), fmWindow);
		gtk_object_set_data(GTK_OBJECT(fmWindow->pPopupMenuColumnsMenuItemSize), "accelKeyId", (gpointer)g_strdup("ColumnSize"));
		fmWindow_accelKeyAddToGtkWidget(fmWindow, GTK_WIDGET(fmWindow->pPopupMenuColumnsMenuItemSize));
		
		fmWindow->pPopupMenuColumnsMenuItemUser = gtk_check_menu_item_new_with_label("User");
		gtk_check_menu_item_set_show_toggle(GTK_CHECK_MENU_ITEM(fmWindow->pPopupMenuColumnsMenuItemUser), TRUE);
		if(g_list_find(fmWindow->pCTreeColumns, (gpointer)COLUMN_USER))
		{
			gtk_check_menu_item_set_state(GTK_CHECK_MENU_ITEM(fmWindow->pPopupMenuColumnsMenuItemUser), TRUE);
		}
		gtk_object_set_data(GTK_OBJECT(fmWindow->pPopupMenuColumnsMenuItemUser), "columnId", (gpointer)COLUMN_USER);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuColumnsMenuItemUser), "toggled", GTK_SIGNAL_FUNC(fmWindow_ctreeMenuItemColumnsToggleColumn), fmWindow);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuColumnsMenuItemUser), "destroy", GTK_SIGNAL_FUNC(fmWindow_destroyMenuItem), fmWindow);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuColumnsMenuItemUser), "add_accelerator", GTK_SIGNAL_FUNC(fmWindow_acceleratorAdd), fmWindow);
		gtk_signal_connect_after(GTK_OBJECT(fmWindow->pPopupMenuColumnsMenuItemUser), "remove_accelerator", GTK_SIGNAL_FUNC(fmWindow_acceleratorRemove), fmWindow);
		gtk_object_set_data(GTK_OBJECT(fmWindow->pPopupMenuColumnsMenuItemUser), "accelKeyId", (gpointer)g_strdup("ColumnUser"));
		fmWindow_accelKeyAddToGtkWidget(fmWindow, GTK_WIDGET(fmWindow->pPopupMenuColumnsMenuItemUser));
		
		fmWindow->pPopupMenuColumnsMenuItemGroup = gtk_check_menu_item_new_with_label("Group");
		gtk_check_menu_item_set_show_toggle(GTK_CHECK_MENU_ITEM(fmWindow->pPopupMenuColumnsMenuItemGroup), TRUE);
		if(g_list_find(fmWindow->pCTreeColumns, (gpointer)COLUMN_GROUP))
		{
			gtk_check_menu_item_set_state(GTK_CHECK_MENU_ITEM(fmWindow->pPopupMenuColumnsMenuItemGroup), TRUE);
		}
		gtk_object_set_data(GTK_OBJECT(fmWindow->pPopupMenuColumnsMenuItemGroup), "columnId", (gpointer)COLUMN_GROUP);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuColumnsMenuItemGroup), "toggled", GTK_SIGNAL_FUNC(fmWindow_ctreeMenuItemColumnsToggleColumn), fmWindow);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuColumnsMenuItemGroup), "destroy", GTK_SIGNAL_FUNC(fmWindow_destroyMenuItem), fmWindow);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuColumnsMenuItemGroup), "add_accelerator", GTK_SIGNAL_FUNC(fmWindow_acceleratorAdd), fmWindow);
		gtk_signal_connect_after(GTK_OBJECT(fmWindow->pPopupMenuColumnsMenuItemGroup), "remove_accelerator", GTK_SIGNAL_FUNC(fmWindow_acceleratorRemove), fmWindow);
		gtk_object_set_data(GTK_OBJECT(fmWindow->pPopupMenuColumnsMenuItemGroup), "accelKeyId", (gpointer)g_strdup("ColumnGroup"));
		fmWindow_accelKeyAddToGtkWidget(fmWindow, GTK_WIDGET(fmWindow->pPopupMenuColumnsMenuItemGroup));
		
		fmWindow->pPopupMenuColumnsMenuItemPermissions = gtk_check_menu_item_new_with_label("Permissions");
		gtk_check_menu_item_set_show_toggle(GTK_CHECK_MENU_ITEM(fmWindow->pPopupMenuColumnsMenuItemPermissions), TRUE);
		if(g_list_find(fmWindow->pCTreeColumns, (gpointer)COLUMN_PERMISSIONS))
		{
			gtk_check_menu_item_set_state(GTK_CHECK_MENU_ITEM(fmWindow->pPopupMenuColumnsMenuItemPermissions), TRUE);
		}
		gtk_object_set_data(GTK_OBJECT(fmWindow->pPopupMenuColumnsMenuItemPermissions), "columnId", (gpointer)COLUMN_PERMISSIONS);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuColumnsMenuItemPermissions), "toggled", GTK_SIGNAL_FUNC(fmWindow_ctreeMenuItemColumnsToggleColumn), fmWindow);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuColumnsMenuItemPermissions), "destroy", GTK_SIGNAL_FUNC(fmWindow_destroyMenuItem), fmWindow);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuColumnsMenuItemPermissions), "add_accelerator", GTK_SIGNAL_FUNC(fmWindow_acceleratorAdd), fmWindow);
		gtk_signal_connect_after(GTK_OBJECT(fmWindow->pPopupMenuColumnsMenuItemPermissions), "remove_accelerator", GTK_SIGNAL_FUNC(fmWindow_acceleratorRemove), fmWindow);
		gtk_object_set_data(GTK_OBJECT(fmWindow->pPopupMenuColumnsMenuItemPermissions), "accelKeyId", (gpointer)g_strdup("ColumnPermissions"));
		fmWindow_accelKeyAddToGtkWidget(fmWindow, GTK_WIDGET(fmWindow->pPopupMenuColumnsMenuItemPermissions));
		
		fmWindow->pPopupMenuColumnsMenuItemModificationTime = gtk_check_menu_item_new_with_label("Modification Time");
		gtk_check_menu_item_set_show_toggle(GTK_CHECK_MENU_ITEM(fmWindow->pPopupMenuColumnsMenuItemModificationTime), TRUE);
		if(g_list_find(fmWindow->pCTreeColumns, (gpointer)COLUMN_MODIFICATION_TIME))
		{
			gtk_check_menu_item_set_state(GTK_CHECK_MENU_ITEM(fmWindow->pPopupMenuColumnsMenuItemModificationTime), TRUE);
		}
		gtk_object_set_data(GTK_OBJECT(fmWindow->pPopupMenuColumnsMenuItemModificationTime), "columnId", (gpointer)COLUMN_MODIFICATION_TIME);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuColumnsMenuItemModificationTime), "toggled", GTK_SIGNAL_FUNC(fmWindow_ctreeMenuItemColumnsToggleColumn), fmWindow);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuColumnsMenuItemModificationTime), "destroy", GTK_SIGNAL_FUNC(fmWindow_destroyMenuItem), fmWindow);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuColumnsMenuItemModificationTime), "add_accelerator", GTK_SIGNAL_FUNC(fmWindow_acceleratorAdd), fmWindow);
		gtk_signal_connect_after(GTK_OBJECT(fmWindow->pPopupMenuColumnsMenuItemModificationTime), "remove_accelerator", GTK_SIGNAL_FUNC(fmWindow_acceleratorRemove), fmWindow);
		gtk_object_set_data(GTK_OBJECT(fmWindow->pPopupMenuColumnsMenuItemModificationTime), "accelKeyId", (gpointer)g_strdup("ColumnModificationTime"));
		fmWindow_accelKeyAddToGtkWidget(fmWindow, GTK_WIDGET(fmWindow->pPopupMenuColumnsMenuItemModificationTime));
		
		fmWindow->pPopupMenuColumnsMenuItemChangeTime = gtk_check_menu_item_new_with_label("Change Time");
		gtk_check_menu_item_set_show_toggle(GTK_CHECK_MENU_ITEM(fmWindow->pPopupMenuColumnsMenuItemChangeTime), TRUE);
		if(g_list_find(fmWindow->pCTreeColumns, (gpointer)COLUMN_CHANGE_TIME))
		{
			gtk_check_menu_item_set_state(GTK_CHECK_MENU_ITEM(fmWindow->pPopupMenuColumnsMenuItemChangeTime), TRUE);
		}
		gtk_object_set_data(GTK_OBJECT(fmWindow->pPopupMenuColumnsMenuItemChangeTime), "columnId", (gpointer)COLUMN_CHANGE_TIME);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuColumnsMenuItemChangeTime), "toggled", GTK_SIGNAL_FUNC(fmWindow_ctreeMenuItemColumnsToggleColumn), fmWindow);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuColumnsMenuItemChangeTime), "destroy", GTK_SIGNAL_FUNC(fmWindow_destroyMenuItem), fmWindow);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuColumnsMenuItemChangeTime), "add_accelerator", GTK_SIGNAL_FUNC(fmWindow_acceleratorAdd), fmWindow);
		gtk_signal_connect_after(GTK_OBJECT(fmWindow->pPopupMenuColumnsMenuItemChangeTime), "remove_accelerator", GTK_SIGNAL_FUNC(fmWindow_acceleratorRemove), fmWindow);
		gtk_object_set_data(GTK_OBJECT(fmWindow->pPopupMenuColumnsMenuItemChangeTime), "accelKeyId", (gpointer)g_strdup("ColumnChangeTime"));
		fmWindow_accelKeyAddToGtkWidget(fmWindow, GTK_WIDGET(fmWindow->pPopupMenuColumnsMenuItemChangeTime));
		
		fmWindow->pPopupMenuItemSeperator1 = gtk_menu_item_new();
		
		fmWindow->pPopupMenuItemAbout = gtk_menu_item_new_with_label("About...");
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuItemAbout), "activate", GTK_SIGNAL_FUNC(fmWindow_ctreeMenuItemAbout), fmWindow);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuItemAbout), "destroy", GTK_SIGNAL_FUNC(fmWindow_destroyMenuItem), fmWindow);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuItemAbout), "add_accelerator", GTK_SIGNAL_FUNC(fmWindow_acceleratorAdd), fmWindow);
		gtk_signal_connect_after(GTK_OBJECT(fmWindow->pPopupMenuItemAbout), "remove_accelerator", GTK_SIGNAL_FUNC(fmWindow_acceleratorRemove), fmWindow);
		gtk_object_set_data(GTK_OBJECT(fmWindow->pPopupMenuItemAbout), "accelKeyId", (gpointer)g_strdup("About"));
		fmWindow_accelKeyAddToGtkWidget(fmWindow, GTK_WIDGET(fmWindow->pPopupMenuItemAbout));

		fmWindow->pPopupMenuItemPreferences = gtk_menu_item_new_with_label("Preferences...");
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuItemPreferences), "activate", GTK_SIGNAL_FUNC(fmWindow_ctreeMenuItemPreferences), fmWindow);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuItemPreferences), "destroy", GTK_SIGNAL_FUNC(fmWindow_destroyMenuItem), fmWindow);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuItemPreferences), "add_accelerator", GTK_SIGNAL_FUNC(fmWindow_acceleratorAdd), fmWindow);
		gtk_signal_connect_after(GTK_OBJECT(fmWindow->pPopupMenuItemPreferences), "remove_accelerator", GTK_SIGNAL_FUNC(fmWindow_acceleratorRemove), fmWindow);
		gtk_object_set_data(GTK_OBJECT(fmWindow->pPopupMenuItemPreferences), "accelKeyId", (gpointer)g_strdup("Preferences"));
		fmWindow_accelKeyAddToGtkWidget(fmWindow, GTK_WIDGET(fmWindow->pPopupMenuItemPreferences));

		fmWindow->pPopupMenuItemSeperator2 = gtk_menu_item_new();
		
		fmWindow->pPopupMenuItemQuit = gtk_menu_item_new_with_label("Quit");
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuItemQuit), "activate", GTK_SIGNAL_FUNC(fmWindow_ctreeMenuItemQuit), fmWindow);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuItemQuit), "destroy", GTK_SIGNAL_FUNC(fmWindow_destroyMenuItem), fmWindow);
		gtk_signal_connect(GTK_OBJECT(fmWindow->pPopupMenuItemQuit), "add_accelerator", GTK_SIGNAL_FUNC(fmWindow_acceleratorAdd), fmWindow);
		gtk_signal_connect_after(GTK_OBJECT(fmWindow->pPopupMenuItemQuit), "remove_accelerator", GTK_SIGNAL_FUNC(fmWindow_acceleratorRemove), fmWindow);
		gtk_object_set_data(GTK_OBJECT(fmWindow->pPopupMenuItemQuit), "accelKeyId", (gpointer)g_strdup("Quit"));
		fmWindow_accelKeyAddToGtkWidget(fmWindow, GTK_WIDGET(fmWindow->pPopupMenuItemQuit));

		// add the items to the popup menu
		gtk_menu_append(GTK_MENU(fmWindow->pPopupMenu), fmWindow->pPopupMenuItemSelection);
        gtk_menu_item_set_submenu(GTK_MENU_ITEM(fmWindow->pPopupMenuItemSelection), fmWindow->pPopupMenuSelection);
		gtk_menu_append(GTK_MENU(fmWindow->pPopupMenuSelection), fmWindow->pPopupMenuSelectionMenuItemClear);
		gtk_menu_append(GTK_MENU(fmWindow->pPopupMenuSelection), fmWindow->pPopupMenuSelectionMenuItemSelectDirectory);
		gtk_menu_append(GTK_MENU(fmWindow->pPopupMenuSelection), fmWindow->pPopupMenuSelectionMenuItemDeselectDirectory);
		gtk_menu_append(GTK_MENU(fmWindow->pPopupMenuSelection), fmWindow->pPopupMenuSelectionMenuItemCopyToXClipboard);
		gtk_menu_append(GTK_MENU(fmWindow->pPopupMenuSelection), fmWindow->pPopupMenuSelectionMenuItemCopyFocusToComboEntry);
		gtk_menu_append(GTK_MENU(fmWindow->pPopupMenu), fmWindow->pPopupMenuItemDirectory);
		gtk_menu_item_set_submenu(GTK_MENU_ITEM(fmWindow->pPopupMenuItemDirectory), fmWindow->pPopupMenuDirectory);
		gtk_menu_append(GTK_MENU(fmWindow->pPopupMenuDirectory), fmWindow->pPopupMenuDirectoryMenuItemRefresh);
		gtk_menu_append(GTK_MENU(fmWindow->pPopupMenuDirectory), fmWindow->pPopupMenuDirectoryMenuItemZoom);
		gtk_menu_append(GTK_MENU(fmWindow->pPopupMenuDirectory), fmWindow->pPopupMenuDirectoryMenuItemOpenInNewWindow);
		gtk_menu_append(GTK_MENU(fmWindow->pPopupMenu), fmWindow->pPopupMenuItemBrowser);
		gtk_menu_item_set_submenu(GTK_MENU_ITEM(fmWindow->pPopupMenuItemBrowser), fmWindow->pPopupMenuBrowser);
		gtk_menu_append(GTK_MENU(fmWindow->pPopupMenuBrowser), fmWindow->pPopupMenuBrowserMenuItemRefresh);
		gtk_menu_append(GTK_MENU(fmWindow->pPopupMenuBrowser), fmWindow->pPopupMenuBrowserMenuItemNew);
		gtk_menu_append(GTK_MENU(fmWindow->pPopupMenuBrowser), fmWindow->pPopupMenuBrowserMenuItemClose);
		gtk_menu_append(GTK_MENU(fmWindow->pPopupMenuBrowser), fmWindow->pPopupMenuBrowserMenuItemCollapseEverything);
		gtk_menu_append(GTK_MENU(fmWindow->pPopupMenuBrowser), fmWindow->pPopupMenuBrowserMenuItemUnZoom);
		gtk_menu_append(GTK_MENU(fmWindow->pPopupMenuBrowser), fmWindow->pPopupMenuBrowserMenuItemCopyState);
		gtk_menu_append(GTK_MENU(fmWindow->pPopupMenuBrowser), fmWindow->pPopupMenuBrowserMenuItemPasteState);
		gtk_menu_append(GTK_MENU(fmWindow->pPopupMenu), fmWindow->pPopupMenuItemColumns);
		gtk_menu_item_set_submenu(GTK_MENU_ITEM(fmWindow->pPopupMenuItemColumns), fmWindow->pPopupMenuColumns);
		gtk_menu_append(GTK_MENU(fmWindow->pPopupMenuColumns), fmWindow->pPopupMenuColumnsMenuItemSize);
		gtk_menu_append(GTK_MENU(fmWindow->pPopupMenuColumns), fmWindow->pPopupMenuColumnsMenuItemUser);
		gtk_menu_append(GTK_MENU(fmWindow->pPopupMenuColumns), fmWindow->pPopupMenuColumnsMenuItemGroup);
		gtk_menu_append(GTK_MENU(fmWindow->pPopupMenuColumns), fmWindow->pPopupMenuColumnsMenuItemPermissions);
		gtk_menu_append(GTK_MENU(fmWindow->pPopupMenuColumns), fmWindow->pPopupMenuColumnsMenuItemModificationTime);
		gtk_menu_append(GTK_MENU(fmWindow->pPopupMenuColumns), fmWindow->pPopupMenuColumnsMenuItemChangeTime);
		gtk_menu_append(GTK_MENU(fmWindow->pPopupMenu), fmWindow->pPopupMenuItemSeperator1);
		gtk_menu_append(GTK_MENU(fmWindow->pPopupMenu), fmWindow->pPopupMenuItemAbout);
		gtk_menu_append(GTK_MENU(fmWindow->pPopupMenu), fmWindow->pPopupMenuItemPreferences);
		gtk_menu_append(GTK_MENU(fmWindow->pPopupMenu), fmWindow->pPopupMenuItemSeperator2);
		gtk_menu_append(GTK_MENU(fmWindow->pPopupMenu), fmWindow->pPopupMenuItemQuit);

		// show the gtk widgets
		gtk_widget_show(fmWindow->pPopupMenuItemQuit);
		gtk_widget_show(fmWindow->pPopupMenuItemSeperator2);
		gtk_widget_show(fmWindow->pPopupMenuItemPreferences);
		gtk_widget_show(fmWindow->pPopupMenuItemAbout);
		gtk_widget_show(fmWindow->pPopupMenuItemSeperator1);
		gtk_widget_show(fmWindow->pPopupMenuColumnsMenuItemChangeTime);
		gtk_widget_show(fmWindow->pPopupMenuColumnsMenuItemModificationTime);
		gtk_widget_show(fmWindow->pPopupMenuColumnsMenuItemPermissions);
		gtk_widget_show(fmWindow->pPopupMenuColumnsMenuItemGroup);
		gtk_widget_show(fmWindow->pPopupMenuColumnsMenuItemUser);
		gtk_widget_show(fmWindow->pPopupMenuColumnsMenuItemSize);
		gtk_widget_show(fmWindow->pPopupMenuColumns);
		gtk_widget_show(fmWindow->pPopupMenuItemColumns);
		gtk_widget_show(fmWindow->pPopupMenuBrowserMenuItemPasteState);
		gtk_widget_show(fmWindow->pPopupMenuBrowserMenuItemCopyState);
		gtk_widget_show(fmWindow->pPopupMenuBrowserMenuItemUnZoom);
		gtk_widget_show(fmWindow->pPopupMenuBrowserMenuItemCollapseEverything);
		gtk_widget_show(fmWindow->pPopupMenuBrowserMenuItemClose);
		gtk_widget_show(fmWindow->pPopupMenuBrowserMenuItemNew);
		gtk_widget_show(fmWindow->pPopupMenuBrowserMenuItemRefresh);
		gtk_widget_show(fmWindow->pPopupMenuBrowser);
		gtk_widget_show(fmWindow->pPopupMenuItemBrowser);
		gtk_widget_show(fmWindow->pPopupMenuDirectoryMenuItemOpenInNewWindow);
		gtk_widget_show(fmWindow->pPopupMenuDirectoryMenuItemZoom);
		gtk_widget_show(fmWindow->pPopupMenuDirectoryMenuItemRefresh);
		gtk_widget_show(fmWindow->pPopupMenuDirectory);
		gtk_widget_show(fmWindow->pPopupMenuItemDirectory);
		gtk_widget_show(fmWindow->pPopupMenuSelectionMenuItemCopyFocusToComboEntry);
		gtk_widget_show(fmWindow->pPopupMenuSelectionMenuItemCopyToXClipboard);
		gtk_widget_show(fmWindow->pPopupMenuSelectionMenuItemDeselectDirectory);
		gtk_widget_show(fmWindow->pPopupMenuSelectionMenuItemSelectDirectory);
		gtk_widget_show(fmWindow->pPopupMenuSelectionMenuItemClear);
		gtk_widget_show(fmWindow->pPopupMenuSelection);
		gtk_widget_show(fmWindow->pPopupMenuItemSelection);
		gtk_widget_show(fmWindow->pPopupMenu);
		gtk_widget_show(fmWindow->pLabelSelectionSize);
		gtk_widget_show(fmWindow->pAlignmentSelectionSize);
		gtk_widget_show(fmWindow->pFrameSelectionSize);
		gtk_widget_show(fmWindow->pLabelGeneralStatus);
		gtk_widget_show(fmWindow->pFrameGeneralStatus);
		gtk_widget_show(fmWindow->pHBoxBottomStatus);
		gtk_widget_show(fmWindow->pBookmarkButtons[0].pButtonBookmark);
		gtk_widget_show(fmWindow->pBookmarkButtons[9].pButtonBookmark);
		gtk_widget_show(fmWindow->pBookmarkButtons[8].pButtonBookmark);
		gtk_widget_show(fmWindow->pBookmarkButtons[7].pButtonBookmark);
		gtk_widget_show(fmWindow->pBookmarkButtons[6].pButtonBookmark);
		gtk_widget_show(fmWindow->pBookmarkButtons[5].pButtonBookmark);
		gtk_widget_show(fmWindow->pBookmarkButtons[4].pButtonBookmark);
		gtk_widget_show(fmWindow->pBookmarkButtons[3].pButtonBookmark);
		gtk_widget_show(fmWindow->pBookmarkButtons[2].pButtonBookmark);
		gtk_widget_show(fmWindow->pBookmarkButtons[1].pButtonBookmark);
		gtk_widget_show(fmWindow->pHBoxBottomButtons);
		gtk_widget_show(fmWindow->pPixmapPanWindow);
		gtk_widget_show(fmWindow->pEventBoxPanWindow);
		gtk_widget_show(fmWindow->pVScrollbarScrolledWindow);
		gtk_widget_show(fmWindow->pVBoxScrolledWindowRight);
		gtk_widget_show(fmWindow->pCTree);
		gtk_widget_show(fmWindow->pHScrollbarScrolledWindow);
		gtk_widget_show(fmWindow->pVBoxScrolledWindowLeft);
		gtk_widget_show(fmWindow->pHBoxScrolledWindow);
		gtk_widget_show(fmWindow->pComboBoxPopupMenu);
		gtk_widget_show(fmWindow->pComboBox);
		gtk_widget_show(fmWindow->pVBox);
		gtk_widget_show(fmWindow->pButtonHideDotfiles);
		gtk_widget_show(fmWindow->pHBoxTop);
		gtk_widget_show(fmWindow->pWindow);

		// create to visited links array
		fmWindow->pVisitedLinks = g_array_new(FALSE, FALSE, sizeof(glong));
		
		// copy the global bookmarks to the new window if there were already windows defined
		if(g_list_length(pgFmWindows))
		{
			for(counter = 0; counter < 10; counter++)
			{
				if(((struct FmWindow *)pgFmWindows->data)->pBookmarks[counter])
				{
					fmWindow_bookmarkSet(fmWindow, ((struct FmWindow *)pgFmWindows->data)->pBookmarks[counter], counter);
				}
			}
		}
		
		// set the window position
		if(fmWindow->initialX >= 0 && fmWindow->initialY >= 0)
		{
			gtk_widget_set_uposition(GTK_WIDGET(fmWindow->pWindow), fmWindow->initialX, fmWindow->initialY);
		}
	}
}

gboolean fmWindow_insertCTreeDir(struct FmWindow *fmWindow, GNode *currentGNode, GtkCTreeNode *parentNode, GtkCTreeNode *siblingNode, gchar* representation)
{
    struct NodeInfo	*current_node_info = (struct NodeInfo*)currentGNode->data;

//	printf("insertCTreeDir %s\n", ((struct NodeInfo *)current_node_info)->pFullName);
	
	// if the gnode is to rootnode, don't display it, but display immediately level 1
	if(currentGNode == fmWindow->pRootNode)
	{
        fmWindow_fillDir(fmWindow, currentGNode);
		return FALSE;
	}
	else
	{
		// create empty node to support expansion of the directory
		current_node_info->pCtreeNode =  gtk_ctree_insert_node(GTK_CTREE(fmWindow->pCTree), GTK_CTREE_NODE(parentNode), GTK_CTREE_NODE(siblingNode), NULL, 0, fileTypeGet(pgPreferences->pFileTypes, currentGNode)->pRowStyle->pPixmapClosed, fileTypeGet(pgPreferences->pFileTypes, currentGNode)->pRowStyle->pMaskClosed, fileTypeGet(pgPreferences->pFileTypes, currentGNode)->pRowStyle->pPixmapOpened, fileTypeGet(pgPreferences->pFileTypes, currentGNode)->pRowStyle->pMaskOpened, FALSE, FALSE);
		if(current_node_info->dirHasChildren)
		{
			// directory has got entries
			current_node_info->pDirEmptyCtreeNode = gtk_ctree_insert_node(GTK_CTREE(fmWindow->pCTree), GTK_CTREE_NODE(current_node_info->pCtreeNode), NULL, NULL, 0, NULL, NULL, NULL, NULL, TRUE, FALSE);
			gtk_ctree_node_set_text(GTK_CTREE(fmWindow->pCTree), GTK_CTREE_NODE(current_node_info->pDirEmptyCtreeNode), 0, "");
		}
	}
	
	return TRUE;
}

gboolean fmWindow_insertCTreeFile(struct FmWindow *fmWindow, GNode *currentGNode, GtkCTreeNode *parentNode, GtkCTreeNode *siblingNode, gchar* representation)
{
    struct NodeInfo	*current_node_info = (struct NodeInfo*)currentGNode->data;

//	printf("insertCTreeFile %s\n", ((struct NodeInfo *)current_node_info)->pFullName);
	
	current_node_info->pCtreeNode =  gtk_ctree_insert_node(GTK_CTREE(fmWindow->pCTree), GTK_CTREE_NODE(parentNode), GTK_CTREE_NODE(siblingNode), NULL, 0, NULL, NULL, NULL, NULL, TRUE, FALSE);
	
	return TRUE;
}

gboolean fmWindow_insertCTreeLink(struct FmWindow *fmWindow, GNode *currentGNode, GtkCTreeNode *parentNode, GtkCTreeNode *siblingNode, gchar* representation)
{
    struct NodeInfo	*current_node_info = (struct NodeInfo*)currentGNode->data;
	GNode			*target_gnode = getLinkTarget(currentGNode);

//	printf("insertCTreeLink %s\n", ((struct NodeInfo *)current_node_info)->pFullName);

	// check if the link isn't broken
	if(target_gnode && !isLink(target_gnode))
	{
		fmWindow_insertCTreeNode(fmWindow, target_gnode, parentNode, siblingNode, current_node_info->pLinkRepresentation);
        current_node_info->pCtreeNode = ((struct NodeInfo*)target_gnode->data)->pCtreeNode;
		
		return FALSE;
	}
	else
	{
		// broken link
		current_node_info->pCtreeNode =  gtk_ctree_insert_node(GTK_CTREE(fmWindow->pCTree), GTK_CTREE_NODE(parentNode), GTK_CTREE_NODE(siblingNode), NULL, 0, NULL, NULL, NULL, NULL, TRUE, FALSE);
		
		return TRUE;
	}
}

void fmWindow_insertCTreeNode(struct FmWindow *fmWindow, GNode *currentGNode, GtkCTreeNode *parentNode, GtkCTreeNode *siblingNode, gchar* representation)
{
	gboolean		return_value = FALSE;
	struct FileType	*file_type = NULL;
	GNode			*source_gnode = NULL;
    struct NodeInfo	*source_node_info = NULL;
    struct NodeInfo	*target_node_info = NULL;
	gint			bookmark_index = -1;
	GList			*column = NULL;
	guint			counter = 0;

//	printf("insertCTreeNode %s\n", ((struct NodeInfo *)currentGNode->data)->pFullName);
	if(isLink(currentGNode))
	{
		return_value = fmWindow_insertCTreeLink(fmWindow, currentGNode, parentNode, siblingNode, representation);
	}
	else if(isDirectory(currentGNode))
	{
		return_value = fmWindow_insertCTreeDir(fmWindow, currentGNode, parentNode, siblingNode, representation);
	}
	else
	{
		return_value = fmWindow_insertCTreeFile(fmWindow, currentGNode, parentNode, siblingNode, representation);
	}

	if(return_value)
	{
		file_type = fileTypeGet(pgPreferences->pFileTypes, currentGNode);
		if(file_type)
		{
            source_gnode = getLinkSource(currentGNode);
			source_node_info = (struct NodeInfo*)source_gnode->data;
			target_node_info = (struct NodeInfo*)getLinkTarget(currentGNode)->data;
//			printf("insertCTreeNode contents %s filetype %d\n", source_node_info->pFullName, file_type->basicType);
			if(representation == NULL)
			{
				representation = source_node_info->pShortName;
			}
			// check if the node is not part of the selection
			if(fmWindow_selectionGetListNode(fmWindow, source_gnode))
			{
				// yes, select the node
				fmWindow_selectNode(fmWindow, source_gnode);
			}
			else
			{
				// no, apply the general file type coloring rules
				gtk_ctree_node_set_background(GTK_CTREE(fmWindow->pCTree), GTK_CTREE_NODE(target_node_info->pCtreeNode), &(file_type->pRowStyle->background));
				gtk_ctree_node_set_foreground(GTK_CTREE(fmWindow->pCTree), GTK_CTREE_NODE(target_node_info->pCtreeNode), &(file_type->pRowStyle->foreground));
			}
            bookmark_index = fmWindow_bookmarkGetIndex(fmWindow, source_gnode);
			if(bookmark_index > -1)
			{
                fmWindow_bookmarkDrawIndication(fmWindow, getLinkTarget(source_gnode), bookmark_index);
			}
            column = g_list_first(pgPreferences->pCtreeColumns);
            counter = 0;
			while(column)
			{
				if(g_list_find(fmWindow->pCTreeColumns,(gpointer)((struct CtreeColumn*)column->data)->type))
				{
					if(((struct CtreeColumn*)column->data)->type == COLUMN_FILENAME)
					{
						gtk_ctree_node_set_pixtext(GTK_CTREE(fmWindow->pCTree), GTK_CTREE_NODE(target_node_info->pCtreeNode), counter, representation, 0, file_type->pRowStyle->pPixmapClosed, file_type->pRowStyle->pMaskClosed);
					}
					else if(((struct CtreeColumn*)column->data)->type == COLUMN_SIZE)
					{
						gtk_ctree_node_set_text(GTK_CTREE(fmWindow->pCTree), GTK_CTREE_NODE(target_node_info->pCtreeNode), counter, source_node_info->pSizeString);
					}
					else if(((struct CtreeColumn*)column->data)->type == COLUMN_USER)
					{
						gtk_ctree_node_set_text(GTK_CTREE(fmWindow->pCTree), GTK_CTREE_NODE(target_node_info->pCtreeNode), counter, source_node_info->pUserString);
					}
					else if(((struct CtreeColumn*)column->data)->type == COLUMN_GROUP)
					{
						gtk_ctree_node_set_text(GTK_CTREE(fmWindow->pCTree), GTK_CTREE_NODE(target_node_info->pCtreeNode), counter, source_node_info->pGroupString);
					}
					else if(((struct CtreeColumn*)column->data)->type == COLUMN_PERMISSIONS)
					{
						gtk_ctree_node_set_text(GTK_CTREE(fmWindow->pCTree), GTK_CTREE_NODE(target_node_info->pCtreeNode), counter, source_node_info->pPermissionsString);
					}
					else if(((struct CtreeColumn*)column->data)->type == COLUMN_MODIFICATION_TIME)
					{
						gtk_ctree_node_set_text(GTK_CTREE(fmWindow->pCTree), GTK_CTREE_NODE(target_node_info->pCtreeNode), counter, source_node_info->pModificationTimeString);
					}
					else if(((struct CtreeColumn*)column->data)->type == COLUMN_CHANGE_TIME)
					{
						gtk_ctree_node_set_text(GTK_CTREE(fmWindow->pCTree), GTK_CTREE_NODE(target_node_info->pCtreeNode), counter, source_node_info->pChangeTimeString);
					}
					counter++;
				}
                column = g_list_next(column);
			}
			gtk_ctree_node_set_row_data(GTK_CTREE(fmWindow->pCTree), GTK_CTREE_NODE(target_node_info->pCtreeNode), source_gnode);
		}
	}
}

gboolean fmWindow_isNodeParentOfNode(struct FmWindow *fmWindow, GNode *parentNode, GNode *childNode)
{
	while(childNode)
	{
		if(childNode == parentNode)
		{
			return TRUE;
		}
		else
		{
			childNode = getLinkSource(childNode->parent);
		}
	}
	
	return FALSE;
}

void fmWindow_listCTreeDirChildren(struct FmWindow *fmWindow, GtkCTreeNode *ctreeNode)
{
	GNode			*source_gnode = NULL;
	struct NodeInfo	*source_node_info = NULL;
	GtkCTreeNode	*temp_empty_ctree_node = NULL;
	GtkCTreeNode	*temp_ctree_node = NULL;
	GtkCTreeNode	*temp_ctree_node_previous = NULL;
	GNode			*temp_ctree_node_gnode = NULL;

	source_gnode = (GNode *)gtk_ctree_node_get_row_data(GTK_CTREE(fmWindow->pCTree), GTK_CTREE_NODE(ctreeNode));
	source_node_info = source_gnode->data;

	temp_empty_ctree_node = gtk_ctree_insert_node(GTK_CTREE(fmWindow->pCTree), GTK_CTREE_NODE(ctreeNode), NULL, NULL, 0, NULL, NULL, NULL, NULL, TRUE, FALSE);
	gtk_ctree_node_set_text(GTK_CTREE(fmWindow->pCTree), GTK_CTREE_NODE(temp_empty_ctree_node), 0, "");

	temp_ctree_node_previous = GTK_CTREE_NODE_PREV(temp_empty_ctree_node);
	while(temp_ctree_node_previous != ctreeNode)
	{
        temp_ctree_node = temp_ctree_node_previous;
        temp_ctree_node_gnode = (GNode *)gtk_ctree_node_get_row_data(GTK_CTREE(fmWindow->pCTree), GTK_CTREE_NODE(temp_ctree_node));
		if(temp_ctree_node_gnode && (temp_ctree_node_gnode->parent == source_gnode || ( ((struct NodeInfo*)temp_ctree_node_gnode->data)->pLinkFromNode && ((struct NodeInfo*)temp_ctree_node_gnode->data)->pLinkFromNode->parent == source_gnode )))
		{
			printf("%s\n", ((struct NodeInfo*)temp_ctree_node_gnode->data)->pFullName);
		}
		temp_ctree_node_previous = GTK_CTREE_NODE_PREV(temp_ctree_node);
	}

	gtk_ctree_remove_node(GTK_CTREE(fmWindow->pCTree), GTK_CTREE_NODE(temp_empty_ctree_node));
}

void fmWindow_moveToCTreeRow(struct FmWindow *fmWindow, guint row)
{
//	printf("row = %d visibility = %d\n", row, gtk_clist_row_is_visible(GTK_CLIST(fmWindow->pCTree), row));
	if(gtk_clist_row_is_visible(GTK_CLIST(fmWindow->pCTree), row) != GTK_VISIBILITY_FULL)
	{
		gtk_clist_moveto(GTK_CLIST(fmWindow->pCTree), row, fmWindow->focusedColumn, 0.5, 0);
	}
	else
	{
		gtk_clist_moveto(GTK_CLIST(fmWindow->pCTree), -1, fmWindow->focusedColumn, 0.5, 0);
	}
}

void fmWindow_moveToNode(struct FmWindow *fmWindow, GNode *sourceNode)
{
	guint row = 0;
	
	row = gtk_clist_find_row_from_data(GTK_CLIST(fmWindow->pCTree), sourceNode);
	fmWindow_moveToCTreeRow(fmWindow, row);
}

struct FmWindow* fmWindow_new(struct FmWindow *exampleFmWindow)
{
	struct FmWindow			*fm_window = NULL;
	gint					counter = 0;

	fm_window = g_new(struct FmWindow, 1);
	if(fm_window)
	{
		//initialize the values
		fm_window->pExpandedDirs = NULL;
		fm_window->pFocusedNode = NULL;
		fm_window->build = 0;
		fm_window->focusedColumn = 0;
		fm_window->freed = 0;
		fm_window->hideDotFiles = TRUE;
		fm_window->pInitialFocus = NULL;
		fm_window->initialX = -1;
		fm_window->initialY = -1;
		fm_window->initialWidth = -1;
		fm_window->initialHeight = -1;
		fm_window->sortDirection = SORT_ASCENDING;
		fm_window->sortType = SORT_FILENAME;
		
		fm_window->pComboBoxStrings = NULL;
		fm_window->pCTreeColumns = NULL;
		fm_window->pRootNode = NULL;
		fm_window->pSelection = NULL;
		fm_window->pSelectionTextSnapshot = NULL;
		fm_window->pShortcut = g_string_new("");
		fm_window->pVisitedLinks = NULL;
		fm_window->pZoomDirectory = g_strdup("/");
		
		fm_window->pAccelGroup = NULL;
		fm_window->pAdjustmentHScrollbar = NULL;
		fm_window->pAdjustmentVScrollbar = NULL;
		fm_window->pAlignmentSelectionSize = NULL;
		fm_window->pButtonHideDotfiles = NULL;
		fm_window->pComboBox = NULL;
        fm_window->pComboBoxPopupMenu = NULL;
		fm_window->pCTree = NULL;
		fm_window->pEventBoxPanWindow = NULL;
		fm_window->pFrameGeneralStatus = NULL;
		fm_window->pFrameSelectionSize = NULL;
		fm_window->pHBoxBottomButtons = NULL;
		fm_window->pHBoxBottomStatus = NULL;
		fm_window->pHBoxScrolledWindow = NULL;
		fm_window->pHBoxTop = NULL;
		fm_window->pHScrollbarScrolledWindow = NULL;
		fm_window->pLabelGeneralStatus = NULL;
		fm_window->pLabelSelectionSize = NULL;
		fm_window->pPixmapPanWindow = NULL;
		fm_window->pPopupMenuItemSelection = NULL;
		fm_window->pPopupMenuSelection = NULL;
		fm_window->pPopupMenuSelectionMenuItemClear = NULL;
		fm_window->pPopupMenuSelectionMenuItemSelectDirectory = NULL;
		fm_window->pPopupMenuSelectionMenuItemDeselectDirectory = NULL;
		fm_window->pPopupMenuSelectionMenuItemCopyToXClipboard = NULL;
		fm_window->pPopupMenuSelectionMenuItemCopyFocusToComboEntry = NULL;
		fm_window->pPopupMenuItemDirectory = NULL;
		fm_window->pPopupMenuDirectory = NULL;
		fm_window->pPopupMenuDirectoryMenuItemRefresh = NULL;
		fm_window->pPopupMenuDirectoryMenuItemZoom = NULL;
		fm_window->pPopupMenuDirectoryMenuItemOpenInNewWindow = NULL;
		fm_window->pPopupMenuItemBrowser = NULL;
		fm_window->pPopupMenuBrowser = NULL;
		fm_window->pPopupMenuBrowserMenuItemRefresh = NULL;
		fm_window->pPopupMenuBrowserMenuItemNew = NULL;
		fm_window->pPopupMenuBrowserMenuItemClose = NULL;
		fm_window->pPopupMenuBrowserMenuItemCollapseEverything = NULL;
		fm_window->pPopupMenuBrowserMenuItemUnZoom = NULL;
		fm_window->pPopupMenuBrowserMenuItemCopyState = NULL;
		fm_window->pPopupMenuBrowserMenuItemPasteState = NULL;
		fm_window->pPopupMenuItemColumns = NULL;
		fm_window->pPopupMenuColumns = NULL;
		fm_window->pPopupMenuColumnsMenuItemSize = NULL;
		fm_window->pPopupMenuColumnsMenuItemUser = NULL;
		fm_window->pPopupMenuColumnsMenuItemGroup = NULL;
		fm_window->pPopupMenuColumnsMenuItemPermissions = NULL;
		fm_window->pPopupMenuColumnsMenuItemModificationTime = NULL;
		fm_window->pPopupMenuColumnsMenuItemChangeTime = NULL;
		fm_window->pPopupMenuItemSeperator1 = NULL;
		fm_window->pPopupMenuItemAbout = NULL;
		fm_window->pPopupMenuItemPreferences = NULL;
		fm_window->pPopupMenuItemSeperator2 = NULL;
		fm_window->pPopupMenuItemQuit = NULL;
		fm_window->pToolTips = NULL;
		fm_window->pVBox = NULL;
		fm_window->pVBoxScrolledWindowLeft = NULL;
		fm_window->pVBoxScrolledWindowRight = NULL;
		fm_window->pVScrollbarScrolledWindow = NULL;
		fm_window->pWindow = NULL;
		
		fm_window->pAcceleratedWidgets = NULL;
		fm_window->pBookmarks = NULL;

		// add the minimum required columns if no example fm window has been defined, otherwise copy the column ids of the example window
		if(exampleFmWindow)
		{
			for(counter = g_list_length(exampleFmWindow->pCTreeColumns)-1; counter >= 0; counter--)
			{
				fm_window->pCTreeColumns = g_list_prepend(fm_window->pCTreeColumns, g_list_nth_data(exampleFmWindow->pCTreeColumns, counter));
			}
		}
		else
		{
			fm_window->pCTreeColumns = g_list_prepend(fm_window->pCTreeColumns, (gpointer)COLUMN_BOOKMARK);
			fm_window->pCTreeColumns = g_list_prepend(fm_window->pCTreeColumns, (gpointer)COLUMN_FILENAME);
		}
		
		// initialize the bookmarks array
		fm_window->pBookmarks = g_new0(gchar*, 10);
	}

	return fm_window;
}

void fmWindow_openDirectoryInNewWindow(struct FmWindow *fmWindow, GNode *sourceGnode)
{
	struct FmWindow	*new_fmWindow = NULL;
	gchar			*temp_string1 = NULL;
	gchar			*temp_string2 = NULL;

	if(isDirectoryFilled(getLinkTarget(sourceGnode)))
	{
		new_fmWindow = fmWindow_new(fmWindow);
		if(new_fmWindow)
		{
			if(isLink(sourceGnode))
			{
				if(g_path_is_absolute(((struct NodeInfo*)sourceGnode->data)->pLinkTarget))
				{
					new_fmWindow->pZoomDirectory = compressPath(((struct NodeInfo*)sourceGnode->data)->pLinkTarget);
				}
				else
				{
					temp_string1 = fmWindow_getFilenameToRemember(fmWindow, sourceGnode);
					temp_string2 = g_dirname(temp_string1);
					g_free(temp_string1);
					temp_string1 = g_strjoin("/", temp_string2, ((struct NodeInfo*)sourceGnode->data)->pLinkTarget, NULL);
					g_free(temp_string2);
					new_fmWindow->pZoomDirectory = compressPath(temp_string1);
					g_free(temp_string1);
				}
			}
			else
			{
				new_fmWindow->pZoomDirectory = fmWindow_getFilenameToRemember(fmWindow, sourceGnode);
			}
			buildNewFmWindow(new_fmWindow);
		}
	}
}

gint fmWindow_panTimeout(struct FmWindow *fmWindow)
{
	gint x = 0;
	gint y = 0;

    gtk_widget_get_pointer(GTK_WIDGET(fmWindow->pPixmapPanWindow), &x, &y);

	if(x > PAN_RIGHT_EDGE)
	{
		fmWindow_ctreeMoveHorizontal(fmWindow, x - PAN_RIGHT_EDGE);
	}
	else if(x < PAN_LEFT_EDGE)
	{
		fmWindow_ctreeMoveHorizontal(fmWindow, x - PAN_LEFT_EDGE);
	}

	if(y > PAN_BOTTOM_EDGE)
	{
		fmWindow_ctreeMoveVertical(fmWindow, y - PAN_BOTTOM_EDGE);
	}
	else if(y < PAN_TOP_EDGE)
	{
		fmWindow_ctreeMoveVertical(fmWindow, y - PAN_TOP_EDGE);
	}
  
	if(y >= PAN_TOP_EDGE && y <= PAN_BOTTOM_EDGE && x >= PAN_LEFT_EDGE && x <= PAN_RIGHT_EDGE)
	{
		gtk_pixmap_set(GTK_PIXMAP(fmWindow->pPixmapPanWindow), pgPixmapPanWindowCenter, pgMaskPanWindowCenter);
	}
	else if(y < PAN_TOP_EDGE && x >= PAN_LEFT_EDGE && x <= PAN_RIGHT_EDGE)
	{
		gtk_pixmap_set(GTK_PIXMAP(fmWindow->pPixmapPanWindow), pgPixmapPanWindowN, pgMaskPanWindowN);
	}
	else if(y < PAN_TOP_EDGE && x > PAN_RIGHT_EDGE)
	{
		gtk_pixmap_set(GTK_PIXMAP(fmWindow->pPixmapPanWindow), pgPixmapPanWindowNE, pgMaskPanWindowNE);
	}
	else if(y >= PAN_TOP_EDGE && y <= PAN_BOTTOM_EDGE && x > PAN_RIGHT_EDGE)
	{
		gtk_pixmap_set(GTK_PIXMAP(fmWindow->pPixmapPanWindow), pgPixmapPanWindowE, pgMaskPanWindowE);
	}
	else if(y > PAN_BOTTOM_EDGE && x > PAN_RIGHT_EDGE)
	{
		gtk_pixmap_set(GTK_PIXMAP(fmWindow->pPixmapPanWindow), pgPixmapPanWindowSE, pgMaskPanWindowSE);
	}
	else if(y > PAN_BOTTOM_EDGE && x >= PAN_LEFT_EDGE && x <= PAN_RIGHT_EDGE)
	{
		gtk_pixmap_set(GTK_PIXMAP(fmWindow->pPixmapPanWindow), pgPixmapPanWindowS, pgMaskPanWindowS);
	}
	else if(y > PAN_BOTTOM_EDGE && x < PAN_LEFT_EDGE )
	{
		gtk_pixmap_set(GTK_PIXMAP(fmWindow->pPixmapPanWindow), pgPixmapPanWindowSW, pgMaskPanWindowSW);
	}
	else if(y >= PAN_TOP_EDGE && y <= PAN_BOTTOM_EDGE && x < PAN_LEFT_EDGE )
	{
		gtk_pixmap_set(GTK_PIXMAP(fmWindow->pPixmapPanWindow), pgPixmapPanWindowW, pgMaskPanWindowW);
	}
	else if(y < PAN_TOP_EDGE && x < PAN_LEFT_EDGE)
	{
		gtk_pixmap_set(GTK_PIXMAP(fmWindow->pPixmapPanWindow), pgPixmapPanWindowNW, pgMaskPanWindowNW);
	}
	
	return TRUE;
}

void fmWindow_popupMenu(struct FmWindow *fmWindow, GdkEventButton *event)
{
	if(isDirectoryFilled(getLinkTarget(fmWindow->pFocusedNode)))
	{
		gtk_widget_set_sensitive(fmWindow->pPopupMenuDirectoryMenuItemZoom, TRUE);
		gtk_widget_set_sensitive(fmWindow->pPopupMenuDirectoryMenuItemOpenInNewWindow, TRUE);
	}
	else
	{
		gtk_widget_set_sensitive(fmWindow->pPopupMenuDirectoryMenuItemZoom, FALSE);
		gtk_widget_set_sensitive(fmWindow->pPopupMenuDirectoryMenuItemOpenInNewWindow, FALSE);
	}
	if(gBrowserState.pExpandedDirs != NULL || gBrowserState.pZoomDirectory != NULL || gBrowserState.pFocusedFile != NULL)
	{
		gtk_widget_set_sensitive(fmWindow->pPopupMenuBrowserMenuItemPasteState, TRUE);
	}
	else
	{
		gtk_widget_set_sensitive(fmWindow->pPopupMenuBrowserMenuItemPasteState, FALSE);
	}

	gtk_menu_popup(GTK_MENU(fmWindow->pPopupMenu), NULL, NULL, NULL, NULL, event->button, event->time);
}

void fmWindow_rebuildCTree(struct FmWindow *fmWindow)
{
	gchar	*focused_file = fmWindow_getFilenameToRemember(fmWindow, fmWindow->pFocusedNode);

	gtk_widget_destroy(GTK_WIDGET(fmWindow->pCTree));
	fmWindow->pCTree = fmWindow_buildCtree(fmWindow);
	gtk_clist_set_hadjustment(GTK_CLIST(fmWindow->pCTree), GTK_ADJUSTMENT(fmWindow->pAdjustmentHScrollbar));
	gtk_clist_set_vadjustment(GTK_CLIST(fmWindow->pCTree), GTK_ADJUSTMENT(fmWindow->pAdjustmentVScrollbar));
	gtk_box_pack_start(GTK_BOX(fmWindow->pVBoxScrolledWindowLeft), GTK_WIDGET(fmWindow->pCTree), TRUE, TRUE, 0);
	
	fmWindow_updateCTreeTitles(fmWindow);
	fmWindow_ctreeColumnsApplyAlignment(fmWindow);
	gtk_widget_show(fmWindow->pCTree);
	
	if(fmWindow->pRootNode)
	{
		fmWindow_freeNode(fmWindow, fmWindow->pRootNode);
	}
	fmWindow->pRootNode = fmWindow_buildNode(fmWindow, fmWindow->pZoomDirectory);
	gtk_clist_freeze(GTK_CLIST(fmWindow->pCTree));
	fmWindow_insertCTreeNode(fmWindow, fmWindow->pRootNode, NULL, NULL, NULL);
	fmWindow_resizeColumns(fmWindow);

	// let all the changements execute before continuing
	while(gtk_events_pending())
	{
		gtk_main_iteration();
	}

	if(focused_file)
	{
		fmWindow_focusFileName(fmWindow, focused_file);
		fmWindow_selectMoveToCTreeFocus(fmWindow);
		g_free(focused_file);
	}
	else if(fmWindow->pInitialFocus)
	{
		fmWindow_focusFileName(fmWindow, fmWindow->pInitialFocus);
		fmWindow_selectMoveToCTreeFocus(fmWindow);
	}
	else
	{
		fmWindow_focusFirstRow(fmWindow);
	}

	gtk_clist_thaw(GTK_CLIST(fmWindow->pCTree));
};

void fmWindow_refreshCTree(struct FmWindow *fmWindow)
{
	gchar	*focused_node_name = NULL;
	
	focused_node_name = fmWindow_getFilenameToRemember(fmWindow, fmWindow->pFocusedNode);
    fmWindow_refreshDirectory(fmWindow, fmWindow->pRootNode, focused_node_name);
};

void fmWindow_refreshDirectory(struct FmWindow *fmWindow, GNode *targetNode, gchar *focusedNodeName)
{
	gint	hadjustment = GTK_CTREE(fmWindow->pCTree)->clist.hadjustment->value;
	gint	vadjustment = GTK_CTREE(fmWindow->pCTree)->clist.vadjustment->value;
	GNode	*focused_source_gnode = NULL;
	
	gtk_clist_freeze(GTK_CLIST(fmWindow->pCTree));
	fmWindow_removeCTreeDirChildren(fmWindow, ((struct NodeInfo*)targetNode->data)->pCtreeNode);
	fmWindow_freeDirEntries(fmWindow, targetNode);
	fmWindow_fillDir(fmWindow, targetNode);
	focused_source_gnode = fmWindow_getNodeFromFilename(fmWindow, focusedNodeName, TRUE);
	if(focused_source_gnode == NULL)
	{
		if(targetNode == fmWindow->pRootNode)
		{
			focused_source_gnode = targetNode->children;
		}
		else
		{
			focused_source_gnode = targetNode;
		}
	}
	gtk_clist_thaw(GTK_CLIST(fmWindow->pCTree));
	fmWindow_focusNode(fmWindow, focused_source_gnode);
	fmWindow_selectNode(fmWindow, focused_source_gnode);
	fmWindow_resizeColumns(fmWindow);
	gtk_adjustment_set_value(GTK_CTREE(fmWindow->pCTree)->clist.hadjustment, hadjustment);
	gtk_adjustment_set_value(GTK_CTREE(fmWindow->pCTree)->clist.vadjustment, vadjustment);
}

void fmWindow_refreshParentDirectory(struct FmWindow *fmWindow, GNode *sourceNode)
{
	GNode	*parent_target_gnode = NULL;
	gchar	*focused_node_name = NULL;

	parent_target_gnode = sourceNode->parent;
	focused_node_name = fmWindow_getFilenameToRemember(fmWindow, fmWindow->pFocusedNode);
	fmWindow_refreshDirectory(fmWindow, parent_target_gnode, focused_node_name);
};

void fmWindow_removeCTreeDirChildren(struct FmWindow *fmWindow, GtkCTreeNode *ctreeNode) 
{
	GNode			*source_gnode = NULL;
	GNode			*target_gnode = NULL;
	struct NodeInfo	*target_node_info = NULL;
	GtkCTreeNode	*temp_ctree_node = NULL;
	GtkCTreeNode	*temp_ctree_node_previous = NULL;
	GNode			*temp_ctree_node_gnode = NULL;

	if(ctreeNode != NULL)
	{
		source_gnode = (GNode *)gtk_ctree_node_get_row_data(GTK_CTREE(fmWindow->pCTree), GTK_CTREE_NODE(ctreeNode));
	}
	else
	{
		source_gnode = fmWindow->pRootNode;
	}
	
	target_gnode = getLinkTarget(source_gnode);
	target_node_info = target_gnode->data;

	// insert the empty child at the end
	target_node_info->pDirEmptyCtreeNode = gtk_ctree_insert_node(GTK_CTREE(fmWindow->pCTree), GTK_CTREE_NODE(ctreeNode), NULL, NULL, 0, NULL, NULL, NULL, NULL, TRUE, FALSE);
	gtk_ctree_node_set_text(GTK_CTREE(fmWindow->pCTree), GTK_CTREE_NODE(target_node_info->pDirEmptyCtreeNode), 0, "");

	temp_ctree_node_previous = GTK_CTREE_NODE_PREV(target_node_info->pDirEmptyCtreeNode);
	while(temp_ctree_node_previous != ctreeNode)
	{
		temp_ctree_node = temp_ctree_node_previous;
		temp_ctree_node_gnode = (GNode *)gtk_ctree_node_get_row_data(GTK_CTREE(fmWindow->pCTree), GTK_CTREE_NODE(temp_ctree_node));
		if(temp_ctree_node_gnode->parent == target_gnode)
		{
			gtk_ctree_remove_node(GTK_CTREE(fmWindow->pCTree), GTK_CTREE_NODE(temp_ctree_node));
		}
		temp_ctree_node_previous = GTK_CTREE_NODE_PREV(temp_ctree_node);
	}

	target_node_info->pDirFirstCtreeNode = NULL;
}

void fmWindow_resizeColumns(struct FmWindow *fmWindow)
{
	gint	column = 0;
	guint	column_width = 0;
	guint	total_column_width = 0;
	guint	new_column_width = 0;

	gtk_clist_freeze(GTK_CLIST(fmWindow->pCTree));
	for(column = 0; column < GTK_CLIST(fmWindow->pCTree)->columns; column++)
	{
		column_width = gtk_clist_optimal_column_width(GTK_CLIST(fmWindow->pCTree), column);
		if(((struct CtreeColumn*)g_list_nth_data(pgPreferences->pCtreeColumns, column))->type != COLUMN_BOOKMARK)
		{
			column_width += 2*CELL_HORIZ_PADDING;
		}
//		printf("column %d width = %d\n", column, column_width+(2*COLUMN_INSET));
		gtk_clist_set_column_width(GTK_CLIST(fmWindow->pCTree), column, column_width);
		total_column_width += column_width+(2*COLUMN_INSET);
	}
//	printf("GTK_CLIST(fmWindow->pCTree)->clist_window_width = %d total_column_width = %d\n", GTK_CLIST(fmWindow->pCTree)->clist_window_width, total_column_width);
	if(GTK_CLIST(fmWindow->pCTree)->clist_window_width > total_column_width)
	{
        new_column_width = gtk_clist_optimal_column_width(GTK_CLIST(fmWindow->pCTree), 1)+(2*CELL_HORIZ_PADDING)+(GTK_CLIST(fmWindow->pCTree)->clist_window_width-total_column_width)-(2*COLUMN_INSET);
//		printf("new column width = %d\n", new_column_width);
		gtk_clist_set_column_width(GTK_CLIST(fmWindow->pCTree), 1, new_column_width-COLUMN_INSET);
	}
	
	gtk_clist_thaw(GTK_CLIST(fmWindow->pCTree));
}

void fmWindow_selectionAddNode(struct FmWindow *fmWindow, GNode *sourceNode)
{
	gchar	*filename_to_remember = NULL;

	// don't add the selection twice
	if(fmWindow_selectionGetListNode(fmWindow, sourceNode) == NULL)
	{
		filename_to_remember = fmWindow_getFilenameToRemember(fmWindow, sourceNode);
		fmWindow->pSelection = g_list_prepend(fmWindow->pSelection, filename_to_remember);
		fmWindow_selectNode(fmWindow, sourceNode);
		fmWindow_selectionUpdateSizeLabel(fmWindow);
	}
}

void fmWindow_selectionAddAllDirEntries(struct FmWindow *fmWindow, GNode *sourceNode)
{
	GNode	*parent_target_gnode = sourceNode->parent;
	GNode	*child_source_gnode = NULL;
	
	if(parent_target_gnode)
	{
		child_source_gnode = g_node_first_child(parent_target_gnode);
		while(child_source_gnode)
		{
			fmWindow_selectionAddNode(fmWindow, child_source_gnode);
			child_source_gnode = child_source_gnode->next;
		}
	}
}

void fmWindow_selectionDestroy(struct FmWindow *fmWindow)
{
	gchar	*temp_string = NULL;

	if(fmWindow)
	{
		while(g_list_nth(fmWindow->pSelection, 0))
		{
			temp_string = g_list_nth_data(fmWindow->pSelection, 0);
			fmWindow->pSelection = g_list_remove(fmWindow->pSelection, temp_string);
			g_free(temp_string);
		}
		g_list_free(fmWindow->pSelection);

		if(fmWindow->pSelectionTextSnapshot)
		{
			g_free(fmWindow->pSelectionTextSnapshot);
            fmWindow->pSelectionTextSnapshot = NULL;
		}
	}
}

GList* fmWindow_selectionGetListNode(struct FmWindow *fmWindow, GNode *sourceNode)
{
	gchar	*filename_to_remember = NULL;
	GList	*selection_node = NULL;

	filename_to_remember = fmWindow_getFilenameToRemember(fmWindow, sourceNode);
	if(filename_to_remember)
	{
		selection_node = g_list_find_custom(fmWindow->pSelection, filename_to_remember, (GCompareFunc)strcmp);
		g_free(filename_to_remember);
	}
	
	return selection_node;
}

void fmWindow_selectionRemoveAllDirEntries(struct FmWindow *fmWindow, GNode *sourceNode)
{
	GNode	*parent_gnode = sourceNode->parent;
	GNode	*child_source_gnode = NULL;
	
	if(parent_gnode)
	{
		child_source_gnode = g_node_first_child(parent_gnode);
		while(child_source_gnode)
		{
			fmWindow_selectionRemoveNode(fmWindow, child_source_gnode);
			child_source_gnode = child_source_gnode->next;
		}
	}
}

void fmWindow_selectionRemoveNode(struct FmWindow *fmWindow, GNode *sourceNode)
{
	GList			*selection_node = NULL;
	GtkCTreeNode	*focused_node = NULL;
	GtkCTreeNode	*ctree_node = NULL;

	if(sourceNode)
	{
		// check if the node to remove is still present in the selection glist
		selection_node = fmWindow_selectionGetListNode(fmWindow, sourceNode);
		if(selection_node && selection_node->data)
		{
			// yes, remove it
			fmWindow->pSelection = g_list_remove(fmWindow->pSelection, selection_node->data);
			g_free(selection_node->data);
		}
		
		// check if the node corresponds to a ctree node
		ctree_node = ((struct NodeInfo*)sourceNode->data)->pCtreeNode;
		if(ctree_node)
		{
			// yes, check if the node is not the currently focused one
			focused_node = gtk_ctree_node_nth(GTK_CTREE(fmWindow->pCTree), GTK_CLIST(fmWindow->pCTree)->focus_row);
			if(ctree_node != focused_node)
			{
				// no, just unselect the node
				fmWindow_unselectNode(fmWindow, sourceNode);
			}
			else
			{
				// yes, apply the focus row selection style
				fmWindow_selectNode(fmWindow, sourceNode);
			}
		}
		fmWindow_selectionUpdateSizeLabel(fmWindow);
	}
}

gchar* fmWindow_selectionTextSnapshotString(struct FmWindow *fmWindow, GList *selectionList)
{
	GList			*temp_selection_list = NULL;
	GString			*selection_string = g_string_new("");
	guint			ctree_selection_length = g_list_length(selectionList);
	guint			counter = 0;
	gchar			*text_snapshot = NULL;
    
	if(ctree_selection_length)
	{
        temp_selection_list = g_list_copy(selectionList);
		if(temp_selection_list)
		{
			temp_selection_list = g_list_sort(temp_selection_list, (GCompareFunc)strcmp);
			while(counter < ctree_selection_length)
			{
//				selection_string = g_string_append(selection_string, "\"");
				selection_string = g_string_append(selection_string, (gchar*)g_list_nth_data(temp_selection_list, counter));
//				selection_string = g_string_append(selection_string, "\"");
				if(counter < ctree_selection_length-1 && pgPreferences->pSelectionTextElementSeperator)
				{
					selection_string = g_string_append(selection_string, pgPreferences->pSelectionTextElementSeperator);
				}
				counter++;
			}
			g_list_free(temp_selection_list);
		}
	}
	else
	{
		selection_string = g_string_append(selection_string, fmWindow_getFilenameToRemember(fmWindow, fmWindow->pFocusedNode));
	}

	text_snapshot = selection_string->str;
	g_string_free(selection_string, FALSE);
	return(text_snapshot);
}

gchar* fmWindow_selectionTextSnapshotURL(struct FmWindow *fmWindow, GList *selectionList)
{
	GList			*temp_selection_list = NULL;
	GString			*selection_string = g_string_new("");
	guint			ctree_selection_length = g_list_length(selectionList);
	gchar			*text_snapshot = NULL;
    
	if(ctree_selection_length)
	{
        temp_selection_list = g_list_copy(selectionList);
		if(temp_selection_list)
		{
			temp_selection_list = g_list_sort(temp_selection_list, (GCompareFunc)strcmp);
			selection_string = g_string_append(selection_string, "file:");
			selection_string = g_string_append(selection_string, (gchar*)g_list_nth_data(temp_selection_list, 0));
			g_list_free(temp_selection_list);
		}
	}
	else
	{
		selection_string = g_string_append(selection_string, fmWindow_getFilenameToRemember(fmWindow, fmWindow->pFocusedNode));
	}

	text_snapshot = selection_string->str;
	g_string_free(selection_string, FALSE);
	return(text_snapshot);
}

void fmWindow_selectionToggleNode(struct FmWindow *fmWindow, GNode *sourceNode)
{
	if(fmWindow_selectionGetListNode(fmWindow, sourceNode))
	{
		fmWindow_selectionRemoveNode(fmWindow, sourceNode);
	}
	else
	{
		fmWindow_selectionAddNode(fmWindow, sourceNode);
	}
}

void fmWindow_selectionUnselectAll(struct FmWindow *fmWindow)
{
	GtkCTreeNode	*ctree_node = NULL;
	GNode			*source_gnode = NULL;
	
	// remove all the other selections
	fmWindow_selectionDestroy(fmWindow);
	
	// remove the visible selected rows
	ctree_node = gtk_ctree_node_nth(GTK_CTREE(fmWindow->pCTree), 0);
	while(ctree_node)
	{
		source_gnode = gtk_ctree_node_get_row_data(GTK_CTREE(fmWindow->pCTree), ctree_node);
		fmWindow_selectionRemoveNode(fmWindow, source_gnode);
		ctree_node = GTK_CTREE_NODE_NEXT(ctree_node);
	}
}

void fmWindow_selectionUpdateSizeLabel(struct FmWindow *fmWindow)
{
	gchar	*selection_size = NULL;

	selection_size = fmWindow_getSelectionSize(fmWindow);
	if(selection_size)
	{
		gtk_label_set_text(GTK_LABEL(fmWindow->pLabelSelectionSize), selection_size);
		g_free(selection_size);
	}
}

void fmWindow_selectCTreeNode(struct FmWindow *fmWindow, GtkCTreeNode *ctreeNode)
{
	GtkCTreeNode	*focus_ctree_node = NULL;
	struct RowStyle	*row_style = NULL;
	
	if(ctreeNode)
	{
		focus_ctree_node = gtk_ctree_node_nth(GTK_CTREE(fmWindow->pCTree), GTK_CLIST(fmWindow->pCTree)->focus_row);
		if(focus_ctree_node == ctreeNode)
		{
			if(fmWindow_selectionGetListNode(fmWindow, gtk_ctree_node_get_row_data(GTK_CTREE(fmWindow->pCTree), ctreeNode)))
			{
				row_style = rowStyleFind(pgPreferences->pRowStyles, "SelectedFocus");
			}
			else
			{
				row_style = rowStyleFind(pgPreferences->pRowStyles, "Focus");
			}
		}
		else
		{
			row_style = rowStyleFind(pgPreferences->pRowStyles, "Selected");
		}
		if(row_style)
		{
			gtk_widget_draw_focus(GTK_WIDGET(fmWindow->pCTree));
			gtk_ctree_node_set_background(GTK_CTREE(fmWindow->pCTree), ctreeNode, &(row_style->background));
			gtk_ctree_node_set_foreground(GTK_CTREE(fmWindow->pCTree), ctreeNode, &(row_style->foreground));
		}
	}
}

void fmWindow_selectCTreeRow(struct FmWindow *fmWindow, guint row)
{
	fmWindow_selectCTreeNode(fmWindow, gtk_ctree_node_nth(GTK_CTREE(fmWindow->pCTree), row));
}

void fmWindow_selectMoveToCTreeFocus(struct FmWindow *fmWindow)
{
	guint child_row = 0;
	
	child_row = GTK_CLIST(fmWindow->pCTree)->focus_row;
	fmWindow_selectCTreeRow(fmWindow, child_row);
	fmWindow_moveToCTreeRow(fmWindow, child_row);
}

void fmWindow_selectNode(struct FmWindow *fmWindow, GNode *sourceNode)
{
	fmWindow_selectCTreeNode(fmWindow, ((struct NodeInfo*)getLinkTarget(sourceNode)->data)->pCtreeNode);
}

void fmWindow_shortcutClear(struct FmWindow *fmWindow)
{
	fmWindow->pShortcut = g_string_assign(fmWindow->pShortcut, "");
}

gboolean fmWindow_shortcutCompare(struct FmWindow *fmWindow, GNode *sourceNode)
{
//	printf("sourceNode = 0x%x name = %s\n", (guint)sourceNode, ((struct NodeInfo*)sourceNode->data)->pShortName);
	if(pgPreferences->characterBrowsingIgnoreCase)
	{
		if(strncasecmp(fmWindow->pShortcut->str, ((struct NodeInfo*)sourceNode->data)->pShortName, fmWindow->pShortcut->len) == 0)
		{
			return TRUE;
		}
	}
	else
	{
		if(strncmp(fmWindow->pShortcut->str, ((struct NodeInfo*)sourceNode->data)->pShortName, fmWindow->pShortcut->len) == 0)
		{
			return TRUE;
		}
	}

	return FALSE;
}

GNode* fmWindow_shortcutFindFirstNode(struct FmWindow *fmWindow, GNode *sourceNode)
{
	GNode	*child_source_gnode = NULL;

	if(sourceNode)
	{
		child_source_gnode = sourceNode->parent->children;
		while(child_source_gnode)
		{
            if(fmWindow_shortcutCompare(fmWindow, child_source_gnode))
			{
				return child_source_gnode;
			}
            child_source_gnode = child_source_gnode->next;
		}
	}

	return NULL;
}

GNode* fmWindow_shortcutFindFollowingNode(struct FmWindow *fmWindow, GNode *sourceNode)
{
	GNode	*child_source_gnode = NULL;

	if(sourceNode)
	{
		child_source_gnode = sourceNode->next;
		while(child_source_gnode)
		{
            if(fmWindow_shortcutCompare(fmWindow, child_source_gnode))
			{
				return child_source_gnode;
			}
            child_source_gnode = child_source_gnode->next;
		}
	}

	return fmWindow_shortcutFindFirstNode(fmWindow, sourceNode);
}

GNode* fmWindow_shortcutFindPreviousNode(struct FmWindow *fmWindow, GNode *sourceNode)
{
	GNode	*child_source_gnode = NULL;

	if(sourceNode)
	{
		child_source_gnode = sourceNode->prev;
		while(child_source_gnode)
		{
            if(fmWindow_shortcutCompare(fmWindow, child_source_gnode))
			{
				return child_source_gnode;
			}
            child_source_gnode = child_source_gnode->prev;
		}
	}

	return fmWindow_shortcutFindFirstNode(fmWindow, sourceNode);
}

gboolean fmWindow_toggleNode(struct FmWindow *fmWindow, GNode *targetNode)
{
	struct NodeInfo *target_node_info = NULL;
	
	if(fmWindow && targetNode)
	{
		target_node_info = (struct NodeInfo*)targetNode->data;
		if(target_node_info->pDirEmptyCtreeNode)
		{
			fmWindow_expandNode(fmWindow, targetNode);
		}
		else if(target_node_info->pDirFirstCtreeNode)
		{
			fmWindow_collapseNode(fmWindow, targetNode);
		}
	}

	return FALSE;
}

void fmWindow_unselectNode(struct FmWindow *fmWindow, GNode *sourceNode)
{
	struct FileType	*file_type = NULL;
	GtkCTreeNode	*ctree_node = NULL;
	
	if(sourceNode)
	{
		ctree_node = ((struct NodeInfo*)sourceNode->data)->pCtreeNode;
		file_type = fileTypeGet(pgPreferences->pFileTypes, getLinkTarget(sourceNode));
		gtk_ctree_node_set_background(GTK_CTREE(fmWindow->pCTree), ctree_node, &(file_type->pRowStyle->background));
		gtk_ctree_node_set_foreground(GTK_CTREE(fmWindow->pCTree), ctree_node, &(file_type->pRowStyle->foreground));
	}
}

void fmWindow_unselectCTreeRow(struct FmWindow *fmWindow, guint row)
{
	GtkCTreeNode	*ctree_node = NULL;
	GNode			*source_gnode = NULL;
	
	ctree_node = gtk_ctree_node_nth(GTK_CTREE(fmWindow->pCTree), row);
	if(ctree_node)
	{
		source_gnode = gtk_ctree_node_get_row_data(GTK_CTREE(fmWindow->pCTree), ctree_node);
		fmWindow_unselectNode(fmWindow, source_gnode);
	}
}

void fmWindow_updateCTreeTitles(struct FmWindow *fmWindow)
{
	GList			*column = NULL;
	guint			counter = 0;
	GtkWidget		*previous_column_pixmap = NULL;
	GtkWidget		*previous_column_arrow = NULL;
	GtkWidget		*previous_column_hbox = NULL;
	GtkWidget		*previous_column_label = NULL;
	GdkPixmap		*pixmap = NULL;
	GdkPixmap		*mask = NULL;
	GtkWidget		*column_pixmap = NULL;
	GtkWidget		*column_arrow = NULL;
	GtkWidget		*column_hbox = NULL;
	GtkWidget		*column_label = NULL;

	column = g_list_first(pgPreferences->pCtreeColumns);
	counter = 0;

	while(column)
	{
		if(g_list_find(fmWindow->pCTreeColumns,(gpointer)((struct CtreeColumn*)column->data)->type))
		{
			if(((struct CtreeColumn*)column->data)->type == COLUMN_BOOKMARK)
			{
				// clear the eventual previous widget and its members
				if(gtk_clist_get_column_title(GTK_CLIST(fmWindow->pCTree), counter) == NULL)
				{
					previous_column_pixmap = gtk_clist_get_column_widget(GTK_CLIST(fmWindow->pCTree), counter);
					if(previous_column_pixmap)
					{
						gtk_widget_destroy(previous_column_pixmap);
					}
				}
				pixmap = gdk_pixmap_colormap_create_from_xpm_d(NULL, gdk_colormap_get_system(), &(mask), NULL, pgsReloadXpm);
				column_pixmap = gtk_pixmap_new(pixmap, mask);
				gtk_widget_show(GTK_WIDGET(column_pixmap));
				gtk_clist_set_column_widget(GTK_CLIST(fmWindow->pCTree), counter, column_pixmap);
			}
			else
			{
				// clear the eventual previous widget and its members
				if(gtk_clist_get_column_title(GTK_CLIST(fmWindow->pCTree), counter) == NULL)
				{
					previous_column_hbox = gtk_clist_get_column_widget(GTK_CLIST(fmWindow->pCTree), counter);
					if(previous_column_hbox)
					{
						previous_column_arrow = ((GtkBoxChild*)g_list_nth_data(GTK_BOX(previous_column_hbox)->children, 0))->widget;
						previous_column_label = ((GtkBoxChild*)g_list_nth_data(GTK_BOX(previous_column_hbox)->children, 1))->widget;
						gtk_widget_destroy(previous_column_label);
						gtk_widget_destroy(previous_column_arrow);
						gtk_widget_destroy(previous_column_hbox);
					}
				}
				// set the column titles and widget
				if(((struct CtreeColumn*)column->data)->type == fmWindow->sortType)
				{
					column_hbox = gtk_hbox_new(FALSE, 0);
					gtk_widget_show(GTK_WIDGET(column_hbox));
					column_label = gtk_label_new(pgColumnTitles[((struct CtreeColumn*)column->data)->type]);
					gtk_box_pack_start(GTK_BOX(column_hbox), GTK_WIDGET(column_label), TRUE, TRUE, 0);
					gtk_widget_show(GTK_WIDGET(column_label));
					if(fmWindow->sortDirection == SORT_ASCENDING)
					{
						column_arrow = gtk_arrow_new(GTK_ARROW_DOWN, GTK_SHADOW_ETCHED_IN);
					}
					else
					{
						column_arrow = gtk_arrow_new(GTK_ARROW_UP, GTK_SHADOW_ETCHED_IN);
					}
					gtk_box_pack_start(GTK_BOX(column_hbox), GTK_WIDGET(column_arrow), FALSE, FALSE, 2);
					gtk_widget_show(GTK_WIDGET(column_arrow));
					gtk_clist_set_column_widget(GTK_CLIST(fmWindow->pCTree), counter, column_hbox);
				}
				else
				{
					gtk_clist_set_column_title(GTK_CLIST(fmWindow->pCTree), counter, pgColumnTitles[((struct CtreeColumn*)column->data)->type]);
				}
			}
			gtk_clist_set_column_resizeable(GTK_CLIST(fmWindow->pCTree), counter, FALSE);
			counter++;
		}
		column = g_list_next(column);
	}
}

gboolean fmWindow_useDirEntry(struct FmWindow *fmWindow, struct dirent *dirEntry)
{
	if(!fmWindow->hideDotFiles || (fmWindow->hideDotFiles && dirEntry->d_name[0] != '.'))
	{
		if(strcmp(dirEntry->d_name, ".") != 0 && strcmp(dirEntry->d_name, ".."))
		{
			return TRUE;
		}
	}

	return FALSE;
}

GNode* fmWindow_validateActionOnNode(struct FmWindow *fmWindow, GNode *sourceNode)
{
	gchar		*file_name = NULL;
	struct stat	temp_stat;
	GNode		*parent_gnode = NULL;
	GNode		*return_gnode = NULL;
	gchar		*parent_file_name = NULL;
	gchar		*focused_file_name = NULL;

	file_name = fmWindow_getFilenameToRemember(fmWindow, sourceNode);
	
	if(file_name)
	{
		// check if file exists
		if(!lstat(file_name, &temp_stat))
		{
			// file exists
			// check if file hasn't been modified in the meantime
			if(temp_stat.st_mtime != ((struct NodeInfo*)sourceNode->data)->pStat->st_mtime || temp_stat.st_ctime != ((struct NodeInfo*)sourceNode->data)->pStat->st_ctime)
			{
//				printf("fmWindow_validateActionOnNode : '%s' has been modified, refreshing parent\n", file_name);
				fmWindow_refreshParentDirectory(fmWindow, sourceNode);
				fmWindow_moveToNode(fmWindow, fmWindow->pFocusedNode);
				return_gnode = fmWindow_getNodeFromFilename(fmWindow, file_name, TRUE);
			}
			else
			{
//				printf("fmWindow_validateActionOnNode : '%s' is perfectly ok\n", file_name);
				return_gnode = sourceNode;
			}
		}
		else
		{
			// file does not exist
//			printf("fmWindow_validateActionOnNode : '%s' does not exist, moving up 'till existing parent\n", file_name);
            parent_gnode = getLinkSource(sourceNode->parent);
			while(parent_gnode)
			{
				parent_file_name = fmWindow_getFilenameToRemember(fmWindow, parent_gnode);
//				printf("fmWindow_validateActionOnNode : trying parent '%s'\n", parent_file_name);
				if(parent_file_name)
				{
					// check if parent exists
					if(!lstat(parent_file_name, &temp_stat))
					{
						// check whether parent has been modified in the meantime
						if(temp_stat.st_mtime == ((struct NodeInfo*)parent_gnode->data)->pStat->st_mtime && temp_stat.st_ctime == ((struct NodeInfo*)parent_gnode->data)->pStat->st_ctime)
						{
//							printf("fmWindow_validateActionOnNode : parent '%s' has not been modified\n", parent_file_name);
							// parent exists and has not been modified
							if(focused_file_name)
							{
//								printf("fmWindow_validateActionOnNode : refreshing '%s' and focussing '%s'\n", parent_file_name, focused_file_name);
								fmWindow_refreshDirectory(fmWindow, getLinkTarget(parent_gnode), focused_file_name);
								fmWindow_moveToNode(fmWindow, fmWindow->pFocusedNode);
							}
							else
							{
//								printf("fmWindow_validateActionOnNode : refreshing '%s' and focussing '%s'\n", parent_file_name, parent_file_name);
								fmWindow_refreshDirectory(fmWindow, getLinkTarget(parent_gnode), parent_file_name);
								fmWindow_moveToNode(fmWindow, fmWindow->pFocusedNode);
							}
							break;
						}
						else
						{
//							printf("fmWindow_validateActionOnNode : parent '%s' has been modified\n", parent_file_name);
							if(focused_file_name)
							{
								g_free(focused_file_name);
							}
							focused_file_name = g_strdup(parent_file_name);
							// we're at the root directory
							if(!parent_gnode->parent)
							{
								fmWindow_refreshDirectory(fmWindow, getLinkTarget(parent_gnode), parent_file_name);
								break;
							}
						}
					}
					g_free(parent_file_name);
				}
//				printf("fmWindow_validateActionOnNode : parent_gnode->parent=0x%x\n", (guint)parent_gnode->parent);
				parent_gnode = getLinkSource(parent_gnode->parent);
			}
			if(parent_file_name)
			{
				g_free(parent_file_name);
			}
			if(focused_file_name)
			{
				g_free(focused_file_name);
			}
		}

		g_free(file_name);
	}

//	printf("fmWindow_validateActionOnNode : return_gnode=0x%x\n", (guint)return_gnode);

	return return_gnode;
}



struct FmWindow* buildNewFmWindow(struct FmWindow *fmWindow)
{
	if(fmWindow)
	{
		// if there's no default window geometry, apply the default position and size if this is specified in the settings
		if(pgPreferences->fmWindowDefaultPosition && (fmWindow->initialX < 0 || fmWindow->initialY < 0))
		{
			fmWindow->initialX = pgPreferences->fmWindowX;
			fmWindow->initialY = pgPreferences->fmWindowY;
		}

		if(fmWindow->initialWidth < 0 || fmWindow->initialHeight < 0)
		{
			fmWindow->initialWidth = pgPreferences->fmWindowWidth;
			fmWindow->initialHeight = pgPreferences->fmWindowHeight;
		}

		// initialize the window
		fmWindow_initialize(fmWindow);
		
		// add the window to the current list of windows
		pgFmWindows = g_list_prepend(pgFmWindows, fmWindow);

		// build the directory
		fmWindow->pRootNode = fmWindow_buildNode(fmWindow, fmWindow->pZoomDirectory);
		gtk_clist_freeze(GTK_CLIST(fmWindow->pCTree));
		fmWindow_insertCTreeNode(fmWindow, fmWindow->pRootNode, NULL, NULL, NULL);
		fmWindow_resizeColumns(fmWindow);
		gtk_clist_thaw(GTK_CLIST(fmWindow->pCTree));
		// let all the changements execute before continuing
		while(gtk_events_pending())
		{
			gtk_main_iteration();
		}
		if(fmWindow->pInitialFocus)
		{
			fmWindow_focusFileName(fmWindow, fmWindow->pInitialFocus);
			fmWindow_selectMoveToCTreeFocus(fmWindow);
		}
		else
		{
			fmWindow_focusFirstRow(fmWindow);
		}
	}

	return fmWindow;
}

void closeFmWindow(struct FmWindow *fmWindow)
{
	GList	*found_fm_window_node = NULL;

//	printf("closeFmWindow() : fmWindow = 0x%x\n", (guint)fmWindow);
	if(fmWindow)
	{
		found_fm_window_node = g_list_find(pgFmWindows, fmWindow);
		if(found_fm_window_node)
		{
			pgFmWindows = g_list_remove(pgFmWindows, fmWindow);
			fmWindow_destroy(fmWindow);
		}
	}
}

void closeAllFmWindows()
{
	struct	FmWindow	*fm_window = NULL;

//	printf("closeAllFmWindows() : number_of_fm_windows = %d\n", g_list_length(pgFmWindows));

	while(g_list_length(pgFmWindows))
	{
		fm_window = g_list_nth_data(pgFmWindows, 0);
        closeFmWindow(fm_window);
	}
}

GNode* findGNodeShortName(GNode *parentGNode, gchar *shortName)
{
	guint	number_of_children = 0;
	guint	counter = 0;
	GNode	*child_node = NULL;
	gchar	*child_node_short_name = NULL;

	if(parentGNode && shortName)
	{
		number_of_children = g_node_n_children(parentGNode);
		for(counter = 0; counter < number_of_children; counter++)
		{
            child_node = g_node_nth_child(parentGNode, counter);
            child_node_short_name = ((struct NodeInfo*)child_node->data)->pShortName;
//			printf("%d %s\n", counter, child_node_short_name);
			if(strcmp(child_node_short_name, shortName) == 0)
			{
				return child_node;
			}
		}
	}

	return NULL;
}

void rebuildAllFmWindowsCTrees()
{
	struct	FmWindow	*fm_window = NULL;
	guint				counter = 0;

//	printf("rebuildAllFmWindowsCTrees() : number_of_fm_windows = %d\n", g_list_length(pgFmWindows));

	for(counter = 0; counter < g_list_length(pgFmWindows); counter++)
	{
		fm_window = g_list_nth_data(pgFmWindows, counter);

        fmWindow_rebuildCTree(fm_window);
	}
}

gint sortNodesByChangeTime(GNode *node1, GNode *node2)
{
	gint			return_value = 0;

	return_value = ((struct NodeInfo*)node1->data)->pStat->st_ctime - ((struct NodeInfo*)node2->data)->pStat->st_ctime;
	
	if(return_value == 0)
	{
		return_value = sortNodesByFilename(node1, node2);
	}

	return return_value;
}

gint sortNodesByChangeTimeAscending(gconstpointer a, gconstpointer b)
{
	/// reverse the data since the items will be inserted in the tree with the preprend function
	return sortNodesByChangeTime((GNode*)b, (GNode*)a);
}

gint sortNodesByChangeTimeDescending(gconstpointer a, gconstpointer b)
{
	/// reverse the data since the items will be inserted in the tree with the preprend function
	return sortNodesByChangeTime((GNode*)a, (GNode*)b);
}

gint sortNodesByFilename(GNode *node1, GNode *node2)
{
	gint			return_value = 0;

	if(pgPreferences->sortIgnoreCase)
	{
		return_value = strcasecmp(((struct NodeInfo*)node1->data)->pShortName, ((struct NodeInfo*)node2->data)->pShortName);
	}
	else
	{
		return_value = strcmp(((struct NodeInfo*)node1->data)->pShortName, ((struct NodeInfo*)node2->data)->pShortName);
	}

	return return_value;
}

gint sortNodesByFilenameAscending(gconstpointer a, gconstpointer b)
{
	/// reverse the data since the items will be inserted in the tree with the preprend function
	return sortNodesByFilename((GNode*)b, (GNode*)a);
}

gint sortNodesByFilenameDescending(gconstpointer a, gconstpointer b)
{
	/// reverse the data since the items will be inserted in the tree with the preprend function
	return sortNodesByFilename((GNode*)a, (GNode*)b);
}

gint sortNodesByGroup(GNode *node1, GNode *node2)
{
	gint			return_value = 0;

	if(pgPreferences->sortIgnoreCase)
	{
		return_value = strcasecmp(((struct NodeInfo*)node1->data)->pGroupString, ((struct NodeInfo*)node2->data)->pGroupString);
	}
	else
	{
		return_value = strcmp(((struct NodeInfo*)node1->data)->pGroupString, ((struct NodeInfo*)node2->data)->pGroupString);
	}

	if(return_value == 0)
	{
		return_value = sortNodesByFilename(node1, node2);
	}
	
	return return_value;
}

gint sortNodesByGroupAscending(gconstpointer a, gconstpointer b)
{
	/// reverse the data since the items will be inserted in the tree with the preprend function
	return sortNodesByGroup((GNode*)b, (GNode*)a);
}

gint sortNodesByGroupDescending(gconstpointer a, gconstpointer b)
{
	/// reverse the data since the items will be inserted in the tree with the preprend function
	return sortNodesByGroup((GNode*)a, (GNode*)b);
}

gint sortNodesByModificationTime(GNode *node1, GNode *node2)
{
	gint			return_value = 0;

	return_value = ((struct NodeInfo*)node1->data)->pStat->st_mtime - ((struct NodeInfo*)node2->data)->pStat->st_mtime;

	if(return_value == 0)
	{
		return_value = sortNodesByFilename(node1, node2);
	}
	
	return return_value;
}

gint sortNodesByModificationTimeAscending(gconstpointer a, gconstpointer b)
{
	/// reverse the data since the items will be inserted in the tree with the preprend function
	return sortNodesByModificationTime((GNode*)b, (GNode*)a);
}

gint sortNodesByModificationTimeDescending(gconstpointer a, gconstpointer b)
{
	/// reverse the data since the items will be inserted in the tree with the preprend function
	return sortNodesByModificationTime((GNode*)a, (GNode*)b);
}

gint sortNodesByPermissions(GNode *node1, GNode *node2)
{
	gint			return_value = 0;

	return_value = strcmp(((struct NodeInfo*)node1->data)->pPermissionsString, ((struct NodeInfo*)node2->data)->pPermissionsString);

	if(return_value == 0)
	{
		return_value = sortNodesByFilename(node1, node2);
	}
	
	return return_value;
}

gint sortNodesByPermissionsAscending(gconstpointer a, gconstpointer b)
{
	/// reverse the data since the items will be inserted in the tree with the preprend function
	return sortNodesByPermissions((GNode*)b, (GNode*)a);
}

gint sortNodesByPermissionsDescending(gconstpointer a, gconstpointer b)
{
	/// reverse the data since the items will be inserted in the tree with the preprend function
	return sortNodesByPermissions((GNode*)a, (GNode*)b);
}

gint sortNodesBySize(GNode *node1, GNode *node2)
{
	gint			return_value = 0;

	return_value =  ((struct NodeInfo*)node1->data)->pStat->st_size - ((struct NodeInfo*)node2->data)->pStat->st_size;

	if(return_value == 0)
	{
		return_value = sortNodesByFilename(node1, node2);
	}
	
	return return_value;
}

gint sortNodesBySizeAscending(gconstpointer a, gconstpointer b)
{
	/// reverse the data since the items will be inserted in the tree with the preprend function
	return sortNodesBySize((GNode*)b, (GNode*)a);
}

gint sortNodesBySizeDescending(gconstpointer a, gconstpointer b)
{
	/// reverse the data since the items will be inserted in the tree with the preprend function
	return sortNodesBySize((GNode*)a, (GNode*)b);
}

gint sortNodesByUser(GNode *node1, GNode *node2)
{
	gint			return_value = 0;

	if(pgPreferences->sortIgnoreCase)
	{
		return_value = strcasecmp(((struct NodeInfo*)node1->data)->pUserString, ((struct NodeInfo*)node2->data)->pUserString);
	}
	else
	{
		return_value = strcmp(((struct NodeInfo*)node1->data)->pUserString, ((struct NodeInfo*)node2->data)->pUserString);
	}

	if(return_value == 0)
	{
		return_value = sortNodesByFilename(node1, node2);
	}
	
	return return_value;
}

gint sortNodesByUserAscending(gconstpointer a, gconstpointer b)
{
	/// reverse the data since the items will be inserted in the tree with the preprend function
	return sortNodesByUser((GNode*)b, (GNode*)a);
}

gint sortNodesByUserDescending(gconstpointer a, gconstpointer b)
{
	/// reverse the data since the items will be inserted in the tree with the preprend function
	return sortNodesByUser((GNode*)a, (GNode*)b);
}


